




 





















#include "analyze.h"

static LIST ana_CalculatePredicatePrecedence(LIST, LIST);
static LIST ana_CalculateFunctionPrecedence(LIST, LIST);
static void ana_AddTPTPSetupFof();
static void ana_AddTPTPSetupCls(LIST);





LIST ana_FINITEMONADICPREDICATES;  

BOOL ana_EQUATIONS;                
static BOOL ana_PEQUATIONS;        
static BOOL ana_NEQUATIONS;        
static BOOL ana_FUNCTIONS;         
static BOOL ana_PROP;              
static BOOL ana_GROUND;            
static BOOL ana_NONUNIT;           
static BOOL ana_MONADIC;           
static BOOL ana_NONMONADIC;        
BOOL ana_SORTRES;                  
BOOL ana_USORTRES;                 
static BOOL ana_FINDOMAIN;         
static BOOL ana_NONTRIVDOMAIN;     
static BOOL ana_CONGROUND;         

static BOOL ana_PUREEQUATIONAL;    
static BOOL ana_PUREPROPOSITIONAL; 

BOOL ana_SORTDECEQUATIONS;         
BOOL ana_SORTMANYEQUATIONS;        

static NAT  ana_PROP1;      
static NAT  ana_PROP2;        
static NAT  ana_PROP3;    
static NAT  ana_PROP4;      
static NAT  ana_PROP5;        
static NAT  ana_PROP6;         
static NAT  ana_PROP7;           
static NAT  ana_PROP8;           
static NAT  ana_PROP9;             
static NAT  ana_PROP10;        
static NAT  ana_PROP11;          
static NAT  ana_PROP12;            
static NAT  ana_PROP13;            
static BOOL ana_PROP14;             





void ana_Init()
{
  ana_FINITEMONADICPREDICATES = list_Nil();
}

void ana_Free()
{
  list_Delete(ana_FINITEMONADICPREDICATES);
}

static BOOL ana_PropertyEqual(NAT prop1, NAT prop2)
{
  return (prop1 == prop2);
}





void ana_AnalyzeProblem(PROOFSEARCH Search, LIST Clauses)

{
  LIST   Scan;
  CLAUSE Clause;

  ana_EQUATIONS       = FALSE;  
  ana_PEQUATIONS      = FALSE;             
  ana_NEQUATIONS      = FALSE; 
  ana_FUNCTIONS       = FALSE;
  ana_FINDOMAIN       = FALSE;
  ana_NONTRIVDOMAIN   = FALSE;
  ana_MONADIC         = FALSE;
  ana_NONMONADIC      = FALSE;
  ana_PROP            = FALSE;
  ana_GROUND          = FALSE;
  ana_SORTRES         = FALSE;
  ana_USORTRES        = FALSE;
  ana_NONUNIT         = FALSE;
  ana_CONGROUND       = TRUE;

  ana_PROP1    = 0; 
  ana_PROP2      = 0;
  ana_PROP3  = 0;

  list_Delete(ana_FINITEMONADICPREDICATES);
  ana_FINITEMONADICPREDICATES = list_Nil();

  if (list_Empty(Clauses))
    return;
  Scan = Clauses;

  ana_FINITEMONADICPREDICATES = clause_FiniteMonadicPredicates(Clauses);

  while(!list_Empty(Clauses)) {
    Clause = (CLAUSE)list_Car(Clauses);
    clause_SetWeight(Clause, clause_ComputeWeight(Clause));

    if (clause_GetFlag(Clause,CONCLAUSE))
      ana_PROP2++;
    else
      ana_PROP1++;

    if (clause_NumOfSuccLits(Clause) > 1)
      ana_PROP3++;

    if (ana_CONGROUND && clause_GetFlag(Clause,CONCLAUSE) &&
	clause_MaxVar(Clause) != symbol_GetInitialStandardVarCounter())
      ana_CONGROUND = FALSE;
    if (!ana_PEQUATIONS && clause_ContainsPositiveEquations(Clause)) {
      ana_PEQUATIONS = TRUE;
    }
    if (!ana_NEQUATIONS && clause_ContainsNegativeEquations(Clause)) {
      ana_NEQUATIONS = TRUE;
    }
    if (!ana_MONADIC || !ana_NONMONADIC || !ana_PROP || !ana_GROUND)
      clause_ContainsFolAtom(Clause,&ana_PROP,&ana_GROUND,&ana_MONADIC,&ana_NONMONADIC);

    if (!ana_FUNCTIONS && clause_ContainsFunctions(Clause)) {
      ana_FUNCTIONS = TRUE;
    }
    if (!ana_FINDOMAIN && clause_ImpliesFiniteDomain(Clause)) {
      ana_FINDOMAIN = TRUE;
    }
    if (!ana_NONTRIVDOMAIN && clause_ImpliesNonTrivialDomain(Clause)) {
      prfs_SetNonTrivClauseNumber(Search, clause_Number(Clause));
      ana_NONTRIVDOMAIN =  TRUE;
    }
    if (!ana_NONUNIT && clause_Length(Clause) > 1) {
      ana_NONUNIT = TRUE;
    }
    if (!ana_SORTRES || !ana_USORTRES) 
      clause_ContainsSortRestriction(Clause,&ana_SORTRES,&ana_USORTRES);
    
    Clauses = list_Cdr(Clauses);
  }

  ana_PUREEQUATIONAL    = ((ana_PEQUATIONS || ana_NEQUATIONS) && !ana_MONADIC && !ana_NONMONADIC 
			   && !ana_PROP && !ana_GROUND);
  ana_EQUATIONS         = (ana_PEQUATIONS || ana_NEQUATIONS);
  ana_PUREPROPOSITIONAL = (!ana_PEQUATIONS && !ana_NEQUATIONS &&!ana_MONADIC && !ana_NONMONADIC && ana_PROP);

  return;
}

void ana_AnalyzeSortStructure(LIST Clauses)
{
  if (ana_PEQUATIONS && ana_SORTRES) {
    STR Result;
    Result                = sort_AnalyzeSortStructure(Clauses);
    ana_SORTMANYEQUATIONS = (Result == SORTEQMANY);
    ana_SORTDECEQUATIONS  = (Result == SORTEQMANY || Result == SORTEQDECR);
  }
}

void ana_Print(void)
{
  char* Horn;

  if (ana_PROP3 == 0)
    Horn = "Horn";
  else
    Horn = "Non-Horn";
  
  if (ana_MONADIC && !ana_NONMONADIC) {
    printf("\n This is a monadic %s problem",Horn);
    if (ana_PEQUATIONS || ana_NEQUATIONS)
      printf(" with equality.");
    else
      printf(" without equality.");
  }
  else
    if (ana_PEQUATIONS || ana_NEQUATIONS) {
      if (ana_NONMONADIC || ana_MONADIC || ana_PROP)
	printf("\n This is a first-order %s problem containing equality.",Horn);
      else
	if (ana_NONUNIT)
	  printf("\n This is a pure equality %s problem.",Horn);
	else
	  printf("\n This is a unit equality problem.");
    }
    else
      if (ana_NONMONADIC || ana_MONADIC)
	printf("\n This is a first-order %s problem without equality.",Horn);

  if (ana_PUREPROPOSITIONAL)
    printf("\n This is a propositional %s problem.",Horn);
  else
    if (ana_FINDOMAIN || !ana_FUNCTIONS) {
      printf("\n This is a problem that has, if any, a finite domain model.");
      if (ana_FINDOMAIN)
	printf("\n There is a finite domain clause.");
      if (!ana_FUNCTIONS)
	printf("\n There are no function symbols.");
    }

  if (ana_NONTRIVDOMAIN)
    printf("\n This is a problem that has, if any, a non-trivial domain model.");
      
  
  if (ana_SORTRES) {
    printf("\n This is a problem that contains sort information.");
    if (ana_PEQUATIONS) {
      if (ana_SORTMANYEQUATIONS)
	printf("\n All equations are many sorted.");
      else {
	if (ana_SORTDECEQUATIONS)
	  printf("\n All equations are sort-decreasing.");
      }
    }
  }

  if (ana_PROP2 > 0 && ana_CONGROUND && !ana_PUREPROPOSITIONAL)
    printf("\n The conjecture is ground.");

  if (!list_Empty(ana_FINITEMONADICPREDICATES)) {
    LIST Scan;
    printf("\n The following monadic predicates have finite extensions: ");
    for(Scan=ana_FINITEMONADICPREDICATES;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      symbol_Print((SYMBOL)list_Car(Scan));
      if (!list_Empty(list_Cdr(Scan)))
	printf(", ");
    }
    printf(".");
  }

  printf("\n Axiom clauses: %d Conjecture clauses: %d",ana_PROP1,ana_PROP2);

  flag_PrintInferenceRules();
  flag_PrintReductionRules();
  printf("\n Extras    : ");
  if (flag_GetFlagValue(flag_SATINPUT))
    printf("Input Saturation, ");
  else
    printf("No Input Saturation, ");
  if (flag_GetFlagValue(flag_SELECT) == 0)
    printf("No Selection, ");
  else
    if (flag_GetFlagValue(flag_SELECT) == 1)
      printf("Dynamic Selection, ");
    else
      printf("Always Selection, ");
  if (flag_GetFlagValue(flag_SPLITS) == -1)
    printf("Full Splitting, ");
  else
    if (flag_GetFlagValue(flag_SPLITS) == 0)
      printf("No Splitting, ");
    else
      printf("Maximum of %d Splits, ",flag_GetFlagValue(flag_SPLITS));
  if (flag_GetFlagValue(flag_FULLRED))
    printf("Full Reduction, ");
  else
    printf("Lazy Reduction, ");
  printf(" Ratio: %d, FuncWeight: %d, VarWeight: %d",flag_GetFlagValue(flag_WDRATIO),
	 flag_GetFlagValue(flag_FUNCWEIGHT), flag_GetFlagValue(flag_VARWEIGHT));
  printf("\n Precedence: ");
  fol_PrintPrecedence();
  printf("\n Ordering  : %s",(flag_GetFlagValue(flag_ORD) == flag_ORDKBO ? "KBO" : "RPOS"));
}

void ana_AutoConfiguration(LIST Clauses)
{
  LIST       Scan, Functions, Predicates, Constants;
  SYMBOL     Symbol;
  SIGNATURE  Entry;
  int        i;

  Functions  = list_Nil();            
  Predicates = list_Nil();

  for (i=1;i<symbol_ACTINDEX;i++) {
    if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)i)) {
      Entry  = symbol_Signature(i);
      Symbol = Entry->info;
      if (symbol_IsPredicate(Symbol))
	Predicates = list_Cons((POINTER)Symbol,Predicates);
      else
	if (symbol_IsFunction(Symbol))
	  Functions = list_Cons((POINTER)Symbol,Functions);
    }
  }
  
  Predicates = ana_CalculatePredicatePrecedence(Predicates, Clauses);
  Functions  = ana_CalculateFunctionPrecedence(Functions, Clauses);
  Constants  = list_Nil();

  for(Scan=Functions;!list_Empty(Scan);Scan=list_Cdr(Scan))
    if (symbol_IsConstant((SYMBOL)list_Car(Scan)))
      Constants = list_Cons(list_Car(Scan),Constants);
  Functions = list_NPointerDifference(Functions,Constants);
  Constants = list_NReverse(Constants);

  for ( ; !list_Empty(Functions); Functions = list_Pop(Functions))
    symbol_SetOrdering((SYMBOL)list_Car(Functions),
		       symbol_GetIncreasedOrderingCounter());
  
  for ( ; !list_Empty(Predicates); Predicates = list_Pop(Predicates))
    symbol_SetOrdering((SYMBOL)list_Car(Predicates),
		       symbol_GetIncreasedOrderingCounter());
  
  
  for ( ; !list_Empty(Constants); Constants = list_Pop(Constants))
    symbol_SetOrdering((SYMBOL)list_Car(Constants),
		       symbol_GetIncreasedOrderingCounter());

  flag_ClearInferenceRules();
  flag_ClearReductionRules();

  flag_SetFlagValue(flag_ROBV,1);
  flag_SetFlagValue(flag_RTAUT,1);
  flag_SetFlagValue(flag_RFSUB,1);
  flag_SetFlagValue(flag_RBSUB,1);
  flag_SetFlagValue(flag_RFCLR,1);
  flag_SetFlagValue(flag_RBCLR,1);
  flag_SetFlagValue(flag_RUNC,1);
  flag_SetFlagValue(flag_FULLRED,1);
  
  flag_SetFlagValue(flag_WDRATIO,5);

  if (ana_NEQUATIONS) {
    flag_SetFlagValue(flag_IEQR,1);
    if (ana_NONUNIT) {
      if (ana_NONTRIVDOMAIN)
	flag_SetFlagValue(flag_RRRE,2);
      else
	flag_SetFlagValue(flag_RRRE,1);
    }
  }

  if (ana_PEQUATIONS) {
    flag_SetFlagValue(flag_ISPR,1);
    flag_SetFlagValue(flag_ISPL,1);
    if (ana_PROP3 >0)
      flag_SetFlagValue(flag_IEQF,1);
    if (ana_NONUNIT) 
      flag_SetFlagValue(flag_RCON,1);
    flag_SetFlagValue(flag_RFREW,1);
    flag_SetFlagValue(flag_RBREW,1);
  }
  
  if (ana_SORTRES) {
    flag_SetFlagValue(flag_SORTS,1);
    flag_SetFlagValue(flag_IEMS,1);
    flag_SetFlagValue(flag_ISOR,1);
    flag_SetFlagValue(flag_RSSI,1);
    if (!ana_PEQUATIONS || ana_SORTMANYEQUATIONS)
      flag_SetFlagValue(flag_RSST,1);
  }
  else
    flag_SetFlagValue(flag_SORTS,0);

  if (ana_MONADIC || ana_NONMONADIC) { 
    flag_SetFlagValue(flag_IORE,1);
    if (ana_PROP3 > 0)
      flag_SetFlagValue(flag_IOFC,1);
    if (ana_NONUNIT) 
      flag_SetFlagValue(flag_RCON,1);
  }


  if (!ana_FUNCTIONS)
    flag_SetFlagValue(flag_SELECT,2);
  else
    if (ana_NONUNIT)
      flag_SetFlagValue(flag_SELECT,1);
    else
      flag_SetFlagValue(flag_SELECT,0);

  if (ana_PROP2 < ana_PROP1 || (ana_CONGROUND && !ana_PUREPROPOSITIONAL))
    flag_SetFlagValue(flag_SATINPUT,1);
  else
    flag_SetFlagValue(flag_SATINPUT,0);

  if (ana_PROP3 > 0)
    flag_SetFlagValue(flag_SPLITS,-1);
  else
    flag_SetFlagValue(flag_SPLITS,0);
  if (ana_PROP14)
    ana_AddTPTPSetupFof();
  else
    ana_AddTPTPSetupCls(Clauses);
}

void ana_ExploitSortAnalysis()

{
  if (ana_SORTRES && (!ana_PEQUATIONS || ana_SORTMANYEQUATIONS))
    flag_SetFlagValue(flag_RSST,1);
}


static LIST ana_CalculatePredicatePrecedence(LIST Predicates, LIST Clauses)

{
  GRAPH  graph;
  LIST   result, scan;
  int    i, j;
  NAT    count;
  SYMBOL s;

  

  if (list_Empty(Predicates)) {
    return Predicates;
  }

  graph = graph_Create();

  
  for ( ; !list_Empty(Predicates); Predicates = list_Pop(Predicates))
    graph_AddNode(graph, symbol_Index((SYMBOL)list_Car(Predicates)));

  
  
  for (scan = Clauses; !list_Empty(scan); scan = list_Cdr(scan)) {
    CLAUSE c = list_Car(scan);

    for (i = clause_FirstLitIndex(); i < clause_FirstSuccedentLitIndex(c); i++) {
      SYMBOL negPred = term_TopSymbol(clause_GetLiteralAtom(c, i));
      if (!symbol_Equal(negPred, fol_Equality())) { 
	for (j = clause_FirstSuccedentLitIndex(c); j < clause_Length(c); j++) {
	  SYMBOL posPred = term_TopSymbol(clause_GetLiteralAtom(c, j));
	  if (!symbol_Equal(posPred, fol_Equality()) && 
	      negPred != posPred) {  
	    graph_AddEdge(graph_GetNode(graph, symbol_Index(negPred)),
			  graph_GetNode(graph, symbol_Index(posPred)));
	  }
	}
      }
    }
  }

  

  
  count = graph_StronglyConnectedComponents(graph);

  
  
  
  
  
  result = list_Nil();
  for (i = count - 1; i >= 0; i--) {
    for (scan = graph_Nodes(graph); !list_Empty(scan); scan = list_Cdr(scan)) {
      GRAPHNODE n = list_Car(scan);
      if (graph_NodeCompNum(n) == i) {
	
	s = symbol_GetSigSymbol(graph_NodeNumber(n));
	result = list_Cons((POINTER)s, result);
      }
    }
  }

  

  graph_Delete(graph);

  return result;
}


static __inline__ NAT ana_NodeDegree(GRAPHNODE Node)
{
  return (NAT)graph_NodeInfo(Node);
}

static __inline__ void ana_IncNodeDegree(GRAPHNODE Node)
{
  graph_NodeSetInfo(Node, (POINTER)(ana_NodeDegree(Node)+1));
}

static BOOL ana_NodeGreater(GRAPHNODE N1, GRAPHNODE N2)

{
  return (symbol_Arity(symbol_GetSigSymbol(graph_NodeNumber(N1))) >
	  symbol_Arity(symbol_GetSigSymbol(graph_NodeNumber(N2))));
}

static BOOL ana_BidirectionalDistributivity(LIST SymbolPairs)

{
  LIST scan, actPair, nextPair;

  for ( ; !list_Empty(SymbolPairs); SymbolPairs = list_Cdr(SymbolPairs)) {
    actPair = list_Car(SymbolPairs);
    
    for (scan = list_Cdr(SymbolPairs); !list_Empty(scan); scan = list_Cdr(scan)) {
      nextPair = list_Car(scan);
      if (symbol_Equal((SYMBOL)list_PairFirst(actPair),(SYMBOL)list_PairSecond(nextPair)) &&
	  symbol_Equal((SYMBOL)list_PairSecond(actPair),(SYMBOL)list_PairFirst(nextPair)))
	return TRUE;
    }
  }
  return FALSE;
}

static LIST ana_CalculateFunctionPrecedence(LIST Functions, LIST Clauses)

{
  GRAPH     graph;
  GRAPHNODE n1, n2;
  LIST      result, scan, scan2, distrPairs;
  NAT       i, j;
  SYMBOL    s, Add, Mult;

  if (list_Empty(Functions))
    return Functions;   
  else if (!ana_PEQUATIONS) {
    Functions = list_NumberSort(Functions, symbol_PositiveArity);;
    return Functions;
  }

  graph = graph_Create();
  
  for (; !list_Empty(Functions); Functions = list_Pop(Functions))
    graph_AddNode(graph, symbol_Index((SYMBOL)list_Car(Functions)));

  
  graph_SortNodes(graph, ana_NodeGreater);

  
  distrPairs = list_Nil();

  
  
  
  for (scan = Clauses; !list_Empty(scan); scan = list_Cdr(scan)) {
    CLAUSE c = list_Car(scan);
    for (i = clause_FirstSuccedentLitIndex(c);
	 i <= clause_LastSuccedentLitIndex(c); i++) {
      if (clause_LiteralIsEquality(clause_GetLiteral(c, i))) {
	
	TERM t1, t2;

	if (fol_DistributiveEquation(clause_GetLiteralAtom(c,i), &Add, &Mult)) {
	  
	  distrPairs = list_Cons(list_PairCreate((POINTER)Add, (POINTER)Mult),
				 distrPairs);
	  
	}

	t1 = term_FirstArgument(clause_GetLiteralAtom(c, i));
	t2 = term_SecondArgument(clause_GetLiteralAtom(c, i));

	if  (!term_IsVariable(t1) && !term_IsVariable(t2) &&
	     !term_EqualTopSymbols(t1, t2) &&  
	     !term_HasSubterm(t1, t2) &&       
	     !term_HasSubterm(t2, t1)) {
	  n1 = graph_GetNode(graph, symbol_Index(term_TopSymbol(t1)));
	  n2 = graph_GetNode(graph, symbol_Index(term_TopSymbol(t2)));
	  
	  graph_AddEdge(n1, n2);
	  graph_AddEdge(n2, n1);
	  
	  ana_IncNodeDegree(n1);
	  ana_IncNodeDegree(n2);
	}
      }
    }
  }
  
  

  graph_DeleteDuplicateEdges(graph);
  
  
  for (scan = graph_Nodes(graph); !list_Empty(scan); scan = list_Cdr(scan)) {
    n1 = list_Car(scan);
    result = list_Nil();  
    for (scan2 = graph_NodeNeighbors(n1); !list_Empty(scan2);
	 scan2 = list_Cdr(scan2)) {
      int a1, a2;
      n2 = list_Car(scan2);
      
      i  = ana_NodeDegree(n1);
      j  = ana_NodeDegree(n2);
      a1 = symbol_Arity(symbol_GetSigSymbol(graph_NodeNumber(n1)));
      a2 = symbol_Arity(symbol_GetSigSymbol(graph_NodeNumber(n2)));

      if (i > j || (i==j && a1 >= a2)) {
	
	result = list_Cons(n2, result);
      }
      if (i < j || (i==j && a1 <= a2)) {
	
	graph_DeleteEdge(n2, n1);
      }
      
    }
    
    for ( ; !list_Empty(result); result = list_Pop(result))
      graph_DeleteEdge(n1, list_Car(result));
  }

  if (!list_Empty(distrPairs) && !ana_BidirectionalDistributivity(distrPairs)) {
    
    flag_SetFlagValue(flag_ORD, flag_ORDRPOS);
  }

  
  
  for ( ; !list_Empty(distrPairs); distrPairs = list_Pop(distrPairs)) {
    scan = list_Car(distrPairs); 
    
    n1 = graph_GetNode(graph,
		       symbol_Index((SYMBOL)list_PairFirst(scan)));
    
    n2 = graph_GetNode(graph, 
		       symbol_Index((SYMBOL)list_PairSecond(scan)));
    
    graph_DeleteEdge(n1, n2);
    graph_DeleteEdge(n2, n1);
    
    graph_AddEdge(n1, n2);
    list_PairFree(scan);
  }

  

  
  
  i = graph_StronglyConnectedComponents(graph);

  
  
  
  
  
  result = list_Nil();
  while (i-- > 0) {   
    for (scan = graph_Nodes(graph); !list_Empty(scan); scan = list_Cdr(scan)) {
      n1 = list_Car(scan);
      if (graph_NodeCompNum(n1) == i) {
	
	s = symbol_GetSigSymbol(graph_NodeNumber(n1));
	result = list_Cons((POINTER)s, result);
      }
    }
  }

  

  graph_Delete(graph);

  return result;
}

BOOL ana_CheckProp(NAT Axioms, NAT Cons, NAT Axsymbols, NAT Consymbols,
			 NAT Axvars, NAT Convars, NAT Axdepth, NAT Condepth,
			 NAT Funcs, NAT Preds)
{
  return (ana_PropertyEqual(ana_PROP8,Axioms) &&
	  ana_PropertyEqual(ana_PROP9,Cons) &&
	  ana_PropertyEqual(ana_PROP4,Axsymbols) &&
	  ana_PropertyEqual(ana_PROP5,Consymbols) &&
	  ana_PropertyEqual(ana_PROP6,Axvars) &&
	  ana_PropertyEqual(ana_PROP7, Convars) &&
	  ana_PropertyEqual(ana_PROP10, Axdepth) &&
	  ana_PropertyEqual(ana_PROP11, Condepth) &&
	  ana_PropertyEqual(ana_PROP13, Funcs) &&
	  ana_PropertyEqual(ana_PROP12, Preds));
	  
}


void ana_AddTPTPSetupCls(LIST Clauses)
{
  
  if (!(ana_PEQUATIONS || ana_NEQUATIONS) &&
      (ana_CheckProp(fol_SC14,fol_SC4,fol_SC157,fol_SC18,fol_SC61,fol_SC0,fol_SC4,fol_SC3,fol_SC9,fol_SC5) || ana_CheckProp(fol_SC13,fol_SC4,fol_SC157,fol_SC21,fol_SC58,fol_SC2,fol_SC4,fol_SC3,fol_SC9,fol_SC4) || ana_CheckProp(fol_SC13,fol_SC4,fol_SC157,fol_SC21,fol_SC58,fol_SC0,fol_SC4,fol_SC3,fol_SC9,fol_SC4))) {
    flag_ClearInferenceRules();  
    flag_ClearReductionRules();
    flag_SetFlagValue(flag_SORTS,0);
    flag_SetFlagValue(flag_ISRE,1);  
    flag_SetFlagValue(flag_SOS,1);
    flag_SetFlagValue(flag_SPLITS,0);
    flag_SetFlagValue(flag_ISFC,1);
    flag_ActivateStdRedNoEq();
    
    return;
  }
  
  if (ana_CheckProp(fol_SC22,fol_SC1,fol_SC284,fol_SC6,fol_SC172,fol_SC0,fol_SC3,fol_SC4,fol_SC6,fol_SC5) 
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC53,fol_SC35,fol_SC25,fol_SC0,fol_SC5,fol_SC6,fol_SC6,fol_SC3)
      || ana_CheckProp(fol_SC8,fol_SC1,fol_SC84,fol_SC14,fol_SC34,fol_SC0,fol_SC6,fol_SC6,fol_SC5,fol_SC5)
      ) {
    flag_SetFlagValue(flag_RTER,2);
    flag_SetFlagValue(flag_SELECT,2);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC10,fol_SC1,fol_SC96,fol_SC14,fol_SC52,fol_SC0,fol_SC5,fol_SC4,fol_SC6,fol_SC3)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC29,fol_SC13,fol_SC13,fol_SC5,fol_SC5,fol_SC5,fol_SC5,fol_SC3)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC43,fol_SC10,fol_SC11,fol_SC0,fol_SC9,fol_SC4,fol_SC5,fol_SC3)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC43,fol_SC10,fol_SC11,fol_SC0,fol_SC7,fol_SC4,fol_SC5,fol_SC3)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC43,fol_SC10,fol_SC11,fol_SC0,fol_SC8,fol_SC4,fol_SC5,fol_SC3)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC43,fol_SC10,fol_SC11,fol_SC0,fol_SC6,fol_SC4,fol_SC5,fol_SC3)
      || ana_CheckProp(fol_SC11,fol_SC1,fol_SC99,fol_SC14,fol_SC42,fol_SC0,fol_SC5,fol_SC7,fol_SC9,fol_SC3)
      || ana_CheckProp(fol_SC14,fol_SC1,fol_SC110,fol_SC12,fol_SC52,fol_SC0,fol_SC4,fol_SC4,fol_SC8,fol_SC3)
      || ana_CheckProp(fol_SC14,fol_SC1,fol_SC110,fol_SC12,fol_SC52,fol_SC0,fol_SC4,fol_SC4,fol_SC8,fol_SC3)
      || ana_CheckProp(fol_SC20,fol_SC2,fol_SC355,fol_SC15,fol_SC234,fol_SC3,fol_SC3,fol_SC2,fol_SC10,fol_SC5)
      || ana_CheckProp(fol_SC20,fol_SC2,fol_SC355,fol_SC14,fol_SC234,fol_SC3,fol_SC3,fol_SC2,fol_SC10,fol_SC5)
      || ana_CheckProp(fol_SC2,fol_SC1,fol_SC27,fol_SC31,fol_SC12,fol_SC0,fol_SC8,fol_SC5,fol_SC10,fol_SC3)
      ) {
    flag_SetFlagValue(flag_ORD,1);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC18,fol_SC6,fol_SC187,fol_SC42,fol_SC108,fol_SC12,fol_SC3,fol_SC2,fol_SC9,fol_SC6)
      || ana_CheckProp(fol_SC7,fol_SC7,fol_SC73,fol_SC88,fol_SC30,fol_SC16,fol_SC4,fol_SC4,fol_SC7,fol_SC3)
      || ana_CheckProp(fol_SC45,fol_SC13,fol_SC613,fol_SC80,fol_SC151,fol_SC0,fol_SC4,fol_SC4,fol_SC21,fol_SC8)
      || ana_CheckProp(fol_SC18,fol_SC5,fol_SC320,fol_SC24,fol_SC211,fol_SC0,fol_SC3,fol_SC2,fol_SC14,fol_SC5)
      || ana_CheckProp(fol_SC18,fol_SC5,fol_SC320,fol_SC21,fol_SC211,fol_SC0,fol_SC3,fol_SC2,fol_SC12,fol_SC5)
      || ana_CheckProp(fol_SC18,fol_SC3,fol_SC320,fol_SC16,fol_SC211,fol_SC0,fol_SC3,fol_SC3,fol_SC11,fol_SC5)
      || ana_CheckProp(fol_SC7,fol_SC6,fol_SC53,fol_SC48,fol_SC26,fol_SC0,fol_SC4,fol_SC5,fol_SC9,fol_SC4)
      || ana_CheckProp(fol_SC19,fol_SC2,fol_SC253,fol_SC25,fol_SC116,fol_SC0,fol_SC4,fol_SC3,fol_SC16,fol_SC4)) {
    flag_SetFlagValue(flag_ISHY,1);
    flag_SetFlagValue(flag_SELECT,2);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC36,fol_SC2,fol_SC542,fol_SC37,fol_SC214,fol_SC0,fol_SC5,fol_SC5,fol_SC11,fol_SC4)) {
    flag_SetFlagValue(flag_ORD,1);
    flag_SetFlagValue(flag_ISHY,1);
    flag_SetFlagValue(flag_SELECT,2);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC20,fol_SC4,fol_SC353,fol_SC63,fol_SC194,fol_SC6,fol_SC15,fol_SC2,fol_SC25,fol_SC5)
      || ana_CheckProp(fol_SC21,fol_SC4,fol_SC374,fol_SC63,fol_SC212,fol_SC6,fol_SC15,fol_SC2,fol_SC25,fol_SC6)) {
    flag_SetFlagValue(flag_WDRATIO,10);
    flag_SetFlagValue(flag_SELECT,2);
    flag_SetFlagValue(flag_RTER,2);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC5,fol_SC1,fol_SC71,fol_SC6,fol_SC24,fol_SC0,fol_SC7,fol_SC3,fol_SC7,fol_SC3)
      || ana_CheckProp(fol_SC13,fol_SC1,fol_SC103,fol_SC14,fol_SC39,fol_SC0,fol_SC6,fol_SC5,fol_SC9,fol_SC3)
      || ana_CheckProp(fol_SC13,fol_SC1,fol_SC103,fol_SC12,fol_SC39,fol_SC0,fol_SC6,fol_SC4,fol_SC8,fol_SC3)) {
    flag_SetFlagValue(flag_ORD,1);
    flag_SetFlagValue(flag_ISPM,1);
    flag_SetFlagValue(flag_SELECT,2);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC3,fol_SC1,fol_SC36,fol_SC17,fol_SC14,fol_SC0,fol_SC5,fol_SC6,fol_SC4,fol_SC4)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC36,fol_SC17,fol_SC14,fol_SC0,fol_SC5,fol_SC7,fol_SC4,fol_SC4)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC36,fol_SC17,fol_SC14,fol_SC0,fol_SC5,fol_SC9,fol_SC4,fol_SC4)
      || ana_CheckProp(fol_SC2,fol_SC1,fol_SC22,fol_SC13,fol_SC10,fol_SC5,fol_SC5,fol_SC5,fol_SC4,fol_SC3)
      || ana_CheckProp(fol_SC2,fol_SC1,fol_SC24,fol_SC6,fol_SC11,fol_SC2,fol_SC5,fol_SC3,fol_SC4,fol_SC3)
      || ana_CheckProp(fol_SC2,fol_SC1,fol_SC28,fol_SC6,fol_SC13,fol_SC2,fol_SC5,fol_SC3,fol_SC4,fol_SC3)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC41,fol_SC13,fol_SC19,fol_SC5,fol_SC5,fol_SC5,fol_SC5,fol_SC3)
      || ana_CheckProp(fol_SC11,fol_SC1,fol_SC99,fol_SC10,fol_SC42,fol_SC0,fol_SC5,fol_SC5,fol_SC10,fol_SC3)
      || ana_CheckProp(fol_SC22,fol_SC5,fol_SC284,fol_SC20,fol_SC172,fol_SC0,fol_SC3,fol_SC2,fol_SC12,fol_SC5)
      || ana_CheckProp(fol_SC5,fol_SC1,fol_SC54,fol_SC25,fol_SC24,fol_SC3,fol_SC5,fol_SC7,fol_SC6,fol_SC4)
      || ana_CheckProp(fol_SC10,fol_SC1,fol_SC94,fol_SC12,fol_SC40,fol_SC0,fol_SC5,fol_SC5,fol_SC9,fol_SC3)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC37,fol_SC12,fol_SC18,fol_SC0,fol_SC5,fol_SC5,fol_SC5,fol_SC3)
      || ana_CheckProp(fol_SC78,fol_SC5,fol_SC1173,fol_SC22,fol_SC769,fol_SC0,fol_SC4,fol_SC2,fol_SC14,fol_SC5)
      || ana_CheckProp(fol_SC18,fol_SC3,fol_SC320,fol_SC13,fol_SC211,fol_SC0,fol_SC3,fol_SC2,fol_SC11,fol_SC5)
      || ana_CheckProp(fol_SC19,fol_SC7,fol_SC202,fol_SC28,fol_SC109,fol_SC0,fol_SC3,fol_SC2,fol_SC11,fol_SC5)
      || ana_CheckProp(fol_SC29,fol_SC1,fol_SC367,fol_SC5,fol_SC222,fol_SC0,fol_SC3,fol_SC2,fol_SC5,fol_SC6)
      || ana_CheckProp(fol_SC31,fol_SC3,fol_SC375,fol_SC14,fol_SC224,fol_SC0,fol_SC3,fol_SC3,fol_SC8,fol_SC6)
      || ana_CheckProp(fol_SC17,fol_SC3,fol_SC237,fol_SC13,fol_SC146,fol_SC3,fol_SC3,fol_SC2,fol_SC7,fol_SC5)
      || ana_CheckProp(fol_SC19,fol_SC3,fol_SC263,fol_SC13,fol_SC162,fol_SC3,fol_SC3,fol_SC2,fol_SC7,fol_SC5)
      || ana_CheckProp(fol_SC23,fol_SC3,fol_SC273,fol_SC13,fol_SC158,fol_SC3,fol_SC4,fol_SC2,fol_SC7,fol_SC5)
      || ana_CheckProp(fol_SC19,fol_SC3,fol_SC245,fol_SC13,fol_SC148,fol_SC3,fol_SC3,fol_SC2,fol_SC7,fol_SC5)
      || ana_CheckProp(fol_SC17,fol_SC1,fol_SC146,fol_SC12,fol_SC68,fol_SC0,fol_SC4,fol_SC4,fol_SC6,fol_SC3)
      || ana_CheckProp(fol_SC18,fol_SC3,fol_SC163,fol_SC51,fol_SC77,fol_SC16,fol_SC6,fol_SC5,fol_SC8,fol_SC3)
      || ana_CheckProp(fol_SC14,fol_SC4,fol_SC157,fol_SC18,fol_SC61,fol_SC2,fol_SC4,fol_SC3,fol_SC9,fol_SC5)
      || ana_CheckProp(fol_SC15,fol_SC3,fol_SC139,fol_SC12,fol_SC72,fol_SC0,fol_SC5,fol_SC3,fol_SC10,fol_SC6)
      || ana_CheckProp(fol_SC14,fol_SC4,fol_SC157,fol_SC18,fol_SC61,fol_SC2,fol_SC4,fol_SC3,fol_SC9,fol_SC5)
      || ana_CheckProp(fol_SC3,fol_SC2,fol_SC41,fol_SC39,fol_SC22,fol_SC10,fol_SC4,fol_SC6,fol_SC3,fol_SC3)
      || ana_CheckProp(fol_SC18,fol_SC5,fol_SC187,fol_SC29,fol_SC108,fol_SC6,fol_SC3,fol_SC2,fol_SC9,fol_SC6)
      || ana_CheckProp(fol_SC56,fol_SC12,fol_SC761,fol_SC67,fol_SC182,fol_SC0,fol_SC4,fol_SC5,fol_SC27,fol_SC8)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC32,fol_SC39,fol_SC12,fol_SC4,fol_SC7,fol_SC5,fol_SC11,fol_SC3)
      || ana_CheckProp(fol_SC93,fol_SC55,fol_SC566,fol_SC634,fol_SC148,fol_SC197,fol_SC2,fol_SC2,fol_SC30,fol_SC16)
      || ana_CheckProp(fol_SC17,fol_SC1,fol_SC237,fol_SC5,fol_SC146,fol_SC0,fol_SC3,fol_SC2,fol_SC5,fol_SC5)
      || ana_CheckProp(fol_SC7,fol_SC1,fol_SC101,fol_SC5,fol_SC62,fol_SC0,fol_SC3,fol_SC2,fol_SC4,fol_SC5)
      || ana_CheckProp(fol_SC17,fol_SC2,fol_SC237,fol_SC9,fol_SC146,fol_SC3,fol_SC3,fol_SC2,fol_SC5,fol_SC5)
      || ana_CheckProp(fol_SC0,fol_SC84,fol_SC0,fol_SC1588,fol_SC0,fol_SC424,fol_SC0,fol_SC3,fol_SC11,fol_SC5)
      ) {
    flag_SetFlagValue(flag_IURR,1);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC1,fol_SC1,fol_SC12,fol_SC39,fol_SC6,fol_SC4,fol_SC5,fol_SC5,fol_SC9,fol_SC3)
      || ana_CheckProp(fol_SC1,fol_SC1,fol_SC12,fol_SC39,fol_SC6,fol_SC4,fol_SC6,fol_SC5,fol_SC9,fol_SC3)
      || ana_CheckProp(fol_SC1,fol_SC1,fol_SC12,fol_SC39,fol_SC6,fol_SC4,fol_SC7,fol_SC5,fol_SC9,fol_SC3)
      || ana_CheckProp(fol_SC2,fol_SC1,fol_SC23,fol_SC39,fol_SC10,fol_SC4,fol_SC8,fol_SC5,fol_SC10,fol_SC3)
      || ana_CheckProp(fol_SC2,fol_SC1,fol_SC23,fol_SC39,fol_SC10,fol_SC4,fol_SC9,fol_SC5,fol_SC10,fol_SC3)) {
    flag_SetFlagValue(flag_ISPM,1);
    flag_SetFlagValue(flag_SELECT,0);
    flag_SetFlagValue(flag_RTER,2);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC13,fol_SC7,fol_SC97,fol_SC48,fol_SC49,fol_SC0,fol_SC4,fol_SC4,fol_SC10,fol_SC3)
      || ana_CheckProp(fol_SC20,fol_SC11,fol_SC222,fol_SC45,fol_SC138,fol_SC0,fol_SC3,fol_SC2,fol_SC14,fol_SC5)
      || ana_CheckProp(fol_SC26,fol_SC11,fol_SC286,fol_SC45,fol_SC174,fol_SC0,fol_SC3,fol_SC2,fol_SC14,fol_SC5)
      || ana_CheckProp(fol_SC2,fol_SC1,fol_SC20,fol_SC12,fol_SC11,fol_SC0,fol_SC4,fol_SC4,fol_SC5,fol_SC3)) {
    flag_SetFlagValue(flag_ISPM,1);
    flag_SetFlagValue(flag_SELECT,2);
    flag_SetFlagValue(flag_ISHY,1);
    
    return;
  }
  
  if(ana_CheckProp(fol_SC4,fol_SC1,fol_SC38,fol_SC13,fol_SC16,fol_SC0,fol_SC5,fol_SC5,fol_SC5,fol_SC4)
     || ana_CheckProp(fol_SC6,fol_SC1,fol_SC58,fol_SC15,fol_SC27,fol_SC0,fol_SC6,fol_SC5,fol_SC5,fol_SC4)
     || ana_CheckProp(fol_SC6,fol_SC1,fol_SC58,fol_SC11,fol_SC27,fol_SC0,fol_SC6,fol_SC5,fol_SC4,fol_SC4)
     || ana_CheckProp(fol_SC4,fol_SC1,fol_SC40,fol_SC13,fol_SC18,fol_SC0,fol_SC5,fol_SC5,fol_SC5,fol_SC4)
     || ana_CheckProp(fol_SC4,fol_SC1,fol_SC50,fol_SC17,fol_SC24,fol_SC0,fol_SC7,fol_SC6,fol_SC5,fol_SC4)
     || ana_CheckProp(fol_SC5,fol_SC1,fol_SC50,fol_SC13,fol_SC23,fol_SC0,fol_SC5,fol_SC5,fol_SC5,fol_SC4)
     || ana_CheckProp(fol_SC5,fol_SC1,fol_SC50,fol_SC11,fol_SC23,fol_SC0,fol_SC5,fol_SC5,fol_SC4,fol_SC4)
     || ana_CheckProp(fol_SC2,fol_SC1,fol_SC26,fol_SC21,fol_SC12,fol_SC0,fol_SC7,fol_SC6,fol_SC5,fol_SC4)
     || ana_CheckProp(fol_SC2,fol_SC1,fol_SC26,fol_SC17,fol_SC12,fol_SC0,fol_SC7,fol_SC7,fol_SC5,fol_SC4)
     || ana_CheckProp(fol_SC8,fol_SC1,fol_SC84,fol_SC12,fol_SC34,fol_SC0,fol_SC6,fol_SC6,fol_SC4,fol_SC5)
     || ana_CheckProp(fol_SC8,fol_SC1,fol_SC84,fol_SC15,fol_SC34,fol_SC0,fol_SC6,fol_SC7,fol_SC4,fol_SC5)
     || ana_CheckProp(fol_SC8,fol_SC1,fol_SC84,fol_SC14,fol_SC34,fol_SC0,fol_SC6,fol_SC7,fol_SC4,fol_SC5)
     || ana_CheckProp(fol_SC12,fol_SC1,fol_SC128,fol_SC12,fol_SC70,fol_SC0,fol_SC4,fol_SC4,fol_SC5,fol_SC4)
     || ana_CheckProp(fol_SC4,fol_SC12,fol_SC65,fol_SC42,fol_SC38,fol_SC2,fol_SC4,fol_SC3,fol_SC18,fol_SC5)
     || ana_CheckProp(fol_SC4,fol_SC1,fol_SC38,fol_SC9,fol_SC16,fol_SC0,fol_SC5,fol_SC5,fol_SC4,fol_SC4)
     || ana_CheckProp(fol_SC5,fol_SC1,fol_SC40,fol_SC15,fol_SC18,fol_SC0,fol_SC5,fol_SC5,fol_SC5,fol_SC4)
     || ana_CheckProp(fol_SC6,fol_SC1,fol_SC52,fol_SC13,fol_SC22,fol_SC0,fol_SC5,fol_SC5,fol_SC5,fol_SC4)
     || ana_CheckProp(fol_SC4,fol_SC1,fol_SC36,fol_SC13,fol_SC16,fol_SC0,fol_SC5,fol_SC5,fol_SC5,fol_SC4)
     
     || ana_CheckProp(fol_SC3,fol_SC1,fol_SC38,fol_SC17,fol_SC18,fol_SC0,fol_SC6,fol_SC7,fol_SC5,fol_SC4)
     || ana_CheckProp(fol_SC2,fol_SC1,fol_SC26,fol_SC13,fol_SC12,fol_SC0,fol_SC7,fol_SC7,fol_SC4,fol_SC4)
     || ana_CheckProp(fol_SC2,fol_SC1,fol_SC26,fol_SC17,fol_SC12,fol_SC0,fol_SC6,fol_SC7,fol_SC5,fol_SC4)
     || ana_CheckProp(fol_SC8,fol_SC1,fol_SC84,fol_SC16,fol_SC34,fol_SC0,fol_SC6,fol_SC6,fol_SC5,fol_SC5)) {
    flag_ClearInferenceRules();  
    flag_ClearReductionRules();
    flag_SetFlagValue(flag_SORTS,0);
    flag_SetFlagValue(flag_ISHY,1); 
    flag_SetFlagValue(flag_RFSUB,1);
    flag_SetFlagValue(flag_RBSUB,1);  
    flag_SetFlagValue(flag_RTER,2);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC4,fol_SC1,fol_SC46,fol_SC13,fol_SC21,fol_SC0,fol_SC6,fol_SC5,fol_SC5,fol_SC4)
      || ana_CheckProp(fol_SC38,fol_SC10,fol_SC164,fol_SC83,fol_SC18,fol_SC24,fol_SC2,fol_SC2,fol_SC10,fol_SC8)
      || ana_CheckProp(fol_SC17,fol_SC10,fol_SC237,fol_SC54,fol_SC146,fol_SC16,fol_SC3,fol_SC3,fol_SC12,fol_SC5)) {
    flag_SetFlagValue(flag_RTER,1);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC8,fol_SC1,fol_SC77,fol_SC11,fol_SC35,fol_SC0,fol_SC5,fol_SC4,fol_SC7,fol_SC4)) {
    flag_SetFlagValue(flag_ORD,1);
    flag_SetFlagValue(flag_SELECT,2);
    flag_SetFlagValue(flag_RTER,2);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC112,fol_SC3,fol_SC1246,fol_SC13,fol_SC498,fol_SC0,fol_SC7,fol_SC3,fol_SC49,fol_SC13)
      || ana_CheckProp(fol_SC3,fol_SC2,fol_SC45,fol_SC29,fol_SC24,fol_SC0,fol_SC4,fol_SC5,fol_SC7,fol_SC3)
      || ana_CheckProp(fol_SC34,fol_SC1,fol_SC406,fol_SC6,fol_SC248,fol_SC0,fol_SC3,fol_SC4,fol_SC6,fol_SC5)
      || ana_CheckProp(fol_SC4,fol_SC1,fol_SC42,fol_SC10,fol_SC20,fol_SC0,fol_SC5,fol_SC5,fol_SC6,fol_SC3)
      || ana_CheckProp(fol_SC37,fol_SC5,fol_SC515,fol_SC21,fol_SC346,fol_SC0,fol_SC3,fol_SC2,fol_SC13,fol_SC5)
      || ana_CheckProp(fol_SC18,fol_SC3,fol_SC320,fol_SC21,fol_SC211,fol_SC0,fol_SC3,fol_SC3,fol_SC14,fol_SC5)
      || ana_CheckProp(fol_SC18,fol_SC2,fol_SC320,fol_SC16,fol_SC211,fol_SC0,fol_SC3,fol_SC3,fol_SC10,fol_SC5)
      || ana_CheckProp(fol_SC64,fol_SC5,fol_SC864,fol_SC22,fol_SC562,fol_SC0,fol_SC4,fol_SC2,fol_SC13,fol_SC5)
      || ana_CheckProp(fol_SC66,fol_SC5,fol_SC926,fol_SC24,fol_SC603,fol_SC0,fol_SC4,fol_SC2,fol_SC15,fol_SC5)
      || ana_CheckProp(fol_SC67,fol_SC7,fol_SC953,fol_SC34,fol_SC621,fol_SC0,fol_SC4,fol_SC2,fol_SC17,fol_SC5)
      || ana_CheckProp(fol_SC68,fol_SC4,fol_SC992,fol_SC18,fol_SC647,fol_SC0,fol_SC4,fol_SC2,fol_SC13,fol_SC5)
      || ana_CheckProp(fol_SC28,fol_SC3,fol_SC354,fol_SC14,fol_SC212,fol_SC0,fol_SC3,fol_SC3,fol_SC8,fol_SC6)
      || ana_CheckProp(fol_SC31,fol_SC5,fol_SC387,fol_SC23,fol_SC230,fol_SC0,fol_SC3,fol_SC3,fol_SC11,fol_SC6)
      || ana_CheckProp(fol_SC26,fol_SC3,fol_SC345,fol_SC14,fol_SC208,fol_SC0,fol_SC3,fol_SC3,fol_SC8,fol_SC6)
      || ana_CheckProp(fol_SC15,fol_SC10,fol_SC189,fol_SC54,fol_SC116,fol_SC16,fol_SC3,fol_SC3,fol_SC12,fol_SC5)
      || ana_CheckProp(fol_SC157,fol_SC2,fol_SC1726,fol_SC14,fol_SC701,fol_SC0,fol_SC7,fol_SC4,fol_SC42,fol_SC12)
      ) {
    flag_SetFlagValue(flag_RTER,2);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC112,fol_SC3,fol_SC1246,fol_SC12,fol_SC498,fol_SC0,fol_SC7,fol_SC3,fol_SC49,fol_SC13)) {
    flag_SetFlagValue(flag_ISHY,1);
    flag_SetFlagValue(flag_RTER,2);
    
    return;
  }
  
  if ( ana_CheckProp(fol_SC112,fol_SC2,fol_SC1246,fol_SC15,fol_SC498,fol_SC0,fol_SC7,fol_SC5,fol_SC49,fol_SC13)
	     || ana_CheckProp(fol_SC112,fol_SC1,fol_SC1246,fol_SC7,fol_SC498,fol_SC0,fol_SC7,fol_SC3,fol_SC48,fol_SC13)
	     
	     || ana_CheckProp(fol_SC112,fol_SC2,fol_SC1246,fol_SC9,fol_SC498,fol_SC0,fol_SC7,fol_SC3,fol_SC47,fol_SC13)
	     || ana_CheckProp(fol_SC112,fol_SC2,fol_SC1246,fol_SC11,fol_SC498,fol_SC0,fol_SC7,fol_SC4,fol_SC48,fol_SC13)) {
    flag_SetFlagValue(flag_ISRE,1);
    flag_SetFlagValue(flag_IORE,0);
    flag_SetFlagValue(flag_SELECT,0);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC2,fol_SC1,fol_SC22,fol_SC20,fol_SC10,fol_SC7,fol_SC5,fol_SC6,fol_SC5,fol_SC3)|| ana_CheckProp(fol_SC3,fol_SC1,fol_SC36,fol_SC19,fol_SC14,fol_SC0,fol_SC5,fol_SC9,fol_SC4,fol_SC4)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC36,fol_SC19,fol_SC14,fol_SC0,fol_SC5,fol_SC7,fol_SC4,fol_SC4)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC36,fol_SC19,fol_SC14,fol_SC0,fol_SC5,fol_SC8,fol_SC4,fol_SC4)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC36,fol_SC19,fol_SC14,fol_SC0,fol_SC5,fol_SC6,fol_SC4,fol_SC4)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC35,fol_SC13,fol_SC16,fol_SC5,fol_SC5,fol_SC5,fol_SC5,fol_SC3)
      || ana_CheckProp(fol_SC90,fol_SC2,fol_SC1334,fol_SC29,fol_SC876,fol_SC0,fol_SC4,fol_SC2,fol_SC13,fol_SC6)
      || ana_CheckProp(fol_SC74,fol_SC2,fol_SC1087,fol_SC17,fol_SC711,fol_SC0,fol_SC4,fol_SC3,fol_SC16,fol_SC5)
      || ana_CheckProp(fol_SC1,fol_SC1,fol_SC18,fol_SC31,fol_SC8,fol_SC0,fol_SC10,fol_SC5,fol_SC9,fol_SC3)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC38,fol_SC17,fol_SC18,fol_SC0,fol_SC6,fol_SC7,fol_SC5,fol_SC4)
      || ana_CheckProp(fol_SC8,fol_SC2,fol_SC77,fol_SC22,fol_SC35,fol_SC0,fol_SC5,fol_SC3,fol_SC8,fol_SC4)
      
      || ana_CheckProp(fol_SC112,fol_SC1,fol_SC1246,fol_SC6,fol_SC498,fol_SC0,fol_SC7,fol_SC3,fol_SC46,fol_SC13)) {
    flag_SetFlagValue(flag_ISRE,1);
    flag_SetFlagValue(flag_IORE,0);
    flag_SetFlagValue(flag_PREFCON,2);
    flag_SetFlagValue(flag_SORTS,0);
    flag_SetFlagValue(flag_SPLITS,0);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC4,fol_SC1,fol_SC37,fol_SC8,fol_SC24,fol_SC0,fol_SC4,fol_SC4,fol_SC4,fol_SC3)
      || ana_CheckProp(fol_SC11,fol_SC6,fol_SC90,fol_SC51,fol_SC40,fol_SC12,fol_SC4,fol_SC4,fol_SC7,fol_SC5)
      || ana_CheckProp(fol_SC2,fol_SC1,fol_SC22,fol_SC20,fol_SC10,fol_SC7,fol_SC5,fol_SC6,fol_SC6,fol_SC3)
      || ana_CheckProp(fol_SC36,fol_SC5,fol_SC488,fol_SC24,fol_SC328,fol_SC0,fol_SC3,fol_SC2,fol_SC15,fol_SC5)
      || ana_CheckProp(fol_SC38,fol_SC3,fol_SC537,fol_SC16,fol_SC360,fol_SC0,fol_SC3,fol_SC3,fol_SC12,fol_SC5)
      || ana_CheckProp(fol_SC60,fol_SC1,fol_SC792,fol_SC30,fol_SC518,fol_SC0,fol_SC4,fol_SC3,fol_SC13,fol_SC5)
      || ana_CheckProp(fol_SC53,fol_SC4,fol_SC718,fol_SC17,fol_SC478,fol_SC0,fol_SC4,fol_SC2,fol_SC13,fol_SC5)
      || ana_CheckProp(fol_SC62,fol_SC1,fol_SC278,fol_SC14,fol_SC47,fol_SC9,fol_SC2,fol_SC2,fol_SC6,fol_SC8)
      || ana_CheckProp(fol_SC61,fol_SC1,fol_SC274,fol_SC14,fol_SC44,fol_SC9,fol_SC2,fol_SC2,fol_SC6,fol_SC8)
      || ana_CheckProp(fol_SC4,fol_SC1,fol_SC36,fol_SC11,fol_SC16,fol_SC0,fol_SC5,fol_SC5,fol_SC4,fol_SC4)
      || ana_CheckProp(fol_SC18,fol_SC2,fol_SC125,fol_SC19,fol_SC41,fol_SC0,fol_SC4,fol_SC5,fol_SC14,fol_SC4)
      || ana_CheckProp(fol_SC92,fol_SC2,fol_SC1016,fol_SC13,fol_SC417,fol_SC0,fol_SC7,fol_SC5,fol_SC42,fol_SC12)
      
      || ana_CheckProp(fol_SC112,fol_SC1,fol_SC1246,fol_SC7,fol_SC498,fol_SC0,fol_SC7,fol_SC4,fol_SC47,fol_SC13)
      || ana_CheckProp(fol_SC109,fol_SC6,fol_SC1633,fol_SC94,fol_SC939,fol_SC0,fol_SC4,fol_SC3,fol_SC38,fol_SC25)
      || ana_CheckProp(fol_SC57,fol_SC1,fol_SC331,fol_SC14,fol_SC87,fol_SC9,fol_SC2,fol_SC2,fol_SC6,fol_SC9)
      ) {
    flag_SetFlagValue(flag_FULLRED,0);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC20,fol_SC1,fol_SC117,fol_SC14,fol_SC34,fol_SC9,fol_SC2,fol_SC2,fol_SC4,fol_SC6) 
      || ana_CheckProp(fol_SC30,fol_SC1,fol_SC157,fol_SC14,fol_SC40,fol_SC9,fol_SC2,fol_SC2,fol_SC4,fol_SC8)
      || ana_CheckProp(fol_SC41,fol_SC1,fol_SC268,fol_SC14,fol_SC81,fol_SC9,fol_SC2,fol_SC2,fol_SC5,fol_SC9)
      || ana_CheckProp(fol_SC22,fol_SC3,fol_SC161,fol_SC42,fol_SC54,fol_SC27,fol_SC2,fol_SC2,fol_SC4,fol_SC6)|| ana_CheckProp(fol_SC42,fol_SC2,fol_SC272,fol_SC46,fol_SC84,fol_SC28,fol_SC2,fol_SC2,fol_SC5,fol_SC9)|| ana_CheckProp(fol_SC23,fol_SC2,fol_SC165,fol_SC46,fol_SC57,fol_SC28,fol_SC2,fol_SC2,fol_SC4,fol_SC6)|| ana_CheckProp(fol_SC47,fol_SC1,fol_SC337,fol_SC14,fol_SC121,fol_SC9,fol_SC2,fol_SC2,fol_SC5,fol_SC11)
      || ana_CheckProp(fol_SC28,fol_SC1,fol_SC230,fol_SC14,fol_SC94,fol_SC9,fol_SC2,fol_SC2,fol_SC4,fol_SC8)
      || ana_CheckProp(fol_SC21,fol_SC1,fol_SC121,fol_SC14,fol_SC37,fol_SC9,fol_SC2,fol_SC2,fol_SC4,fol_SC6)
      || ana_CheckProp(fol_SC31,fol_SC1,fol_SC161,fol_SC14,fol_SC43,fol_SC9,fol_SC2,fol_SC2,fol_SC4,fol_SC8)
      || ana_CheckProp(fol_SC42,fol_SC1,fol_SC272,fol_SC14,fol_SC84,fol_SC9,fol_SC2,fol_SC2,fol_SC5,fol_SC9)
      || ana_CheckProp(fol_SC23,fol_SC3,fol_SC165,fol_SC42,fol_SC57,fol_SC27,fol_SC2,fol_SC2,fol_SC4,fol_SC6)
      || ana_CheckProp(fol_SC30,fol_SC2,fol_SC159,fol_SC46,fol_SC39,fol_SC28,fol_SC2,fol_SC2,fol_SC5,fol_SC6)
      || ana_CheckProp(fol_SC45,fol_SC2,fol_SC214,fol_SC46,fol_SC45,fol_SC28,fol_SC2,fol_SC2,fol_SC5,fol_SC8)
      || ana_CheckProp(fol_SC32,fol_SC2,fol_SC211,fol_SC46,fol_SC63,fol_SC28,fol_SC2,fol_SC2,fol_SC5,fol_SC6)
      || ana_CheckProp(fol_SC30,fol_SC1,fol_SC159,fol_SC14,fol_SC39,fol_SC9,fol_SC2,fol_SC2,fol_SC5,fol_SC6)
      || ana_CheckProp(fol_SC45,fol_SC1,fol_SC214,fol_SC14,fol_SC45,fol_SC9,fol_SC2,fol_SC2,fol_SC5,fol_SC8)
      
      || ana_CheckProp(fol_SC32,fol_SC3,fol_SC211,fol_SC42,fol_SC63,fol_SC27,fol_SC2,fol_SC2,fol_SC5,fol_SC6)
      || ana_CheckProp(fol_SC31,fol_SC3,fol_SC207,fol_SC42,fol_SC60,fol_SC27,fol_SC2,fol_SC2,fol_SC5,fol_SC6)
      || ana_CheckProp(fol_SC29,fol_SC1,fol_SC155,fol_SC14,fol_SC36,fol_SC9,fol_SC2,fol_SC2,fol_SC5,fol_SC6)
      || ana_CheckProp(fol_SC44,fol_SC1,fol_SC210,fol_SC14,fol_SC42,fol_SC9,fol_SC2,fol_SC2,fol_SC5,fol_SC8)
      || ana_CheckProp(fol_SC29,fol_SC2,fol_SC155,fol_SC46,fol_SC36,fol_SC28,fol_SC2,fol_SC2,fol_SC5,fol_SC6)
      || ana_CheckProp(fol_SC2,fol_SC1,fol_SC22,fol_SC13,fol_SC10,fol_SC0,fol_SC6,fol_SC5,fol_SC4,fol_SC4)
      || ana_CheckProp(fol_SC4,fol_SC1,fol_SC46,fol_SC8,fol_SC21,fol_SC0,fol_SC6,fol_SC5,fol_SC3,fol_SC4)
      || ana_CheckProp(fol_SC2,fol_SC1,fol_SC28,fol_SC13,fol_SC13,fol_SC0,fol_SC6,fol_SC5,fol_SC4,fol_SC4)
      || ana_CheckProp(fol_SC8,fol_SC1,fol_SC84,fol_SC18,fol_SC34,fol_SC0,fol_SC6,fol_SC7,fol_SC5,fol_SC5)
      || ana_CheckProp(fol_SC8,fol_SC1,fol_SC84,fol_SC21,fol_SC34,fol_SC0,fol_SC6,fol_SC7,fol_SC5,fol_SC5)
      || ana_CheckProp(fol_SC8,fol_SC1,fol_SC84,fol_SC24,fol_SC34,fol_SC0,fol_SC6,fol_SC7,fol_SC6,fol_SC5)
      || ana_CheckProp(fol_SC0,fol_SC724,fol_SC0,fol_SC8266,fol_SC0,fol_SC2209,fol_SC0,fol_SC2,fol_SC212,fol_SC99)
      || ana_CheckProp(fol_SC0,fol_SC700,fol_SC0,fol_SC7465,fol_SC0,fol_SC1878,fol_SC0,fol_SC2,fol_SC199,fol_SC96)
      || ana_CheckProp(fol_SC0,fol_SC364,fol_SC0,fol_SC2153,fol_SC0,fol_SC282,fol_SC0,fol_SC2,fol_SC76,fol_SC84)
      || ana_CheckProp(fol_SC0,fol_SC340,fol_SC0,fol_SC2052,fol_SC0,fol_SC273,fol_SC0,fol_SC2,fol_SC70,fol_SC78)
      || ana_CheckProp(fol_SC0,fol_SC281,fol_SC0,fol_SC3612,fol_SC0,fol_SC1074,fol_SC0,fol_SC2,fol_SC89,fol_SC43)
      ) {
    flag_SetFlagValue(flag_SORTS,0);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC8,fol_SC8,fol_SC75,fol_SC47,fol_SC44,fol_SC10,fol_SC3,fol_SC3,fol_SC10,fol_SC4)
      || ana_CheckProp(fol_SC16,fol_SC1,fol_SC115,fol_SC5,fol_SC41,fol_SC0,fol_SC4,fol_SC3,fol_SC13,fol_SC4)
      || ana_CheckProp(fol_SC20,fol_SC5,fol_SC270,fol_SC23,fol_SC164,fol_SC0,fol_SC3,fol_SC3,fol_SC11,fol_SC5)
      ) {
    flag_SetFlagValue(flag_SORTS,0);
    flag_SetFlagValue(flag_SPLITS,0);
    flag_SetFlagValue(flag_SELECT,2);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC62,fol_SC1,fol_SC400,fol_SC14,fol_SC126,fol_SC9,fol_SC2,fol_SC2,fol_SC6,fol_SC11)
      || ana_CheckProp(fol_SC0,fol_SC55,fol_SC0,fol_SC968,fol_SC0,fol_SC348,fol_SC0,fol_SC3,fol_SC12,fol_SC5)
      || ana_CheckProp(fol_SC0,fol_SC316,fol_SC0,fol_SC1933,fol_SC0,fol_SC276,fol_SC0,fol_SC2,fol_SC64,fol_SC72)
      || ana_CheckProp(fol_SC22,fol_SC3,fol_SC284,fol_SC15,fol_SC172,fol_SC0,fol_SC3,fol_SC3,fol_SC9,fol_SC5)
      || ana_CheckProp(fol_SC24,fol_SC3,fol_SC312,fol_SC15,fol_SC184,fol_SC0,fol_SC4,fol_SC3,fol_SC9,fol_SC5)
      || ana_CheckProp(fol_SC7,fol_SC6,fol_SC73,fol_SC61,fol_SC30,fol_SC8,fol_SC4,fol_SC4,fol_SC7,fol_SC3)
      || ana_CheckProp(fol_SC4,fol_SC1,fol_SC42,fol_SC12,fol_SC20,fol_SC0,fol_SC5,fol_SC5,fol_SC5,fol_SC3)
      || ana_CheckProp(fol_SC13,fol_SC28,fol_SC158,fol_SC136,fol_SC97,fol_SC14,fol_SC3,fol_SC3,fol_SC14,fol_SC5)
      || ana_CheckProp(fol_SC15,fol_SC28,fol_SC188,fol_SC136,fol_SC117,fol_SC14,fol_SC3,fol_SC3,fol_SC14,fol_SC5)
      || ana_CheckProp(fol_SC11,fol_SC13,fol_SC132,fol_SC109,fol_SC85,fol_SC31,fol_SC3,fol_SC6,fol_SC10,fol_SC5)
      || ana_CheckProp(fol_SC57,fol_SC2,fol_SC331,fol_SC46,fol_SC87,fol_SC28,fol_SC2,fol_SC2,fol_SC6,fol_SC9)
      || ana_CheckProp(fol_SC9,fol_SC7,fol_SC88,fol_SC28,fol_SC50,fol_SC0,fol_SC3,fol_SC2,fol_SC11,fol_SC5)
      || ana_CheckProp(fol_SC18,fol_SC2,fol_SC128,fol_SC11,fol_SC42,fol_SC0,fol_SC4,fol_SC3,fol_SC14,fol_SC4)
      || ana_CheckProp(fol_SC18,fol_SC2,fol_SC128,fol_SC15,fol_SC42,fol_SC0,fol_SC4,fol_SC5,fol_SC14,fol_SC4)
      || ana_CheckProp(fol_SC19,fol_SC5,fol_SC214,fol_SC23,fol_SC127,fol_SC2,fol_SC3,fol_SC2,fol_SC7,fol_SC7)
      || ana_CheckProp(fol_SC10,fol_SC5,fol_SC123,fol_SC23,fol_SC75,fol_SC2,fol_SC3,fol_SC2,fol_SC7,fol_SC6)
      || ana_CheckProp(fol_SC5,fol_SC1,fol_SC53,fol_SC18,fol_SC21,fol_SC1,fol_SC4,fol_SC16,fol_SC2,fol_SC5)
      || ana_CheckProp(fol_SC12,fol_SC3,fol_SC111,fol_SC14,fol_SC56,fol_SC0,fol_SC5,fol_SC3,fol_SC10,fol_SC5)
      || ana_CheckProp(fol_SC32,fol_SC3,fol_SC295,fol_SC12,fol_SC161,fol_SC0,fol_SC3,fol_SC3,fol_SC8,fol_SC7)
      ) {
    flag_SetFlagValue(flag_SORTS,0);
    flag_SetFlagValue(flag_SELECT,2);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC20,fol_SC4,fol_SC355,fol_SC17,fol_SC234,fol_SC0,fol_SC3,fol_SC2,fol_SC13,fol_SC5)
      || ana_CheckProp(fol_SC1,fol_SC1,fol_SC20,fol_SC31,fol_SC8,fol_SC0,fol_SC11,fol_SC5,fol_SC9,fol_SC3)
      || ana_CheckProp(fol_SC5,fol_SC1,fol_SC43,fol_SC8,fol_SC28,fol_SC0,fol_SC4,fol_SC4,fol_SC4,fol_SC3)
      || ana_CheckProp(fol_SC100,fol_SC5,fol_SC1404,fol_SC24,fol_SC920,fol_SC0,fol_SC4,fol_SC2,fol_SC16,fol_SC6)
      || ana_CheckProp(fol_SC1,fol_SC1,fol_SC20,fol_SC31,fol_SC8,fol_SC0,fol_SC9,fol_SC5,fol_SC9,fol_SC3)
      || ana_CheckProp(fol_SC1,fol_SC1,fol_SC22,fol_SC31,fol_SC8,fol_SC0,fol_SC12,fol_SC5,fol_SC9,fol_SC3)
      || ana_CheckProp(fol_SC2,fol_SC1,fol_SC24,fol_SC31,fol_SC12,fol_SC0,fol_SC7,fol_SC5,fol_SC10,fol_SC3)
      || ana_CheckProp(fol_SC5,fol_SC11,fol_SC33,fol_SC126,fol_SC14,fol_SC4,fol_SC4,fol_SC5,fol_SC6,fol_SC3)
      || ana_CheckProp(fol_SC20,fol_SC1,fol_SC165,fol_SC8,fol_SC82,fol_SC0,fol_SC4,fol_SC4,fol_SC8,fol_SC3)
      || ana_CheckProp(fol_SC15,fol_SC5,fol_SC134,fol_SC30,fol_SC72,fol_SC0,fol_SC4,fol_SC4,fol_SC8,fol_SC3)
      || ana_CheckProp(fol_SC9,fol_SC1,fol_SC65,fol_SC16,fol_SC32,fol_SC0,fol_SC5,fol_SC5,fol_SC9,fol_SC3)
      || ana_CheckProp(fol_SC3,fol_SC1,fol_SC26,fol_SC16,fol_SC12,fol_SC0,fol_SC5,fol_SC5,fol_SC6,fol_SC3)
      || ana_CheckProp(fol_SC13,fol_SC1,fol_SC97,fol_SC8,fol_SC55,fol_SC0,fol_SC4,fol_SC3,fol_SC5,fol_SC3)
      || ana_CheckProp(fol_SC2,fol_SC0,fol_SC127,fol_SC0,fol_SC38,fol_SC0,fol_SC12,fol_SC0,fol_SC5,fol_SC3)
      || ana_CheckProp(fol_SC2,fol_SC1,fol_SC22,fol_SC13,fol_SC10,fol_SC0,fol_SC5,fol_SC5,fol_SC4,fol_SC4) 
      || ana_CheckProp(fol_SC2,fol_SC1,fol_SC22,fol_SC13,fol_SC10,fol_SC0,fol_SC6,fol_SC6,fol_SC4,fol_SC4)
      || ana_CheckProp(fol_SC16,fol_SC1,fol_SC146,fol_SC20,fol_SC70,fol_SC0,fol_SC5,fol_SC6,fol_SC10,fol_SC3)
      || ana_CheckProp(fol_SC158,fol_SC3,fol_SC1790,fol_SC11,fol_SC731,fol_SC0,fol_SC7,fol_SC3,fol_SC65,fol_SC19)
      || ana_CheckProp(fol_SC158,fol_SC4,fol_SC1790,fol_SC20,fol_SC731,fol_SC0,fol_SC7,fol_SC5,fol_SC65,fol_SC19)
      || ana_CheckProp(fol_SC158,fol_SC4,fol_SC1790,fol_SC21,fol_SC731,fol_SC0,fol_SC7,fol_SC5,fol_SC65,fol_SC19)
      || ana_CheckProp(fol_SC158,fol_SC2,fol_SC1790,fol_SC8,fol_SC731,fol_SC0,fol_SC7,fol_SC3,fol_SC63,fol_SC19)
      || ana_CheckProp(fol_SC158,fol_SC1,fol_SC1790,fol_SC4,fol_SC731,fol_SC0,fol_SC7,fol_SC2,fol_SC62,fol_SC19)
      || ana_CheckProp(fol_SC162,fol_SC4,fol_SC1858,fol_SC17,fol_SC904,fol_SC0,fol_SC5,fol_SC3,fol_SC62,fol_SC16)
      || ana_CheckProp(fol_SC140,fol_SC3,fol_SC1487,fol_SC29,fol_SC617,fol_SC0,fol_SC7,fol_SC7,fol_SC41,fol_SC12)
      || ana_CheckProp(fol_SC92,fol_SC4,fol_SC1016,fol_SC19,fol_SC417,fol_SC0,fol_SC7,fol_SC5,fol_SC42,fol_SC12)
      || ana_CheckProp(fol_SC15,fol_SC3,fol_SC134,fol_SC22,fol_SC72,fol_SC0,fol_SC4,fol_SC4,fol_SC9,fol_SC3)
      ) {
    flag_SetFlagValue(flag_BOUNDMODE,1);
    flag_SetFlagValue(flag_BOUNDSTART,5);
    flag_SetFlagValue(flag_BOUNDLOOPS,1000);
    flag_SetFlagValue(flag_SPLITS,0);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC16,fol_SC5,fol_SC162,fol_SC20,fol_SC96,fol_SC0,fol_SC3,fol_SC2,fol_SC12,fol_SC5)) {
    flag_SetFlagValue(flag_SATINPUT,0);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC16,fol_SC1,fol_SC115,fol_SC8,fol_SC41,fol_SC0,fol_SC4,fol_SC3,fol_SC12,fol_SC4)
      || ana_CheckProp(fol_SC16,fol_SC1,fol_SC118,fol_SC8,fol_SC42,fol_SC0,fol_SC4,fol_SC3,fol_SC12,fol_SC4)
      || ana_CheckProp(fol_SC16,fol_SC1,fol_SC118,fol_SC5,fol_SC42,fol_SC0,fol_SC4,fol_SC3,fol_SC13,fol_SC4)
      || ana_CheckProp(fol_SC127,fol_SC1,fol_SC1002,fol_SC100,fol_SC254,fol_SC0,fol_SC2,fol_SC2,fol_SC30,fol_SC16)
)
    {
      flag_SetFlagValue(flag_IURR,1);
      flag_SetFlagValue(flag_BOUNDMODE,1);
      flag_SetFlagValue(flag_BOUNDSTART,5);
      flag_SetFlagValue(flag_BOUNDLOOPS,1000);
      
      return;
    }
}

void ana_AddTPTPSetupFof()
{
  
  if (ana_CheckProp(fol_SC10,fol_SC1,fol_SC350,fol_SC102,fol_SC206,fol_SC63,fol_SC20,fol_SC21,fol_SC0,fol_SC11)
|| ana_CheckProp(fol_SC12,fol_SC1,fol_SC446,fol_SC119,fol_SC258,fol_SC72,fol_SC20,fol_SC23,fol_SC2,fol_SC11)
|| ana_CheckProp(fol_SC19,fol_SC1,fol_SC479,fol_SC33,fol_SC206,fol_SC12,fol_SC9,fol_SC9,fol_SC9,fol_SC14)
|| ana_CheckProp(fol_SC31,fol_SC1,fol_SC707,fol_SC48,fol_SC253,fol_SC20,fol_SC13,fol_SC13,fol_SC13,fol_SC8)
|| ana_CheckProp(fol_SC29,fol_SC1,fol_SC748,fol_SC97,fol_SC270,fol_SC39,fol_SC15,fol_SC21,fol_SC12,fol_SC7)
|| ana_CheckProp(fol_SC27,fol_SC1,fol_SC685,fol_SC75,fol_SC243,fol_SC29,fol_SC15,fol_SC17,fol_SC12,fol_SC7)
|| ana_CheckProp(fol_SC0,fol_SC1,fol_SC0,fol_SC4226,fol_SC0,fol_SC727,fol_SC0,fol_SC109,fol_SC159,fol_SC35)
|| ana_CheckProp(fol_SC0,fol_SC1,fol_SC0,fol_SC6930,fol_SC0,fol_SC1253,fol_SC0,fol_SC145,fol_SC253,fol_SC35)
|| ana_CheckProp(fol_SC0,fol_SC1,fol_SC0,fol_SC2598,fol_SC0,fol_SC372,fol_SC0,fol_SC131,fol_SC64,fol_SC72)
|| ana_CheckProp(fol_SC0,fol_SC1,fol_SC0,fol_SC2661,fol_SC0,fol_SC416,fol_SC0,fol_SC128,fol_SC59,fol_SC67)
|| ana_CheckProp(fol_SC0,fol_SC1,fol_SC0,fol_SC2977,fol_SC0,fol_SC647,fol_SC0,fol_SC78,fol_SC90,fol_SC20)
|| ana_CheckProp(fol_SC0,fol_SC1,fol_SC0,fol_SC2703,fol_SC0,fol_SC400,fol_SC0,fol_SC128,fol_SC65,fol_SC75)
) {
    flag_SetFlagValue(flag_SORTS,0);
    flag_SetFlagValue(flag_SELECT,2);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC19,fol_SC1,fol_SC469,fol_SC39,fol_SC166,fol_SC15,fol_SC16,fol_SC12,fol_SC7,fol_SC8)
|| ana_CheckProp(fol_SC19,fol_SC1,fol_SC458,fol_SC50,fol_SC162,fol_SC19,fol_SC13,fol_SC14,fol_SC7,fol_SC8)
|| ana_CheckProp(fol_SC30,fol_SC1,fol_SC745,fol_SC30,fol_SC273,fol_SC12,fol_SC16,fol_SC10,fol_SC10,fol_SC8)
) {
    flag_SetFlagValue(flag_BOUNDMODE,1);
    flag_SetFlagValue(flag_BOUNDSTART,5);
    flag_SetFlagValue(flag_BOUNDLOOPS,1000);
    flag_SetFlagValue(flag_SPLITS,0);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC0,fol_SC1,fol_SC0,fol_SC2639,fol_SC0,fol_SC348,fol_SC0,fol_SC140,fol_SC71,fol_SC79)) {
    flag_SetFlagValue(flag_FULLRED,0);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC26,fol_SC1,fol_SC576,fol_SC25,fol_SC202,fol_SC9,fol_SC13,fol_SC9,fol_SC10,fol_SC8)) {
    flag_SetFlagValue(flag_ORD,1);
    
    return;
  }
  
  if (ana_CheckProp(fol_SC26,fol_SC1,fol_SC579,fol_SC36,fol_SC202,fol_SC13,fol_SC13,fol_SC11,fol_SC10,fol_SC8)
|| ana_CheckProp(fol_SC26,fol_SC1,fol_SC580,fol_SC40,fol_SC203,fol_SC15,fol_SC13,fol_SC12,fol_SC10,fol_SC8)) {
    flag_SetFlagValue(flag_ISRE,1);
    flag_SetFlagValue(flag_IORE,0);
    flag_SetFlagValue(flag_PREFCON,2);
    flag_SetFlagValue(flag_SORTS,0);
    flag_SetFlagValue(flag_SPLITS,0);
    
    return;
  }
}

void ana_SetInputStatistic(LIST Axioms, LIST Conjectures, LIST Clauses)
{
  LIST   Scan,Symbols;
  TERM   Term;
  NAT    Help;
  CLAUSE Clause;

  ana_PROP4 = 0; 
  ana_PROP5   = 0;   
  ana_PROP6    = 0;    
  ana_PROP7      = 0;      
  ana_PROP8      = 0;      
  ana_PROP9        = 0;        
  ana_PROP10   = 0;   
  ana_PROP11     = 0;     
  ana_PROP12       = 0;       
  ana_PROP13       = 0;       

  if (!(list_Empty(Axioms) && list_Empty(Conjectures))) {

    ana_PROP14 = TRUE;

    for (Scan = Axioms;	!list_Empty(Scan); Scan = list_Cdr(Scan)) {
      Term = (TERM)list_PairSecond(list_Car(Scan));
	
      ana_PROP4 += term_ComputeSize(Term);
      ana_PROP6    += term_NumberOfVarOccs(Term);
      Help              = term_Depth(Term);

      if (Help > ana_PROP10)
	ana_PROP10 = Help;
    }

    for (Scan = Conjectures; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      Term = (TERM)list_PairSecond(list_Car(Scan));
	
      ana_PROP5 += term_ComputeSize(Term);
      ana_PROP7    += term_NumberOfVarOccs(Term);
      Help              = term_Depth(Term);

      if (Help > ana_PROP11)
	ana_PROP11 = Help;
    }
    ana_PROP8 = list_Length(Axioms);
    ana_PROP9   = list_Length(Conjectures);
  }
  else { 

    ana_PROP14 = FALSE;

    for (Scan = Clauses; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	
      Clause = (CLAUSE)list_Car(Scan);
      Help   = clause_ComputeTermDepth(Clause);
	
      if (clause_GetFlag(Clause, CONCLAUSE)) {
	
	ana_PROP5 += clause_ComputeSize(Clause);
	ana_PROP7    += clause_NumberOfVarOccs(Clause);
	ana_PROP9++;

	if (Help > ana_PROP11)
	  ana_PROP11 = Help;
      }
      else {

	ana_PROP4 += clause_ComputeSize(Clause);
	ana_PROP6    += clause_NumberOfVarOccs(Clause);

	if (Help > ana_PROP10)
	  ana_PROP10 = Help;

	ana_PROP8++;
      }
    }
  }
  Symbols = symbol_GetAllSymbols();
  for(Scan=Symbols;!list_Empty(Scan); Scan = list_Cdr(Scan)) {
    if (symbol_IsFunction((SYMBOL)list_Car(Scan)))
      ana_PROP13++;
    else {
      if (symbol_IsPredicate((SYMBOL)list_Car(Scan)))
	ana_PROP12++;
    }
  }
  list_Delete(Symbols);
}




void ana_AutoFof(void)
{
  if (ana_PROP14) {
    
    if (ana_CheckProp(fol_SC12,fol_SC1,fol_SC177,fol_SC18,fol_SC95,fol_SC10,fol_SC6,fol_SC5,fol_SC3,fol_SC5)
|| ana_CheckProp(fol_SC7,fol_SC1,fol_SC105,fol_SC13,fol_SC55,fol_SC7,fol_SC6,fol_SC5,fol_SC2,fol_SC5)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC15,fol_SC89,fol_SC8,fol_SC6,fol_SC6,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC11,fol_SC89,fol_SC6,fol_SC6,fol_SC4,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC18,fol_SC89,fol_SC10,fol_SC6,fol_SC5,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC23,fol_SC89,fol_SC12,fol_SC6,fol_SC6,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC41,fol_SC1,fol_SC649,fol_SC11,fol_SC269,fol_SC6,fol_SC8,fol_SC4,fol_SC26,fol_SC8)
|| ana_CheckProp(fol_SC41,fol_SC1,fol_SC649,fol_SC15,fol_SC269,fol_SC6,fol_SC8,fol_SC5,fol_SC26,fol_SC8)
|| ana_CheckProp(fol_SC41,fol_SC1,fol_SC649,fol_SC19,fol_SC269,fol_SC6,fol_SC8,fol_SC6,fol_SC26,fol_SC8)
|| ana_CheckProp(fol_SC41,fol_SC1,fol_SC649,fol_SC18,fol_SC269,fol_SC10,fol_SC8,fol_SC5,fol_SC26,fol_SC8)
|| ana_CheckProp(fol_SC6,fol_SC1,fol_SC86,fol_SC16,fol_SC45,fol_SC9,fol_SC6,fol_SC5,fol_SC1,fol_SC5)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC16,fol_SC89,fol_SC9,fol_SC6,fol_SC5,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC8,fol_SC1,fol_SC115,fol_SC20,fol_SC61,fol_SC11,fol_SC6,fol_SC6,fol_SC2,fol_SC5)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC26,fol_SC89,fol_SC14,fol_SC6,fol_SC6,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC8,fol_SC1,fol_SC115,fol_SC18,fol_SC61,fol_SC10,fol_SC6,fol_SC5,fol_SC2,fol_SC5)
|| ana_CheckProp(fol_SC8,fol_SC1,fol_SC115,fol_SC11,fol_SC61,fol_SC6,fol_SC6,fol_SC5,fol_SC2,fol_SC5)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC8,fol_SC89,fol_SC3,fol_SC6,fol_SC5,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC12,fol_SC89,fol_SC6,fol_SC6,fol_SC5,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC14,fol_SC89,fol_SC6,fol_SC6,fol_SC5,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC10,fol_SC1,fol_SC149,fol_SC26,fol_SC78,fol_SC13,fol_SC6,fol_SC7,fol_SC3,fol_SC5)
|| ana_CheckProp(fol_SC6,fol_SC1,fol_SC90,fol_SC22,fol_SC47,fol_SC12,fol_SC6,fol_SC6,fol_SC2,fol_SC5)
|| ana_CheckProp(fol_SC8,fol_SC1,fol_SC123,fol_SC15,fol_SC64,fol_SC8,fol_SC6,fol_SC6,fol_SC2,fol_SC5)
|| ana_CheckProp(fol_SC8,fol_SC1,fol_SC123,fol_SC13,fol_SC64,fol_SC7,fol_SC6,fol_SC5,fol_SC2,fol_SC5)
|| ana_CheckProp(fol_SC9,fol_SC1,fol_SC134,fol_SC13,fol_SC71,fol_SC7,fol_SC6,fol_SC5,fol_SC3,fol_SC5)
|| ana_CheckProp(fol_SC9,fol_SC1,fol_SC134,fol_SC18,fol_SC71,fol_SC10,fol_SC6,fol_SC5,fol_SC3,fol_SC5)
|| ana_CheckProp(fol_SC10,fol_SC1,fol_SC143,fol_SC15,fol_SC71,fol_SC7,fol_SC6,fol_SC5,fol_SC3,fol_SC6)
|| ana_CheckProp(fol_SC10,fol_SC1,fol_SC152,fol_SC19,fol_SC80,fol_SC10,fol_SC6,fol_SC5,fol_SC3,fol_SC5)
|| ana_CheckProp(fol_SC7,fol_SC1,fol_SC105,fol_SC16,fol_SC55,fol_SC9,fol_SC6,fol_SC5,fol_SC2,fol_SC5)
|| ana_CheckProp(fol_SC7,fol_SC1,fol_SC105,fol_SC18,fol_SC55,fol_SC10,fol_SC6,fol_SC5,fol_SC2,fol_SC5)
|| ana_CheckProp(fol_SC5,fol_SC1,fol_SC76,fol_SC18,fol_SC40,fol_SC10,fol_SC6,fol_SC5,fol_SC1,fol_SC5)
|| ana_CheckProp(fol_SC8,fol_SC1,fol_SC123,fol_SC16,fol_SC64,fol_SC9,fol_SC6,fol_SC5,fol_SC2,fol_SC5)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC20,fol_SC89,fol_SC11,fol_SC6,fol_SC6,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC13,fol_SC89,fol_SC7,fol_SC6,fol_SC5,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC24,fol_SC89,fol_SC13,fol_SC6,fol_SC6,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC15,fol_SC89,fol_SC7,fol_SC6,fol_SC6,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC24,fol_SC89,fol_SC12,fol_SC6,fol_SC7,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC24,fol_SC89,fol_SC13,fol_SC6,fol_SC7,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC26,fol_SC89,fol_SC14,fol_SC6,fol_SC7,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC29,fol_SC89,fol_SC16,fol_SC6,fol_SC7,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC11,fol_SC1,fol_SC175,fol_SC13,fol_SC89,fol_SC6,fol_SC6,fol_SC5,fol_SC9,fol_SC6)
|| ana_CheckProp(fol_SC28,fol_SC1,fol_SC809,fol_SC26,fol_SC464,fol_SC17,fol_SC13,fol_SC6,fol_SC15,fol_SC18)
|| ana_CheckProp(fol_SC28,fol_SC1,fol_SC809,fol_SC31,fol_SC464,fol_SC21,fol_SC13,fol_SC7,fol_SC15,fol_SC18)
|| ana_CheckProp(fol_SC28,fol_SC1,fol_SC809,fol_SC48,fol_SC464,fol_SC28,fol_SC13,fol_SC10,fol_SC15,fol_SC18)
|| ana_CheckProp(fol_SC28,fol_SC1,fol_SC809,fol_SC29,fol_SC464,fol_SC19,fol_SC13,fol_SC7,fol_SC15,fol_SC18)
|| ana_CheckProp(fol_SC28,fol_SC1,fol_SC809,fol_SC31,fol_SC464,fol_SC18,fol_SC13,fol_SC7,fol_SC15,fol_SC18)
|| ana_CheckProp(fol_SC28,fol_SC1,fol_SC809,fol_SC33,fol_SC464,fol_SC19,fol_SC13,fol_SC6,fol_SC15,fol_SC18)
|| ana_CheckProp(fol_SC28,fol_SC1,fol_SC809,fol_SC22,fol_SC464,fol_SC13,fol_SC13,fol_SC6,fol_SC15,fol_SC18)
|| ana_CheckProp(fol_SC28,fol_SC1,fol_SC809,fol_SC24,fol_SC464,fol_SC12,fol_SC13,fol_SC7,fol_SC15,fol_SC18)) {
      flag_SetFlagValue(flag_APPLYDEFS,10);
      
      return;
    }
  }
}
