




 





















#include "cnf.h"
#include "rules-inf.h"
#include "rules-red.h"


static TERM   cnf_AntiPrenex(TERM);
static TERM   cnf_AntiPrenexPath(TERM, TERM);
static TERM   cnf_ApplyDefinitionInternOnce(TERM, TERM, TERM, TERM, BOOL*);
static void   cnf_ApplyDistributivity(TERM term);

static TERM   cnf_Cnf(TERM term, LIST* symblist);
static BOOL   cnf_ContainsPredicateIntern(TERM, SYMBOL, int, TERM*, TERM*, LIST*, LIST*);

static TERM   cnf_DistributivFormula(TERM);
static void   cnf_DistrQuantorNoVarSub(TERM Term);
static void   cnf_DistrQuantorNoVarSubPath(TERM, TERM);

static TERM   cnf_FlattenPath(TERM, TERM);
static LIST   cnf_FreeVariablesBut(TERM Term, LIST Symbols);

static SYMBOL cnf_GetDualSymbol(SYMBOL symbol);
static int    cnf_GetFormulaPolarity(TERM term, TERM subterm);
static LIST   cnf_GetSymbolList(LIST varlist);
static LIST   cnf_GetUsedTerms(CLAUSE, PROOFSEARCH, HASH);

static BOOL   cnf_HasDeeperVariable(LIST, LIST);
static BOOL   cnf_HaveProofOptSkolem(PROOFSEARCH, TERM, char*, TERM, LIST*, LIST*, HASH);

static TERM   cnf_IsDefinition(TERM);

static LIST   cnf_MakeClauseList(TERM,BOOL,BOOL);
static TERM   cnf_MakeOneAnd(TERM term);
static TERM   cnf_MakeOneAndPredicate(TERM term);
static TERM   cnf_MakeOneAndTerm(TERM term);
static TERM   cnf_MakeOneOr(TERM term);
static TERM   cnf_MakeOneOrPredicate(TERM term);
static TERM   cnf_MakeOneOrTerm(TERM term);
static TERM   cnf_MakeSkolemfunction(LIST varlist);
static TERM   cnf_MoveEquivVariablesUp(TERM, TERM);
static TERM   cnf_MovePredicateVariablesUp(TERM, TERM, LIST);
static TERM   cnf_MoveProvedTermToTopLevel(TERM, TERM, TERM, LIST, LIST);

static TERM   cnf_NegationNormalFormulaPath(TERM, TERM);

static void   cnf_OptimizedSkolemFormula(PROOFSEARCH, TERM, char*, BOOL, TERM, LIST*, LIST*, BOOL, HASH, int);
static LIST   cnf_OptimizedSkolemization(PROOFSEARCH, TERM, char*, LIST*, LIST*, BOOL, BOOL, HASH);

static void   cnf_PopAllQuantifier(TERM term);
static int    cnf_PredicateOccurrences(TERM, SYMBOL);

static TERM   cnf_QuantifyAndNegate(TERM term, LIST VarList, LIST FreeList);
static TERM   cnf_QuantMakeOneVar(TERM term);

static TERM   cnf_RemoveImplFromFormulaPath(TERM, TERM);
static TERM   cnf_RemoveEquivImplFromFormula(TERM);
static TERM   cnf_RemoveTrivialOperators(TERM Term);
static void   cnf_RplacVar(TERM, LIST, LIST);
static void   cnf_RplacVarsymbFunction(TERM term, SYMBOL varsymb, TERM function);

static LIST   cnf_SatUnit(PROOFSEARCH, LIST); 
static TERM   cnf_SimplifyQuantors(TERM Term);
static TERM   cnf_SkolemFormula(TERM,LIST*);
static void   cnf_SkolemFunctionFormula(TERM, LIST, LIST, LIST*);
static void   cnf_Skolemize(TERM Term, LIST FreeList);
static void   cnf_StrongSkolemization(PROOFSEARCH, TERM, char*, BOOL, TERM, LIST*, LIST*, BOOL, HASH, int);
static LIST   cnf_SubsumeClauseList(LIST clauselist);

static BOOL   cnf_TopIsAnd(LIST termlist);

int* cnf_VARIABLEDEPTHARRAY; 
PROOFSEARCH SearchCopy;      

PROOFSEARCH cnf_HaveProof_PS; 
                              

void cnf_Init(void)

{
  
  if(flag_GetFlagValue(flag_CNFSTRSKOLEM))
    cnf_VARIABLEDEPTHARRAY = (int*) memory_Malloc(sizeof(int) * (symbol__MAXSTANDARDVAR + 1));
  else
    cnf_VARIABLEDEPTHARRAY = NULL;
  SearchCopy       = prfs_Create();
  cnf_HaveProof_PS = prfs_Create();
}

void cnf_Free(void)

{
  
  if(flag_GetFlagValue(flag_CNFSTRSKOLEM)) {
    memory_Free(cnf_VARIABLEDEPTHARRAY, sizeof(int) * (symbol__NOOFSTANDARDVAR + 1));
    cnf_VARIABLEDEPTHARRAY = NULL;
  }
  prfs_Delete(SearchCopy);
  SearchCopy = NULL;
  prfs_Delete(cnf_HaveProof_PS);
  cnf_HaveProof_PS = NULL;
}



static int cnf_GetFormulaPolarity(TERM term, TERM subterm)

{  
  LIST   scan;
  TERM   term1;
  int    polterm1,bottom; 

  bottom = vec_ActMax();
  vec_Push((POINTER) 1);
  vec_Push(term);

  do {
    term1    = (TERM)vec_PopResult();
    polterm1 = (int)vec_PopResult();
    if (term1 == subterm){
      vec_SetMax(bottom);
       return polterm1;
    }else
      if (symbol_Equal(term_TopSymbol(term1),fol_Not())){
	vec_Push((POINTER) (- polterm1));
	vec_Push(list_Car(term_ArgumentList(term1)));
      }
    if (symbol_Equal(term_TopSymbol(term1),fol_Exist()) ||
	symbol_Equal(term_TopSymbol(term1),fol_All())){
      vec_Push((POINTER)polterm1);
      vec_Push(list_Second(term_ArgumentList(term1)));
    }
    else
      if (symbol_Equal(term_TopSymbol(term1),fol_Implies())){
	vec_Push((POINTER) (- polterm1));
	vec_Push(list_Car(term_ArgumentList(term1)));
	vec_Push((POINTER) polterm1);
	vec_Push(list_Second(term_ArgumentList(term1)));
      }
    else
      if (symbol_Equal(term_TopSymbol(term1),fol_Equiv())){
	vec_Push(0);
	vec_Push(list_Car(term_ArgumentList(term1)));
	vec_Push(0);
	vec_Push(list_Second(term_ArgumentList(term1)));
      }
    else
      if ((symbol_Equal(term_TopSymbol(term1),fol_And())) || 
	  (symbol_Equal(term_TopSymbol(term1),fol_Or()))){
	for (scan = term_ArgumentList(term1);
	     !list_Empty(scan);
	     scan = list_Cdr(scan)) {
	  vec_Push((POINTER) polterm1);
	  vec_Push(list_Car(scan));
	}
      }
  } while (bottom != vec_ActMax());
  vec_SetMax(bottom);
  misc_StartErrorReport();
  misc_ErrorReport("\n In cnf_GetFormulaPolarity: Wrong arguments !\n");
  misc_FinishErrorReport();
  return(-2);
}

static BOOL cnf_ContainsDefinitionIntern(TERM TopDef, TERM Def, int Polarity, TERM* FoundPred)

{
  
  

  if((symbol_Equal(term_TopSymbol(Def),fol_And()) && (Polarity == 1))
     || (symbol_Equal(term_TopSymbol(Def),fol_Or()) && (Polarity == -1)))
    return FALSE;

  if(symbol_Equal(term_TopSymbol(Def), fol_And()) || symbol_Equal(term_TopSymbol(Def),fol_Or())) {
    
    LIST l;
    for(l=term_ArgumentList(Def); !list_Empty(l); l=list_Cdr(l))
      if(cnf_ContainsDefinitionIntern(TopDef, (TERM) list_Car(l), Polarity, FoundPred))
	return TRUE;
    return FALSE;
  }

  
  if(fol_IsQuantifier(term_TopSymbol(Def)))
    return cnf_ContainsDefinitionIntern(TopDef, term_SecondArgument(Def), Polarity, FoundPred);
  
  
  if(symbol_Equal(term_TopSymbol(Def),fol_Not()))
    return cnf_ContainsDefinitionIntern(TopDef, term_FirstArgument(Def), -Polarity, FoundPred);
  
  
  if(symbol_Equal(term_TopSymbol(Def),fol_Implies())) {
    if(Polarity==1) {
      if(cnf_ContainsDefinitionIntern(TopDef, term_FirstArgument(Def), -Polarity, FoundPred))
	return TRUE;
      return cnf_ContainsDefinitionIntern(TopDef, term_SecondArgument(Def), Polarity, FoundPred);
    }
    return FALSE;
  }
  
  
  if(symbol_Equal(term_TopSymbol(Def),fol_Equiv()) && (Polarity==1)) {
    
    TERM defpredicate;
    defpredicate = cnf_IsDefinition(Def);
    if(defpredicate != (TERM) NULL) {
      LIST predicate_vars, l, defpath;
      BOOL allquantifierfound;
      TERM super;
      int pol;

      
      if(cnf_PredicateOccurrences(TopDef, term_TopSymbol(defpredicate)) > 1) {}
      


      
      
      predicate_vars = term_ListOfVariables(defpredicate);
      list_DeleteDuplicates(predicate_vars, term_Equal);

      
      allquantifierfound = FALSE;
      
      
      
      defpath = list_Nil();
      super = Def;
      while(super != (TERM) NULL) {
	defpath = list_Cons(super, defpath);
	super = term_Superterm(super);
      }
      
      pol = 1;
      for(l=defpath; !list_Empty(l); l=list_Cdr(l)) {
	list_Rplaca(l, list_PairCreate((TERM) list_Car(l), (LIST) pol));
	if(symbol_Equal(term_TopSymbol((TERM) list_Car(l)), fol_Not()))
	  pol = -pol;
	else {
	  if(symbol_Equal(term_TopSymbol((TERM) list_Car(l)), fol_Implies())
	     && (term_FirstArgument((TERM) list_Car(l)) == (TERM) list_Car(list_Cdr(l))))
	    pol = -pol;
	  else 
	    if(symbol_Equal(term_TopSymbol((TERM) list_Car(l)), fol_Equiv()))
	      pol = 0;
	}
      }
      
      for(l=defpath; !list_Empty(l) && !list_Empty(predicate_vars); l=list_Cdr(l)) {
	LIST pair;
	TERM t;
	int p;
	
	pair = (LIST) list_Car(l);
	t = (TERM) list_PairFirst(pair);
	p = (int) list_PairSecond(pair);

	if(fol_IsQuantifier(term_TopSymbol(t))) {
	  
	  
	  if((symbol_Equal(term_TopSymbol(t), fol_All()) && (p==1))
	     || (symbol_Equal(term_TopSymbol(t), fol_Exist()) && (p==-1))) {
	    LIST scan;
	    allquantifierfound = TRUE;
	    for(scan=fol_QuantifierVariables(t); !list_Empty(scan); scan=list_Cdr(scan))
	      predicate_vars = list_DeleteElement(predicate_vars, (TERM) list_Car(scan), term_Equal);
	  }
	  else {
	    
	    
	    if(allquantifierfound == TRUE) {
	      list_Delete(predicate_vars);
	      list_DeleteWithElement(defpath, list_Free);
	      return FALSE;
	    }
	    else {
	      LIST scan;
	    
	      for(scan=fol_QuantifierVariables(t); !list_Empty(scan); scan=list_Cdr(scan)) {
		if(list_Member(predicate_vars, (TERM) list_Car(scan), term_Equal)) {
		  list_Delete(predicate_vars);
		  list_DeleteWithElement(defpath, list_Free);
		  return FALSE;
		}
	      }
	    }
	  }
	}
      }
      
#ifdef DEBUG
      if(!list_Empty(predicate_vars)) {
	list_Delete(predicate_vars);
	misc_StartErrorReport();
	misc_ErrorReport("\n In cnf_ContainsDefinitionIntern: Definition has free variables.\n");
	misc_FinishErrorReport();
      }
#endif
      list_DeleteWithElement(defpath, list_Free);
      *FoundPred = defpredicate;
      return TRUE;
    }
  }
  
  return FALSE;
}


BOOL cnf_ContainsDefinition(TERM Def, TERM* FoundPred)

{
  BOOL result;
#ifdef DEBUG
  fol_CheckFatherLinks(Def);
#endif
  result = cnf_ContainsDefinitionIntern(Def, Def, 1, FoundPred);
#ifdef DEBUG
  fol_CheckFatherLinks(Def);
#endif
  return result;
}


static TERM cnf_IsDefinition(TERM Def)

{
  LIST l,freevars, predicatevars;

#ifdef DEBUG
  if(Def == NULL) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In cnf_IsDefinition: Empty formula.\n");
    misc_FinishErrorReport();
  }
  if(!symbol_Equal(term_TopSymbol(Def),fol_Equiv())) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In cnf_IsDefinition: Formula is no equivalence.\n");
    misc_FinishErrorReport();
  }
#endif
  
  
  if((!symbol_IsPredicate(term_TopSymbol(term_FirstArgument(Def))))) {
    TERM arg1;
    arg1 = term_FirstArgument(Def);
    term_RplacFirstArgument(Def, term_SecondArgument(Def));
    term_RplacSecondArgument(Def, arg1);
  }


  
  if (!symbol_IsPredicate(term_TopSymbol(term_FirstArgument(Def)))) 
    return NULL;

  
  


  
  
  freevars = fol_FreeVariables(term_SecondArgument(Def));
  freevars = list_DeleteDuplicates(freevars, term_Equal);
  predicatevars = term_ListOfVariables(term_FirstArgument(Def));
  predicatevars = list_DeleteDuplicates(predicatevars, term_Equal);

  for(l=predicatevars; !list_Empty(l); l=list_Cdr(l)) {
    freevars = list_DeleteElement(freevars, list_Car(l), term_Equal);
  }

  if (!list_Empty(freevars)) {
    list_Delete(freevars);
    list_Delete(predicatevars);
    return NULL;
  }

  list_Delete(predicatevars);
  return term_FirstArgument(Def);
}


static BOOL cnf_ContainsPredicateIntern(TERM Target, SYMBOL Predicate, 
					int Polarity, TERM* TargetPredicate, TERM* ToTopLevel,
					LIST* TargetVars,
					LIST* VarsForTopLevel)

{
  
  

  if((symbol_Equal(term_TopSymbol(Target),fol_And()) && (Polarity == 1)) 
     || (symbol_Equal(term_TopSymbol(Target),fol_Or()) && (Polarity == -1))
     || (symbol_Equal(term_TopSymbol(Target),fol_Implies()) && (Polarity != 1))
     || (symbol_Equal(term_TopSymbol(Target), fol_Equiv()))) {
    TERM s;
    LIST l;
    
    s = term_FindSubterm(Target, Predicate);
    if(s == NULL)
      return FALSE;
    
    
    for(l=term_ArgumentList(s); !list_Empty(l); l = list_Cdr(l))
      *TargetVars = list_Nconc(fol_FreeVariables((TERM) list_Car(l)), *TargetVars);
    *TargetVars = list_DeleteDuplicates(*TargetVars, term_Equal);
    
    *TargetPredicate = s;
    *ToTopLevel = Target;
    return TRUE;
  }
  
  
  if(symbol_Equal(term_TopSymbol(Target),fol_And()) || symbol_Equal(term_TopSymbol(Target),fol_Or())) {
    
    LIST l;
    for(l=term_ArgumentList(Target); !list_Empty(l); l = list_Cdr(l))
      if(cnf_ContainsPredicateIntern((TERM) list_Car(l), Predicate, Polarity, TargetPredicate, ToTopLevel,
				     TargetVars, VarsForTopLevel))
	return TRUE;
    return FALSE;
  }

  
  if(fol_IsQuantifier(term_TopSymbol(Target))) {
    if(cnf_ContainsPredicateIntern(term_SecondArgument(Target), Predicate, Polarity, TargetPredicate, ToTopLevel,
				   TargetVars, VarsForTopLevel)) {
      
      if((symbol_Equal(term_TopSymbol(Target), fol_All()) && (Polarity == 1))
	 || (symbol_Equal(term_TopSymbol(Target), fol_Exist()) && (Polarity == -1))) {
	LIST l;
	
	for(l=*TargetVars; !list_Empty(l); l=list_Cdr(l)) {
	  if(list_Member(fol_QuantifierVariables(Target), (TERM) list_Car(l), term_Equal))
	    *VarsForTopLevel = list_Cons(list_Car(l), *VarsForTopLevel);
	}
      }
      return TRUE;
    }
    return FALSE;
  }

  
  if(symbol_Equal(term_TopSymbol(Target),fol_Not()))
    return cnf_ContainsPredicateIntern(term_FirstArgument(Target), Predicate, -Polarity,
				       TargetPredicate, ToTopLevel, TargetVars, VarsForTopLevel);
  
  if(symbol_Equal(term_TopSymbol(Target),fol_Implies())) {
    
    if(cnf_ContainsPredicateIntern(term_FirstArgument(Target), Predicate, -Polarity,
				   TargetPredicate, ToTopLevel, TargetVars, VarsForTopLevel))
      return TRUE;
    return cnf_ContainsPredicateIntern(term_SecondArgument(Target), Predicate, Polarity,
				       TargetPredicate, ToTopLevel, TargetVars, VarsForTopLevel);
  }

  
  if(symbol_Equal(term_TopSymbol(Target), Predicate)) {
    LIST l;
    for(l=term_ArgumentList(Target); !list_Empty(l); l = list_Cdr(l))
      *TargetVars = list_Nconc(fol_FreeVariables((TERM) list_Car(l)), *TargetVars);
    *TargetVars = list_DeleteDuplicates(*TargetVars, term_Equal);
    
    *TargetPredicate = Target;
    *ToTopLevel      = Target;
    return TRUE;
  }
  
  
  return FALSE;
}
  

  
BOOL cnf_ContainsPredicate(TERM Target, SYMBOL Predicate, TERM* TargetPredicate, TERM* ToTopLevel,
				  LIST* TargetVars, LIST* VarsForTopLevel)

{
  BOOL result;
#ifdef DEBUG
  fol_CheckFatherLinks(Target);
#endif
  result = cnf_ContainsPredicateIntern(Target, Predicate, 1, TargetPredicate, ToTopLevel,
				       TargetVars, VarsForTopLevel);
#ifdef DEBUG
  fol_CheckFatherLinks(Target);
#endif
  return result;
}


static int cnf_PredicateOccurrences(TERM Term, SYMBOL P)
			       
{
  
  if(fol_IsQuantifier(term_TopSymbol(Term)))
    return cnf_PredicateOccurrences(term_SecondArgument(Term), P);

  
  if(fol_IsJunctor(term_TopSymbol(Term))
     || symbol_Equal(term_TopSymbol(Term),fol_Not())){
    LIST scan;
    int count;
    count = 0;
    for(scan=term_ArgumentList(Term); !list_Empty(scan); scan=list_Cdr(scan)) {
      count += cnf_PredicateOccurrences((TERM) list_Car(scan), P);
      
      if (count > 1)
	return count;
    }
    return count;
  }
  
  if(symbol_Equal(term_TopSymbol(Term), P))
    return 1;
  return 0;
} 
  
TERM cnf_NegationNormalFormulaPath(TERM Term, TERM PredicateTerm)

{
  TERM   subterm, termL, term1;
  LIST   scan;
  SYMBOL symbol;
  BOOL   set;

  term1 = Term;
  while (term1 != NULL) {
    set = FALSE;
    if (symbol_Equal(term_TopSymbol(term1),fol_Not())) {	    
      subterm = (TERM)list_Car(term_ArgumentList(term1));
      if (symbol_Equal(term_TopSymbol(subterm),fol_Not())) {
	LIST l;
	term_RplacTop(term1,term_TopSymbol(term_FirstArgument(subterm)));
	list_Delete(term_ArgumentList(term1));
	term_RplacArgumentList(term1,term_ArgumentList(term_FirstArgument(subterm)));
	term_Free(term_FirstArgument(subterm));
	list_Delete(term_ArgumentList(subterm));
	term_Free(subterm);
	
	for(l=term_ArgumentList(term1); !list_Empty(l); l=list_Cdr(l))
	  term_RplacSuperterm((TERM) list_Car(l), term1);
	
	set = TRUE;
      }
      else {
	if (fol_IsQuantifier(term_TopSymbol(subterm))) {
	  LIST l;
	  symbol = (SYMBOL)cnf_GetDualSymbol(term_TopSymbol(subterm));
	  termL  = term_CreateAddFather(fol_Not(),
			       list_List(term_SecondArgument(subterm)));
	  list_RplacSecond(term_ArgumentList(subterm), termL);
	  term_RplacSuperterm(termL, subterm);
	  term_RplacTop(term1,symbol);
	  list_Delete(term_ArgumentList(term1));
	  term_RplacArgumentList(term1, term_ArgumentList(subterm));
	  for(l=term_ArgumentList(term1); !list_Empty(l); l = list_Cdr(l))
	    term_RplacSuperterm((TERM) list_Car(l), term1);
	  term_RplacArgumentList(subterm, list_Nil());
	  term_Delete(subterm);
	  term1 = termL;
	  
	  set = TRUE;
	} 
	else {
	  if (symbol_Equal(term_TopSymbol(subterm),fol_Or()) || 
	      (symbol_Equal(term_TopSymbol(subterm),fol_And()))) {
	    LIST l;
	    symbol = (SYMBOL)cnf_GetDualSymbol(term_TopSymbol(subterm));
	    for (scan = term_ArgumentList(subterm); !list_Empty(scan);
		 scan = list_Cdr(scan)) {
	      TERM new;
	      termL = (TERM)list_Car(scan);
	      new = term_CreateAddFather(fol_Not(),list_List(termL));
	      list_Rplaca(scan, new);
	      term_RplacSuperterm(new, subterm);
	    }
	    term_RplacTop(term1,symbol);
	    list_Delete(term_ArgumentList(term1));
	    term_RplacArgumentList(term1,term_ArgumentList(subterm));
	    for(l=term_ArgumentList(term1); !list_Empty(l); l=list_Cdr(l))
	      term_RplacSuperterm((TERM) list_Car(l), term1);
	    term_RplacArgumentList(subterm, list_Nil());
	    term_Delete(subterm);
	  }
	}
      }
    }
    if(set == FALSE) {
      if (!list_Empty(term_ArgumentList(term1)))
	for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	  if(term_HasProperSuperterm(PredicateTerm, (TERM) list_Car(scan))) {
	    term1 = (TERM) list_Car(scan);
	    set = TRUE;
	    break;
	  }
      if(set == FALSE)
	term1 = NULL;
    }
  }
  return Term;
}

TERM cnf_ApplyDefinitionOnce(TERM Predicate, TERM Formula, TERM TargetTerm,
			     TERM TargetPredicate)

{
  SYMBOL maxvar, maxvar_temp;
  LIST bound, scan;
  BOOL success;
  
  
  maxvar = term_MaxVar(TargetTerm);
  maxvar_temp = term_MaxVar(Formula);
  if(maxvar_temp > maxvar)
    maxvar = maxvar_temp;
  symbol_SetStandardVarCounter(maxvar);

  
  bound = fol_BoundVariables(Formula);
  for(scan=bound;!list_Empty(scan);scan=list_Cdr(scan)) {
    
    if(term_ContainsSymbol(TargetTerm, term_TopSymbol((TERM) list_Car(scan)))) 
      term_ExchangeVariable(Formula, term_TopSymbol((TERM) list_Car(scan)),
			    symbol_CreateStandardVariable());
  }
  list_Delete(bound);
  TargetTerm = cnf_ApplyDefinitionInternOnce(Predicate, Formula, TargetTerm, 
				       TargetPredicate,&success);
  if(flag_GetFlagValue(flag_PAPPLYDEFS)) {
    if(success == TRUE) {
      fputs("\nTarget after applying def:\n", stdout);
      fol_PrettyPrint(TargetTerm);
      printf("\n\n");
    }
  }
  
  return TargetTerm;
}

static TERM cnf_ApplyDefinitionInternOnce(TERM Predicate, TERM Formula, 
					  TERM TargetTerm, TERM TargetPredicate,
					  BOOL* Success)

{
   
  if(fol_IsQuantifier(term_TopSymbol(TargetTerm))) {
    term_RplacSecondArgument(TargetTerm, 
			     cnf_ApplyDefinitionInternOnce(Predicate, 
							   Formula, 
							   term_SecondArgument(TargetTerm),
							   TargetPredicate,
							   Success));
    term_RplacSuperterm(term_SecondArgument(TargetTerm), TargetTerm);
    return TargetTerm;
  }

  
  if(fol_IsJunctor(term_TopSymbol(TargetTerm))
     || symbol_Equal(term_TopSymbol(TargetTerm),fol_Not())){
    LIST scan;
    for(scan=term_ArgumentList(TargetTerm); !list_Empty(scan); scan=list_Cdr(scan)) {
      list_Rplaca(scan, cnf_ApplyDefinitionInternOnce(Predicate, Formula, 
						      (TERM) list_Car(scan), 
						      TargetPredicate,
						      Success));
      term_RplacSuperterm((TERM) list_Car(scan), TargetTerm);
    }
    return TargetTerm;
  }
  
  if(symbol_Equal(term_TopSymbol(TargetTerm), term_TopSymbol(Predicate))) {
    if(TargetTerm == TargetPredicate) {
      TERM result;
      result = Formula;
      cnf_RplacVar(result, term_ArgumentList(Predicate), 
		   term_ArgumentList(TargetTerm));
      term_AddFatherLinks(result);
      term_Delete(TargetTerm);
      *Success = TRUE;
      return result;
    }
  }
  
  return TargetTerm;
} 

TERM cnf_RemoveEquivImplFromFormula(TERM term)

{
  TERM term1,termL,termR,termLneg,termRneg;
  LIST scan;
  int  bottom,pol;
  
  bottom = vec_ActMax();
  vec_Push(term);

  while (!(bottom == vec_ActMax())){
    term1 = (TERM)vec_PopResult();
    if (symbol_Equal(term_TopSymbol(term1),fol_Implies())){
      term_RplacTop(term1, fol_Or());
      list_Rplaca(term_ArgumentList(term1),
        term_Create(fol_Not(), list_List(list_Car(term_ArgumentList(term1)))));
    }else
      if (symbol_Equal(term_TopSymbol(term1),fol_Equiv())){
	pol      = cnf_GetFormulaPolarity(term,term1);
	termL    = (TERM)list_Car(term_ArgumentList(term1));
	termR    = (TERM)list_Second(term_ArgumentList(term1));
	termLneg = term_Create(fol_Not(),list_List(term_Copy(termL)));
	termRneg = term_Create(fol_Not(),list_List(term_Copy(termR)));
	if ((pol == 1)|| (pol == 0)){
	  term_RplacTop(term1, fol_And());
	  list_Rplaca(term_ArgumentList(term1), term_Create(fol_Or(),list_Cons(termLneg,list_List(termR))));
	  list_RplacSecond(term_ArgumentList(term1), term_Create(fol_Or(),list_Cons(termRneg,list_List(termL))));
	}else
	  if(pol == -1){
	    term_RplacTop(term1, fol_Or());
	    list_Rplaca(term_ArgumentList(term1), term_Create(fol_And(),list_Cons(termLneg,list_List(termRneg))));
	    list_RplacSecond(term_ArgumentList(term1), term_Create(fol_And(),list_Cons(termL,list_List(termR))));
	  }
      }  
    if (!list_Empty(term_ArgumentList(term1)))
      for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	vec_Push(list_Car(scan));
  }
  vec_SetMax(bottom);
  return term;
}

static TERM cnf_MoveEquivVariablesUp(TERM term, TERM PredicateTerm)

{
  TERM equiv, term1;
  LIST freevars, scan;
  int bottom;

  equiv = term_Superterm(PredicateTerm);
  freevars = fol_FreeVariables(equiv);

  bottom = vec_ActMax();
  vec_Push(term);
  
  while (!(bottom == vec_ActMax())) {
    term1 = (TERM)vec_PopResult();
    if (!list_Empty(term_ArgumentList(term1)))
      for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan)) {
	TERM arg;
	arg = (TERM) list_Car(scan);
	if(term_HasProperSuperterm(PredicateTerm, arg)) {
	  if(symbol_Equal(term_TopSymbol(arg), fol_All())) {
	    LIST freevarscan, quantifiervars;
	    quantifiervars = fol_QuantifierVariables(arg);
	    for(freevarscan=freevars; !list_Empty(freevarscan);
		freevarscan = list_Cdr(freevarscan))
	      quantifiervars = list_DeleteElementFree(quantifiervars, 
						      (TERM) list_Car(freevarscan),
						      term_Equal, term_Delete);
	    if(!list_Empty(quantifiervars))
	      term_RplacArgumentList(term_FirstArgument(arg), quantifiervars);
	    else {
	      TERM subterm;
	      subterm = term_SecondArgument(arg);
	      term_Free(term_FirstArgument(arg));
	      list_Delete(term_ArgumentList(arg));
	      term_Free(arg);
	      list_Rplaca(scan, subterm);
	      term_RplacSuperterm(subterm, term1);
	    }
	  }
	  vec_Push((TERM) list_Car(scan));
	}
      }
  }
  
  
  if(!list_Empty(freevars)) {
    for(scan=freevars; !list_Empty(scan); scan = list_Cdr(scan))
      list_Rplaca(scan, term_Copy((TERM) list_Car(scan)));
    if (symbol_Equal(term_TopSymbol(term), fol_All())) {
      LIST vars;
      vars = fol_QuantifierVariables(term);
      vars = list_Nconc(vars, freevars);
      vars = list_DeleteDuplicatesFree(vars, term_Equal, term_Delete);
      term_RplacArgumentList(term_FirstArgument(term), vars);
    }
    else {
      TERM newtop;
      newtop = fol_CreateQuantifier(fol_All(), freevars, list_List(term));
      term_RplacSuperterm(term, newtop);
      term = newtop;
    }
  }
  vec_SetMax(bottom);
  return term;
}
 
static TERM cnf_MovePredicateVariablesUp(TERM Term, TERM TargetPredicateTerm, LIST VarsForTopLevel)

{
  TERM term1;
  LIST scan;
  int  bottom;

  bottom = vec_ActMax();
  vec_Push(Term);
  
  while (!(bottom == vec_ActMax())) {
    term1 = (TERM)vec_PopResult();
    if (!list_Empty(term_ArgumentList(term1)))
      for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan)) {
	TERM arg;
	arg = (TERM) list_Car(scan);
	if(term_HasProperSuperterm(TargetPredicateTerm, arg)) {
	  if(symbol_Equal(term_TopSymbol(arg), fol_All())) {
	    LIST predicatevarscan, quantifiervars;
	    quantifiervars = fol_QuantifierVariables(arg);
	    for(predicatevarscan=VarsForTopLevel; !list_Empty(predicatevarscan);
		predicatevarscan = list_Cdr(predicatevarscan))
	      quantifiervars = list_DeleteElementFree(quantifiervars, 
						  (TERM) list_Car(predicatevarscan),
						  term_Equal, term_Delete);
	    if(!list_Empty(quantifiervars))
	      term_RplacArgumentList(term_FirstArgument(arg), quantifiervars);
	    else {
	      TERM subterm;
	      subterm = term_SecondArgument(arg);
	      term_Free(term_FirstArgument(arg));
	      list_Delete(term_ArgumentList(arg));
	      term_Free(arg);
	      list_Rplaca(scan, subterm);
	      term_RplacSuperterm(subterm, term1);
	    }
	  }
	  vec_Push((TERM) list_Car(scan));
	}
      }
  }
  
  for(scan=VarsForTopLevel; !list_Empty(scan); scan = list_Cdr(scan))
    list_Rplaca(scan, term_Copy((TERM) list_Car(scan)));
  if (symbol_Equal(term_TopSymbol(Term), fol_All())) {
    LIST vars;
    vars = fol_QuantifierVariables(Term);
    vars = list_Nconc(vars, list_Copy(VarsForTopLevel));
    vars = list_DeleteDuplicatesFree(vars, term_Equal, term_Delete);
    term_RplacArgumentList(term_FirstArgument(Term), vars);
  }
  else {
    TERM newtop;
    newtop = fol_CreateQuantifier(fol_All(), list_Copy(VarsForTopLevel), list_List(Term));
    term_RplacSuperterm(Term, newtop);
    Term = newtop;
  }
  vec_SetMax(bottom);
  return Term;
}

TERM cnf_RemoveImplFromFormulaPath(TERM Term, TERM PredicateTerm)

{
  TERM term1;
  LIST scan;
  int  bottom;
  
  bottom = vec_ActMax();
  vec_Push(Term);
  
  while (!(bottom == vec_ActMax())) {
    term1 = (TERM)vec_PopResult();
    if(term_HasProperSuperterm(PredicateTerm, term1)) {
      if (symbol_Equal(term_TopSymbol(term1),fol_Implies())) {
	TERM newterm;
	term_RplacTop(term1, fol_Or());
	newterm = term_CreateAddFather(fol_Not(), list_List(list_Car(term_ArgumentList(term1))));
	list_Rplaca(term_ArgumentList(term1), newterm);
	term_RplacSuperterm(newterm, term1);
      }
      
      if (!list_Empty(term_ArgumentList(term1)))
	for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	  vec_Push(list_Car(scan));
    }
  }
  vec_SetMax(bottom);
  return Term;
}


static SYMBOL cnf_GetDualSymbol(SYMBOL symbol)

{
  SYMBOL dual;

  dual = symbol;
  if (symbol_Equal(symbol,fol_All()))
    dual = fol_Exist();
  else
    if (symbol_Equal(symbol,fol_Exist()))
      dual = fol_All();
    else
      if (symbol_Equal(symbol,fol_Or()))
	dual = fol_And();
      else
	if (symbol_Equal(symbol,fol_And()))
	  dual =  fol_Or();
  return dual;
}


TERM cnf_NegationNormalFormula(TERM term)

{
  TERM   term1,subterm,termL;
  LIST   scan;
  SYMBOL symbol;
  int    bottom;
  
  bottom = vec_ActMax();
  vec_Push(term);

  while (!(bottom == vec_ActMax())){
    term1 = (TERM)vec_PopResult();
    if (symbol_Equal(term_TopSymbol(term1),fol_Not())){	    
      subterm = (TERM)list_Car(term_ArgumentList(term1));
      if (symbol_Equal(term_TopSymbol(subterm),fol_Not())){
	term_RplacTop(term1,term_TopSymbol(term_FirstArgument(subterm)));
	list_Delete(term_ArgumentList(term1));
	term_RplacArgumentList(term1,term_ArgumentList(term_FirstArgument(subterm)));
	term_Free(term_FirstArgument(subterm));
	list_Delete(term_ArgumentList(subterm));
	term_Free(subterm);
	vec_Push(term1);
      }else 
	if (fol_IsQuantifier(term_TopSymbol(subterm))){
	  symbol = (SYMBOL)cnf_GetDualSymbol(term_TopSymbol(subterm));
	  termL  = term_Create(fol_Not(),
			       list_List(term_SecondArgument(subterm)));
	  list_RplacSecond(term_ArgumentList(subterm), termL);
	  term_RplacTop(term1,symbol);
	  list_Delete(term_ArgumentList(term1));
	  term_RplacArgumentList(term1,list_CopyWithElement(
              term_ArgumentList(subterm),(POINTER (*)())term_Copy));
	  term_Delete(subterm);
	} else
	  if (symbol_Equal(term_TopSymbol(subterm),fol_Or()) || 
	      (symbol_Equal(term_TopSymbol(subterm),fol_And()))){
	    symbol = (SYMBOL)cnf_GetDualSymbol(term_TopSymbol(subterm));
	    for (scan = term_ArgumentList(subterm);
		 !list_Empty(scan);
		 scan = list_Cdr(scan)) {
	      termL = (TERM)list_Car(scan);
	      list_Rplaca(scan, term_Create(fol_Not(),list_List(termL)));
	    }
	    term_RplacTop(term1,symbol);
	    list_Delete(term_ArgumentList(term1));
	    term_RplacArgumentList(term1,list_CopyWithElement(
                term_ArgumentList(subterm), (POINTER (*)())term_Copy));
	    term_Delete(subterm); 
	  }
    }
    if (!list_Empty(term_ArgumentList(term1)))
      for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	vec_Push(list_Car(scan));
  }
  vec_SetMax(bottom);
  return term;
}


static TERM cnf_QuantMakeOneVar(TERM term)

{
  TERM   term1,termL;
  SYMBOL quantor;
  LIST   scan,varlist;
  int    bottom;
  
  bottom = vec_ActMax();
  vec_Push(term);       

  while (bottom != vec_ActMax()) {
    term1 = (TERM)vec_PopResult();
    if (fol_IsQuantifier(term_TopSymbol(term1))){
      quantor = term_TopSymbol(term1);
      if (list_Length(term_ArgumentList(term_FirstArgument(term1))) > 1) {
	varlist =
	  list_Copy(list_Cdr(term_ArgumentList(term_FirstArgument(term1))));
	for (scan=varlist;!list_Empty(scan);scan=list_Cdr(scan)){
	  termL = term_SecondArgument(term1);
	  term_RplacSecondArgument(term1, fol_CreateQuantifier(quantor,list_List(list_Car(scan)),list_List(termL)));
	}
	for (scan=varlist;!list_Empty(scan);scan=list_Cdr(scan)) {
	  term_RplacArgumentList(term_FirstArgument(term1), 
				 list_PointerDeleteElement(term_ArgumentList(term_FirstArgument(term1)),list_Car(scan)));
	}
	list_Delete(varlist);
      }
    }
    if (!list_Empty(term_ArgumentList(term1)))
      for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	vec_Push(list_Car(scan));
  }
  vec_SetMax(bottom);
  return term;
}


static LIST cnf_GetSymbolList(LIST varlist)

{
  LIST scan,result;

  result = list_Nil();
  for (scan=varlist;!list_Empty(scan);scan=list_Cdr(scan))
    result = list_Cons((POINTER)term_TopSymbol((TERM)list_Car(scan)),result);

  return result;
}


static BOOL cnf_TopIsAnd(LIST termlist)

{
  LIST scan;

  for (scan=termlist;!list_Empty(scan);scan=list_Cdr(scan))
    if (term_TopSymbol(list_Car(scan)) == fol_And())
      return TRUE;
  return FALSE;
}


static TERM cnf_MakeOneOr(TERM term)

{

  LIST   scan;

  if (symbol_Equal(term_TopSymbol(term),fol_Or())){
    TERM   argterm;
    scan=term_ArgumentList(term);
    while (!list_Empty(scan)) {
      argterm = (TERM)list_Car(scan);
      cnf_MakeOneOr(argterm);
      if (symbol_Equal(term_TopSymbol(argterm),fol_Or())) {
	scan = list_Cdr(scan);
	term_RplacArgumentList(term,
	  list_Nconc(term_ArgumentList(argterm),list_PointerDeleteElement(term_ArgumentList(term),argterm)));
	term_Free(argterm);
      }
      else
	scan = list_Cdr(scan);
    }
  } else if (!symbol_IsPredicate(term_TopSymbol(term)))
    for (scan=term_ArgumentList(term);!list_Empty(scan);scan=list_Cdr(scan))
      cnf_MakeOneOr(list_Car(scan));
  
  return term;
}


static TERM cnf_MakeOneOrPredicate(TERM term)

{

  LIST   scan,scan1,predlist;

  if (cnf_TopIsAnd(term_ArgumentList(term))){
    
    for (scan1=term_ArgumentList(term);
	 !(list_Empty(scan1) || symbol_Equal(term_TopSymbol(list_Car(scan1)),fol_And()));
	 scan1=list_Cdr(scan1));

    if (!(list_Empty(scan1))){
      predlist = list_Nil();
      for (scan=scan1;!list_Empty(scan);scan=list_Cdr(scan)){
	if ((symbol_IsPredicate(term_TopSymbol(list_Car(scan)))) ||
	    (fol_IsNegativeLiteral(list_Car(scan)))){
	  predlist = list_Cons(list_Car(scan),predlist);
	}
      }
      for (scan=predlist;!list_Empty(scan);scan=list_Cdr(scan))
	term_RplacArgumentList(term,
	  list_PointerDeleteElement(term_ArgumentList(term),list_Car(scan)));
      
      if (!list_Empty(predlist))
	term_RplacArgumentList(term, list_Nconc(predlist,term_ArgumentList(term)));
    }
  }
  return term;
}


static TERM cnf_MakeOneOrTerm(TERM term)

{
  return cnf_MakeOneOrPredicate(cnf_MakeOneOr(term));
}


static TERM cnf_MakeOneAnd(TERM term)

{
  LIST   scan;

  if (symbol_Equal(term_TopSymbol(term),fol_And())) {
    TERM   argterm;
    scan=term_ArgumentList(term);
    while (!list_Empty(scan)) {
      argterm = (TERM)list_Car(scan);
      cnf_MakeOneAnd(argterm);
      if (symbol_Equal(term_TopSymbol(argterm),fol_And())) {
	scan = list_Cdr(scan);
	term_RplacArgumentList(term,
	  list_Nconc(term_ArgumentList(argterm),list_PointerDeleteElement(term_ArgumentList(term),argterm)));
	term_Free(argterm);
      }
      else
	scan = list_Cdr(scan);
    }
  } else if (!symbol_IsPredicate(term_TopSymbol(term)))
    for (scan=term_ArgumentList(term);!list_Empty(scan);scan=list_Cdr(scan))
      cnf_MakeOneAnd(list_Car(scan));

  return term;
}


static TERM cnf_MakeOneAndPredicate(TERM term)

{
  LIST   scan,scan1,predlist;
  
  for (scan1=term_ArgumentList(term);
       !(list_Empty(scan1) || symbol_Equal(term_TopSymbol(list_Car(scan1)),fol_Or()));
       scan1=list_Cdr(scan1));
  
  if (!list_Empty(scan1)) {
    
    predlist = list_Nil();
    for (scan=scan1; !list_Empty(scan); scan=list_Cdr(scan)){
      if (symbol_IsPredicate(term_TopSymbol(list_Car(scan))) &&
	  fol_IsNegativeLiteral(list_Car(scan))) {
	predlist = list_Cons(list_Car(scan),predlist);
      }
    }
    for (scan=predlist; !list_Empty(scan); scan=list_Cdr(scan))
      term_RplacArgumentList(term, list_PointerDeleteElement(term_ArgumentList(term),list_Car(scan)));
    
    if (!list_Empty(predlist))
      term_RplacArgumentList(term, list_Nconc(predlist,term_ArgumentList(term)));
  }
  return term;
}


static TERM cnf_MakeOneAndTerm(TERM term)

{
  return cnf_MakeOneAndPredicate(cnf_MakeOneAnd(term));
}


static void cnf_ApplyDistributivity(TERM term)

{
  if (symbol_Equal(term_TopSymbol(term),fol_Or())) {
    TERM andterm;
    LIST Scan;

    for (Scan=term_ArgumentList(term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      cnf_ApplyDistributivity(list_Car(Scan));

    andterm = (TERM)NULL;

    for (Scan=term_ArgumentList(term);!(list_Empty(Scan) || andterm);Scan=list_Cdr(Scan))
      if (symbol_Equal(term_TopSymbol(list_Car(Scan)),fol_And()))
	andterm = (TERM)list_Car(Scan);

    if (andterm) {
      term_RplacArgumentList(term, list_PointerDeleteElement(term_ArgumentList(term),andterm));
      for (Scan=term_ArgumentList(andterm);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	list_Rplaca(Scan, 
	  term_Create(fol_Or(), 
	    list_Cons(list_Car(Scan),list_CopyWithElement(term_ArgumentList(term),
							    (POINTER (*)())term_Copy))));
      }
      term_RplacTop(term, fol_And());
      list_DeleteWithElement(term_ArgumentList(term),term_Delete);
      term_RplacArgumentList(term, term_ArgumentList(andterm));
      term_Free(andterm);
      cnf_ApplyDistributivity(term);
    }
  }
  else
    if (!symbol_Equal(term_TopSymbol(term),fol_Not()) &&
	!symbol_IsPredicate(term_TopSymbol(term))) {
      LIST Scan;
      for (Scan=term_ArgumentList(term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	cnf_ApplyDistributivity(list_Car(Scan));
    }
}


TERM cnf_DistributivFormula(TERM term)

{
  cnf_ApplyDistributivity(term);
  cnf_MakeOneAnd(term);
  cnf_MakeOneOr(term);

  return term;
}


void cnf_FPrintClause(TERM term, FILE* file)

{
    
 TERM   term1;
 LIST   scan;
 int    bottom;
  
  bottom = vec_ActMax();
  vec_Push(term);       

  while (!(bottom == vec_ActMax())){
    term1 = (TERM)vec_PopResult();
    if (symbol_Equal(term_TopSymbol(term1),fol_Or())){
      for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	vec_Push(list_Car(scan));
    }else
      term_FPrint(file,term1);
  }
 fprintf(file,".\n");
 vec_SetMax(bottom);
}


void cnf_FPrint(TERM term, FILE* file)

{
  TERM   term1;
  LIST   scan;
  int    bottom;
  
  bottom = vec_ActMax();
  vec_Push(term);       

  while (!(bottom == vec_ActMax())){
    term1 = (TERM)vec_PopResult();
    if (symbol_Equal(term_TopSymbol(term1),fol_And())){
      for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	vec_Push(list_Car(scan));
    }else
      if ((symbol_Equal(term_TopSymbol(term1),fol_Or())) ||
	  (symbol_IsPredicate(term_TopSymbol(term1)))  ||
	  (symbol_Equal(term_TopSymbol(term1),fol_Not())))
	cnf_FPrintClause(term1,file);
  }
  vec_SetMax(bottom);
}


void cnf_StdoutPrint(TERM term)

{
  LIST termlist,scan,scan1;

  for (scan=term_ArgumentList(term);!list_Empty(scan);scan=list_Cdr(scan)){
    termlist = list_Nil();
    if (!((symbol_IsPredicate(term_TopSymbol(list_Car(scan)))) ||
	  (fol_IsNegativeLiteral(list_Car(scan)))))
      termlist = term_ArgumentList(list_Car(scan));
    
    if (!list_Empty(termlist)){
      for (scan1=termlist;!list_Empty(scan1);scan1=list_Cdr(scan1))
	term_Print(list_Car(scan1));
      printf(".\n");
    }else{
      term_Print(list_Car(scan));
      printf(".\n");
    }
  }
}


void cnf_FilePrint(TERM term, FILE* file)

{
  LIST termlist,scan,scan1;

  for (scan=term_ArgumentList(term);!list_Empty(scan);scan=list_Cdr(scan)){
    termlist = list_Nil();
    if (!((symbol_IsPredicate(term_TopSymbol(list_Car(scan)))) ||
	  (fol_IsNegativeLiteral(list_Car(scan)))))
      termlist = term_ArgumentList(list_Car(scan));
    
    if (!list_Empty(termlist)){
      for (scan1=termlist;!list_Empty(scan1);scan1=list_Cdr(scan1))
	term_FPrint(file,list_Car(scan1));
      fprintf(file,".\n");
    }else{
      term_FPrint(file,list_Car(scan));
      fprintf(file,".\n");
    }
  }
  
}


void cnf_FilePrintPrefix(TERM term, FILE* file)

{
  LIST termlist,scan,scan1;

  for (scan=term_ArgumentList(term);!list_Empty(scan);scan=list_Cdr(scan)){
    termlist = list_Nil();
    if (!((symbol_IsPredicate(term_TopSymbol(list_Car(scan)))) ||
	  (fol_IsNegativeLiteral(list_Car(scan)))))
      termlist = term_ArgumentList(list_Car(scan));
    
    if (!list_Empty(termlist)){
      for (scan1=termlist;!list_Empty(scan1);scan1=list_Cdr(scan1)) {
	term_FPrintPrefix(file,list_Car(scan1));
	if (!list_Empty(list_Cdr(scan1)))
	  fprintf(file," | ");
      }
      fprintf(file,".\n");
    } else {
      term_FPrintPrefix(file,list_Car(scan));
      fprintf(file,".\n");
    }
  }
}



static LIST cnf_SubsumeClauseList(LIST clauselist)

{
  LIST     scan,result;
  st_INDEX stindex;
  CLAUSE   actclause;

  stindex = st_IndexCreate();
  result  = list_Nil();

  for (scan=clauselist;!list_Empty(scan);scan=list_Cdr(scan))
    res_InsertClauseIndex(list_Car(scan),stindex);

  for (scan=clauselist;!list_Empty(scan);scan=list_Cdr(scan)){
    actclause     = list_Car(scan);
    res_DeleteClauseIndex(actclause,stindex);
    if (!(res_BackSubWithLength(actclause,stindex))){
      res_InsertClauseIndex(actclause,stindex);
      result = list_Cons(actclause,result);
    }
  }
  if (!(list_Length(result) == list_Length(clauselist))){
    for (scan=result;!list_Empty(scan);scan=list_Cdr(scan))
      clauselist = list_PointerDeleteElement(clauselist,list_Car(scan));
    if (!list_Empty(result))
      list_DeleteWithElement(clauselist,clause_Delete);
  }else
    list_Delete(clauselist);
  st_IndexDelete(stindex);
  return result;
}


LIST cnf_MakeClauseList(TERM term, BOOL Sorts, BOOL Conclause)

{
  LIST   termlist,scan,clauselist,newclauselist,delclauselist,condlist;
  CLAUSE clause;
  int    j;

  termlist    = list_Nil();
  clauselist  = list_Nil();

  if (fol_IsTrue(term))
    return  clauselist;

  if ((fol_IsNegativeLiteral(term))||
      (symbol_IsPredicate(term_TopSymbol(term)))){
    termlist = list_List(term_Copy(term));
    clause = clause_CreateFromLiterals(termlist,Sorts,Conclause,TRUE);
    clauselist = list_Nconc(clauselist,list_List(clause));
    term_StartMinRenaming();
    term_Rename(clause_GetLiteralTerm(clause,0));
    list_Delete(termlist);
    return clauselist;
  }

  delclauselist = list_Nil();
  term          = cnf_MakeOneAndTerm(term);
  if (symbol_Equal(term_TopSymbol(term), fol_And())){
    for (scan=term_ArgumentList(term);!list_Empty(scan);scan=list_Cdr(scan)) {
      list_Rplaca(scan, cnf_MakeOneOrTerm(list_Car(scan)));
      termlist = list_Nil();
      if (!(symbol_IsPredicate(term_TopSymbol(list_Car(scan))) ||
	    fol_IsNegativeLiteral(list_Car(scan)))) {
	termlist = list_CopyWithElement(term_ArgumentList(list_Car(scan)),
					(POINTER (*)())term_Copy);
	termlist = list_DeleteDuplicatesFree(termlist,term_Equal,term_Delete);
      }else
	termlist = list_List(term_Copy(list_Car(scan)));
  
      if (!list_Empty(termlist)){
	clause     = clause_CreateFromLiterals(termlist,Sorts,Conclause,TRUE);
	term_StartMinRenaming();
	for (j=0;j<clause_Length(clause);j++)
	  term_Rename(clause_GetLiteralTerm(clause,j));
	clauselist = list_Cons(clause,clauselist);
	list_Delete(termlist);
      }
    }
  } else {
    
    term = cnf_MakeOneOrTerm(term);
    if (!((symbol_IsPredicate(term_TopSymbol(term))) ||
	  (fol_IsNegativeLiteral(term)))){
      termlist = list_CopyWithElement(term_ArgumentList(term),
				      (POINTER (*)())term_Copy);
      termlist = list_DeleteDuplicatesFree(termlist,term_Equal,term_Delete);
    } else
      termlist = list_List(term_Copy(term));
    
    if (!list_Empty(termlist)){
      clause     = clause_CreateFromLiterals(termlist,Sorts,Conclause,TRUE);
      term_StartMinRenaming();
      for (j = 0; j < clause_Length(clause); j++)
	term_Rename(clause_GetLiteralTerm(clause,j));
      clauselist = list_Cons(clause,clauselist);
      list_Delete(termlist);
    }
  }

  for (scan=clauselist; !list_Empty(scan); scan=list_Cdr(scan)){
    condlist = cond_CondFast(list_Car(scan));
    if (!list_Empty(condlist))
      clause_DeleteLiterals(list_Car(scan), condlist);
    list_Delete(condlist);
  }
  clauselist = cnf_SubsumeClauseList(clauselist);
  newclauselist = list_Nil();
  while (!list_Empty(clauselist)){
    clause = res_SelectLightestClause(clauselist);
    newclauselist = list_Nconc(newclauselist,list_List(clause));
    clauselist = list_PointerDeleteElement(clauselist,clause);
  }
  list_Delete(clauselist);
  for (scan=newclauselist;!list_Empty(scan);scan=list_Cdr(scan)){
    if (res_HasTautology(list_Car(scan)))
      delclauselist = list_Cons(list_Car(scan),delclauselist);
  }
  for (scan=delclauselist;!list_Empty(scan);scan=list_Cdr(scan))
    newclauselist = list_PointerDeleteElement(newclauselist,list_Car(scan));
  list_DeleteWithElement(delclauselist,clause_Delete);
    
  return newclauselist;
}

TERM cnf_Flatten(TERM Term, SYMBOL Symbol)

{
  LIST   Scan1,Scan2;

  if (symbol_Equal(term_TopSymbol(Term), Symbol)) {
    TERM   Argterm;
    Scan1 =term_ArgumentList(Term);
    while (!list_Empty(Scan1)) {
      Argterm = (TERM)list_Car(Scan1);
      Scan2   = list_Cdr(Scan1);
      if (symbol_Equal(term_TopSymbol(Argterm),Symbol)) {
	cnf_Flatten(Argterm,Symbol);
	list_NInsert(Scan1,list_Cdr(term_ArgumentList(Argterm)));   
	list_Rplaca(Scan1,list_Car(term_ArgumentList(Argterm)));
	list_Free(term_ArgumentList(Argterm));
	term_Free(Argterm);
      }
      Scan1  = Scan2;
    }
  }
  return Term;
}

TERM cnf_FlattenPath(TERM Term, TERM PredicateTerm)

{
  TERM   subterm;
  
  subterm = Term;
  while (symbol_Equal(term_TopSymbol(subterm), fol_All()))
    subterm = term_SecondArgument(subterm);

  if (symbol_Equal(term_TopSymbol(subterm), fol_Or())) {
    TERM Argterm;
    LIST scan1;
    scan1 = term_ArgumentList(subterm);
    while (!list_Empty(scan1)) {
      LIST scan2;
      Argterm = (TERM)list_Car(scan1);
      scan2   = list_Cdr(scan1);
      if(term_HasProperSuperterm(PredicateTerm, Argterm)) {
	if (symbol_Equal(term_TopSymbol(Argterm),fol_Or())) {
	  LIST l;
	  cnf_Flatten(Argterm,fol_Or());
	  for(l=term_ArgumentList(Argterm); !list_Empty(l); l=list_Cdr(l))
	    term_RplacSuperterm((TERM) list_Car(l), subterm);
	  list_NInsert(scan1,list_Cdr(term_ArgumentList(Argterm)));   
	  list_Rplaca(scan1,list_Car(term_ArgumentList(Argterm)));
	  list_Free(term_ArgumentList(Argterm));
	  term_Free(Argterm);
	}
      }
      scan1  = scan2;
    }
  }
  return Term;
}

static void cnf_DistrQuantorNoVarSub(TERM Term)

{
  LIST   Variables,Subformulas,Scan1,Scan2,Rest;
  TERM   Subterm,Var,NewForm;
  SYMBOL Subtop,Top;

  Top       = term_TopSymbol(Term);
  Variables = list_Copy(fol_QuantifierVariables(Term));
  Subterm   = term_SecondArgument(Term);
  Subtop    = term_TopSymbol(Subterm);
  Subterm   = cnf_Flatten(Subterm,Subtop);

  for (Scan1=Variables;!list_Empty(Scan1);Scan1=list_Cdr(Scan1)) {
    Subformulas = list_Nil();
    Var         = (TERM)list_Car(Scan1);
    for(Scan2=term_ArgumentList(Subterm);!list_Empty(Scan2);Scan2=list_Cdr(Scan2))
      if (!fol_VarOccursFreely(Var,list_Car(Scan2)))
	Subformulas = list_Cons(list_Car(Scan2),Subformulas);
    if (!list_Empty(Subformulas)) {
      Rest     = list_NPointerDifference(term_ArgumentList(Subterm),Subformulas);
      if (list_Empty(list_Cdr(Rest))) { 
	if (symbol_Equal(Top,term_TopSymbol(list_Car(Rest)))) { 
	  NewForm = (TERM)list_Car(Rest);
	  term_RplacArgumentList(term_FirstArgument(NewForm),
				 list_Cons((POINTER)Var,fol_QuantifierVariables(NewForm)));
	  list_Delete(Rest);
	}
	else
	  NewForm = fol_CreateQuantifier(Top,list_List((POINTER)Var),Rest);
      }
      else
	NewForm = fol_CreateQuantifier(Top,list_List((POINTER)Var),list_List(term_Create(Subtop,Rest)));
      term_RplacArgumentList(Subterm,list_Cons(NewForm,Subformulas));
      term_RplacArgumentList(term_FirstArgument(Term),
			     list_PointerDeleteElement(fol_QuantifierVariables(Term),(POINTER)Var));
    } 
  }

  if (list_Empty(fol_QuantifierVariables(Term))) { 
    term_Free(term_FirstArgument(Term));
    list_Delete(term_ArgumentList(Term));
    term_RplacTop(Term,Subtop);
    term_RplacArgumentList(Term,term_ArgumentList(Subterm));
    term_Free(Subterm);
  }
  list_Delete(Variables);
}

static TERM cnf_AntiPrenex(TERM Term)

{
  LIST   Scan;
  SYMBOL Top;

  Top = term_TopSymbol(Term);

  if (fol_IsQuantifier(Top)) {
    TERM   Subterm,Actterm;
    SYMBOL DistrSymb,Subtop;          
    
    Subterm = term_SecondArgument(Term);
    Subtop  = term_TopSymbol(Subterm);

    if (!symbol_IsPredicate(Subtop) && 
	!symbol_Equal(Subtop,fol_Not())) { 
      if (symbol_Equal(Top,fol_All()))
	DistrSymb = fol_And();
      else
	DistrSymb = fol_Or();
      if (fol_IsQuantifier(Subtop)) {
	cnf_AntiPrenex(Subterm);
	Subtop = term_TopSymbol(Subterm);
      }
      if (symbol_Equal(Subtop,DistrSymb)) {
	LIST Variables;	
	LIST NewVars;
	Variables = fol_QuantifierVariables(Term);
	Subterm   = cnf_Flatten(Subterm,DistrSymb);
	for (Scan=term_ArgumentList(Subterm);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	  Actterm = (TERM)list_Car(Scan);
	  NewVars = list_NIntersect(fol_FreeVariables(Actterm),Variables,term_Equal);
	  if (!list_Empty(NewVars)) {
	    if (symbol_Equal(Top,term_TopSymbol(Actterm))) {        
	      list_NMapCar(NewVars,(POINTER (*)())term_Copy);
	      term_RplacArgumentList(term_FirstArgument(Actterm),
				     list_Nconc(fol_QuantifierVariables(Actterm),
						NewVars));
	    }
	    else {
	      list_NMapCar(NewVars,(POINTER (*)())term_Copy);
	      list_Rplaca(Scan,fol_CreateQuantifier(Top, NewVars, list_List(Actterm)));
	    }
	  }
	}
	term_Delete(term_FirstArgument(Term));   
	list_Delete(term_ArgumentList(Term)); 
	term_RplacTop(Term,DistrSymb);
	term_RplacArgumentList(Term,term_ArgumentList(Subterm));
	term_Free(Subterm);
	for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan)) 
	  list_Rplaca(Scan,cnf_AntiPrenex(list_Car(Scan)));
      }
      else 
	if (!fol_IsQuantifier(Subtop)) {
	  cnf_DistrQuantorNoVarSub(Term);
	  for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	    cnf_AntiPrenex(list_Car(Scan));
	}
    }
  }
  else
    if (!symbol_Equal(Top,fol_Not()) && !symbol_IsPredicate(Top))
      for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	cnf_AntiPrenex(list_Car(Scan));

  return Term;
}

static void cnf_DistrQuantorNoVarSubPath(TERM Term, TERM PredicateTerm)

{
  LIST   Variables,Subformulas,Scan1,Scan2,Rest;
  TERM   Subterm,Var,NewForm;
  SYMBOL Subtop,Top;

  

  Top       = term_TopSymbol(Term);
  Variables = list_Copy(fol_QuantifierVariables(Term));
  Subterm   = term_SecondArgument(Term);
  Subtop    = term_TopSymbol(Subterm);
  Subterm   = cnf_Flatten(Subterm,Subtop);

  

  for (Scan1=Variables;!list_Empty(Scan1);Scan1=list_Cdr(Scan1)) {
    Subformulas = list_Nil();
    Var         = (TERM)list_Car(Scan1);
    
    for(Scan2=term_ArgumentList(Subterm);!list_Empty(Scan2);Scan2=list_Cdr(Scan2))
      if (!fol_VarOccursFreely(Var,list_Car(Scan2)))
	Subformulas = list_Cons(list_Car(Scan2),Subformulas);
    if (!list_Empty(Subformulas)) {
      
      Rest     = list_NPointerDifference(term_ArgumentList(Subterm),Subformulas);
      if (list_Empty(list_Cdr(Rest))) { 
	if (symbol_Equal(Top,term_TopSymbol(list_Car(Rest)))) { 
	  NewForm = (TERM)list_Car(Rest);
	  
	  term_RplacArgumentList(term_FirstArgument(NewForm),
				 list_Cons((POINTER)Var,fol_QuantifierVariables(NewForm)));
	  list_Delete(Rest);
	}
	else {
	  NewForm = fol_CreateQuantifierAddFather(Top,list_List((POINTER)Var),
						  Rest);
	}
      }
      else {
	TERM t;
	t = term_CreateAddFather(Subtop,Rest);
	NewForm = fol_CreateQuantifierAddFather(Top,list_List((POINTER)Var),list_List(t));
      }
      if(term_HasProperSuperterm(PredicateTerm, NewForm))
	NewForm = cnf_AntiPrenexPath(NewForm, PredicateTerm);
      term_RplacArgumentList(Subterm,list_Cons(NewForm,Subformulas));
      term_RplacSuperterm(NewForm, Subterm);
      term_RplacArgumentList(term_FirstArgument(Term),
			     list_PointerDeleteElement(fol_QuantifierVariables(Term),(POINTER)Var));
    } 
  }

  if (list_Empty(fol_QuantifierVariables(Term))) { 
    LIST l;
    term_Free(term_FirstArgument(Term));
    list_Delete(term_ArgumentList(Term));
    term_RplacTop(Term,Subtop);
    term_RplacArgumentList(Term,term_ArgumentList(Subterm));
    term_Free(Subterm);
    for(l=term_ArgumentList(Term); !list_Empty(l); l=list_Cdr(l))
      term_RplacSuperterm((TERM) list_Car(l), Term);
  }
  list_Delete(Variables);
}

static TERM cnf_AntiPrenexPath(TERM Term, TERM PredicateTerm)

{
  LIST   Scan;
  SYMBOL Top;

  Top = term_TopSymbol(Term);

  if(symbol_Equal(Top, fol_Exist())) {
    TERM   Subterm,Actterm;
    SYMBOL DistrSymb,Subtop;          
    
    Subterm = term_SecondArgument(Term);
    Subtop  = term_TopSymbol(Subterm);

    if (!symbol_Equal(Subtop,fol_Not()) && !symbol_IsPredicate(Subtop)) { 
      if (symbol_Equal(Top,fol_All()))
	DistrSymb = fol_And();
      else
	DistrSymb = fol_Or();
      if (fol_IsQuantifier(Subtop)) {
	cnf_AntiPrenexPath(Subterm, PredicateTerm);
	Subtop = term_TopSymbol(Subterm);
      }
      if (symbol_Equal(Subtop,DistrSymb)) {
	LIST Variables;	
	LIST NewVars;
	LIST l;
	Variables = fol_QuantifierVariables(Term);
	Subterm   = cnf_Flatten(Subterm,DistrSymb);
	for(l=term_ArgumentList(Subterm); !list_Empty(l); l=list_Cdr(l))
	  term_RplacSuperterm((TERM) list_Car(l), Subterm);

	for (Scan=term_ArgumentList(Subterm);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	  Actterm = (TERM)list_Car(Scan);
	  NewVars = list_NIntersect(fol_FreeVariables(Actterm),Variables,term_Equal);
	  if (!list_Empty(NewVars)) {
	    if (symbol_Equal(Top,term_TopSymbol(Actterm))) {         
	      list_NMapCar(NewVars,(POINTER (*)())term_Copy);
	      term_RplacArgumentList(term_FirstArgument(Actterm),
				     list_Nconc(fol_QuantifierVariables(Actterm), NewVars));
	    }
	    else {
	      TERM newquantifier;
	      list_NMapCar(NewVars,(POINTER (*)())term_Copy);
	      newquantifier = fol_CreateQuantifierAddFather(Top, NewVars, list_List(Actterm));
	      list_Rplaca(Scan,newquantifier);
	      term_RplacSuperterm(newquantifier, Subterm);
	    }
	  }
	}
	term_Delete(term_FirstArgument(Term));   
	list_Delete(term_ArgumentList(Term)); 
	term_RplacTop(Term,DistrSymb);
	term_RplacArgumentList(Term,term_ArgumentList(Subterm));
	term_Free(Subterm);
	for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	  term_RplacSuperterm((TERM) list_Car(Scan), Term);
	  if(term_HasProperSuperterm(PredicateTerm, (TERM) list_Car(Scan))) {
	    list_Rplaca(Scan,cnf_AntiPrenexPath(list_Car(Scan), PredicateTerm));
	    term_RplacSuperterm((TERM) list_Car(Scan), Term);
	  }
	}
      }
      else 
	if (!fol_IsQuantifier(Subtop))
	  cnf_DistrQuantorNoVarSubPath(Term, PredicateTerm);
    }
  }
  else
    if (!symbol_Equal(Top,fol_Not()) && !symbol_IsPredicate(Top))
      for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	if(term_HasProperSuperterm(PredicateTerm, (TERM) list_Car(Scan)))
	  cnf_AntiPrenexPath(list_Car(Scan), PredicateTerm);
  
  return Term;
}

static TERM  cnf_RemoveTrivialOperators(TERM Term)

{
  SYMBOL Top;
  LIST   Scan;

  Top = term_TopSymbol(Term);

  if (symbol_IsPredicate(Top))
    return Term;

  if ((symbol_Equal(Top,fol_And()) || symbol_Equal(Top,fol_Or())) &&
      list_Empty(list_Cdr(term_ArgumentList(Term)))) {
    TERM Result;
    Result = term_FirstArgument(Term);
    term_RplacSuperterm(Result, term_Superterm(Term));
    list_Delete(term_ArgumentList(Term));
    term_Free(Term);
    return cnf_RemoveTrivialOperators(Result);
  }

  for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    list_Rplaca(Scan,cnf_RemoveTrivialOperators(list_Car(Scan)));
    term_RplacSuperterm((TERM) list_Car(Scan), Term);
  }
  
  return Term;
}

static TERM  cnf_SimplifyQuantors(TERM Term)

{
   SYMBOL Top;
   LIST   Scan;

  Top = term_TopSymbol(Term);

  if (symbol_IsPredicate(Top) || symbol_Equal(Top,fol_Varlist()))
    return Term;

  if (fol_IsQuantifier(Top)) {
    LIST Vars;
    TERM Var,Subterm,Aux;
    Vars    = list_Nil();
    Subterm = term_SecondArgument(Term);

    while (symbol_Equal(term_TopSymbol(Subterm),Top)) {
      term_RplacArgumentList(term_FirstArgument(Term),
			     list_Nconc(fol_QuantifierVariables(Term),fol_QuantifierVariables(Subterm)));
      term_Free(term_FirstArgument(Subterm));
      Aux = term_SecondArgument(Subterm);
      list_Delete(term_ArgumentList(Subterm));
      term_Free(Subterm);
      list_Rplaca(list_Cdr(term_ArgumentList(Term)),Aux);
      Subterm = Aux;
    }
    
    for(Scan=fol_QuantifierVariables(Term);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Var = (TERM)list_Car(Scan);
      if (!fol_VarOccursFreely(Var,Subterm))
	Vars = list_Cons(Var,Vars);
    }
    if (!list_Empty(Vars)) {
      Subterm = term_FirstArgument(Term);
      term_RplacArgumentList(Subterm,list_NPointerDifference(term_ArgumentList(Subterm),Vars));
      list_DeleteWithElement(Vars,term_Delete);
      if (list_Empty(term_ArgumentList(Subterm))) {
	Subterm = term_SecondArgument(Term);
	term_Delete(term_FirstArgument(Term));
	list_Delete(term_ArgumentList(Term));
	term_Free(Term);
	return cnf_SimplifyQuantors(Subterm);
      }
    }
  }
  
  for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
    list_Rplaca(Scan,cnf_SimplifyQuantors(list_Car(Scan)));
  
  return Term; 
}

TERM  cnf_RemoveTrivialAtoms(TERM Term)

{
  SYMBOL Top,Subtop;
  LIST   Scan;
  TERM   Result;
  BOOL   Update;

  if (!term_IsComplex(Term))
    return Term;
  
  Top    = term_TopSymbol(Term);
  Update = FALSE;

  if (symbol_Equal(Top,fol_And())) {
    Scan  = term_ArgumentList(Term);
    while (!list_Empty(Scan)) {
      Result = cnf_RemoveTrivialAtoms(list_Car(Scan));
      Subtop = term_TopSymbol(Result);
      if (symbol_Equal(Subtop,fol_True()))
	Update = TRUE;
      else
	if (symbol_Equal(Subtop,fol_False())) {
	  term_RplacTop(Term,fol_False());
	  list_DeleteWithElement(term_ArgumentList(Term),term_Delete);
	  term_RplacArgumentList(Term,list_Nil());
	  return Term;
	}
      Scan = list_Cdr(Scan);
    }
    if (Update) {
      term_RplacArgumentList(Term,list_DeleteElementIfFree(term_ArgumentList(Term),fol_IsTrue,term_Delete));
      if (list_Empty(term_ArgumentList(Term))) {
	term_RplacTop(Term,fol_True());
	return Term;
      }
    }
  }
  else if (symbol_Equal(Top,fol_Or())) {
    Scan  = term_ArgumentList(Term);
    while (!list_Empty(Scan)) {
      Result = cnf_RemoveTrivialAtoms(list_Car(Scan));
      Subtop = term_TopSymbol(Result);
      if (symbol_Equal(Subtop,fol_False()))
	Update = TRUE;
      else
	if (symbol_Equal(Subtop,fol_True())) {
	  term_RplacTop(Term,fol_True());
	  list_DeleteWithElement(term_ArgumentList(Term),term_Delete);
	  term_RplacArgumentList(Term,list_Nil());
	  return Term;
	}
      Scan = list_Cdr(Scan);
    }
    if (Update) {
      term_RplacArgumentList(Term,list_DeleteElementIfFree(term_ArgumentList(Term),fol_IsFalse,term_Delete));
      if (list_Empty(term_ArgumentList(Term))) {
	term_RplacTop(Term,fol_False());
	return Term;
      }
    }
  }
  else if (fol_IsQuantifier(Top) || symbol_Equal(Top,fol_Not())) {
    if (fol_IsQuantifier(Top))
      Result = cnf_RemoveTrivialAtoms(term_SecondArgument(Term));
    else
      Result = cnf_RemoveTrivialAtoms(term_FirstArgument(Term));
    Subtop = term_TopSymbol(Result);
    if ((symbol_Equal(Subtop,fol_False()) && symbol_Equal(Top,fol_Not())) ||
	(symbol_Equal(Subtop,fol_True()) && fol_IsQuantifier(Top))) {
      term_RplacTop(Term,fol_True());
      list_DeleteWithElement(term_ArgumentList(Term),term_Delete);
      term_RplacArgumentList(Term,list_Nil());
      return Term;
    }
    else
      if ((symbol_Equal(Subtop,fol_True()) && symbol_Equal(Top,fol_Not())) ||
	  (symbol_Equal(Subtop,fol_False()) && fol_IsQuantifier(Top))) {
	term_RplacTop(Term,fol_False());
	list_DeleteWithElement(term_ArgumentList(Term),term_Delete);
	term_RplacArgumentList(Term,list_Nil());
	return Term;
      }
  }
  else if (symbol_Equal(Top,fol_Implies())) {
    Result = cnf_RemoveTrivialAtoms(term_FirstArgument(Term));
    Subtop = term_TopSymbol(Result);
    if (symbol_Equal(Subtop,fol_False())) {
      term_RplacTop(Term,fol_True());
      list_DeleteWithElement(term_ArgumentList(Term),term_Delete);
      term_RplacArgumentList(Term,list_Nil());
      return Term;
    }
    else if (symbol_Equal(Subtop,fol_True())) {
      term_Delete(Result);
      Result = term_SecondArgument(Term);
      list_Delete(term_ArgumentList(Term));
      term_RplacTop(Term,term_TopSymbol(Result));
      term_RplacArgumentList(Term,term_ArgumentList(Result));
      term_Free(Result);
      return cnf_RemoveTrivialAtoms(Term);
    }
    Result = cnf_RemoveTrivialAtoms(term_SecondArgument(Term));
    Subtop = term_TopSymbol(Result);
    if (symbol_Equal(Subtop,fol_True())) {
      term_RplacTop(Term,fol_True());
      list_DeleteWithElement(term_ArgumentList(Term),term_Delete);
      term_RplacArgumentList(Term,list_Nil());
      return Term;
    }
    else if (symbol_Equal(Subtop,fol_False())) {
      term_RplacTop(Term,fol_Not());
      term_RplacArgumentList(Term,list_DeleteElementIfFree(term_ArgumentList(Term),fol_IsFalse,term_Delete));
    }
  }
  else if (symbol_Equal(Top,fol_Equiv())) {
    Result = cnf_RemoveTrivialAtoms(term_FirstArgument(Term));
    Subtop = term_TopSymbol(Result);
    if (symbol_Equal(Subtop,fol_False())) {
      term_RplacTop(Term,fol_Not());
      term_RplacArgumentList(Term,list_DeleteElementIfFree(term_ArgumentList(Term),fol_IsFalse,term_Delete));
      return cnf_RemoveTrivialAtoms(Term);
    }
    else if (symbol_Equal(Subtop,fol_True())) {
      term_Delete(Result);
      Result = term_SecondArgument(Term);
      list_Delete(term_ArgumentList(Term));
      term_RplacTop(Term,term_TopSymbol(Result));
      term_RplacArgumentList(Term,term_ArgumentList(Result));
      term_Free(Result);
      return cnf_RemoveTrivialAtoms(Term);
    }
    Result = cnf_RemoveTrivialAtoms(term_SecondArgument(Term));
    Subtop = term_TopSymbol(Result);
    if (symbol_Equal(Subtop,fol_False())) {
      term_RplacTop(Term,fol_Not());
      term_RplacArgumentList(Term,list_DeleteElementIfFree(term_ArgumentList(Term),fol_IsFalse,term_Delete));
    }
    else if (symbol_Equal(Subtop,fol_True())) {
      term_Delete(Result);
      Result = term_FirstArgument(Term);
      list_Delete(term_ArgumentList(Term));
      term_RplacTop(Term,term_TopSymbol(Result));
      term_RplacArgumentList(Term,term_ArgumentList(Result));
      term_Free(Result);
    }
  }

  return Term;
}

TERM  cnf_ObviousSimplifications(TERM Term)

{

  Term = cnf_RemoveTrivialAtoms(Term);
  Term = cnf_RemoveTrivialOperators(Term);
  Term = cnf_SimplifyQuantors(Term);

  return Term;
}

TERM cnf_SkolemFormula(TERM Term, LIST* Symblist)

{
  SYMBOL Top,SkolemSymbol;
  TERM   Subterm,SkolemTerm;
  LIST   Varlist,Scan;
  NAT    Arity;

  Top = term_TopSymbol(Term);
    
  if (fol_IsQuantifier(term_TopSymbol(Term))) {
    if (symbol_Equal(Top,fol_All())){
      term_Delete(term_FirstArgument(Term));
      Subterm = term_SecondArgument(Term);
      list_Delete(term_ArgumentList(Term));
      term_RplacTop(Term,term_TopSymbol(Subterm));
      term_RplacArgumentList(Term,term_ArgumentList(Subterm));
      term_Free(Subterm);
      return cnf_SkolemFormula(Term,Symblist);
    } 
    else {
      Varlist = fol_FreeVariables(Term);
      Arity   = list_Length(Varlist);
      for (Scan=fol_QuantifierVariables(Term);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	SkolemSymbol = symbol_CreateSkolemFunction(Arity);
	*Symblist  = list_Cons((POINTER)SkolemSymbol,*Symblist);
	SkolemTerm = term_Create(SkolemSymbol,Varlist); 
	fol_ReplaceVariable(term_SecondArgument(Term),term_TopSymbol(list_Car(Scan)),SkolemTerm);
	term_Free(SkolemTerm);
      }
      list_Delete(Varlist);
      term_Delete(term_FirstArgument(Term));
      Subterm = term_SecondArgument(Term);
      list_Delete(term_ArgumentList(Term));
      term_RplacTop(Term,term_TopSymbol(Subterm));
      term_RplacArgumentList(Term,term_ArgumentList(Subterm));
      term_Free(Subterm);
      return cnf_SkolemFormula(Term,Symblist);
    } 
  }
  else
      if (symbol_Equal(Top,fol_And()) || symbol_Equal(Top,fol_Or()))
	for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	  cnf_SkolemFormula(list_Car(Scan), Symblist);
  return Term;
}

static TERM cnf_Cnf(TERM term, LIST* symblist)

{
  
  term     = cnf_ObviousSimplifications(term);
  term_AddFatherLinks(term);
  term     = ren_Rename(term,symblist,FALSE, FALSE);
  term     = cnf_RemoveEquivImplFromFormula(term);
  term     = cnf_NegationNormalFormula(term);
  term     = cnf_SkolemFormula(cnf_AntiPrenex(term), symblist);
  term     = cnf_DistributivFormula(term);

  return term;
}

static LIST cnf_GetUsedTerms(CLAUSE C, PROOFSEARCH Search, HASH InputClauseToTermLabellist) 
{
  LIST UsedTerms, Used2, Scan;
  UsedTerms = list_Copy(hsh_Get(InputClauseToTermLabellist, C));
  UsedTerms = list_DeleteDuplicates(UsedTerms, cnf_LabelEqual);
  if (!list_Empty(UsedTerms))
    return UsedTerms;

  for (Scan = clause_ParentClauses(C); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    CLAUSE P;
    int ClauseNumber;
    ClauseNumber = (int) list_Car(Scan);
    P = clause_GetNumberedCl(ClauseNumber, prfs_WorkedOffClauses(Search));
    if (P == NULL) {
      P = clause_GetNumberedCl(ClauseNumber, prfs_UsableClauses(Search));
      if (P == NULL)
	P = clause_GetNumberedCl(ClauseNumber, prfs_DocProofClauses(Search));
    }
    Used2 = cnf_GetUsedTerms(P, Search, InputClauseToTermLabellist);
    UsedTerms = list_Nconc(UsedTerms, Used2);
  }
  return UsedTerms;
}

static BOOL cnf_HaveProofOptSkolem(PROOFSEARCH Search, TERM topterm, char* toplabel, TERM term2, LIST* UsedTerms,
				   LIST* Symblist, HASH InputClauseToTermLabellist)

{
    LIST ConClauses, EmptyClauses;
    LIST scan;
    BOOL   found;
    LIST Usables;
    NAT  Depth;

    ConClauses = list_Nil();
    found      = FALSE;
    
    term_AddFatherLinks(term2);
    term2      = cnf_Cnf(term2,Symblist);
    Usables    = cnf_MakeClauseList(term2,FALSE,FALSE);
    Depth      = clause_MaxTermDepthClauseList(Usables);
    term_Delete(term2);

    for (scan=Usables;!list_Empty(scan); scan = list_Cdr(scan)) {
      clause_SetFlag(list_Car(scan),CONCLAUSE);
      if(flag_GetFlagValue(flag_DOCPROOF)) {
#ifdef DEBUG
	hsh_Check(InputClauseToTermLabellist);
#endif
	hsh_Put(InputClauseToTermLabellist, list_Car(scan), toplabel);
#ifdef DEBUG_CNF
	printf("\nUsable : ");
	clause_Print(list_Car(scan));
	printf(" Label %s", toplabel);
#endif
      }
    }
    EmptyClauses = cnf_SatUnit(Search, Usables);
    if (!list_Empty(EmptyClauses)) {
      found = TRUE;
#ifdef DEBUG_CNF
      printf("\nHaveProof : Empty Clause : "); 
      clause_Print((CLAUSE) list_Car(EmptyClauses));
      printf("\n");
#endif
      if (flag_GetFlagValue(flag_DOCPROOF))
	*UsedTerms = list_Nconc(*UsedTerms, cnf_GetUsedTerms((CLAUSE) list_Car(EmptyClauses), Search, InputClauseToTermLabellist));
      EmptyClauses = list_PointerDeleteDuplicates(EmptyClauses);
      list_DeleteWithElement(EmptyClauses, clause_Delete);
    }
    
    
    ConClauses = list_Copy(prfs_UsableClauses(Search));
    for(scan = ConClauses; !list_Empty(scan); scan = list_Cdr(scan))
      if(flag_GetFlagValue(flag_DOCPROOF))
	prfs_MoveUsableDocProof(Search, (CLAUSE) list_Car(scan));
      else
	prfs_DeleteUsable(Search, (CLAUSE) list_Car(scan));
    list_Delete(ConClauses); 
	
    return found;
}

BOOL cnf_HaveProof(LIST TermList, TERM ToProve)

{
  PROOFSEARCH search;
  LIST        scan, usables, symblist, emptyclauses;
  BOOL        found;

  
  search  = cnf_HaveProof_PS;
  usables = symblist = list_Nil();

  
  for(scan=TermList; !list_Empty(scan); scan=list_Cdr(scan)) {
    TERM t;
    t       = term_Copy((TERM) list_Car(scan));
    t       = cnf_Cnf(t, &symblist);

    usables = list_Nconc(cnf_MakeClauseList(t, FALSE, FALSE), usables);
    term_Delete(t); 
  }

  
  ToProve = term_Create(fol_Not(), list_List(term_Copy(ToProve)));
  term_AddFatherLinks(ToProve);
  ToProve = cnf_Cnf(ToProve, &symblist);
  usables = list_Nconc(cnf_MakeClauseList(ToProve, FALSE, FALSE), usables);
  term_Delete(ToProve);

  
  for (scan=usables;!list_Empty(scan); scan = list_Cdr(scan))
    clause_SetFlag(list_Car(scan),CONCLAUSE);

  emptyclauses = cnf_SatUnit(search, usables);
  
  if (!list_Empty(emptyclauses)) {
    found = TRUE;
    emptyclauses = list_PointerDeleteDuplicates(emptyclauses);
    list_DeleteWithElement(emptyclauses, clause_Delete);
  }
  else
    found = FALSE;
  prfs_Clean(search);
  list_DeleteWithElement(symblist,symbol_Delete); 

  return found;
}

static void cnf_RplacVarsymbFunction(TERM term, SYMBOL varsymb,
					TERM function)

{
  int  bottom;
  TERM term1;
  LIST scan;

  bottom = vec_ActMax();
  vec_Push(term);

  while (!(bottom == vec_ActMax())){
    term1 = (TERM)vec_PopResult();
    if (symbol_Equal(term_TopSymbol(term1),varsymb)){
      term_RplacTop(term1,term_TopSymbol(function));
      term_RplacArgumentList(term1,list_CopyWithElement(term_ArgumentList(function),( POINTER (*)())term_Copy));
    }else
      if (!list_Empty(term_ArgumentList(term1)))
	for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	  vec_Push(list_Car(scan));
  }
  vec_SetMax(bottom);
}


static void cnf_RplacVar(TERM Term, LIST Varlist,
			 LIST Termlist)

{
  int  bottom;
  TERM term1;
  LIST scan,scan2;

  bottom = vec_ActMax();
  vec_Push(Term);

  while (!(bottom == vec_ActMax())){
    term1 = (TERM)vec_PopResult();
    if(symbol_IsVariable(term_TopSymbol(term1))) {
      BOOL done;
      done = FALSE;
      for(scan=Varlist, scan2=Termlist; !list_Empty(scan) && (done == FALSE); 
	  scan = list_Cdr(scan), scan2 = list_Cdr(scan2)) {
	if (symbol_Equal(term_TopSymbol(term1),term_TopSymbol((TERM) list_Car(scan)))) {
	  term_RplacTop(term1,term_TopSymbol((TERM) list_Car(scan2)));
	  term_RplacArgumentList(term1,
				 list_CopyWithElement(term_ArgumentList((TERM) list_Car(scan2)),
						      ( POINTER (*)())term_Copy));
	  done = TRUE;
	}
      }
    }
    else
      if (!list_Empty(term_ArgumentList(term1)))
	for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	  vec_Push(list_Car(scan));
  }
  vec_SetMax(bottom);
}

static TERM cnf_MakeSkolemfunction(LIST varlist)

{
  TERM   term;
  SYMBOL skolem;
  int    n;
  
  n = list_Length(varlist);
  skolem    = symbol_CreateSkolemFunction(n);
  term      = term_Create(skolem,list_CopyWithElement(varlist,(POINTER (*)())term_Copy));
  return term;
}


static void cnf_PopAllQuantifier(TERM term)

{
    TERM SubTerm;
    LIST VarList;
#ifdef DEBUG
    if (term_TopSymbol(term) != fol_All()) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In cnf_PopAllQuantifie: Top symbol is not fol_All !\n");
      misc_FinishErrorReport();
    }
#endif
    VarList = term_ArgumentList(term_FirstArgument(term));
    list_DeleteWithElement(VarList, term_Delete);
    term_Free(term_FirstArgument(term));
    SubTerm = term_SecondArgument(term);
    list_Delete(term_ArgumentList(term));
    term_RplacTop(term,term_TopSymbol(SubTerm));
    term_RplacArgumentList(term,term_ArgumentList(SubTerm));
    term_Free(SubTerm);
}

static TERM cnf_QuantifyAndNegate(TERM term, LIST VarList, LIST FreeList)

{
    TERM Result;
    TERM TermCopy;
    LIST VarListCopy;
    LIST FreeListCopy;

    TermCopy = term_Copy(term);
    VarListCopy = list_CopyWithElement(VarList, (POINTER (*) ()) term_Copy);
    Result = fol_CreateQuantifier(fol_Exist(),VarListCopy,list_List(TermCopy));
    FreeListCopy = list_Nil();

    FreeList = fol_FreeVariables(Result);
    if (!list_Empty(FreeList)) {
      FreeListCopy = list_CopyWithElement(FreeList, 
					  (POINTER (*) ()) term_Copy);
      list_Delete(FreeList);
      Result = fol_CreateQuantifier(fol_All(), FreeListCopy, 
				    list_List(Result));
    }
    Result = term_Create(fol_Not(), list_List(Result));
    return Result;
}

static TERM cnf_MoveProvedTermToTopLevel(TERM Term, TERM Term1, TERM Proved,
				     LIST VarList, LIST FreeList)

{
    TERM termR;
    TERM skolemfunction;
    SYMBOL varsymb;
    LIST scan;
    termR = term_SecondArgument(Term1); 
    term_RplacArgumentList(termR,
			   list_PointerDeleteElement(term_ArgumentList(termR),
						     Proved));
    if (list_Length(term_ArgumentList(termR)) < 2) {
	TERM termRL = term_FirstArgument(termR);     
	list_Delete(term_ArgumentList(termR));
	term_RplacTop(termR, term_TopSymbol(termRL));
	term_RplacArgumentList(termR,term_ArgumentList(termRL));
	term_Free(termRL);
    }

    for(scan = VarList; scan != list_Nil(); scan = list_Cdr(scan)) {
	varsymb = term_TopSymbol(list_Car(scan));
	skolemfunction = cnf_MakeSkolemfunction(FreeList);
	cnf_RplacVarsymbFunction(termR,varsymb,skolemfunction);
	cnf_RplacVarsymbFunction(Proved,varsymb,skolemfunction); 
	term_Delete(skolemfunction); 
    }

    if (!list_Empty(FreeList)) {
	Proved = 
	    fol_CreateQuantifier(fol_All(),
				 list_CopyWithElement(FreeList,
						      (POINTER(*)())term_Copy),
				 list_List(Proved));
	if (list_Length(FreeList) > 1)
	    Proved = cnf_QuantMakeOneVar(Proved);
    }
 
    term_Delete(term_FirstArgument(Term1)); 
    list_Delete(term_ArgumentList(Term1));
    term_RplacTop(Term1,term_TopSymbol(termR));
    term_RplacArgumentList(Term1,term_ArgumentList(termR));
    term_Free(termR);
    
    term_RplacArgumentList(Term, list_Cons(Proved, term_ArgumentList(Term)));
    return Proved;
}

static void cnf_Skolemize(TERM Term, LIST FreeList)

{
    LIST exlist;
    TERM subterm;
    LIST symblist;
    symblist = list_Nil();
    exlist = cnf_GetSymbolList(term_ArgumentList(term_FirstArgument(Term)));
    term_Delete(term_FirstArgument(Term));
    subterm = term_SecondArgument(Term);
    list_Delete(term_ArgumentList(Term));
    term_RplacTop(Term,term_TopSymbol(subterm));
    term_RplacArgumentList(Term,term_ArgumentList(subterm));
    term_Free(subterm);
    cnf_SkolemFunctionFormula(Term, FreeList, exlist, &symblist);
    list_Delete(exlist);
    list_Delete(symblist);
}

static LIST cnf_FreeVariablesBut(TERM Term, LIST Symbols)

{
  LIST follist, Scan;
  follist = fol_FreeVariables(Term);
  for(Scan = follist; !list_Empty(Scan); Scan = list_Cdr(Scan))
    if(list_Member(Symbols, (POINTER)term_TopSymbol(list_Car(Scan)), symbol_Equal))
      list_Rplaca(Scan,NULL);
  follist = list_PointerDeleteElement(follist,NULL);
    
  return follist;
}

static void cnf_SkolemFunctionFormulaMapped(TERM term, LIST allist, LIST map)

{
  TERM   term1;
  LIST   scan,scan1;
  SYMBOL skolem, symbol;
  int    bottom,n;

  bottom = vec_ActMax();
  n = list_Length(allist);
  
  for (scan1=map;!list_Empty(scan1);scan1=list_Cdr(scan1)){
    vec_Push(term);
    symbol = (SYMBOL) list_PairFirst((LIST) list_Car(scan1));
    skolem = (SYMBOL) list_PairSecond((LIST) list_Car(scan1));

#ifdef DEBUG_STRSKOLEM
    if(flag_GetFlagValue(flag_PSTRSKOLEM)) {
      fputs("\nVariable : ", stdout);
      symbol_Print(symbol);
      fputs("\nFunction : ", stdout);
      symbol_Print(skolem);
    }
#endif
    while (!(bottom == vec_ActMax())){
      term1 = (TERM)vec_PopResult();

      if (symbol_Equal(term_TopSymbol(term1),symbol)){
	term_RplacTop(term1,skolem);
	list_Delete(term_ArgumentList(term1));
	term_RplacArgumentList(term1,list_CopyWithElement(allist,(POINTER (*)())term_Copy));
      }
      if (!list_Empty(term_ArgumentList(term1)))
	for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan)) {
	  vec_Push(list_Car(scan));
	}
    }
  }
  vec_SetMax(bottom);
}
     
static BOOL cnf_HasDeeperVariable(LIST List1, LIST List2)
     
{
  LIST scan;
  int maxdepth1;
  
  
  maxdepth1 = 0;
  for(scan=List1; !list_Empty(scan); scan=list_Cdr(scan)) {
    int i;
    i = cnf_VARIABLEDEPTHARRAY[term_TopSymbol((TERM) list_Car(scan))];
#ifdef DEBUG_STRSKOLEM
    if(flag_GetFlagValue(flag_PSTRSKOLEM)) {
      fputs("\nFor variable ", stdout);
      symbol_Print(term_TopSymbol((TERM) list_Car(scan)));
      printf(" depth is %d.", i);
    }
#endif
    if (i > maxdepth1)
      maxdepth1 = i;
  }

  
  for(scan=List2; !list_Empty(scan); scan=list_Cdr(scan)) {
    int i;
    i = cnf_VARIABLEDEPTHARRAY[term_TopSymbol((TERM) list_Car(scan))];
#ifdef DEBUG_STRSKOLEM
    if(flag_GetFlagValue(flag_PSTRSKOLEM)) {
      fputs("\nFor variable ", stdout);
      symbol_Print(term_TopSymbol((TERM) list_Car(scan)));
      printf(" depth is %d.", i);
    }
#endif
    if (i >= maxdepth1)
      return FALSE;
  }
  return  TRUE;
}

static void cnf_StrongSkolemization(PROOFSEARCH Search, 
				    TERM Topterm, char* Toplabel, 
				    BOOL TopAnd, TERM Term, 
				    LIST* UsedTerms, 
				    LIST* Symblist, BOOL Result1, 
				    HASH InputClauseToTermLabellist, int Depth)

{
  LIST exlist;    
  LIST pairlist;  
  LIST allfreevariables;
  LIST newvariables;       
  LIST mapping;
  int numberofallfreevariables, acc_length, i;
  LIST pair, pairscan, pairscan_pred, scan, accumulatedvariables;
  TERM subterm, w;
  BOOL strskolemsuccess;  

  
  if(flag_GetFlagValue(flag_CNFSTRSKOLEM))
    symbol_SetStandardVarCounter(term_MaxVar(Topterm));

  
  exlist = cnf_GetSymbolList(term_ArgumentList(term_FirstArgument(Term)));
  
  term_Delete(term_FirstArgument(Term));
  subterm = term_SecondArgument(Term);
  list_Delete(term_ArgumentList(Term));
  term_RplacTop(Term,term_TopSymbol(subterm));
  term_RplacArgumentList(Term,term_ArgumentList(subterm));
  term_Free(subterm);

  
  pairlist = list_Nil();
  for(scan=term_ArgumentList(Term); !list_Empty(scan); scan = list_Cdr(scan)) {
    pair = list_PairCreate((TERM) list_Car(scan), 
			   cnf_FreeVariablesBut((TERM) list_Car(scan), exlist));
    if (list_Empty(pairlist)) 
      pairlist = list_List(pair);
    else {
      
      int pairlength, currentlength;
      pairlength = list_Length((LIST) list_PairSecond(pair));
      pairscan = pairlist; 
      pairscan_pred = list_Nil();
      currentlength = 0;
      while(!list_Empty(pairscan)) {
	currentlength = list_Length((LIST) list_PairSecond((LIST) list_Car(pairscan)));
	if(currentlength < pairlength) {
	  pairscan_pred = pairscan;
	  pairscan = list_Cdr(pairscan);
	}
	else if(currentlength == pairlength) {
	  
	  if(cnf_HasDeeperVariable((LIST) list_PairSecond((LIST) list_Car(pairscan)), 
				   (LIST) list_PairSecond(pair))) {                   
#ifdef DEBUG_STRSKOLEM
	    if(flag_GetFlagValue(flag_PSTRSKOLEM)) {
	      fputs("\nTerm ", stdout);
	      term_Print((TERM) list_PairFirst((LIST) list_Car(pairscan)));
	      fputs("\n has deeper variable than ", stdout);
	      term_Print((TERM) list_PairFirst(pair));
	    }
#endif	    
	    pairscan_pred = pairscan;
	    pairscan = list_Cdr(pairscan);
	  }
	  else
	    break;
	}
	else
	  break;
      }
      
      
      if(list_Empty(pairscan))
	list_Rplacd(pairscan_pred, list_List(pair));
      
      else if(currentlength >= pairlength) { 
	
	if(list_Empty(pairscan_pred))
	  pairlist = list_Cons(pair, pairlist);
	else 
	  list_InsertNext(pairscan_pred, pair);
      }
      
    }
  }

#ifdef DEBUG_STRSKOLEM
  if(flag_GetFlagValue(flag_PSTRSKOLEM)) {
    for(pairscan=pairlist; !list_Empty(pairscan); pairscan = list_Cdr(pairscan)) {
      LIST l;
      fputs("\nSubterm ", stdout);
      term_Print((TERM) list_PairFirst((LIST) list_Car(pairscan)));
      fputs("\n  has free variables ", stdout);
      for(l=(LIST) list_PairSecond((LIST) list_Car(pairscan)); !list_Empty(l); l = list_Cdr(l)) {
	term_Print((TERM) list_Car(l));
	fputs(" ", stdout);
      }
    }
  }
#endif

  
  
  allfreevariables = cnf_FreeVariablesBut(Term, exlist);
  numberofallfreevariables = list_Length(allfreevariables);
    
  mapping  = list_Nil();

  for(scan = exlist; !list_Empty(scan); scan = list_Cdr(scan)) {
    SYMBOL skolem;
    skolem = symbol_CreateSkolemFunction(numberofallfreevariables);
    *Symblist = list_Cons((POINTER)skolem,*Symblist);
    mapping = list_Cons(list_PairCreate(list_Car(scan), (POINTER)skolem), 
			mapping);
  }
  list_Delete(allfreevariables);

  
  newvariables = list_Nil();

  for(i=0; i < numberofallfreevariables; i++) {
    w = term_CreateStandardVariable();
    newvariables = list_Cons(w, newvariables);
  }

#ifdef DEBUG_STRSKOLEM
  if(flag_GetFlagValue(flag_PSTRSKOLEM)) {
    LIST l;
    fputs("\nNew variables : ", stdout);
    for(l=newvariables; !list_Empty(l); l = list_Cdr(l)) {
      term_Print((TERM) list_Car(l));
      fputs(" ", stdout);
    }
  }
#endif
  
  
  accumulatedvariables = list_Nil();
  acc_length = 0;
  strskolemsuccess = FALSE;
  for(pairscan=pairlist; !list_Empty(pairscan); pairscan=list_Cdr(pairscan)) {
    LIST allist;

    
    accumulatedvariables = list_Nconc(accumulatedvariables,
				      (LIST) list_PairSecond((LIST) list_Car(pairscan)));
    accumulatedvariables = list_DeleteDuplicates(accumulatedvariables, term_Equal);

    
    for(i=0; i < list_Length(accumulatedvariables) - acc_length; i++) {
      term_Delete((TERM) list_Top(newvariables));
      newvariables = list_Pop(newvariables);
    }
    acc_length = list_Length(accumulatedvariables);

#ifdef DEBUG_STRSKOLEM
    if(flag_GetFlagValue(flag_PSTRSKOLEM)) {
      LIST l;
      fputs("\n\nSubterm is ", stdout);
      term_Print((TERM) list_PairFirst((LIST) list_Car(pairscan)));
      fputs("\nFree variables : ", stdout);
      for(l=accumulatedvariables; !list_Empty(l); l = list_Cdr(l)) {
	term_Print((TERM) list_Car(l));
	fputs(" ", stdout);
      }
    }
#endif
    if (list_Length(newvariables) > 0)
      strskolemsuccess = TRUE;
    allist = list_Nconc(list_Copy(accumulatedvariables), list_Copy(newvariables));

    cnf_SkolemFunctionFormulaMapped((TERM) list_PairFirst((LIST) list_Car(pairscan)), allist,
					    mapping);
#ifdef DEBUG_STRSKOLEM
    if(flag_GetFlagValue(flag_PSTRSKOLEM)) {
      fputs("\nSubterm after skolemization : ", stdout);
      term_Print(list_PairFirst((LIST) list_Car(pairscan)));
    }
#endif

    list_Delete(allist);
    cnf_OptimizedSkolemFormula(Search, Topterm, Toplabel, TopAnd,
			       (TERM) list_PairFirst((LIST) list_Car(pairscan)),
			       UsedTerms, Symblist, Result1, 
			       InputClauseToTermLabellist, Depth);
  }
  while(!list_Empty(newvariables)) {
    term_Delete((TERM) list_Top(newvariables));
    newvariables = list_Pop(newvariables);
  }
  list_Delete(accumulatedvariables); 
  list_DeleteWithElement(pairlist, list_Free);
  list_Delete(exlist);
  list_DeleteWithElement(mapping, list_Free);
  if((flag_GetFlagValue(flag_PSTRSKOLEM)) && (strskolemsuccess == TRUE)) {
    fputs("\nStrong skolemization applied", stdout);
  }
}
    
static void cnf_OptimizedSkolemFormula(PROOFSEARCH Search, 
				       TERM topterm, char* toplabel, 
				       BOOL TopAnd, TERM term, 
				       LIST* UsedTerms, 
				       LIST* Symblist, BOOL Result1, HASH InputClauseToTermLabellist, int Depth)

{
    TERM   termL2,provedterm;
    LIST   freevariables,scan,varlist;
    SYMBOL top;
    BOOL   result2;
    BOOL   optimized;

    result2 = FALSE;
    freevariables = list_Nil();

    top = term_TopSymbol(term);
    if (fol_IsQuantifier(top)) {
      if (symbol_Equal(top,fol_All())) {
	  
	  if (flag_GetFlagValue(flag_CNFSTRSKOLEM)) {
	     LIST variables;
	     variables = term_ArgumentList(term_FirstArgument(term));
	     for(scan=variables; !list_Empty(scan); scan=list_Cdr(scan)) {
#ifdef DEBUG_STRSKOLEM
	       if(flag_GetFlagValue(flag_PSTRSKOLEM)) {
		 if(cnf_VARIABLEDEPTHARRAY[term_TopSymbol((TERM) list_Car(scan))] != -1) {
		   fputs("\nFor variable ", stderr);
		   term_Print((TERM) list_Car(scan));
		   printf(" depth is already set to %d, now setting it to %d", 
			  cnf_VARIABLEDEPTHARRAY[term_TopSymbol((TERM) list_Car(scan))], Depth);
		 }
	       }
#endif
#ifdef DEBUG_STRSKOLEM
	       if(flag_GetFlagValue(flag_PSTRSKOLEM)) {
		 fputs("\nVariable ", stdout);
		 term_Print((TERM) list_Car(scan));
		 printf(" has depth %d in term\n  ", Depth);
		 term_Print(term);
	       }
#endif
	       cnf_VARIABLEDEPTHARRAY[term_TopSymbol((TERM) list_Car(scan))] = Depth;
	     }
	     Depth++;
	  }
	  cnf_PopAllQuantifier(term);
	  cnf_OptimizedSkolemFormula(Search,topterm, toplabel, TopAnd, term, 
				     UsedTerms, Symblist, Result1,
				     InputClauseToTermLabellist, Depth);
	  return;
      }
      freevariables = fol_FreeVariables(term);
      optimized = FALSE;
      if (symbol_Equal(term_TopSymbol(term_SecondArgument(term)), fol_And())) {
	if(flag_GetFlagValue(flag_CNFOPTSKOLEM)) {
	  scan = term_ArgumentList(term_SecondArgument(term));
	  varlist = term_ArgumentList(term_FirstArgument(term));
	  while((!list_Empty(scan)) && (!optimized)){
	    if (Result1 == FALSE) {
	      if (TopAnd == TRUE) {
		if(flag_GetFlagValue(flag_POPTSKOLEM)) {
		  fputs("\nHaveProof not necessary", stdout);
		}
		result2 = TRUE;
	      }
	      else {
		termL2 = cnf_QuantifyAndNegate((TERM) list_Car(scan), 
					       varlist, freevariables);
		result2 = cnf_HaveProofOptSkolem(Search, 
						 topterm, toplabel, termL2, UsedTerms, Symblist, 
						 InputClauseToTermLabellist);
#ifdef DEBUG_OPTSKOLEM
		if(flag_GetFlagValue(flag_POPTSKOLEM)) {
		  fputs("\nHaveProof result : ", stdout);
		  if(result2==TRUE)
		    fputs("TRUE", stdout);
		  else
		    fputs("FALSE", stdout);
		}
#endif
	      }
	    }
	    
	    if ((Result1 == TRUE) || (result2 == TRUE)) {
	      optimized   = TRUE;
	      if(flag_GetFlagValue(flag_POPTSKOLEM)) {
		printf("\nIn term ");
		term_Print(topterm);
		printf("\n subterm ");
		term_Print((TERM) list_Car(scan));
		printf(" is moved to toplevel.\n");
	      }
	      provedterm 
		= cnf_MoveProvedTermToTopLevel(topterm, term, 
					       (TERM) list_Car(scan), 
					       varlist,
					       freevariables);
	      if(flag_GetFlagValue(flag_POPTSKOLEM)) {
		printf("Result : ");
		term_Print(topterm);
		printf("\n");
	      }
	      
	      cnf_OptimizedSkolemFormula(Search, topterm, toplabel, 
					 TRUE,
					 provedterm,
					 UsedTerms, Symblist, Result1, 
					 InputClauseToTermLabellist, Depth);
	    }
	    else 
	      scan = list_Cdr(scan);
	  }
	}
	if(!optimized) {
	  
	  if(flag_GetFlagValue(flag_CNFSTRSKOLEM)) {
	    optimized = TRUE; 
	    cnf_StrongSkolemization(Search, topterm, toplabel, 
				    TopAnd,
				    term,
				    UsedTerms, Symblist, Result1, 
				    InputClauseToTermLabellist, Depth);
	  }
	}
      }
      else
	TopAnd = FALSE;
      if(!optimized) {
	
	
	cnf_Skolemize(term, freevariables);
      }
      list_Delete(freevariables);
      cnf_OptimizedSkolemFormula(Search, topterm, toplabel, TopAnd,
				 term,
				 UsedTerms, Symblist,
				 Result1, InputClauseToTermLabellist, Depth);
      return;
    } 
    else {
      if ((symbol_Equal(top,fol_And())) || (symbol_Equal(top,fol_Or()))) { 
	if (symbol_Equal(top,fol_Or()))
	  TopAnd = FALSE;
	for (scan=term_ArgumentList(term);!list_Empty(scan);
	     scan=list_Cdr(scan))
	  cnf_OptimizedSkolemFormula(Search, topterm, toplabel, TopAnd,
				     (TERM) list_Car(scan),
				     UsedTerms, Symblist,
				     Result1, InputClauseToTermLabellist, Depth);
      }
    }
    return;
}


static void cnf_SkolemFunctionFormula(TERM term, LIST allist, LIST exlist,
				  LIST* symblist)

{
  TERM   term1;
  LIST   scan,scan1;
  SYMBOL skolem;
  int    bottom,n;

  bottom = vec_ActMax();
  n = list_Length(allist);
  
  for (scan1=exlist;!list_Empty(scan1);scan1=list_Cdr(scan1)){
    vec_Push(term);
    skolem    = symbol_CreateSkolemFunction(n);
    *symblist = list_Cons((POINTER)skolem,*symblist);
    
    while (!(bottom == vec_ActMax())){
      term1 = (TERM)vec_PopResult();
      if (symbol_Equal(term_TopSymbol(term1),(SYMBOL)list_Car(scan1))){
	term_RplacTop(term1,skolem);
	list_Delete(term_ArgumentList(term1));
	term_RplacArgumentList(term1,list_CopyWithElement(allist,(POINTER (*)())term_Copy));
      }
      if (!list_Empty(term_ArgumentList(term1)))
	for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	  vec_Push(list_Car(scan));
    }
  }
  vec_SetMax(bottom);
}


static LIST cnf_OptimizedSkolemization(PROOFSEARCH Search, TERM Term, char* Label, 
				LIST* UsedTerms, 
				LIST* Symblist, BOOL result, BOOL Conjecture, 
				HASH InputClauseToTermLabellist)

{
    LIST Clauses;
    TERM FirstArg;
    int  i;
#ifdef DEBUG
    if (Term == NULL) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In cnf_OptimizedSkolemization: Input term is NULL.\n");
      misc_FinishErrorReport();
    }
#endif
    FirstArg = Term;

    if(flag_GetFlagValue(flag_CNFSTRSKOLEM)) {
      
      for(i=1; i <= symbol__MAXSTANDARDVAR; i++)
	cnf_VARIABLEDEPTHARRAY[i] = -1;
    }

    if(flag_GetFlagValue(flag_POPTSKOLEM) || flag_GetFlagValue(flag_PSTRSKOLEM)) {
      fputs("\nTerm before skolemization : \n ", stdout);
      fol_PrettyPrintDFG(Term);
    }

    if(!fol_IsLiteral(Term)) {
      if((flag_GetFlagValue(flag_CNFOPTSKOLEM)) ||
	 (flag_GetFlagValue(flag_CNFSTRSKOLEM)))  {
	if(flag_GetFlagValue(flag_CNFOPTSKOLEM))
	  Term = term_Create(fol_And(), list_List(Term)); 
	cnf_OptimizedSkolemFormula(Search, Term, Label, TRUE, FirstArg, UsedTerms, 
				   Symblist, result, InputClauseToTermLabellist, 0);
      }
      else {
	LIST Symbols;
	Symbols = list_Nil();
	Term    = cnf_SkolemFormula(Term,&Symbols);
	list_Delete(Symbols);
      }
    }
    if(flag_GetFlagValue(flag_POPTSKOLEM) || flag_GetFlagValue(flag_PSTRSKOLEM)) {
      fputs("\nTerm after skolemization : ", stdout);
      term_Print(Term);
    }
    Term    = cnf_DistributivFormula(Term);
    Clauses = cnf_MakeClauseList(Term,FALSE,Conjecture);
    term_Delete(Term);

    return Clauses;
}


LIST cnf_GetSkolemFunctions(TERM Term, LIST ArgList, LIST* SkolToExVar)
 
{
    LIST Scan;
    SYMBOL Top;
    
    Top = term_TopSymbol(Term);

    if (fol_IsQuantifier(Top)) 
	return cnf_GetSkolemFunctions(term_SecondArgument(Term), ArgList, 
				      SkolToExVar);

    if (symbol_IsFunction(Top) && symbol_HasProperty(Top, SKOLEM)) {
      BOOL found;
      SYMBOL Var = 0;
      int Arity;
      found = FALSE;
      
      
      
      Arity = list_Length(term_ArgumentList(Term));
      if (Arity > list_Length(ArgList)) {
	list_DeleteWithElement(ArgList, term_Delete);
	ArgList = term_ArgumentList(Term);
      }
      else 
	list_DeleteWithElement(term_ArgumentList(Term), term_Delete);
      term_RplacArgumentList(Term, NULL);
      
      
      if (list_Length(*SkolToExVar) > Arity) {
	NAT i;
	LIST SkolScan = *SkolToExVar;
	for(i = 0; i < Arity; i++)
	  SkolScan = list_Cdr(SkolScan);
	for(SkolScan = (LIST) list_Car(SkolScan); 
	    (SkolScan != list_Nil()) && (found == FALSE); 
	    SkolScan = list_Cdr(SkolScan)) {
	  SYMBOL Skol;
	  Skol = (SYMBOL) list_PairFirst((LIST) list_Car(SkolScan));
	  if (Skol == term_TopSymbol(Term)) {
	    Var = (SYMBOL) list_PairSecond((LIST) list_Car(SkolScan));
	    found = TRUE;
	  }
	}
      }
      if (found == FALSE) {
	LIST Pair;
	NAT i;
	LIST SkolScan;

	SkolScan = *SkolToExVar;
	for(i=0; i < Arity; i++) {
	  if (list_Cdr(SkolScan) == list_Nil())
	    list_Rplacd(SkolScan, list_List(NULL));
	  SkolScan = list_Cdr(SkolScan);
	}

	Var = symbol_CreateStandardVariable();
	Pair = list_PairCreate((POINTER) term_TopSymbol(Term), 
			       (POINTER) Var);
	if (list_Car(SkolScan) == list_Nil())
	  list_Rplaca(SkolScan, list_List(Pair));
	else
	  list_Rplaca(SkolScan, list_Nconc((LIST) list_Car(SkolScan), 
					   list_List(Pair)));
      }
      term_RplacTop(Term, Var);
    }
    else {
      for(Scan = term_ArgumentList(Term); Scan != list_Nil(); 
	  Scan = list_Cdr(Scan))
	ArgList = cnf_GetSkolemFunctions((TERM) list_Car(Scan), ArgList, 
					 SkolToExVar);
    }
    return ArgList;
}

void cnf_ReplaceVariable(TERM Term, SYMBOL Old, SYMBOL New)
 
{
  LIST Scan;

#ifdef DEBUG
  if (!symbol_IsVariable(Old)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In cnf_ReplaceVariable: Illegal input symbol.\n");
    misc_FinishErrorReport();
  }
#endif

  if (symbol_Equal(term_TopSymbol(Term), Old))
    term_RplacTop(Term, New);
  else 
      for (Scan = term_ArgumentList(Term); !list_Empty(Scan); 
	   Scan = list_Cdr(Scan))
	  cnf_ReplaceVariable(list_Car(Scan), Old, New);
}

LIST cnf_RemoveSkolemFunctions(CLAUSE Clause, LIST* SkolToExVar, LIST Vars)

{
    LIST Scan;
    LIST TermScan, TermList, ArgList;
    TERM Term;
    int i;

    TermList = list_Nil();

    ArgList = list_Nil();
    for (i = 0; i < clause_Length(Clause); i++) {
	Term = term_Copy(clause_GetLiteralTerm(Clause, i));
	ArgList = cnf_GetSkolemFunctions(Term, ArgList, SkolToExVar);
	TermList = list_Cons(Term, TermList);
    }
    
    if (list_Length(ArgList) == 0)
	return TermList;
    
    
    for(Scan = ArgList; Scan != list_Nil(); Scan = list_Cdr(Scan)) {
	for (TermScan = TermList; TermScan != list_Nil(); 
	     TermScan = list_Cdr(TermScan)) {
	    Term = (TERM) list_Car(TermScan);
	    cnf_ReplaceVariable(Term,
				term_TopSymbol((TERM) list_Car(Scan)),
				(SYMBOL) list_Car(Vars));
	}
	if (list_Cdr(Vars) == list_Nil()) {
	    SYMBOL New = symbol_CreateStandardVariable();
	    Vars = list_Nconc(Vars, list_List((POINTER) New));
	}
	Vars = list_Cdr(Vars);
    }
    list_DeleteWithElement(ArgList, term_Delete);
    return TermList;
}


TERM   cnf_DeSkolemFormula(LIST Clauses)

{
    LIST Scan;
    LIST SkolToExVar;
    LIST Vars;
    LIST FreeVars;
    LIST FreeVarsCopy;
    TERM VarListTerm;
    LIST VarScan;
    LIST TermList;
    TERM TopTerm;
    TERM Term;
    NAT i;
    BOOL First;

    SkolToExVar = list_List(NULL);
    Vars = list_List((POINTER) symbol_CreateStandardVariable());

    TopTerm = term_Create(fol_And(), NULL);

    for(Scan = Clauses; Scan != list_Nil(); Scan = list_Cdr(Scan)) {
	TermList =  cnf_RemoveSkolemFunctions((CLAUSE) list_Car(Scan),
					      &SkolToExVar, Vars);
	Term = term_Create(fol_Or(), TermList);
	FreeVars = fol_FreeVariables(Term);
	if (!list_Empty(FreeVars)) {
	  FreeVarsCopy = list_CopyWithElement(FreeVars, (POINTER (*) ()) term_Copy);
	  list_Delete(FreeVars);
	  Term = fol_CreateQuantifier(fol_All(), FreeVarsCopy,
				      list_List(Term));
	}
	term_RplacArgumentList(TopTerm, list_Cons(Term, 
						  term_ArgumentList(TopTerm)));
    }
    
    i = 0;
    VarScan = Vars;
    First = TRUE;

    for (Scan = SkolToExVar; Scan != list_Nil(); Scan = list_Cdr(Scan)) {
	if (list_Empty(list_Car(Scan))) {
	  if (term_TopSymbol(TopTerm) == fol_All())
	    term_RplacArgumentList(TopTerm, list_Cons(term_Create((SYMBOL) list_Car(VarScan), NULL),  
						      term_ArgumentList(TopTerm)));
	    if (First == FALSE)
	      TopTerm = fol_CreateQuantifier(fol_All(),
					     list_List(term_Create((SYMBOL) list_Car(VarScan), NULL)),
					     list_List(TopTerm)); 
	}
	else {
	    LIST ExVarScan;
	    LIST ExVars = list_Nil();
	    for(ExVarScan = list_Car(Scan); ExVarScan != list_Nil();
		ExVarScan = list_Cdr(ExVarScan)) {
		if (ExVars == list_Nil())
		    ExVars = list_List(term_Create((SYMBOL) list_PairSecond((LIST) list_Car(ExVarScan)), NULL));
		else
		    ExVars = list_Cons(term_Create((SYMBOL) list_PairSecond((LIST) list_Car(ExVarScan)), NULL), ExVars);
		list_PairFree((LIST) list_Car(ExVarScan));
	    }
	    list_Delete((LIST) list_Car(Scan));
	    list_Rplaca(Scan, NULL);

	    if (term_TopSymbol(TopTerm) == fol_Exist()) {
	      VarListTerm = (TERM) list_Car(term_ArgumentList(TopTerm));
	      term_RplacArgumentList(VarListTerm,
				     list_Nconc(term_ArgumentList(VarListTerm),
				     ExVars));
	    }
	    else
	      TopTerm = fol_CreateQuantifier(fol_Exist(), ExVars, list_List(TopTerm));
	    ExVars = list_Nil();

	    if (First == FALSE) 
		TopTerm = fol_CreateQuantifier(fol_All(), 
					       list_List(term_Create((SYMBOL) list_Car(VarScan), NULL)),
					       list_List(TopTerm));
	}
	if (First == FALSE) 
	    VarScan = list_Cdr(VarScan);
	else
	    First = FALSE;
	
    }
    list_Delete(SkolToExVar);
    list_Delete(Vars);

    return TopTerm;
}

#ifdef OPTCHECK
LIST checkOptimizedSkolemization(LIST* AxClauses, LIST* ConClauses,
				 TERM AxTerm, TERM ConTerm,
				 LIST NonConClauses, LIST* SkolemPredicates,
				 SHARED_INDEX ShIndex,
				 BOOL result)

{
  TERM DeSkolemizedAxOpt, DeSkolemizedConOpt, DeSkolemizedAx, DeSkolemizedCon;
  TERM TopOpt, Top, ToProve;
  LIST SkolemFunctions2;

  if (*AxClauses != list_Nil()) {
    DeSkolemizedAxOpt = cnf_DeSkolemFormula(*AxClauses);
    if (*ConClauses != list_Nil()) {
      DeSkolemizedConOpt = cnf_DeSkolemFormula(*ConClauses);
      TopOpt = term_Create(fol_And(), 
			   list_Cons(DeSkolemizedAxOpt, 
				     list_List(DeSkolemizedConOpt)));
    }
    else 
      TopOpt = DeSkolemizedAxOpt;
  }
  else {
    DeSkolemizedConOpt = cnf_DeSkolemFormula(*ConClauses);     
    TopOpt = DeSkolemizedConOpt;
  }
  
  list_DeleteWithElement(*AxClauses,clause_Delete);
  list_DeleteWithElement(*ConClauses,clause_Delete); 
  *AxClauses = list_Nil();
  *ConClauses = list_Nil();
  
  flag_SetFlagValue(flag_CNFOptSkolem, FALSE);  
  if (AxTerm) {
    *AxClauses = cnf_OptimizedSkolemization(term_Copy(AxTerm), ShIndex, NonConClauses, result,FALSE, ClauseToTermLabellist);
  }
  if (ConTerm) {
    *ConClauses = cnf_OptimizedSkolemization(term_Copy(ConTerm), ShIndex, NonConClauses, result,TRUE, ClauseToTermLabellist);
  }
  
  if (*AxClauses != list_Nil()) {
    DeSkolemizedAx = cnf_DeSkolemFormula(*AxClauses);
    if (*ConClauses != list_Nil()) {
      DeSkolemizedCon = cnf_DeSkolemFormula(*ConClauses);
      Top = term_Create(fol_And(), 
			list_Cons(DeSkolemizedAx, 
				  list_List(DeSkolemizedCon)));
    }
    else 
      Top = DeSkolemizedAx;
  }
  else {
    DeSkolemizedCon = cnf_DeSkolemFormula(*ConClauses);    
    Top = DeSkolemizedCon;
  }
  
  list_DeleteWithElement(*AxClauses,clause_Delete);
  list_DeleteWithElement(*ConClauses,clause_Delete); 
  *AxClauses = list_Nil();
  *ConClauses = list_Nil();
  
  ToProve = term_Create(fol_Equiv(), list_Cons(TopOpt, list_List(Top)));
  ToProve = term_Create(fol_Not(), list_List(ToProve));
  fol_NormalizeVars(ToProve); 
  ToProve  = cnf_ObviousSimplifications(ToProve);
  term_AddFatherLinks(ToProve);
  ToProve  = ren_Rename(ToProve,SkolemPredicates,FALSE);
  ToProve  = cnf_RemoveEquivImplFromFormula(ToProve);
  ToProve  = cnf_NegationNormalFormula(ToProve);
  ToProve  = cnf_AntiPrenex(ToProve);
  
  SkolemFunctions2 = list_Nil();
  ToProve = cnf_SkolemFormula(ToProve, &SkolemFunctions2);
  ToProve    = cnf_DistributivFormula(ToProve);
  *ConClauses = cnf_MakeClauseList(ToProve);
  if (ToProve)
    term_Delete(ToProve);
  *AxClauses = list_Nil();
  return SkolemFunctions2;
}
#endif

PROOFSEARCH cnf_Flotter(LIST AxiomList, LIST ConjectureList, LIST* AxClauses, 
		 LIST* AllLabels,
		 HASH  TermLabelToClauselist, 
		 HASH  ClauseToTermLabellist,
		 BOOL  KeepProofSearch,
		 LIST* Symblist)

{
  LIST         Scan, Scan2, FormulaClauses,Usables,SkolemFunctions;
  LIST         SkolemPredicates, EmptyClauses, AllFormulae;
  LIST         UsedTerms;
  TERM         AxTerm,Formula;
  BOOL         Result;
  PROOFSEARCH  Search;
  NAT          Count;
  HASH         InputClauseToTermLabellist;

  Search = prfs_Create();
  if (flag_GetFlagValue(flag_DOCPROOF))
    prfs_AddDocProofSharingIndex(Search);

  AxTerm           = (TERM)NULL;
  SkolemPredicates = Usables = list_Nil();
  Result           = FALSE;
  if(flag_GetFlagValue(flag_DOCPROOF))
    InputClauseToTermLabellist = hsh_Create();
  else 
    InputClauseToTermLabellist = NULL;

  symbol_ReinitGenericNameCounters();

  for(Scan=AxiomList; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    LIST Pair;
    Pair = list_Car(Scan);
    AxTerm = (TERM) list_PairSecond(Pair);
    fol_RemoveImplied(AxTerm);
    term_AddFatherLinks(AxTerm);
    fol_NormalizeVars(AxTerm);
    if (flag_GetFlagValue(flag_CNFFEQREDUCTIONS))
      cnf_PropagateSubstEquations(AxTerm);
    AxTerm = cnf_ObviousSimplifications(AxTerm);
    if (flag_GetFlagValue(flag_CNFRENAMING)) {
      term_AddFatherLinks(AxTerm);
      AxTerm   = ren_Rename(AxTerm,&SkolemPredicates,flag_GetFlagValue(flag_CNFPRENAMING), TRUE);
    }
    AxTerm   = cnf_RemoveEquivImplFromFormula(AxTerm);
    AxTerm   = cnf_NegationNormalFormula(AxTerm);
    AxTerm   = cnf_AntiPrenex(AxTerm);
    list_Rplacd(Pair, (LIST) AxTerm);
  }
  AllFormulae = AxiomList;
  
  

  Count = 0;
  for(Scan=ConjectureList; !list_Empty(Scan); Scan=list_Cdr(Scan)) {
    TERM  ConTerm;
    char* Label;
    char  buf[100];
      
    if (list_PairFirst(list_Car(Scan)) == NULL) {
      sprintf(buf, "conjecture%d", Count);
      Label = memory_Malloc(sizeof(char) * (strlen(buf) + 1));
      sprintf(Label, "conjecture%d", Count);
      list_Rplaca((LIST) list_Car(Scan), Label);
      if(flag_GetFlagValue(flag_DOCPROOF) && flag_GetFlagValue(flag_PLABELS)) {
	printf("\nAdded label %s for conjecture  \n", Label);
	fol_PrettyPrintDFG((TERM) list_PairSecond(list_Car(Scan)));
      }
    }

    ConTerm = (TERM) list_PairSecond((LIST) list_Car(Scan));
    fol_RemoveImplied(ConTerm);
    term_AddFatherLinks(ConTerm);
    fol_NormalizeVars(ConTerm);  
        if (flag_GetFlagValue(flag_CNFFEQREDUCTIONS))
      cnf_PropagateSubstEquations(ConTerm);
        ConTerm  = cnf_ObviousSimplifications(ConTerm); 

    if (flag_GetFlagValue(flag_CNFRENAMING)) {
      term_AddFatherLinks(ConTerm);
      ConTerm   = ren_Rename(ConTerm,&SkolemPredicates,flag_GetFlagValue(flag_CNFPRENAMING), TRUE);
    }
    
    ConTerm  = cnf_RemoveEquivImplFromFormula(ConTerm); 
    ConTerm  = cnf_NegationNormalFormula(ConTerm); 
    
    ConTerm  = cnf_AntiPrenex(ConTerm); 
    
    
    list_Rplacd((LIST) list_Car(Scan), (LIST) ConTerm);

    Count++;
  } 

  AllFormulae = list_Append(ConjectureList, AllFormulae);
  for(Scan=ConjectureList;!list_Empty(Scan); Scan = list_Cdr(Scan))
    list_Rplaca(Scan,list_PairSecond(list_Car(Scan)));

  FormulaClauses = list_Nil();
  SkolemFunctions = list_Nil();
  Count = 0;
  for(Scan = AllFormulae; !list_Empty(Scan); Scan = list_Cdr(Scan), Count++) {
    LIST FormulaClausesTemp;
    Formula         = term_Copy((TERM) list_PairSecond(list_Car(Scan)));
#ifdef DEBUG_CNF
    printf("\nInputFormula : "); term_Print(Formula); 
    printf("\nLabel : %s", (char*) list_PairFirst((list_Car(Scan)))); 
#endif
    Formula             = cnf_SkolemFormula(Formula, &SkolemFunctions);
    Formula             = cnf_DistributivFormula(Formula);
    FormulaClausesTemp  = cnf_MakeClauseList(Formula,FALSE,FALSE);
    if (flag_GetFlagValue(flag_DOCPROOF)) {
      for(Scan2 = FormulaClausesTemp; !list_Empty(Scan2); Scan2 = list_Cdr(Scan2)) {
	hsh_Put(InputClauseToTermLabellist, list_Car(Scan2), list_PairFirst((list_Car(Scan))));
      }
    }
    FormulaClauses = list_Nconc(FormulaClauses, FormulaClausesTemp);
    term_Delete(Formula);
  }

  
 
  
  for(Scan = FormulaClauses; !list_Empty(Scan); Scan = list_Cdr(Scan))
    clause_SetFlag((CLAUSE) list_Car(Scan), CONCLAUSE);
  
  
  EmptyClauses = red_SatUnit(Search, FormulaClauses, flag_GetFlagValue(flag_DOCPROOF));
  if (!list_Empty(EmptyClauses)) {
    Result = TRUE;
    
    list_DeleteWithElement(EmptyClauses, clause_Delete);
  }

  
  FormulaClauses = list_Copy(prfs_UsableClauses(Search));
  for(Scan = FormulaClauses; !list_Empty(Scan); Scan = list_Cdr(Scan))
    prfs_MoveUsableWorkedOff(Search, (CLAUSE) list_Car(Scan));
  list_Delete(FormulaClauses);
  FormulaClauses = list_Nil();

#ifdef DEBUG
  
#endif     


  
  *Symblist = list_Nil();
  for(Scan = AllFormulae; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    LIST Ax, Pair;
    UsedTerms = list_Nil();
    Pair = list_Car(Scan);
#ifdef DEBUG_CNF
    printf("\nFormula : "); 
    term_Print((TERM) list_PairSecond(Pair)); 
    printf("\nLabel : %s", (char*) list_PairFirst(Pair)); 
#endif
    Ax = cnf_OptimizedSkolemization(Search, term_Copy((TERM) list_PairSecond(Pair)), 
				    (char*) list_PairFirst(Pair), &UsedTerms, 
				    Symblist, Result, FALSE, InputClauseToTermLabellist);
    
    if (list_PointerMember(ConjectureList,list_PairSecond(Pair))) {
      LIST l;
      for(l=Ax; !list_Empty(l); l = list_Cdr(l))
	clause_SetFlag((CLAUSE) list_Car(l), CONCLAUSE);
    }
    if (flag_GetFlagValue(flag_DOCPROOF)) {
      hsh_PutListWithCompareFunc(TermLabelToClauselist, list_PairFirst(Pair), list_Copy(Ax), cnf_LabelEqual,
			     hsh_StringHashKey);
      UsedTerms = list_Cons(list_PairFirst(Pair), UsedTerms);
      UsedTerms = list_DeleteDuplicates(UsedTerms, cnf_LabelEqual);
      for(Scan2 = Ax; !list_Empty(Scan2); Scan2 = list_Cdr(Scan2)) {
	hsh_PutList(ClauseToTermLabellist, list_Car(Scan2), list_Copy(UsedTerms));
	hsh_PutList(InputClauseToTermLabellist, list_Car(Scan2), list_Copy(UsedTerms));
      } 
    }
    *AxClauses = list_Nconc(*AxClauses, Ax);
    list_Delete(UsedTerms);
  }

  list_Delete(ConjectureList);
  if (flag_GetFlagValue(flag_DOCPROOF))
    hsh_Delete(InputClauseToTermLabellist);
  if (KeepProofSearch == FALSE) { 
      list_Delete(*Symblist);
  }

  *AllLabels = list_Nil();
  for(Scan = AllFormulae; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    LIST Pair;
    Pair = list_Car(Scan);
    term_Delete((TERM) list_PairSecond(Pair));
    *AllLabels = list_Cons(list_PairFirst(Pair), *AllLabels);
    list_PairFree(Pair);
  }

  list_Delete(AllFormulae);
  list_Delete(SkolemFunctions);
  list_Delete(SkolemPredicates);

  if (KeepProofSearch == FALSE) {
    symbol_ResetSkolemIndex();
    prfs_Delete(Search);
    return NULL;
  }
  else {
    
    prfs_DeleteDocProof(Search);
    return Search;
  }
}

LIST cnf_QueryFlotter(PROOFSEARCH Search, TERM Term, LIST* Symblist)

{
  LIST SkolemPredicates, SkolemFunctions, IndexedClauses, Scan;
  LIST ResultClauses, Dummy, EmptyClauses;
  TERM TermCopy;
  int  Formulae2Clause;
  BOOL Result;

  SkolemPredicates = SkolemFunctions = list_Nil();
  Result = FALSE;
  
  prfs_CopyIndices(Search, SearchCopy);
  Term = term_Create(fol_Not(),list_List(Term));
  fol_NormalizeVars(Term);
  Term = cnf_ObviousSimplifications(Term);
  if (flag_GetFlagValue(flag_CNFRENAMING)) {
    term_AddFatherLinks(Term);
    Term   = ren_Rename(Term,&SkolemPredicates,flag_GetFlagValue(flag_CNFPRENAMING), TRUE);
  }
  Term   = cnf_RemoveEquivImplFromFormula(Term);
  Term   = cnf_NegationNormalFormula(Term);
  Term   = cnf_AntiPrenex(Term);
  
  TermCopy = term_Copy(Term);
  TermCopy = cnf_SkolemFormula(TermCopy, &SkolemFunctions);
  TermCopy = cnf_DistributivFormula(TermCopy);

  IndexedClauses = cnf_MakeClauseList(TermCopy,FALSE,FALSE);
  term_Delete(TermCopy);

  
  for(Scan = IndexedClauses; !list_Empty(Scan); Scan = list_Cdr(Scan))
    clause_SetFlag((CLAUSE) list_Car(Scan), CONCLAUSE);

  EmptyClauses = red_SatUnit(SearchCopy, IndexedClauses, flag_GetFlagValue(flag_DOCPROOF));
  if (!list_Empty(EmptyClauses)) {
    Result = TRUE;
    list_DeleteWithElement(EmptyClauses, clause_Delete);
  }

  while(!list_Empty(prfs_UsableClauses(SearchCopy))) {
    prfs_MoveUsableWorkedOff(SearchCopy, (CLAUSE) list_Car(prfs_UsableClauses(SearchCopy)));
  }
  
  Dummy = list_Nil();
  if (flag_GetFlagValue(flag_DOCPROOF))
    Formulae2Clause = TRUE;
  else
    Formulae2Clause = FALSE;
  flag_SetFlagValue(flag_DOCPROOF, 0);
  ResultClauses = cnf_OptimizedSkolemization(SearchCopy, term_Copy(Term),
					     NULL, &Dummy, 
					     Symblist, Result, FALSE, NULL);

  if (Formulae2Clause == TRUE)
    flag_SetFlagValue(flag_DOCPROOF, 1);
  
  term_Delete(Term);
  list_Delete(SkolemPredicates);
  list_Delete(SkolemFunctions);
  prfs_Clean(SearchCopy);

  
  for(Scan=ResultClauses; !list_Empty(Scan); Scan=list_Cdr(Scan))
    clause_SetFlag((CLAUSE) list_Car(Scan), CONCLAUSE);

  return ResultClauses;
}

#ifdef DEBUG
void checkClauseListsConsistency(SHARED_INDEX ShIndex)

{
  LIST AllClauses, scan;

  AllClauses = clause_AllIndexedClauses(ShIndex);
  for(scan = AllClauses; scan != list_Nil(); scan = list_Cdr(scan)) {
    if (clause_GetFlag((CLAUSE) list_Car(scan), CONCLAUSE)) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In checkClauseListsConsistency: Clause is a CONCLAUSE.\n");
      misc_FinishErrorReport();
    } 
    if (clause_GetFlag((CLAUSE) list_Car(scan), BLOCKED)) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In checkClauseListsConsistency: Clause is BLOCKED.\n");
      misc_FinishErrorReport();
    }
  }
  list_Delete(AllClauses);
}
#endif

LIST cnf_SatUnit(PROOFSEARCH Search, LIST ClauseList) 

{
  CLAUSE Given;
  LIST   Scan, Derivables, EmptyClauses, BackReduced;
  NAT    n,Derived;
  Derived      = flag_GetFlagValue(flag_CNFPROOFSTEPS);
  EmptyClauses = list_Nil();
  ClauseList   = clause_ListSortWeighed(ClauseList);
  
  while (!list_Empty(ClauseList) && list_Empty(EmptyClauses)) {
    Given = (CLAUSE)list_NCar(&ClauseList);
    Given = red_CompleteReductionOnDerivedClause(Search, Given, 
						 flag_GetFlagValue(flag_DOCPROOF), red_ALL);
    if (Given) {
      if (clause_IsEmptyClause(Given))
        EmptyClauses = list_List(Given);
      else {
        
	BackReduced = red_BackReduction(Search, Given, 
					flag_GetFlagValue(flag_DOCPROOF), 
					red_USABLE);
	
        if (Derived != 0) {
	  Derivables = inf_BoundedDepthUnitResolution(Given, prfs_UsableSharingIndex(Search),FALSE);
	  Derivables = list_Nconc(Derivables, 
			      inf_BoundedDepthUnitResolution(Given, prfs_WorkedOffSharingIndex(Search),FALSE));
          n          = list_Length(Derivables);
          if (n > Derived)
            Derived = 0;
          else
            Derived = Derived - n;
        }
        else 
          Derivables = list_Nil();
	
        Derivables  = list_Nconc(BackReduced,Derivables);
        Derivables  = split_ExtractEmptyClauses(Derivables, &EmptyClauses);
	
	prfs_InsertUsableClause(Search, Given);
	
	for(Scan = Derivables; !list_Empty(Scan); Scan = list_Cdr(Scan))
	  ClauseList = clause_InsertWeighed((CLAUSE) list_Car(Scan), ClauseList);
	list_Delete(Derivables);
      }
    }
  }
  clause_DeleteClauseList(ClauseList);
  return EmptyClauses;
}
TERM cnf_DefTargetConvert(TERM Target, TERM ToTopLevel, 
			  TERM ToProveDef, 
			  LIST DefPredArgs, LIST TargetPredArgs,
			  LIST TargetPredVars, LIST VarsForTopLevel,
			  BOOL* LocallyTrue)


{
  TERM orterm, targettoprove;
  SYMBOL maxvar;                    
  LIST l1, l2;
  LIST freevars, vars;                    

  if(flag_GetFlagValue(flag_PAPPLYDEFS)) {
    fputs("\nTarget :\n", stdout);
    fol_PrettyPrint(Target);
  }
#ifdef DEBUG
  fol_CheckFatherLinks(Target);
#endif
  
  *LocallyTrue = FALSE;

  
  Target = cnf_RemoveImplFromFormulaPath(Target, ToTopLevel);
    
  
  Target = cnf_NegationNormalFormulaPath(Target, ToTopLevel);
    
  
  Target = cnf_AntiPrenexPath(Target, ToTopLevel);

  
  Target = cnf_MovePredicateVariablesUp(Target, ToTopLevel, VarsForTopLevel);

  
  Target = cnf_FlattenPath(Target, ToTopLevel);

  
  
  if(symbol_Equal(term_TopSymbol(Target), fol_All())) {
    targettoprove = term_Copy(term_SecondArgument(Target));
    orterm        = term_SecondArgument(Target);
  }
  else {
    targettoprove = term_Copy(Target);
    orterm        = Target;
  }

  
  if(symbol_Equal(term_TopSymbol(targettoprove), fol_Or())) {
    
    LIST arglist;
    arglist = term_ArgumentList(targettoprove);
    for(l1=arglist, l2=term_ArgumentList(orterm); !list_Empty(l1); 
	l1 = list_Cdr(l1), l2 = list_Cdr(l2)) {
      if (term_HasProperSuperterm(ToTopLevel, (TERM) list_Car(l2)) || (ToTopLevel == (TERM) list_Car(l2))) {
	arglist = list_PointerDeleteElementFree(arglist, list_Car(l1), term_Delete);
	break;
      }
    }
    term_RplacArgumentList(targettoprove, arglist);
    
    if (list_Length(term_ArgumentList(targettoprove)) == 0) {
      term_Delete(targettoprove);
      term_Delete(ToProveDef);
#ifdef DEBUG
      fol_CheckFatherLinks(Target);
#endif
      return Target;
    }
  }
  else {
    
    term_Delete(targettoprove);
    term_Delete(ToProveDef);
#ifdef DEBUG
    fol_CheckFatherLinks(Target);
#endif
    return Target;
  }

  
  maxvar = term_MaxVar(targettoprove);
  symbol_SetStandardVarCounter(maxvar);
  vars = fol_BoundVariables(ToProveDef);
  vars = list_DeleteDuplicates(vars, term_Equal);
  for(l1=vars; !list_Empty(l1); l1=list_Cdr(l1))
    term_ExchangeVariable(ToProveDef, term_TopSymbol(list_Car(l1)), symbol_CreateStandardVariable());
  list_Delete(vars);
  
  
  for(l1=DefPredArgs, l2=TargetPredArgs; !list_Empty(l1); l1=list_Cdr(l1), l2=list_Cdr(l2)) 
    term_ReplaceVariable(ToProveDef, term_TopSymbol((TERM) list_Car(l1)), (TERM) list_Car(l2));

  targettoprove = term_Create(fol_Not(), list_List(targettoprove));
  targettoprove = cnf_NegationNormalFormula(targettoprove);
  targettoprove = term_Create(fol_Implies(), 
			      list_Cons(targettoprove, list_List(ToProveDef)));

  
  
  
  freevars = fol_FreeVariables(targettoprove);
  list_NMapCar(freevars, (POINTER (*)())term_Copy);
  targettoprove = fol_CreateQuantifier(fol_All(), freevars, list_List(targettoprove));
  
  if(flag_GetFlagValue(flag_PAPPLYDEFS)) {
    fputs("\nConverted to :\n", stdout);
    fol_PrettyPrint(Target);
  }

  targettoprove = cnf_NegationNormalFormula(targettoprove);
  
  if(flag_GetFlagValue(flag_PAPPLYDEFS)) {
    fputs("\nToProve for this target :\n", stdout);
    fol_PrettyPrint(targettoprove);
  }
 
  *LocallyTrue = cnf_HaveProof(list_Nil(), targettoprove);
  
  term_Delete(targettoprove);

#ifdef DEBUG
      fol_CheckFatherLinks(Target);
#endif

  return Target;
}

TERM cnf_DefConvert(TERM Def, TERM FoundPredicate, TERM* ToProve)

{
  TERM orterm;

#ifdef DEBUG
  fol_CheckFatherLinks(Def);
#endif
  
  
  Def = cnf_RemoveImplFromFormulaPath(Def, FoundPredicate);  
  Def = cnf_NegationNormalFormulaPath(Def, FoundPredicate);  
  Def = cnf_AntiPrenex(Def);                                 
  term_AddFatherLinks(Def);
  Def = cnf_MoveEquivVariablesUp(Def, FoundPredicate);       
  Def = cnf_FlattenPath(Def, FoundPredicate);                

  
  if(symbol_Equal(term_TopSymbol(Def), fol_All())) {
    
    if(symbol_Equal(term_TopSymbol(term_SecondArgument(Def)), fol_Or()) 
       && (list_Length(term_ArgumentList(term_SecondArgument(Def))) == 1)) {
      TERM t;
      t = term_SecondArgument(Def);
      term_RplacSecondArgument(Def, term_FirstArgument(term_SecondArgument(Def)));
      term_Free(t);
      orterm = NULL;
      term_RplacSuperterm(term_SecondArgument(Def), Def);
    }
    else 
      orterm = term_SecondArgument(Def);
  }
  else {
    
    if(symbol_Equal(term_TopSymbol(Def), fol_Or())
       && (list_Length(term_ArgumentList(Def)) == 1)) {
      TERM t;
      t = Def;
      Def = term_FirstArgument(Def);
      term_Free(t);
      orterm = NULL;
      term_RplacSuperterm(term_SecondArgument(Def), Def);
    }
    else 
      orterm = Def;
  }

  
  if(orterm != (TERM) NULL) {
    TERM equiv;
    LIST args;

    equiv = (TERM) NULL;

    
    if(symbol_Equal(term_TopSymbol(orterm), fol_Equiv())) {
      equiv = orterm;
      *ToProve = NULL;
    }
    else {
      TERM t;
      
      args  = term_ArgumentList(orterm);
      equiv = term_Superterm(FoundPredicate);
      
      
      args = list_PointerDeleteElement(args, equiv);
      term_RplacArgumentList(orterm, args);
      
      
      *ToProve = term_Copy(orterm);
      
      
      
      *ToProve = term_Create(fol_Not(), list_List(*ToProve));
      *ToProve = cnf_NegationNormalFormula(*ToProve);
      term_AddFatherLinks(*ToProve);

      
      term_RplacTop(orterm, fol_Implies());
      t = term_Create(fol_Not(), 
		      list_List(term_Create(fol_Or(), 
					    term_ArgumentList(orterm))));
      term_RplacArgumentList(orterm, list_Cons(t, list_List(equiv)));
      Def = cnf_NegationNormalFormula(Def);
      term_AddFatherLinks(Def);
    }    
  }
  
#ifdef DEBUG
  fol_CheckFatherLinks(Def);
#endif
  return Def;
}

LIST cnf_HandleDefinition(PROOFSEARCH Search, LIST Pair, LIST Axioms, LIST Sorts, LIST Conjectures)

{
  TERM   definition, defpredicate, equivterm;
  
  BOOL   alwaysapplicable;     

  
  definition = list_PairSecond(Pair);
  
  
  defpredicate = (TERM) NULL;
  if(cnf_ContainsDefinition(definition, &defpredicate)) { 
    TERM toprove;
    LIST allformulae, scan;

    
    
    allformulae = list_Copy(Axioms);
    allformulae = list_Nconc(allformulae, list_Copy(Sorts));
    allformulae = list_Nconc(allformulae, list_Copy(Conjectures));
#ifdef DEBUG
    for(scan=allformulae; !list_Empty(scan); scan=list_Cdr(scan)) {
      if(!list_Empty((LIST) list_Car(scan))) {
	if(!term_IsTerm((TERM) list_PairSecond((LIST) list_Car(scan))))
	  fol_CheckFatherLinks((TERM) list_PairSecond((LIST) list_Car(scan)));
      }
    }
#endif
    
    
    if(flag_GetFlagValue(flag_PAPPLYDEFS)) {
      fputs("\nPredicate : ", stdout);
      symbol_Print(term_TopSymbol(defpredicate));
    }

    definition = cnf_DefConvert(definition, defpredicate, &toprove);
    if (toprove == NULL)
      alwaysapplicable = TRUE;
    else 
      alwaysapplicable = FALSE;

    prfs_SetDefinitions(Search, list_Cons(term_Copy(definition), 
					  prfs_Definitions(Search)));
    
    if(flag_GetFlagValue(flag_PAPPLYDEFS)) {
      if(alwaysapplicable == TRUE) {
	fputs("\nAlways Applicable     : ", stdout);
	fol_PrettyPrint(definition); 
      } 
    }

    
    equivterm     = term_SecondArgument(term_Superterm(defpredicate));
    
    
    scan=allformulae;
    while(!list_Empty(scan)) {
      BOOL localfound;
      LIST pair, targettermvars;
      
      
      pair = list_Car(scan);

      
      if((pair != NULL)
	 && (definition != (TERM) list_PairSecond(pair))) {
	TERM target, targetpredicate, totoplevel;
	LIST varsfortoplevel;
	target = (TERM) list_PairSecond(pair);
	targettermvars = varsfortoplevel = list_Nil();
	
	
	localfound = FALSE;
	if(alwaysapplicable == FALSE) {
	  if (cnf_ContainsPredicate(target, term_TopSymbol(defpredicate),
				    &targetpredicate, &totoplevel, &targettermvars, 
				    &varsfortoplevel)) {
	    TERM toprovecopy;
	    toprovecopy = term_Copy(toprove);
	    target = cnf_DefTargetConvert(target, totoplevel, toprovecopy,
					  term_ArgumentList(defpredicate),
					  term_ArgumentList(targetpredicate),
					  targettermvars, varsfortoplevel,
					  &localfound);
	    list_Delete(targettermvars);
	    list_Delete(varsfortoplevel);
	    targettermvars = varsfortoplevel = list_Nil();
	    
	    list_Rplacd(pair, (LIST) target);
	    if(localfound == TRUE)
	      list_Rplacd(pair, 
			  (LIST) cnf_ApplyDefinitionOnce(defpredicate, 
							 equivterm,
							 (TERM) list_PairSecond(pair),
							 targetpredicate));
	  }
	}
	else {
	  if(cnf_ContainsPredicate(target, term_TopSymbol(defpredicate), 
				   &targetpredicate, &totoplevel,
				   &targettermvars, 
				   &varsfortoplevel))
	    list_Rplacd(pair, (LIST) cnf_ApplyDefinitionOnce(defpredicate, 
							     equivterm,
							     (TERM) list_PairSecond(pair),
							     targetpredicate));
	  else
	    scan = list_Cdr(scan);
	  list_Delete(targettermvars);
	  list_Delete(varsfortoplevel);
	  targettermvars = varsfortoplevel = list_Nil();
	}
      }
      else
	scan = list_Cdr(scan);
    }
    list_Delete(allformulae);
    
    if(toprove != (TERM) NULL)
      term_Delete(toprove);
    list_Rplacd(Pair, (LIST) definition);
  }
 
  return Pair;
}

LIST cnf_ApplyDefinitionToClause(CLAUSE Clause, TERM Predicate, TERM Expansion)

{
  NAT i;
  BOOL changed;
  LIST args, scan, symblist;
  TERM clauseterm, argument;

  changed = FALSE;
  
  
  args = list_Nil();
  for(i=0; i<clause_Length(Clause); i++) {
    argument = clause_GetLiteralTerm(Clause, i); 
    args = list_Cons(term_Copy(argument), args);
  }
  clauseterm = term_Create(fol_Or(), args);

  for(scan=term_ArgumentList(clauseterm); !list_Empty(scan); scan=list_Cdr(scan)) {
    BOOL isneg;

    argument = (TERM) list_Car(scan);
    if(symbol_Equal(term_TopSymbol(argument), fol_Not())) {
      argument = term_FirstArgument(argument);
      isneg = TRUE;
    }
    else 
      isneg = FALSE;
      
    
    cont_StartBinding();
    if (unify_Match(cont_LeftContext(),
		    Predicate, argument)) {
      SUBST subst;
      TERM  newargument;
      subst = subst_ExtractMatcher();
      newargument = subst_Apply(subst, term_Copy(Expansion));
      subst_Free(subst);
      if(isneg == TRUE)
	newargument = term_Create(fol_Not(), list_List(newargument));
      term_Delete((TERM) list_Car(scan));
      list_Rplaca(scan, newargument);
      changed = TRUE;
    }      
    cont_BackTrack();
  }

  if (changed == TRUE) {
    
    LIST result;
    if(flag_GetFlagValue(flag_PAPPLYDEFS)) {
      fputs("\nClause before applying def :\n", stdout);
      clause_Print(Clause);
      fputs("\nPredicate :\n", stdout);
      fol_PrettyPrint(Predicate);
      fputs("\nExpansion :\n", stdout);
      fol_PrettyPrint(Expansion);
    }
    symblist   = list_Nil();
    clauseterm = cnf_Cnf(clauseterm, &symblist);
    result     = cnf_MakeClauseList(clauseterm, FALSE, FALSE);
    list_Delete(symblist);
    term_Delete(clauseterm);
    if(flag_GetFlagValue(flag_PAPPLYDEFS)) {
      LIST l;
      fputs("\nClauses derived by expanding definition :\n", stdout);
      for(l=result; !list_Empty(l); l=list_Cdr(l)) {
	clause_Print((CLAUSE) list_Car(l));
	fputs("\n", stdout);
      }
    }
    return result;
  }
  else {
    term_Delete(clauseterm);
    return list_Nil();
  }
}

BOOL cnf_PropagateSubstEquations(TERM StartTerm)

{
  LIST         Subequ, Escan;
  TERM         QuantorTerm, Equation, EquationTerm;
  SYMBOL       Variable;
  BOOL         Hit, Substitued;

#ifdef DEBUG
  if (fol_VarBoundTwice(StartTerm)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In cnf_PropagateSubstEquations: Variables of");
    misc_ErrorReport("\n input term are not normalized.");
    misc_FinishErrorReport();
  }
#endif

  Substitued = FALSE;

  Subequ = fol_GetSubstEquations(StartTerm);
  for (Escan = Subequ; !list_Empty(Escan); Escan = list_Cdr(Escan)) {
    Hit          = FALSE;
    Equation     = list_Car(Escan);
    Variable     = symbol_Null();
    QuantorTerm  = term_Null();
    EquationTerm = term_Null();

    if (term_IsVariable(term_FirstArgument(Equation)) &&
	!term_ContainsVariable(term_SecondArgument(Equation),
			       term_TopSymbol(term_FirstArgument(Equation)))) {

      Variable    = term_TopSymbol(term_FirstArgument(Equation));
      QuantorTerm = fol_GetBindingQuantifier(Equation, Variable);

      if (fol_PolarCheck(Equation, QuantorTerm) == TRUE) 
	Hit          = TRUE;
        EquationTerm = term_SecondArgument(Equation);
    }
    if (Hit == FALSE && term_IsVariable(term_SecondArgument(Equation)) &&
	!term_ContainsVariable(term_FirstArgument(Equation),
			       term_TopSymbol(term_SecondArgument(Equation)))) {

      Variable = term_TopSymbol(term_SecondArgument(Equation));
      QuantorTerm = fol_GetBindingQuantifier(Equation, Variable);

      if (fol_PolarCheck(Equation, QuantorTerm) == TRUE)
        Hit = TRUE;
        EquationTerm = term_FirstArgument(Equation);
    }
    if (Hit == TRUE) {
      fol_DeleteQuantifierVariable(QuantorTerm,Variable);
      term_ReplaceVariable(StartTerm, Variable, EquationTerm);
      term_AddFatherLinks(StartTerm);
      fol_SetTrue(Equation);
      Substitued = TRUE;
    }   
  }

  list_Delete(Subequ);

  return Substitued;
}
