




 





















#include "cnf.h"
#include "defs.h"
#include "foldfg.h"

static void def_DeleteFromClauses(DEF);
static void def_DeleteFromTerm(DEF);

DEF def_CreateFromClauses(TERM ExpTerm, TERM PredTerm, LIST Clauses, LIST Lits, BOOL Con)

{
  DEF  result;

#ifdef DEBUG
  if (!term_IsTerm(ExpTerm) || !term_IsTerm(PredTerm)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In def_CreateFromClause: Illegal input.");
    misc_FinishErrorReport();
  }
  if (list_Empty(Clauses)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In def_CreateFromClause: No parent clause given.");
    misc_FinishErrorReport();
  }
#endif

  result                   = (DEF) memory_Malloc(sizeof(DEF_NODE));
  result->expansion        = ExpTerm;
  result->predicate        = PredTerm;
  result->toprove          = (TERM) NULL;
  result->parentclauses    = list_PairCreate(Clauses, Lits);
  result->label            = (char*)NULL;
  result->conjecture       = Con;

  return result;
}

DEF def_CreateFromTerm(TERM ExpTerm, TERM PredTerm, TERM ToProve, char* Label)

{
  DEF  result;

#ifdef DEBUG
  if (!term_IsTerm(ExpTerm) || !term_IsTerm(PredTerm)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In def_CreateFromTerm: Illegal input.");
    misc_FinishErrorReport();
  }  
  if (Label ==  NULL) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In def_CreateFromTerm: No parent clause given.");
    misc_FinishErrorReport();
  }
#endif

  result                   = (DEF) memory_Malloc(sizeof(DEF_NODE));
  result->expansion        = ExpTerm;
  result->predicate        = PredTerm;
  result->toprove          = ToProve;
  result->parentclauses    = list_PairCreate(list_Nil(), list_Nil());
  result->label            = Label;
  result->conjecture       = FALSE;

  return result;
}

static void def_DeleteFromClauses(DEF D) 
  
{
#ifdef DEBUG
  if (!term_IsTerm(def_Expansion(D)) || !term_IsTerm(def_Predicate(D))) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In def_DeleteFormClauses: Illegal input.");
    misc_FinishErrorReport();
  }
#endif

  
  term_Delete(def_Expansion(D));
  term_Delete(def_Predicate(D));
  list_Delete(def_ClauseNumberList(D));
  list_Delete(def_ClauseLitsList(D));
  list_PairFree(D->parentclauses);

  memory_Free(D, sizeof(DEF_NODE));
}

static void def_DeleteFromTerm(DEF D) 
  
{
#ifdef DEBUG
  if (!term_IsTerm(def_Expansion(D)) || !term_IsTerm(def_Predicate(D))) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In def_DeleteFromTerm: Illegal input.");
    misc_FinishErrorReport();
  }
#endif

  
  term_Delete(def_Expansion(D));
  term_Delete(def_Predicate(D));
  if(def_ToProve(D) != (TERM) NULL)
    term_Delete(def_ToProve(D));
  list_PairFree(D->parentclauses);
  memory_Free(D, sizeof(DEF_NODE));
}

void def_Delete(DEF D)
  
{
  if (!list_Empty(def_ClauseNumberList(D)))
    def_DeleteFromClauses(D);
  else
    def_DeleteFromTerm(D);
}

int def_PredicateOccurrences(TERM Term, SYMBOL P)
			       
{
  
  if(fol_IsQuantifier(term_TopSymbol(Term)))
    return def_PredicateOccurrences(term_SecondArgument(Term), P);

  
  if(fol_IsJunctor(term_TopSymbol(Term))
     || symbol_Equal(term_TopSymbol(Term),fol_Not())){
    LIST scan;
    int count;
    count = 0;
    for(scan=term_ArgumentList(Term); !list_Empty(scan); scan=list_Cdr(scan)) {
      count += def_PredicateOccurrences((TERM) list_Car(scan), P);
      
      if (count > 1)
	return count;
    }
    return count;
  }
  
  if(symbol_Equal(term_TopSymbol(Term), P))
    return 1;
  return 0;
} 

LIST def_ExtractDefsFromTerm(TERM Term, char* Label) 

{
  TERM andterm;
  BOOL found;
  int  pol;
  LIST univars, termlist, defslist, scan;
  
  

  andterm = Term;
  found   = FALSE;
  pol     = 1;
  univars = list_Nil();

  
  while(found == FALSE) {
    if ((symbol_Equal(term_TopSymbol(andterm), fol_All()) && (pol == 1))
	|| (symbol_Equal(term_TopSymbol(andterm), fol_Exist()) && (pol == -1))) {
      univars = list_Nconc(univars, list_Copy(fol_QuantifierVariables(andterm)));
      andterm = term_SecondArgument(andterm);
    }
    else {
      if (symbol_Equal(term_TopSymbol(andterm), fol_Not())) {
	pol = -pol;
	andterm = term_FirstArgument(andterm);
      }
      else
	found = TRUE;
    }
  }

  termlist = list_Nil();
  
  if ((symbol_Equal(term_TopSymbol(andterm), fol_And()) && (pol == 1))
      || (symbol_Equal(term_TopSymbol(andterm), fol_Or()) && (pol == -1))) {
    LIST l;
    
    
    andterm = cnf_Flatten(term_Copy(andterm), term_TopSymbol(andterm));
    for(l=term_ArgumentList(andterm); !list_Empty(l); l=list_Cdr(l)) {
      TERM newterm;
      newterm = fol_CreateQuantifierAddFather(fol_All(), list_CopyWithElement(univars, (POINTER (*)()) term_Copy), 
					      list_List((TERM) list_Car(l)));
      termlist = list_Cons(newterm, termlist);
    }
    
    list_Delete(term_ArgumentList(andterm));
    term_Free(andterm);
  }
  else
    termlist = list_List(term_Copy(Term));
  
  list_Delete(univars);

  
  defslist = list_Nil();
  for(scan=termlist; !list_Empty(scan); scan=list_Cdr(scan)) {
    TERM cand;
    TERM foundpred, toprove;

    
    cand = (TERM) list_Car(scan);
    term_AddFatherLinks(cand);

    if(cnf_ContainsDefinition(cand, &foundpred)) {
      DEF def;
      cand = cnf_DefConvert(cand, foundpred, &toprove);
      def  = def_CreateFromTerm(term_Copy(term_SecondArgument(term_Superterm(foundpred))), 
				term_Copy(foundpred), toprove, Label);

      if (def_PredicateOccurrences(cand, term_TopSymbol(foundpred)) > 1)
	def_RemoveAttribute(def, PREDOCCURONCE);
      else 
	def_AddAttribute(def, PREDOCCURONCE);
      if (symbol_Equal(term_TopSymbol(foundpred), fol_Equality()))
	def_AddAttribute(def, ISEQUALITY);
      else
	def_RemoveAttribute(def, ISEQUALITY);

      defslist = list_Cons(def, defslist);
    }
    term_Delete(cand);
  }
  
  list_Delete(termlist);
  return defslist;
}

void def_ExtractDefsFromClauselist(PROOFSEARCH Search, LIST Clauselist)

{
  LIST scan, defslist;

  defslist = list_Nil();
  for(scan=Clauselist; !list_Empty(scan); scan=list_Cdr(scan)) {
    CLAUSE Clause;
    NAT index;
    LIST pair;

    Clause = (CLAUSE) list_Car(scan);
    
    if(clause_ContainsPotPredDef(Clause, &index, &pair)) {
      LIST l, compl, compllits;
      BOOL done;

      compl = compllits = list_Nil();
      done = FALSE;
      
      
      for(l=Clauselist; !list_Empty(l) && (done == FALSE); 
	  l=list_Cdr(l)) {
	int predindex;
	if (clause_IsPartOfDefinition((CLAUSE) list_Car(l), 
				      clause_GetLiteralTerm(Clause, index), 
				      &predindex, pair)) {
	  compl     = list_Cons(list_Car(l), compl);
	  compllits = list_Cons((POINTER) predindex, compllits);
	  
	  if(list_Empty(list_PairFirst(pair))
	     && list_Empty(list_PairSecond(pair)))
	    done = TRUE;
	}
      }
      
       
      if(done == TRUE) {
	LIST l2, clausenumbers, args;
	DEF  def;
	NAT  i;
	TERM defterm, predterm;
	BOOL con;

	clausenumbers = list_Nil();
	con           = clause_GetFlag(Clause, CONCLAUSE);

	for(l2=compl; !list_Empty(l2); l2=list_Cdr(l2)) {
	  clausenumbers = list_Cons((POINTER) clause_Number((CLAUSE) list_Car(l2)), 
				    clausenumbers);	    
	  if(clause_GetFlag((CLAUSE) list_Car(l2), CONCLAUSE))
	    con = TRUE;
	}
	clausenumbers = list_Cons((POINTER) clause_Number(Clause), 
				  clausenumbers);
	compllits = list_Cons((POINTER) index, compllits);

	
	predterm = term_Copy(clause_GetLiteralTerm(Clause, index));
	args     = list_Nil();
	for(i=0; i<clause_Length(Clause); i++) 
	  if(i!=index)
	    args = list_Cons(term_Copy(clause_GetLiteralTerm(Clause, i)), args);
	defterm  = term_CreateAddFather(fol_Or(), args);
	
	defterm  = term_Create(fol_Not(), list_List(defterm));
	defterm  = cnf_NegationNormalFormula(defterm);
	def      = def_CreateFromClauses(defterm, predterm, clausenumbers, compllits, con);
	defslist = list_Cons(def, defslist);
	if(flag_GetFlagValue(flag_PAPPLYDEFS)) {
	  fputs("\nNew definition found :", stdout);
	  def_Print(def);
	}
      }
      else {
	list_Delete(compllits);
	list_Delete(list_PairSecond(pair));
	list_Delete(list_PairFirst(pair));
      }
      list_Delete(compl);
      list_PairFree(pair);
    }
  }

  if(flag_GetFlagValue(flag_PAPPLYDEFS)) {
    if (!list_Empty(defslist)) {
      fputs("\nFound definitions :\n", stdout);
      for(scan=defslist; !list_Empty(scan); 
	  scan=list_Cdr(scan)) {
	def_Print((DEF) list_Car(scan));
	fputs("\n---\n", stdout);
      }
    }
  }

  for(scan=defslist; !list_Empty(scan); scan=list_Cdr(scan))
    symbol_AddProperty(term_TopSymbol(def_Predicate((DEF) list_Car(scan))), ISDEF);
  
  prfs_SetDefinitions(Search, list_Nconc(prfs_Definitions(Search), defslist));
}

TERM def_ApplyDefToTermOnce(DEF Def, TERM Term, BOOL* Complete)

{
  TERM newtarget, oldtarget, targetpredicate, totoplevel, toprove;
  LIST targettermvars, varsfortoplevel;
  BOOL applicable;

  oldtarget = Term;
  *Complete = TRUE;

  while (1 == 1) {
    newtarget = term_Copy(oldtarget);
    term_AddFatherLinks(newtarget);
    targettermvars = varsfortoplevel = list_Nil();

    if (cnf_ContainsPredicate(newtarget, term_TopSymbol(def_Predicate(Def)),
			      &targetpredicate, &totoplevel, &targettermvars, 
			      &varsfortoplevel)) {
      *Complete  = FALSE;
      applicable = FALSE;
      
      if (term_Equal(def_ToProve(Def), term_Null())) { 
	applicable = TRUE;
	newtarget = cnf_ApplyDefinitionOnce(def_Predicate(Def), term_Copy(def_Expansion(Def)),
					 newtarget, targetpredicate);
	if (oldtarget != Term)
	  term_Delete(oldtarget);
	oldtarget = newtarget;
      	list_Delete(targettermvars);
	list_Delete(varsfortoplevel); 	
      }
      else {
	toprove = term_Copy(def_ToProve(Def));
	newtarget = cnf_DefTargetConvert(newtarget, totoplevel, toprove,
				      term_ArgumentList(def_Predicate(Def)),
				      term_ArgumentList(targetpredicate),
				      targettermvars, varsfortoplevel,
				      &applicable);
	list_Delete(targettermvars);
	list_Delete(varsfortoplevel); 	
	if(applicable == TRUE) {
	  newtarget = cnf_ApplyDefinitionOnce(def_Predicate(Def), term_Copy(def_Expansion(Def)),
					   newtarget, targetpredicate);
	  if (oldtarget != Term)
	    term_Delete(oldtarget);
	  oldtarget = newtarget;
	}
	else {
	  
	  term_Delete(newtarget);
	  if(oldtarget == Term)
	    return NULL;
	  else {
	    oldtarget = cnf_ObviousSimplifications(oldtarget);
	    return oldtarget;
	  }
	}
      }
    }
    else {
      *Complete = TRUE;
      
      term_Delete(newtarget);
      
      if(oldtarget == Term)
	return NULL;
      else {
	oldtarget = cnf_ObviousSimplifications(oldtarget);
	return oldtarget;
      }
    }
  }
  return NULL; 
}
    
TERM def_ApplyDefToTermExhaustive(PROOFSEARCH Search, TERM Term)

{
  TERM oldterm, newterm;
  BOOL done, complete;
  
  done = FALSE;
  oldterm = Term;

  while (done == FALSE) {
    LIST l;
    done = TRUE;
    
    for(l=prfs_Definitions(Search); !list_Empty(l); l=list_Cdr(l)) {
      newterm = def_ApplyDefToTermOnce((DEF) list_Car(l), oldterm, &complete);
      if (newterm != NULL) {
	if(oldterm != Term)
	  term_Delete(oldterm);
	oldterm = newterm;
	done = FALSE;
      }
    }
  }
  if (oldterm == Term)
    return NULL;
  else
    return oldterm;
}

LIST def_ApplyDefToClauseOnce(DEF Def, CLAUSE Clause)

{
  LIST result, l;
  
  result = list_List(Clause);

  for(l=result; !list_Empty(l); l=list_Cdr(l)) {
    if(clause_ContainsSymbol((CLAUSE) list_Car(l), term_TopSymbol(def_Predicate(Def)))) {
      result = list_Nconc(result, cnf_ApplyDefinitionToClause((CLAUSE) list_Car(l), 
							      def_Predicate(Def),
							      def_Expansion(Def)));
      
      if((CLAUSE) list_Car(l) != Clause)
	clause_Delete((CLAUSE) list_Car(l));
      list_Rplaca(l, NULL);
    }
  }
  result = list_PointerDeleteElement(result, NULL);

  
  if (!list_Empty(result))
    if (list_First(result) == Clause)
      result = list_Pop(result);
  
  for(l=result; !list_Empty(l); l=list_Cdr(l)) {
    CLAUSE c;
    c = (CLAUSE) list_Car(l);
    if (def_Conjecture(Def) == TRUE)
      clause_SetFlag((CLAUSE) list_Car(l), CONCLAUSE);
    clause_SetFromDefApplication(c);
    clause_SetParentClauses(c, list_Cons((POINTER) clause_Number(Clause), 
					 list_Copy(def_ClauseNumberList(Def))));
    
    clause_SetParentLiterals(c, list_Cons((POINTER) 0, 
					  list_Copy(def_ClauseLitsList(Def))));
  }
  return result;
}

LIST def_ApplyDefToClauseExhaustive(PROOFSEARCH Search, CLAUSE Clause)

{
  LIST newclauses, scan, result;
  CLAUSE orig;
  orig = clause_Copy(Clause);
  newclauses = list_List(orig);
  result = list_Nil();

  while (!list_Empty(newclauses)) {
    
    LIST l, nextlist;

    
    nextlist = list_Nil();

    for(l=newclauses; !list_Empty(l); l=list_Cdr(l)) {
      LIST clauses;
      CLAUSE c;
      
      c = (CLAUSE) list_Car(l);

      

      
      
      clauses  = list_Nil();

      for(scan=prfs_Definitions(Search); !list_Empty(scan); scan=list_Cdr(scan))
	clauses = list_Nconc(clauses, def_ApplyDefToClauseOnce((DEF) list_Car(scan), c));
      
      
      if (!list_Empty(clauses)) {
	
	if (c != Clause) {
	  if(flag_GetFlagValue(flag_DOCPROOF)) {
	    prfs_InsertDocProofClause(Search, c);
	  }
	  else
	    clause_Delete(c);
	}
	nextlist = list_Nconc(nextlist, clauses);
      }
      else {
	
	
	if(c != Clause)
	  result = list_Cons(c, result);
      }
    }
    list_Delete(newclauses);
    newclauses = nextlist;
  }
      
  return result;
}
  

void def_Print(DEF D)

{
  LIST scan, scan2;
  fputs("\n\nAtom: ", stdout);
  fol_PrettyPrint(def_Predicate(D));
  fputs("\nExpansion: \n", stdout);
  fol_PrettyPrint(def_Expansion(D));
  if(!list_Empty(def_ClauseNumberList(D))) {
    fputs("\nParent clauses: ", stdout);
    for(scan=def_ClauseNumberList(D), scan2=def_ClauseLitsList(D); 
	!list_Empty(scan); scan=list_Cdr(scan), scan2=list_Cdr(scan2)) 
      printf("%d.%d ", (NAT) list_Car(scan), (NAT) list_Car(scan2));
    if(D->conjecture == TRUE)
      fputs("\nDerived from conjecture clauses.", stdout);
    else
      fputs("\nNot derived from conjecture clauses.", stdout);
  }
  else {
    printf("\nLabel: %s", def_Label(D));
    fputs("\nGuard:\n", stdout);
    if(def_ToProve(D) != NULL) 
      fol_PrettyPrint(def_ToProve(D));
    else
      fputs("Nothing.", stdout);
  }

  printf("\nAttributes: ");
  if (def_HasAttribute(D, ISEQUALITY) || def_HasAttribute(D, PREDOCCURONCE)) {
    if (def_HasAttribute(D, ISEQUALITY))
      printf(" Equality ");
    if (def_HasAttribute(D, PREDOCCURONCE))
      printf(" No Multiple Occurrences ");
  }
  else {
    printf(" None ");
  }    
} 

LIST def_ApplyDefToClauselist(PROOFSEARCH Search, DEF Def, 
			      LIST Clauselist, BOOL Destructive)

{
  LIST l, newclauses, allnew;

  allnew = list_Nil();
  for(l=Clauselist; !list_Empty(l); l=list_Cdr(l)) {
    newclauses = def_ApplyDefToClauseOnce(Def, (CLAUSE) list_Car(l));
    
    if((Destructive == TRUE)
       && (!list_Empty(newclauses))) {
      if(flag_GetFlagValue(flag_DOCPROOF))
	prfs_InsertDocProofClause(Search, (CLAUSE) list_Car(l));
      else
	clause_Delete((CLAUSE) list_Car(l));
      list_Rplaca(l, NULL);
    }
    allnew = list_Nconc(allnew, newclauses);
  }
  if(Destructive == TRUE)
    Clauselist = list_PointerDeleteElement(Clauselist, NULL);


  if(flag_GetFlagValue(flag_PAPPLYDEFS)) {
    if(!list_Empty(allnew)) {
      fputs("\nNew clauses after applying definitions : \n", stdout);
      clause_ListPrint(allnew);
    }
  }

  Clauselist = list_Nconc(Clauselist, allnew);
  return Clauselist;
}

LIST def_ApplyDefToTermlist(DEF Def, LIST Termlist, BOOL* Complete, 
			    BOOL Destructive)

{
  LIST l, newterms;
  newterms = list_Nil();

  *Complete = TRUE;
  for(l=Termlist; !list_Empty(l); l=list_Cdr(l)) {
    TERM newterm;
    TERM oldterm;
    BOOL complete;
    oldterm = (TERM) list_PairSecond((LIST) list_Car(l));
    newterm = def_ApplyDefToTermOnce(Def, oldterm, &complete);
    if (complete == FALSE)
      *Complete = FALSE;
    
    if(newterm != NULL) {
      newterms = list_Cons(list_PairCreate(NULL, newterm),newterms);
      if(Destructive == TRUE) {
	
	term_Delete((TERM) list_PairSecond((LIST) list_Car(l)));
	if ((char*) list_PairFirst((LIST) list_Car(l)) != NULL)
	  memory_Free((char*) list_PairFirst((LIST) list_Car(l)), 
		      strlen((char*) list_PairFirst((LIST) list_Car(l))) + 1);
	list_PairFree((LIST) list_Car(l));
	list_Rplaca(l, NULL);
      }
    }
  }
  Termlist = list_PointerDeleteElement(Termlist, NULL);

  if(flag_GetFlagValue(flag_PAPPLYDEFS)) {
    if (!list_Empty(newterms)) {
      fputs("\n\nNew terms after applying definitions : \n", stdout);
      for(l=newterms; !list_Empty(l); l=list_Cdr(l)) {
	fputs("\n", stdout);
	fol_PrettyPrint((TERM) list_PairSecond((LIST) list_Car(l)));
      }
    }
  }
  
  Termlist = list_Nconc(Termlist, newterms);
  return Termlist;
}

void def_ExtractDefsFromTermlist(PROOFSEARCH Search, LIST Axioms, LIST Conj)

{ 
  LIST l, deflist;
  deflist = list_Nil();

  for(l=Axioms; !list_Empty(l); l=list_Cdr(l)) {
    deflist = list_Nconc(deflist, 
			 def_ExtractDefsFromTerm((TERM) list_PairSecond((LIST) list_Car(l)), 
						 (char*) list_PairFirst((LIST) list_Car(l))));
  }
  for(l=Conj; !list_Empty(l); l=list_Cdr(l)) {
    deflist = list_Nconc(deflist, 
			 def_ExtractDefsFromTerm((TERM) list_PairSecond((LIST) list_Car(l)), 
						 (char*) list_PairFirst((LIST) list_Car(l))));
  }
  for(l=deflist; !list_Empty(l); l=list_Cdr(l))
    symbol_AddProperty(term_TopSymbol(def_Predicate((DEF) list_Car(l))), ISDEF);

  prfs_SetDefinitions(Search, list_Nconc(prfs_Definitions(Search), deflist));  

  if(flag_GetFlagValue(flag_PAPPLYDEFS)) {
    if (!list_Empty(deflist)) {
      fputs("\nFound definitions :\n", stdout);
      for(l=prfs_Definitions(Search); !list_Empty(l); l=list_Cdr(l)) {
	def_Print((DEF) list_Car(l));
	fputs("\n---\n", stdout);
      }
    }
  }
}

LIST def_FlattenWithOneDefinition(PROOFSEARCH Search, DEF Def)

{
  LIST newdefinitions;
  newdefinitions = list_Nil();
  
  if (def_ToProve(Def) == NULL) {
    LIST definitions, l;
    definitions    = prfs_Definitions(Search);

    for(l=definitions; !list_Empty(l); l=list_Cdr(l)) {
      DEF d;
      d = (DEF) list_Car(l);
      
      if (d != Def) {
	
	if(term_ContainsSymbol(def_Expansion(d), term_TopSymbol(def_Predicate(Def)))) {
	  
	  if (!term_ContainsSymbol(def_Expansion(Def), term_TopSymbol(def_Predicate(d)))) {
	    TERM newexpansion;
	    BOOL complete;
	    DEF newdef;
	    newexpansion = def_ApplyDefToTermOnce(Def, def_Expansion(d), &complete);
	    
	    newdef = def_CreateFromTerm(newexpansion, term_Copy(def_Predicate(d)), 
								term_Copy(def_ToProve(d)), def_Label(d));
	    newdefinitions = list_Cons(newdef, newdefinitions);
	  }
	}
      }
      
    }
  }
  return newdefinitions;
}


void def_FlattenWithOneDefinitionDestructive(PROOFSEARCH Search, DEF Def)

{
  if (def_ToProve(Def) == NULL) {
    LIST definitions, l;
    definitions = prfs_Definitions(Search);
    
    for(l=definitions; !list_Empty(l); l=list_Cdr(l)) {
      DEF d;
      d = (DEF) list_Car(l);
      
      if (d != Def) {
	
	if(term_ContainsSymbol(def_Expansion(d), term_TopSymbol(def_Predicate(Def)))) {
	  
	  if (!term_ContainsSymbol(def_Expansion(Def), term_TopSymbol(def_Predicate(d)))) {
	    TERM newexpansion;
	    BOOL complete;
	    newexpansion = def_ApplyDefToTermOnce(Def, def_Expansion(d), &complete);
	    term_Delete(def_Expansion(d));
	    def_RplacExp(d, newexpansion);
	  }
	  else {
	    symbol_RemoveProperty(term_TopSymbol(def_Predicate(d)), ISDEF);
	    def_Delete(d);
	    list_Rplaca(l, NULL);
	  }
	}
      }
      else {
	
	list_Rplaca(l, NULL);
      }
    }
    symbol_RemoveProperty(term_TopSymbol(def_Predicate(Def)), ISDEF);
    def_Delete(Def);
    definitions = list_PointerDeleteElement(definitions, NULL);
    prfs_SetDefinitions(Search, definitions);
  }
}

void def_FlattenWithOneDefinitionSemiDestructive(PROOFSEARCH Search, DEF Def)

{
  if (def_ToProve(Def) == NULL) {
    LIST definitions, l;
    definitions = prfs_Definitions(Search);
    
    for(l=definitions; !list_Empty(l); l=list_Cdr(l)) {
      DEF d;
      d = (DEF) list_Car(l);
      
      if (d != Def) {
	
	if(term_ContainsSymbol(def_Expansion(d), term_TopSymbol(def_Predicate(Def)))) {
	  
	  if (!term_ContainsSymbol(def_Expansion(Def), term_TopSymbol(def_Predicate(d)))) {
	    TERM newexpansion;
	    BOOL complete;
	    newexpansion = def_ApplyDefToTermOnce(Def, def_Expansion(d), &complete);
	    term_Delete(def_Expansion(d));
	    def_RplacExp(d, newexpansion);
	  }
	  else {
	    symbol_RemoveProperty(term_TopSymbol(def_Predicate(d)), ISDEF);
	    def_Delete(d);
	    list_Rplaca(l, NULL);
	  }
	}
      }
    }
  }
}

void def_FlattenDefinitionsDestructive(PROOFSEARCH Search)

{
  LIST l;
  for(l=prfs_Definitions(Search); !list_Empty(l); l=list_Cdr(l)) {
    DEF d;
    d = (DEF) list_Car(l);
    fol_PrettyPrintDFG(def_Predicate(d));
    if (d != NULL)
      def_FlattenWithOneDefinitionSemiDestructive(Search, d);
  }
  prfs_SetDefinitions(Search, list_PointerDeleteElement(prfs_Definitions(Search), NULL));
}

LIST def_GetTermsForProof(TERM Term, TERM SubTerm, int Polarity)

{
  TERM SuperTerm, AddToList;
  SYMBOL Top;
  LIST Scan1, NewList;

  if (Term == SubTerm){
    return list_Nil();
  }

  SuperTerm = term_Superterm(SubTerm);
  Top       = term_TopSymbol(SuperTerm);
  NewList   = list_Nil();
  AddToList = term_Null();


  if (symbol_Equal(Top, fol_Not()))
    return def_GetTermsForProof(Term, SuperTerm, -1*Polarity);

  if (symbol_Equal(Top, fol_Or()) && Polarity == 1) {
    
    for (Scan1 = term_ArgumentList(SuperTerm); !list_Empty(Scan1); Scan1 = list_Cdr(Scan1)) {
      if (!term_HasPointerSubterm(list_Car(Scan1), SubTerm))
	NewList = list_Cons(list_Car(Scan1), NewList);
    }
    if (list_Length(NewList) == 1) {
      AddToList = term_Create(fol_Not(), NewList);
    }
    else{
      AddToList = term_Create(fol_Or(), NewList);
      AddToList = term_Create(fol_Not(), list_List(AddToList));
    }
    return list_Cons(AddToList, def_GetTermsForProof(Term, SuperTerm, Polarity));
  }
  if (symbol_Equal(Top, fol_And()) && Polarity == -1) {
    
    for (Scan1 = term_ArgumentList(SuperTerm); !list_Empty(Scan1); Scan1 = list_Cdr(Scan1)) {
      if (!term_HasPointerSubterm(list_Car(Scan1), SubTerm))
	NewList = list_Cons(list_Car(Scan1), NewList);
    }
    if (list_Length(NewList) == 1) {
      AddToList = term_Copy(list_Car(NewList));
      list_Delete(NewList);
    }
    else {
      AddToList = term_Create(fol_And(), NewList);
    }
    return list_Cons(AddToList, def_GetTermsForProof(Term, SuperTerm, Polarity));
  }
  if (symbol_Equal(Top, fol_Implies())) {
    if (term_HasPointerSubterm(term_SecondArgument(SuperTerm), SubTerm) && Polarity == 1) {
      AddToList = term_Copy(term_FirstArgument(SuperTerm));
      return list_Cons(AddToList, def_GetTermsForProof(Term, SuperTerm, Polarity));
    }
    if (term_HasPointerSubterm(term_FirstArgument(SuperTerm), SubTerm) && Polarity == -1) {
      AddToList = term_Copy(term_SecondArgument(SuperTerm));
      AddToList = term_Create(fol_Not(), list_List(AddToList));
      return list_Cons(AddToList, def_GetTermsForProof(Term, SuperTerm, -1*Polarity));
    }
  }
  if (symbol_Equal(Top, fol_Implied())) { 
    if (term_HasPointerSubterm(term_SecondArgument(SuperTerm), SubTerm) && Polarity == -1) {
      AddToList = term_Copy(term_FirstArgument(SuperTerm));
      AddToList = term_Create(fol_Not(), list_List(AddToList));
      return list_Cons(AddToList, def_GetTermsForProof(Term, SuperTerm, -1*Polarity)); 
    }
    if (term_HasPointerSubterm(term_FirstArgument(SuperTerm), SubTerm) && Polarity == 1) {
      AddToList = term_Copy(term_SecondArgument(SuperTerm));
      return list_Cons(AddToList, def_GetTermsForProof(Term, SuperTerm, Polarity));
    }
  }
  if (symbol_Equal(Top, fol_All())) {
    return def_GetTermsForProof(Term, SuperTerm, Polarity);
  }
  if (symbol_Equal(Top, fol_Exist())) {
    return def_GetTermsForProof(Term, SuperTerm, Polarity);
  }

  

  return list_Nil();
  
}

BOOL def_FindProofForGuard(TERM Term, TERM Atom, TERM Guard)

{
  BOOL LocallyTrue;
  TERM ToProve;
  LIST ArgList, FreeVars, Scan, Scan1;

  ArgList   = list_Nil();
  FreeVars  = list_Nil();
  ToProve   = term_Null();

  term_AddFatherLinks(Term);
  ArgList = def_GetTermsForProof(Term, Atom, 1);

  if (!list_Empty(ArgList)) {
    ToProve  = term_Create(fol_And(), ArgList);
    ArgList  = list_List(term_Copy(Guard));
    ArgList  = list_Cons(ToProve, ArgList);
    ToProve  = term_Create(fol_Implies(), ArgList);
    Scan1    = fol_FreeVariables(ToProve);
    for(Scan = Scan1; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      if (fol_VarOccursFreely(list_Car(Scan), ToProve)) {
	FreeVars = list_Cons(term_Copy(list_Car(Scan)), FreeVars);
      }
    }
    list_Delete(Scan1);
    ToProve  = fol_CreateQuantifier(fol_All(), FreeVars, list_List(ToProve));

    
    printf("\nToProve: \n"); fol_PrettyPrintDFG(ToProve);
    LocallyTrue = cnf_HaveProof(list_Nil(), ToProve);
    term_Delete(ToProve);

    if(LocallyTrue)
      return TRUE;
  }
  else{
    list_DeleteWithElement(ArgList, term_Delete);
  }

  return FALSE;
}

LIST def_ApplyDefinitionToTermList(LIST Defs, LIST Terms, int Applics)

{
  TERM    ActTerm;      
  TERM    DefPredicate; 
  TERM    Expansion;    
  TERM    Target;       
  LIST    TargetList, Scan1, Scan2, Scan3, BoundVars;
  BOOL    Apply;

  Apply      = TRUE;
  TargetList = list_Nil();
  BoundVars  = list_Nil();

  while (Apply && Applics != 0) {
    Apply      = FALSE;
    for (Scan1=Defs; !list_Empty(Scan1) && Applics != 0; Scan1=list_Cdr(Scan1)) {
      if (!def_HasGuard(list_Car(Scan1))) {                         
	DefPredicate = term_Copy(def_Predicate(list_Car(Scan1)));

	

	for (Scan2=Terms; !list_Empty(Scan2) && Applics != 0; Scan2=list_Cdr(Scan2)) {
	  ActTerm    = list_PairSecond(list_Car(Scan2));
	  TargetList = term_FindAllAtoms(ActTerm, term_TopSymbol(DefPredicate));

	  

	  for (Scan3=TargetList; !list_Empty(Scan3) && Applics != 0; Scan3=list_Cdr(Scan3)) {
	    Target    = list_Car(Scan3);

	    

	    cont_StartBinding();
	    if (unify_Match(cont_LeftContext(), DefPredicate, Target)) {
	      cont_BackTrack();
	      Expansion = term_Copy(def_Expansion(list_Car(Scan1)));
	      fol_NormalizeVarsStartingAt(ActTerm, term_MaxVar(Expansion));
	      if (flag_GetFlagValue(flag_PAPPLYDEFS)) {
		printf("\n\n\tConjecture Term:\n");
		fol_PrettyPrintDFG(ActTerm);
		printf("\n\tDefinition Atom: "); 
		fol_PrettyPrintDFG(DefPredicate);
		printf(" Application Atom: "); 
		fol_PrettyPrintDFG(Target);
	      }
	      unify_Match(cont_LeftContext(), DefPredicate, Target);
	      Apply = fol_ApplyContextToTerm(cont_LeftContext(), Expansion);
	      if (Apply) {                                              
		Apply = TRUE;
		Applics--;		
		term_RplacTop(Target, term_TopSymbol(Expansion));
		list_DeleteWithElement(term_ArgumentList(Target), term_Delete);
		term_RplacArgumentList(Target, term_ArgumentList(Expansion));
		term_RplacArgumentList(Expansion, list_Nil());
		term_AddFatherLinks(ActTerm);
		if (flag_GetFlagValue(flag_PAPPLYDEFS)) {
		  printf("\n\tDefinition applied, new term is:\n");
		  fol_PrettyPrintDFG(ActTerm);
		}
		fol_NormalizeVars(ActTerm);
	      }
	      term_Delete(Expansion);
	    }
	    else {
	      
	    }
	    cont_BackTrack();
	  }
	  list_Delete(TargetList);
	}
	term_Delete(DefPredicate);
      }
    }
  }
  return Terms;
}
