




 





















#ifndef _DEFS_
#define _DEFS_






#include "clause.h"
#include "term.h"
#include "list.h"
#include "search.h"




typedef enum { PREDOCCURONCE = 1, 
               ISEQUALITY    = 2  
} DEF_ATTRIBUTES;

typedef struct DEF_HELP {
  TERM   expansion;        
  TERM   predicate;        
  TERM   toprove;
  LIST   parentclauses;    
  char*  label;            
  BOOL   conjecture;
  NAT    attributes;       
} *DEF, DEF_NODE;




static __inline__ TERM def_Expansion(DEF D)
{
  return D->expansion;
}

static __inline__ void def_RplacExp(DEF D, TERM E)
{
  D->expansion = E;
}

static __inline__ TERM def_Predicate(DEF D)
{
  return D->predicate;
}

static __inline__ void def_RplacPred(DEF D, TERM Pred)
{
  D->predicate = Pred;
}

static __inline__ TERM def_ToProve(DEF D)
{
  return D->toprove;
}

static __inline__ void def_RplacToProve(DEF D, TERM ToProve)
{
  D->toprove = ToProve;
}

static __inline__ LIST def_ClauseNumberList(DEF D)
{
  return list_PairFirst(D->parentclauses);
}

static __inline__ void def_RplacClauseNumberList(DEF D, LIST L)
{
  list_Rplaca(D->parentclauses, L);
}

static __inline__ LIST def_ClauseLitsList(DEF D)
{
  return list_PairSecond(D->parentclauses);
}

static __inline__ void def_RplacClauseLitsList(DEF D, LIST L)
{
  list_Rplacd(D->parentclauses, L);
}

static __inline__ char* def_Label(DEF D)
{
  return D->label;
}

static __inline__ void def_RplacLabel(DEF D, char* L)
{
  D->label = L;
}

static __inline__ BOOL def_Conjecture(DEF D)
{
  return D->conjecture;
}

static __inline__ void def_SetConjecture(DEF D)
{
  D->conjecture = TRUE;
}

static __inline__ void def_AddAttribute(DEF D, DEF_ATTRIBUTES Attribute)
{
  D->attributes = D->attributes | Attribute;
}

static __inline__ NAT def_HasAttribute(DEF D, DEF_ATTRIBUTES Attribute)
{
  return (D->attributes & Attribute);
}

static __inline__ void def_RemoveAttribute(DEF D, DEF_ATTRIBUTES Attribute)
{
  if (D->attributes & Attribute)
    D->attributes = D->attributes - Attribute;
}

static __inline__ BOOL def_HasGuard(DEF D)
{
  return (def_ToProve(D)!=term_Null());
}





DEF  def_CreateFromClauses(TERM, TERM, LIST, LIST, BOOL);
DEF  def_CreateFromTerm(TERM, TERM, TERM, char*);

LIST def_ExtractDefsFromTerm(TERM, char*);
void def_ExtractDefsFromTermlist(PROOFSEARCH, LIST, LIST);
void def_ExtractDefsFromClauselist(PROOFSEARCH, LIST);

TERM def_ApplyDefToTermOnce(DEF, TERM, BOOL*);
TERM def_ApplyDefToTermExhaustive(PROOFSEARCH, TERM);
LIST def_ApplyDefToTermlist(DEF, LIST, BOOL*, BOOL);
LIST def_ApplyDefinitionToTermList(LIST, LIST, int);


LIST def_ApplyDefToClauseOnce(DEF, CLAUSE);
LIST def_ApplyDefToClauseExhaustive(PROOFSEARCH, CLAUSE);
LIST def_ApplyDefToClauselist(PROOFSEARCH, DEF, LIST, BOOL);

LIST def_FlattenWithOneDefinition(PROOFSEARCH, DEF);
void def_FlattenWithOneDefinitionSemiDestructive(PROOFSEARCH, DEF);
void def_FlattenWithOneDefinitionDestructive(PROOFSEARCH, DEF);
void def_FlattenDefinitionsDestructive(PROOFSEARCH);

void def_Delete(DEF);
void def_Print(DEF);

int  def_PredicateOccurrences(TERM, SYMBOL);
#endif
