




 





















%{

#include <ctype.h>
#include "dfg.h"
#include "symbol.h"
#include "term.h"
#include "foldfg.h"


typedef struct {
  char*     name;
  char*     author;
  char*     version;
  char*     logic;
  DFG_STATE status;
  char*     description;
  char*     date;
} DFG_DESCRIPTIONTYPE;

static DFG_DESCRIPTIONTYPE dfg_DESC;
static LIST                dfg_AXIOMLIST;
static LIST                dfg_CONJECLIST;
static LIST                dfg_SORTDECLLIST;
static LIST                dfg_AXCLAUSES;
static LIST                dfg_CONCLAUSES;
static LIST                dfg_PROOFLIST;     
static LIST                dfg_TERMLIST;      

NAT                        dfg_LINENUMBER;  


void yyerror(const char*);
int  yylex(void);		

static void   dfg_SymbolDecl(int, char*, int);
static SYMBOL dfg_Symbol(char*, NAT);
static void   dfg_SubSort(char*, char*);
static void   dfg_SymbolGenerated(SYMBOL, BOOL, LIST);
static void   dfg_Settings(char*);

static __inline__ TERM dfg_TermCreate(char* Name, LIST Arguments)

{
  SYMBOL s;
  NAT    arity;
  arity = list_Length(Arguments);
  s = dfg_Symbol(Name, arity); 
  if (!symbol_IsVariable(s) && !symbol_IsFunction(s)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Line %d: is not a function.\n", dfg_LINENUMBER);
    misc_FinishUserErrorReport();
  }
  return term_Create(s, Arguments);
}

static __inline__ TERM dfg_AtomCreate(char* Name, LIST Arguments)

{
  SYMBOL s;
  s = dfg_Symbol(Name, list_Length(Arguments)); 
  if (symbol_IsVariable(s) || !symbol_IsPredicate(s)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Line %d: Symbol is not a predicate.\n", dfg_LINENUMBER);
    misc_FinishUserErrorReport();
  }
  return term_Create(s, Arguments);
}










static LIST   dfg_SYMBOLLIST;

static void dfg_SymAdd(SYMBOL);
static void dfg_SymCheck(SYMBOL, NAT);
static void dfg_SymCleanUp(void);











static LIST dfg_VARLIST;
static BOOL dfg_VARDECL;

static void   dfg_VarStart(void);
static void   dfg_VarStop(void);
static void   dfg_VarBacktrack(void);
static void   dfg_VarCheck(void);
static SYMBOL dfg_VarLookup(char*);

#define YYERROR_VERBOSE

%}

%union {
  int       number;
  char*     string;
  SYMBOL    symbol;
  TERM      term;
  LIST      list;
  DFG_STATE state;
  BOOL      bool;
}

%token DFG_AND DFG_AUTHOR DFG_AXIOMS DFG_BEGPROB DFG_BY DFG_CLAUSE DFG_CNF DFG_CONJECS
%token DFG_DATE DFG_DESC DFG_DNF DFG_ENDLIST DFG_ENDPROB DFG_EQUAL DFG_EQUIV DFG_EXISTS
%token DFG_FALSE DFG_FORMULA DFG_FORALL DFG_FREELY DFG_FUNC DFG_GENERATED DFG_HYPOTH
%token DFG_IMPLIED DFG_IMPLIES DFG_KIV DFG_LEM DFG_CLSLIST DFG_DECLLIST DFG_DESCLIST
%token DFG_FORMLIST DFG_GENSET DFG_PRFLIST DFG_SETTINGS DFG_SYMLIST DFG_LOGIC DFG_NAME
%token DFG_NOT DFG_OPERAT DFG_OR DFG_OTTER DFG_PRED DFG_PRDICAT DFG_PROTEIN DFG_QUANTIF
%token DFG_RESOLU DFG_SATIS DFG_SATURATE DFG_SETHEO DFG_SORT DFG_SORTS DFG_SPASS
%token DFG_STATUS DFG_SUBSORT DFG_TERMLIST DFG_TRUE DFG_UNKNOWN DFG_UNSATIS DFG_VERSION
%token DFG_3TAP

%token <number> DFG_NUM DFG_MINUS1
%token <string> DFG_ID DFG_TEXT
%type <number>	arity slabel
%type <string>	id labelopt
%type <symbol>  binsymbol nsymbol quantsymbol
%type <term>	atom clause cnfclause cnfclauseopt cnfclausebody formula
%type <term>    formulaopt lit qterm term
%type <list>	arglist termlist atomlist cnfclausesopt formulalistopt labellist
%type <list>    litlist qtermlist sortlist funclist
%type <state>	log_state
%type <bool>    origin freelyopt

%%





problem:	DFG_BEGPROB '(' id ')' '.'
		description
		logicalpart
		settinglistsopt
		DFG_ENDPROB '.'		{ misc_StringFree($3);
					  YYACCEPT;  }
;





description:	DFG_DESCLIST '.'
		name
		author
		versionopt
		logicopt
		status
		desctext
		dateopt
		DFG_ENDLIST '.'
;

name: 		DFG_NAME '(' DFG_TEXT ')' '.'
					{ dfg_DESC.name = $3; }
;

author:		DFG_AUTHOR '(' DFG_TEXT ')' '.'
					{ dfg_DESC.author = $3; }
;

status:		DFG_STATUS '(' log_state ')' '.'
					{ dfg_DESC.status = $3; }
;

desctext:	DFG_DESC '(' DFG_TEXT ')' '.'
					{ dfg_DESC.description = $3; }
;

versionopt:	
		| DFG_VERSION '(' DFG_TEXT ')' '.'
					{ dfg_DESC.version = $3; }
;

logicopt:	
		| DFG_LOGIC '(' DFG_TEXT ')' '.'
					{ dfg_DESC.logic = $3; }
;

dateopt:	
		| DFG_DATE '(' DFG_TEXT ')' '.'
					{ dfg_DESC.date = $3; }
;

log_state:	DFG_SATIS		{ $$ = DFG_SATISFIABLE; }
		| DFG_UNSATIS		{ $$ = DFG_UNSATISFIABLE; }
		| DFG_UNKNOWN		{ $$ = DFG_UNKNOWNSTATE; }
;





logicalpart:	symbollistopt
		declarationlistopt
		formulalistsopt
		clauselistsopt
		prooflistsopt
		listOfTermsopt
;





symbollistopt:	
		| DFG_SYMLIST '.'
		  functionsopt
		  predicatesopt
		  sortsopt
		  operatorsopt
		  quantifiersopt
		  DFG_ENDLIST '.'
;

functionsopt:	
		| DFG_FUNC '[' functionlist ']' '.'
;

functionlist:	func
		| functionlist ',' func
;

func:		id			{ dfg_SymbolDecl(DFG_FUNC, $1, -2); }
		| '(' id ',' arity ')'
					{ dfg_SymbolDecl(DFG_FUNC, $2, $4); }
;

predicatesopt:	
		| DFG_PRDICAT '[' predicatelist ']' '.'
;

predicatelist:	pred
		| predicatelist ',' pred
;

pred:		id			{ dfg_SymbolDecl(DFG_PRDICAT, $1, -2); }
		| '(' id ',' arity ')'	{ dfg_SymbolDecl(DFG_PRDICAT, $2, $4); }
;

sortsopt:	
		| DFG_SORTS '[' sortlist ']' '.'
;

sortlist:	id			{ dfg_SymbolDecl(DFG_PRDICAT, $1, 1); }
		| sortlist ',' id	{ dfg_SymbolDecl(DFG_PRDICAT, $3, 1); }
;

operatorsopt:	
		| DFG_OPERAT '[' operatorlist ']' '.'
;

operatorlist:	op
		| operatorlist ',' op
;

op:		id			{ dfg_SymbolDecl(DFG_OPERAT, $1, -2); }
		| '(' id ',' arity ')'	{ dfg_SymbolDecl(DFG_OPERAT, $2, $4); }
;

quantifiersopt:	
		| DFG_QUANTIF '[' quantifierlist ']' '.'
;

quantifierlist:	quant
		| quantifierlist ',' quant
;

quant:		id			{ dfg_SymbolDecl(DFG_QUANTIF, $1, -2); }
		| '(' id ',' arity ')'	{ dfg_SymbolDecl(DFG_QUANTIF, $2, $4); }
;

arity:		DFG_MINUS1		{ $$ = -1; }
		| DFG_NUM		{ $$ = $1; }
;





declarationlistopt:	
		| DFG_DECLLIST '.' 
		  decllistopt
		  DFG_ENDLIST '.'
;

decllistopt:	
		| decllistopt decl
;

decl:		gendecl
		| DFG_SUBSORT '(' id ',' id ')' '.'
					{ dfg_SubSort($3,$5); }
		| atom '.'		{ dfg_SORTDECLLIST = list_Nconc(dfg_SORTDECLLIST,list_List(list_PairCreate(NULL,$1))); }
		| DFG_PRED '(' id ',' sortdecl ')' '.'	
					{ misc_StringFree($3); }
		| DFG_FORALL '(' '['	{ dfg_VarStart(); }
		  qtermlist 		{ dfg_VarStop();  }
		  ']' ',' atom ')' '.'	{ TERM term;
					  dfg_VarBacktrack();
					  dfg_VarCheck();
					  term = dfg_CreateQuantifier(fol_All(),$5,$9);
					  dfg_SORTDECLLIST = list_Nconc(dfg_SORTDECLLIST,list_List(list_PairCreate(NULL,term)));
					}
;

gendecl:	DFG_SORT id freelyopt DFG_GENERATED DFG_BY '[' funclist ']' '.'
					{ dfg_SymbolGenerated(dfg_Symbol($2,1), $3, $7); }
;

freelyopt:			{ $$ = FALSE; }
		| DFG_FREELY		{ $$ = TRUE; }
;

funclist:	id			{ $$ = list_List($1); }
		| funclist ',' id       { $$ = list_Cons($3, $1); }
;

sortdecl:	id			{ misc_StringFree($1); }
		| sortdecl ',' id	{ misc_StringFree($3); }
;





formulalist:	DFG_FORMLIST '(' origin ')' '.'
		formulalistopt
		DFG_ENDLIST '.'	{ if ($3) 
				    dfg_AXIOMLIST = list_Nconc(dfg_AXIOMLIST, $6);
				  else
				    dfg_CONJECLIST = list_Nconc(dfg_CONJECLIST, $6);
				}
;

origin:		DFG_AXIOMS		{ $$ = TRUE;  }
		| DFG_CONJECS		{ $$ = FALSE; }
;

formulalistsopt:	
		| formulalistsopt formulalist
;

formulalistopt:			{ $$ = list_Nil(); }
		| formulalistopt DFG_FORMULA '(' formulaopt labelopt ')' '.'
					{ LIST pair;
					  if ($4 == NULL) { 
					    if ($5 != NULL)
					      misc_StringFree($5);
					    $$ = $1;
					  } else {
					    pair = list_PairCreate($5, $4);
					    $$ = list_Nconc($1, list_List(pair));
					  }
					  dfg_VarCheck();
					}
;

labelopt:			{ $$ = NULL; }
		| ',' id		{ $$ = $2;   }
;

formula:	atom		{ $$ = $1; }
		| DFG_NOT '(' formula ')'
				{ $$ = term_Create(fol_Not(),list_List($3)); }
		| binsymbol '(' formula ',' formula ')'
				{ $$ = term_Create($1, list_Cons($3, list_List($5))); }
		| nsymbol '(' arglist ')'
				{ $$ = term_Create($1, $3); }
		| quantsymbol '(' '[' 	{ dfg_VarStart(); }
		  qtermlist 		{ dfg_VarStop(); }
		  ']' ',' formula ')'
			  	{ dfg_VarBacktrack();
                                  $$ = dfg_CreateQuantifier($1,$5,$9);
				}
;

formulaopt:			{ $$ = NULL; }
		| formula		{ $$ = $1; }
;

arglist:	formula			{ $$ = list_List($1); }
		| arglist ',' formula	{ $$ = list_Nconc($1, list_List($3)); }
;

binsymbol:	DFG_EQUIV		{ $$ = fol_Equiv();    }
		| DFG_IMPLIED		{ $$ = fol_Implied();  }
		| DFG_IMPLIES		{ $$ = fol_Implies();  }
;

nsymbol:	DFG_AND			{ $$ = fol_And(); }
		| DFG_OR		{ $$ = fol_Or();  }
;

quantsymbol:	DFG_EXISTS		{ $$ = fol_Exist(); }
		| DFG_FORALL		{ $$ = fol_All(); }
;

id:		DFG_ID			{ $$ = $1; }
		| DFG_NUM		{ $$ = misc_IntToString($1); }
;

qtermlist:	qterm			{ $$ = list_List($1); }
		| qtermlist ',' qterm	{ $$ = list_Nconc($1, list_List($3)); }
;

qterm:		id			{ SYMBOL s = dfg_Symbol($1,0);
					  if (!symbol_IsVariable(s)) {
					    misc_StartUserErrorReport();
					    misc_UserErrorReport("\n Line %d: Symbol is not a variable.\n",dfg_LINENUMBER);
					    misc_FinishUserErrorReport();
					  }
					  $$ = term_Create(s, list_Nil());
					}
		| id '(' id ')'		{ SYMBOL p, v;
					  p = dfg_Symbol($1, 1);
					  if (!symbol_IsPredicate(p)) {
					    misc_StartUserErrorReport();
					    misc_UserErrorReport("\n Line %d: Symbol is not a predicate.\n",dfg_LINENUMBER);
					    misc_FinishUserErrorReport();
					  }
					  v = dfg_Symbol($3, 0);
					  if (!symbol_IsVariable(v)) {
					    misc_StartUserErrorReport();
					    misc_UserErrorReport("\n Line %d: Symbol is not a variable.\n",dfg_LINENUMBER);
					    misc_FinishUserErrorReport();
					  }
					  $$ = term_Create(p, list_List(term_Create(v,list_Nil())));
					}
;





clauselistsopt:	
		| clauselistsopt clauselist
;

clauselist:	DFG_CLSLIST '(' origin ',' DFG_CNF ')' '.'
		cnfclausesopt
		DFG_ENDLIST '.'		{ if ($3) 
                                            dfg_AXCLAUSES = list_Nconc(dfg_AXCLAUSES,$8);
                                          else
                                            dfg_CONCLAUSES = list_Nconc(dfg_CONCLAUSES,$8);
					}
		| DFG_CLSLIST '(' origin ',' DFG_DNF ')' '.'
		  dnfclausesopt
		  DFG_ENDLIST '.'	
;

cnfclausesopt:			{ $$ = list_Nil(); }
		| cnfclausesopt DFG_CLAUSE '(' cnfclauseopt labelopt ')' '.'
					{ LIST pair;
					  if ($4 == NULL) { 
					    if ($5 != NULL)
					      misc_StringFree($5);
					    $$ = $1;
					  } else {
					    pair = list_PairCreate($5, $4);
					    $$ = list_Nconc($1, list_List(pair));
					  }
					  dfg_VarCheck();
					}
;

cnfclauseopt:			{ $$ = NULL; }
		| cnfclause		{ $$ = $1; }
;

cnfclause:	cnfclausebody		{ $$ = $1; }
		| DFG_FORALL '(' '['	{ dfg_VarStart(); }
		  qtermlist		{ dfg_VarStop();  }
		  ']' ',' cnfclausebody ')'
					{ dfg_VarBacktrack();
					  $$ = dfg_CreateQuantifier(fol_All(),$5,$9);
					}
;

cnfclausebody:	DFG_OR '(' litlist ')'	{ $$ = term_Create(fol_Or(), $3); }
;

litlist:	lit			{ $$ = list_List($1); }
		| litlist ',' lit	{ $$ = list_Nconc($1, list_List($3)); }
;

lit:		atom			{ $$ = $1; }
		| DFG_NOT '(' atom ')'	{ $$ = term_Create(fol_Not(),list_List($3)); }
;

atomlist:	atom			{ $$ = list_List($1); }
		| atomlist ',' atom	{ $$ = list_Nconc($1, list_List($3)); }
;

atom:		id			{ $$ = dfg_AtomCreate($1,list_Nil()); }
		| DFG_TRUE		{ $$ = term_Create(fol_True(),list_Nil()); }
		| DFG_FALSE		{ $$ = term_Create(fol_False(),list_Nil()); }
		| DFG_EQUAL '(' term ',' term ')'
					{ $$ = term_Create(fol_Equality(), list_Cons($3,list_List($5))); }
		| id '(' termlist ')'	{ $$ = dfg_AtomCreate($1, $3); }
;




dnfclausesopt:	
		| dnfclausesopt DFG_CLAUSE '(' dnfclauseopt labelopt ')' '.'
				{ if ($5 != NULL) 
				    misc_StringFree($5);
				}
;

dnfclauseopt:	
		| dnfclause
;

dnfclause:	dnfclausebody
		| DFG_EXISTS '(' '['	{ dfg_VarStart(); }
		  atomlist		{ dfg_VarStop();  }
		  ']' ',' dnfclausebody ')'
				{ dfg_VarBacktrack();
				  dfg_VarCheck();
				  list_DeleteWithElement($5, term_Delete);
				}
;

dnfclausebody:	DFG_AND '(' litlist ')'
				{ list_DeleteWithElement($3, term_Delete); }
;





term:		id			{ $$ = dfg_TermCreate($1,list_Nil()); }
		| id '(' termlist ')'	{ $$ = dfg_TermCreate($1, $3); }
;

termlist:	term			{ $$ = list_List($1); }
		| termlist ',' term	{ $$ = list_Nconc($1,list_List($3)); }
;





prooflistsopt:	
		| prooflistsopt prooflist
;

prooflist:	DFG_PRFLIST '(' DFG_RESOLU ')' '.'
		prooflistopt
		DFG_ENDLIST '.'
;

prooflistopt:	
		| prooflistopt DFG_CLAUSE '(' clause ',' id ',' id ',' '[' labellist ']' ',' DFG_NUM ')' '.'
				{ LIST tupel;
				  RULE Rule = clause_GetOriginFromString($8);
                                  misc_StringFree($8);
				  if ($4 != NULL) {
				    
				    
				    tupel = list_Cons((POINTER)$14,list_List((POINTER)Rule));
                                    tupel = list_Cons($6,list_Cons($4,list_Cons($11,tupel)));
				    dfg_PROOFLIST = list_Cons(tupel, dfg_PROOFLIST);
				  } else {
				    
				    misc_StringFree($6);
				    list_Apply(misc_StringFree, $11);
				  }
				  dfg_VarCheck();
				}		
;

clause:		cnfclause		{ $$ = $1;   }
		| dnfclause		{ $$ = NULL; }
;

labellist:	id			{ $$ = list_List($1); }
		| labellist ',' id	{ $$ = list_Nconc($1, list_List($3)); }
;





listOfTermsopt: 
                | listOfTermsopt listOfTerms
;

listOfTerms:    DFG_TERMLIST '.'        { dfg_VarStart(); }
                terms DFG_ENDLIST '.' {
                                          dfg_VarStop();
                                          dfg_VarBacktrack();
                                          dfg_VarCheck(); }
;

terms:		
                | terms term '.'	{ dfg_TERMLIST = list_Nconc(dfg_TERMLIST, list_List($2)); }
; 





settinglistsopt:	
		| settinglistsopt settinglist
;

settinglist:	DFG_GENSET '.' gsettings DFG_ENDLIST '.'
		| DFG_SETTINGS '(' slabel ')' '.' DFG_TEXT DFG_ENDLIST '.' 
				{ if ($3 == DFG_SPASS)
				    dfg_Settings($6);
				  misc_StringFree($6);
				}
;

gsettings:	gsetting
		| gsettings gsetting
;

gsetting:	DFG_HYPOTH '[' labellist ']' '.'
				{ list_DeleteWithElement($3, misc_StringFree); }
;

slabel:		DFG_KIV			{ $$ = DFG_KIV;      }
		| DFG_LEM		{ $$ = DFG_LEM;      }
		| DFG_OTTER		{ $$ = DFG_OTTER;    }
		| DFG_PROTEIN		{ $$ = DFG_PROTEIN;  }
		| DFG_SATURATE		{ $$ = DFG_SATURATE; }
		| DFG_3TAP		{ $$ = DFG_3TAP;     }
		| DFG_SETHEO		{ $$ = DFG_SETHEO;   }
		| DFG_SPASS		{ $$ = DFG_SPASS;    }
;

%%

void yyerror(const char *s)
{
  misc_StartUserErrorReport();
  misc_UserErrorReport("\n Line %i: %s\n", dfg_LINENUMBER, s);
  misc_FinishUserErrorReport();
}

static void dfg_Init(FILE* Input)

{
  extern FILE* dfg_in;  

  dfg_in               = Input;
  dfg_LINENUMBER       = 1;
  dfg_AXIOMLIST        = list_Nil();
  dfg_CONJECLIST       = list_Nil();
  dfg_SORTDECLLIST     = list_Nil();
  dfg_AXCLAUSES        = list_Nil();
  dfg_CONCLAUSES       = list_Nil();
  dfg_PROOFLIST        = list_Nil();
  dfg_TERMLIST         = list_Nil();
  dfg_SYMBOLLIST       = list_Nil();
  dfg_VARLIST          = list_Nil();
  dfg_VARDECL          = FALSE;
  dfg_DESC.name        = (char*) NULL;
  dfg_DESC.author      = (char*) NULL;
  dfg_DESC.version     = (char*) NULL;
  dfg_DESC.logic       = (char*) NULL;
  dfg_DESC.status      = DFG_UNKNOWNSTATE;
  dfg_DESC.description = (char*) NULL;
  dfg_DESC.date        = (char*) NULL;
}


void dfg_Free(void)

{
  if (dfg_DESC.name != NULL)
    misc_StringFree(dfg_DESC.name);
  if (dfg_DESC.author != NULL)
    misc_StringFree(dfg_DESC.author);
  if (dfg_DESC.version != NULL)
    misc_StringFree(dfg_DESC.version);
  if (dfg_DESC.logic != NULL)
    misc_StringFree(dfg_DESC.logic);
  if (dfg_DESC.description != NULL)
    misc_StringFree(dfg_DESC.description);
  if(dfg_DESC.date != NULL)
    misc_StringFree(dfg_DESC.date);
}

const char* dfg_ProblemName(void)

{
  return dfg_DESC.name;
}

const char* dfg_ProblemAuthor(void)

{
  return dfg_DESC.author;
}

const char* dfg_ProblemVersion(void)

{
  return dfg_DESC.version;
}

const char* dfg_ProblemLogic(void)

{
  return dfg_DESC.logic;
}

DFG_STATE dfg_ProblemStatus(void)

{
  return dfg_DESC.status;
}

const char* dfg_ProblemStatusString(void)

{
  char* result = "";

  switch (dfg_DESC.status) {
  case DFG_SATISFIABLE:
    result = "satisfiable"; break;
  case DFG_UNSATISFIABLE:
    result = "unsatisfiable"; break;
  case DFG_UNKNOWNSTATE:
    result = "unknown"; break;
  default:
    misc_StartErrorReport();
    misc_ErrorReport("\n In dfg_ProblemStatusString: Invalid status.\n");
    misc_FinishErrorReport();
  }
  return result;
}

const char* dfg_ProblemDescription(void)

{
  return dfg_DESC.description;
}

const char* dfg_ProblemDate(void)

{
  return dfg_DESC.date;
}

void dfg_FPrintDescription(FILE* File)

{
  fputs("list_of_descriptions.\n  name(", File);
  if (dfg_DESC.name != NULL)
    fputs(dfg_DESC.name, File);
  else
    fputs("{* *}", File);
  fputs(").\n  author(", File);
  if (dfg_DESC.author != NULL)
    fputs(dfg_DESC.author, File);
  else
    fputs("{* *}", File);
  fputs(").\n", File);
  if (dfg_DESC.version != NULL) {
    fputs("  version(", File);
    fputs(dfg_DESC.version, File);
    fputs(").\n", File);
  }
  if (dfg_DESC.logic != NULL) {
    fputs("  logic(", File);
    fputs(dfg_DESC.logic, File);
    fputs(").\n", File);
  }
  fputs("  status(", File);
  fputs(dfg_ProblemStatusString(), File);
  fputs(").\n  description(", File);
  if (dfg_DESC.description != NULL)
    fputs(dfg_DESC.description, File);
  else
    fputs("{* *}", File);
  fputs(").\n", File);
  if (dfg_DESC.date != NULL) {
    fputs("  date(", File);
    fputs(dfg_DESC.date, File);
    fputs(").\n", File);
  }
  fputs("end_of_list.", File);
}


LIST dfg_DFGParser(FILE* File, LIST* Axioms, LIST* Conjectures, LIST* SortDecl)

{
  LIST  scan, tupel;
  TERM  clauseTerm;

  dfg_Init(File);      
  dfg_parse();          

  dfg_SymCleanUp();

  if (!flag_GetFlagValue(flag_INTERACTIVE) &&
      list_Empty(dfg_AXIOMLIST) && list_Empty(dfg_CONJECLIST) &&
      list_Empty(dfg_AXCLAUSES) && list_Empty(dfg_CONCLAUSES)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n No formulae and clauses found in input file!\n");
    misc_FinishUserErrorReport();
  }

  
  for (scan = dfg_AXCLAUSES; !list_Empty(scan); scan = list_Cdr(scan)) {
    tupel = list_Car(scan);
    clauseTerm = list_PairSecond(tupel);
    list_Rplaca(scan, dfg_CreateClauseFromTerm(clauseTerm,TRUE));
    if (list_PairFirst(tupel) != NULL)        
      misc_StringFree(list_PairFirst(tupel));  
    list_PairFree(tupel);
  }
  for (scan = dfg_CONCLAUSES; !list_Empty(scan); scan = list_Cdr(scan)) {
    tupel = list_Car(scan);
    clauseTerm = list_PairSecond(tupel);
    list_Rplaca(scan, dfg_CreateClauseFromTerm(clauseTerm,FALSE));
    if (list_PairFirst(tupel) != NULL)        
      misc_StringFree(list_PairFirst(tupel));  
    list_PairFree(tupel);
  }

  
  dfg_DeleteProofList(dfg_PROOFLIST);

  
  list_DeleteWithElement(dfg_TERMLIST, term_Delete);

  scan = list_Nconc(dfg_AXCLAUSES, dfg_CONCLAUSES);

  *Axioms      = list_Nconc(*Axioms, dfg_AXIOMLIST);
  *Conjectures = list_Nconc(*Conjectures, dfg_CONJECLIST);
  *SortDecl    = list_Nconc(*SortDecl, dfg_SORTDECLLIST);

  return scan;
}


LIST dfg_ProofParser(FILE* File)

{
  LIST  scan, tupel;
  TERM  term;

  dfg_Init(File);       
  dfg_parse();          

  dfg_SymCleanUp();

  
  dfg_AXCLAUSES  = list_Nconc(dfg_AXCLAUSES, dfg_CONCLAUSES);
  dfg_CONCLAUSES = list_Nil();
  for (scan = dfg_AXCLAUSES; !list_Empty(scan); scan = list_Cdr(scan)) {
    tupel = list_Car(scan);
    term  = list_PairSecond(tupel);
    if (list_PairFirst(tupel) == NULL) {
      
      term_Delete(term);
      list_PairFree(tupel);
      list_Rplaca(scan, NULL);
    } else
      
      
      
      
      list_Rplacd(tupel, list_Cons(term,list_Cons(list_Nil(),list_Cons(0, list_List((POINTER)INPUT)))));
  }
  
  dfg_AXCLAUSES = list_PointerDeleteElement(dfg_AXCLAUSES, NULL);

  
  dfg_DeleteFormulaPairList(dfg_AXIOMLIST);
  dfg_DeleteFormulaPairList(dfg_CONJECLIST);
  
  dfg_DeleteFormulaPairList(dfg_SORTDECLLIST);
  
  list_DeleteWithElement(dfg_TERMLIST, term_Delete);

  
  dfg_PROOFLIST = list_NReverse(dfg_PROOFLIST);
  dfg_AXCLAUSES = list_Nconc(dfg_AXCLAUSES, dfg_PROOFLIST);

  return dfg_AXCLAUSES;
}


LIST dfg_TermParser(FILE* File)

{
  dfg_Init(File);       
  dfg_parse();          

  dfg_SymCleanUp();

  
  dfg_DeleteFormulaPairList(dfg_AXCLAUSES);
  dfg_DeleteFormulaPairList(dfg_CONCLAUSES);
  
  dfg_DeleteFormulaPairList(dfg_AXIOMLIST);
  dfg_DeleteFormulaPairList(dfg_CONJECLIST);
  
  dfg_DeleteProofList(dfg_PROOFLIST);
  
  dfg_DeleteFormulaPairList(dfg_SORTDECLLIST);

  return dfg_TERMLIST;
}


void dfg_DeleteFormulaPairList(LIST FormulaPairs)

{
  LIST pair;

  for ( ; !list_Empty(FormulaPairs); FormulaPairs = list_Pop(FormulaPairs)) {
    pair = list_Car(FormulaPairs);  
    term_Delete(list_PairSecond(pair));
    if (list_PairFirst(pair) != NULL)
      misc_StringFree(list_PairFirst(pair));  
    list_PairFree(pair);
  }
}

void dfg_StripLabelsFromList(LIST FormulaPairs)

{
  LIST pair, scan;

  for (scan = FormulaPairs; !list_Empty(scan); scan = list_Cdr(scan)) {
    pair = list_Car(scan);  
    list_Rplaca(scan, list_PairSecond(pair));
    if (list_PairFirst(pair) != NULL)
      misc_StringFree(list_PairFirst(pair));  
    list_PairFree(pair);
  }
}

void dfg_DeleteProofList(LIST Proof)

{
  
  for ( ; !list_Empty(Proof); Proof = list_Pop(Proof)) {
    LIST tupel = list_Car(Proof);
    misc_StringFree(list_First(tupel));
    term_Delete(list_Second(tupel));
    list_Apply(misc_StringFree, list_Third(tupel));
    list_Delete(list_Third(tupel));
    list_Delete(tupel);
  }
}





static void dfg_SymbolDecl(int SymbolType, char* Name, int Arity)

{
  NAT    arity, length;
  SYMBOL symbol;

  switch (Arity) {
  case -2: 
    arity = 0;
    break;
  case -1:
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Line %u: symbols with arbitrary arity are not allowed.\n",
	    dfg_LINENUMBER);
    misc_FinishUserErrorReport();
  default:
    arity = Arity;
}

  
  length = strlen(Name);
  if (length >= symbol__SYMBOLMAXLEN)
    Name[symbol__SYMBOLMAXLEN-1] = '\0';

  
  symbol = symbol_Lookup(Name);
  if (symbol != 0) {
    
    
    if ((SymbolType == DFG_FUNC && !symbol_IsFunction(symbol)) ||
	(SymbolType == DFG_PRDICAT && !symbol_IsPredicate(symbol)) ||
	((SymbolType == DFG_OPERAT || SymbolType == DFG_QUANTIF) && 
	 !symbol_IsJunctor(symbol))) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n Line %u: symbol %s was already declared as ",
	      dfg_LINENUMBER, Name);
      switch (symbol_Type(symbol)) {
      case symbol_CONSTANT:
      case symbol_FUNCTION:
	misc_UserErrorReport("function.\n"); break;
      case symbol_PREDICATE:
	misc_UserErrorReport("predicate.\n"); break;
      case symbol_JUNCTOR:
	misc_UserErrorReport("junctor.\n"); break;
      default:
	misc_UserErrorReport("unknown type.\n");
      }
      misc_FinishUserErrorReport();
    }
    
    if (Arity != -2 && Arity != symbol_Arity(symbol)) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n Line %u: symbol %s was already declared with arity %d\n",
			   dfg_LINENUMBER, Name, symbol_Arity(symbol));
      misc_FinishUserErrorReport();
    }
  } else {
    
    switch (SymbolType) {
    case DFG_FUNC:
      symbol = symbol_CreateFunction(Name, arity, symbol_STATLEX);
      break;
    case DFG_PRDICAT:
      symbol = symbol_CreatePredicate(Name, arity, symbol_STATLEX);
      break;
    default:
      symbol = symbol_CreateJunctor(Name, arity, symbol_STATLEX);
    }
    if (Arity == -2)
      
      dfg_SymAdd(symbol);
  }

  if (length >= symbol__SYMBOLMAXLEN) {
    
    Name[symbol__SYMBOLMAXLEN-1] = ' ';  
  }
  misc_StringFree(Name);  
}


static SYMBOL dfg_Symbol(char* Name, NAT Arity)

{
  SYMBOL symbol;
  char   old;
  NAT    length;

  old = ' ';  
  
  length = strlen(Name);
  if (length >= symbol__SYMBOLMAXLEN) {
    old = Name[symbol__SYMBOLMAXLEN-1];
    Name[symbol__SYMBOLMAXLEN-1] = '\0';
  }

  symbol = symbol_Lookup(Name);
  if (length >= symbol__SYMBOLMAXLEN) {
    
    Name[symbol__SYMBOLMAXLEN-1] = old;
  }
  if (symbol != 0) {
    misc_StringFree(Name);
    dfg_SymCheck(symbol, Arity); 
  } else {
    
    if (Arity > 0) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n Line %d: Undefined symbol %s.\n",dfg_LINENUMBER,Name);
      misc_FinishUserErrorReport();
    }
    symbol = dfg_VarLookup(Name);
  }
  return symbol;
}


TERM dfg_CreateQuantifier(SYMBOL Symbol, LIST VarTermList, TERM Term)

{
  LIST varlist, sortlist, scan;
  TERM helpterm;

  
  varlist = sortlist = list_Nil();
  for ( ; !list_Empty(VarTermList); VarTermList = list_Pop(VarTermList)) {
    helpterm = list_Car(VarTermList);
    if (term_IsVariable(helpterm)) {
      varlist = list_Nconc(varlist, list_List((POINTER)term_TopSymbol(helpterm)));
      term_Delete(helpterm);
    } else {
      SYMBOL var = term_TopSymbol(term_FirstArgument(helpterm));
      varlist  = list_Nconc(varlist, list_List((POINTER)var));
      sortlist = list_Nconc(sortlist, list_List(helpterm));
    }
  }

  varlist = list_PointerDeleteDuplicates(varlist);
  
  for (scan = varlist; !list_Empty(scan); scan = list_Cdr(scan))
    list_Rplaca(scan, term_Create((SYMBOL)list_Car(scan), list_Nil()));

  if (!list_Empty(sortlist)) {
    if (symbol_Equal(fol_All(), Symbol)) {
      
      if (symbol_Equal(fol_Or(), term_TopSymbol(Term))) {
	
	
	
	for (scan = sortlist; !list_Empty(scan); scan = list_Cdr(scan))
	  
	  list_Rplaca(scan, term_Create(fol_Not(), list_List(list_Car(scan))));
	sortlist = list_Nconc(sortlist, term_ArgumentList(Term));
	term_RplacArgumentList(Term, sortlist);
      } else {
	
	if (list_Empty(list_Cdr(sortlist))) {
	  
	  list_Rplacd(sortlist, list_List(Term));
	  Term = term_Create(fol_Implies(), sortlist);
	} else {
	  
	  helpterm = term_Create(fol_And(), sortlist);
	  Term = term_Create(fol_Implies(), list_Cons(helpterm, list_List(Term)));
	}
      }
    } else if (symbol_Equal(fol_Exist(), Symbol)) {
      
      if (symbol_Equal(fol_And(), term_TopSymbol(Term))) {
	
	
	sortlist = list_Nconc(sortlist, term_ArgumentList(Term));
	term_RplacArgumentList(Term, sortlist);
      } else {
	sortlist = list_Nconc(sortlist, list_List(Term));
	Term = term_Create(fol_And(), sortlist);
      }
    }
  }
  helpterm = fol_CreateQuantifier(Symbol, varlist, list_List(Term));
  return helpterm;
}


CLAUSE dfg_CreateClauseFromTerm(TERM Clause, BOOL IsAxiom)

{
  LIST   literals, scan;
  TERM   literal;
  CLAUSE result;
  
  if (term_TopSymbol(Clause) == fol_All()) {
    
    literals = term_ArgumentList(term_SecondArgument(Clause));
    term_RplacArgumentList(term_SecondArgument(Clause), list_Nil());
  } else {
    
    literals = term_ArgumentList(Clause);
    term_RplacArgumentList(Clause, list_Nil());
  }
  term_Delete(Clause);

  for (scan = literals; !list_Empty(scan); scan = list_Cdr(scan)) {
    literal = (TERM) list_Car(scan);
    if (symbol_IsPredicate(term_TopSymbol(literal))) {  
      if (fol_IsTrue(literal)) {
	
	list_PointerDeleteElement(literals, NULL);
	
	list_DeleteWithElement(literals, term_Delete);
	return (CLAUSE) NULL;
      } else if (fol_IsFalse(literal)) {
	
	term_Delete(literal);
	list_Rplaca(scan, NULL); 
      }
    } else {
      
      TERM atom = term_FirstArgument(literal);
      if (fol_IsFalse(atom)) {
	
	list_PointerDeleteElement(literals, NULL);
	
	list_DeleteWithElement(literals, term_Delete);
	return (CLAUSE) NULL;
      } else if (fol_IsTrue(atom)) {
	
	term_Delete(literal);
	list_Rplaca(literals, NULL); 
      }
    }
  }

  literals = list_PointerDeleteElement(literals, NULL);
  
  result = clause_CreateFromLiterals(literals, FALSE, !IsAxiom, FALSE);
  
  list_Delete(literals);

  return result;
}


static void dfg_SubSort(char* Name1, char* Name2)

{
  SYMBOL s1, s2;
  TERM   varterm, t1, t2, term;

  s1 = dfg_Symbol(Name1, 1);   
  s2 = dfg_Symbol(Name2, 1);
  if (!symbol_IsPredicate(s1)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Line %d: Symbol is not a sort predicate.\n", dfg_LINENUMBER);
    misc_FinishUserErrorReport();
  }
  if (!symbol_IsPredicate(s2)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Line %d: Symbol is not a sort predicate.\n", dfg_LINENUMBER);
    misc_FinishUserErrorReport();
  }

  varterm = term_Create(symbol_CreateStandardVariable(), list_Nil());
  symbol_ResetStandardVarCounter();
  
  t1   = term_Create(s1, list_List(varterm));
  t2   = term_Create(s2, list_List(term_Copy(varterm)));
  term = term_Create(fol_Implies(), list_Cons(t1, list_List(t2)));
  term = fol_CreateQuantifier(fol_All(), list_List(term_Copy(varterm)),
			      list_List(term));
  dfg_SORTDECLLIST = list_Nconc(dfg_SORTDECLLIST, list_List(list_PairCreate(NULL,term)));
}


void dfg_SymbolGenerated(SYMBOL SortPredicate, BOOL FreelyGenerated, LIST GeneratedBy)

{
  SYMBOL symbol;
  LIST   scan;

  if (!symbol_IsPredicate(SortPredicate)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Line %d: Symbol is not a sort predicate.\n", dfg_LINENUMBER);
    misc_FinishUserErrorReport();
  }
  
  symbol_RemoveProperty(SortPredicate, GENERATED);
  symbol_RemoveProperty(SortPredicate, FREELY);
  list_Delete(symbol_GeneratedBy(SortPredicate));
  
  symbol_AddProperty(SortPredicate, GENERATED);
  if (FreelyGenerated)
    symbol_AddProperty(SortPredicate, FREELY);
  for (scan = GeneratedBy; !list_Empty(scan); scan = list_Cdr(scan)) {
    symbol = symbol_Lookup(list_Car(scan));
    if (symbol == 0) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n Line %d: undefined symbol %s.\n", dfg_LINENUMBER,
			   (char*)list_Car(scan));
      misc_FinishUserErrorReport();
    } else if (!symbol_IsFunction(symbol)) { 
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n Line %d: Symbol is not a function.\n", dfg_LINENUMBER);
      misc_FinishUserErrorReport();
    }
    misc_StringFree(list_Car(scan));
    list_Rplaca(scan, (POINTER)symbol);  
    
    symbol_AddProperty(symbol, GENERATED);
    if (FreelyGenerated)
      symbol_AddProperty(symbol, FREELY);
  }
  symbol_SetGeneratedBy(SortPredicate, GeneratedBy);
}

static char* dfg_SettingsScanner(char* Text, int* Token, char** Id)

{
  char* scan;
  BOOL  loop;

  *Id  = NULL;
  scan = Text;
  loop = TRUE;
  do {
    while (isspace((int)*scan))
      scan++;
    if (*scan == '%')  
      while (*scan != '\n')
	scan++;
    else
      loop = FALSE;
  } while (loop);

  if (*scan == '-' && *(scan+1) == '1') {
    *Token = -1;
    scan += 2;
  } else if (!isalnum((int)*scan) && *scan != '_')  
    *Token = *scan++;
  else {
    
    char help, *idstart;

    idstart = scan;
    while (isalnum((int)*scan) || *scan == '_')
      scan++;

    help = *scan;
    *scan = '\0'; 
    *Id = memory_Malloc(sizeof(char)*(scan-idstart+1));
    strcpy(*Id, idstart);
    *scan = help; 
    *Token = DFG_ID;
  }
  return scan;
}


static void dfg_Match(int RequiredToken, int ActToken)

{
  if (RequiredToken != ActToken) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Parse error in settings: ");
    if (RequiredToken == DFG_ID)
      misc_UserErrorReport("Identifier");
    else
      misc_UserErrorReport("%d",RequiredToken);
    misc_UserErrorReport(" expected, ");
    if (ActToken == DFG_ID)
      misc_UserErrorReport("Identifier");
    else if (ActToken == '\0')
      misc_UserErrorReport("unexpected end");
    else
      misc_UserErrorReport("%d",ActToken);
    misc_UserErrorReport(" found.\n");
    misc_FinishUserErrorReport();
  }
}


static void dfg_Settings(char* Text)

{
  int    token;
  char*  id;
  int    number;
  SYMBOL symbol;

  Text += 2;                   
  Text[strlen(Text)-2] = '\0'; 

  Text = dfg_SettingsScanner(Text, &token, &id);
  while (token != '\0') {
    dfg_Match(DFG_ID, token);   
    if (strcmp(id, "set_precedence") == 0) {
      misc_StringFree(id);
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match('(', token);
      do {
        Text = dfg_SettingsScanner(Text, &token, &id);
        if (token == '(') { 
          Text = dfg_SettingsScanner(Text, &token, &id);
          dfg_Match(DFG_ID, token);
	  symbol = symbol_Lookup(id);
	  if (symbol == 0) {
	    misc_StartUserErrorReport();
	    misc_UserErrorReport("\n Undefined symbol %s in precedence list.\n", id);
	    misc_FinishUserErrorReport();
	  }
	  misc_StringFree(id);
          Text = dfg_SettingsScanner(Text, &token, &id);
          dfg_Match(',', token);
          Text = dfg_SettingsScanner(Text, &token, &id);
          dfg_Match(DFG_ID, token);  
	  number = misc_StringToInt(id);
	  symbol_SetWeight(symbol, number);
	  misc_StringFree(id);
          Text = dfg_SettingsScanner(Text, &token, &id);
          if (token == ',') { 
            Text = dfg_SettingsScanner(Text, &token, &id);
            dfg_Match(DFG_ID, token);
	    if (strcmp(id, "r") == 0)
	      
	      symbol_AddProperty(symbol, ORDRIGHT);
	    else if (strcmp(id, "m") == 0)
	      
	      symbol_AddProperty(symbol, ORDMUL);
	    else if (strcmp(id, "l") != 0) {
	      misc_StartUserErrorReport();
	      misc_UserErrorReport("\n Unknown symbol status \"%s\" in precedence list.\n", id);
	      misc_FinishUserErrorReport();
	    }
	    misc_StringFree(id);
	    Text = dfg_SettingsScanner(Text, &token, &id);
          }
	  dfg_Match(')', token);
	} else {
	  dfg_Match(DFG_ID, token);
	  symbol = symbol_Lookup(id);
	  if (symbol == 0) {
	    misc_StartUserErrorReport();
	    misc_UserErrorReport("\n Undefined symbol %s in precedence list.\n", id);
	    misc_FinishUserErrorReport();
	  }
	  misc_StringFree(id);
	}
	symbol_SetOrdering(symbol, symbol_GetIncreasedOrderingCounter());
        Text = dfg_SettingsScanner(Text, &token, &id);
      } while (token == ',');
      dfg_Match(')', token);
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match('.', token);
    } else if (strcmp(id, "set_flag") == 0) {
      int  flag;
      misc_StringFree(id);
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match('(', token);
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match(DFG_ID, token);
      flag = flag_Id(id);
      if (flag == -1) {
	misc_StartUserErrorReport();
	misc_UserErrorReport("\n Found unkown flag ");
	misc_UserErrorReport("%s.\n",id); 
	misc_FinishUserErrorReport();
      }
      misc_StringFree(id);
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match(',', token);
      Text = dfg_SettingsScanner(Text, &token, &id);
      if (token == -1)
	number = -1;
      else {
	dfg_Match(DFG_ID, token);  
	number = misc_StringToInt(id);
	misc_StringFree(id);
      }
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match(')', token);
      Text = dfg_SettingsScanner(Text, &token, &id);
      dfg_Match('.', token);
      flag_SetFlagValue(flag, number);
    } else if (strcmp(id, "set_DomPred") == 0) {
      misc_StringFree(id); 
      Text = dfg_SettingsScanner(Text, &token, &id); 
      dfg_Match('(', token); 
      do {
	Text = dfg_SettingsScanner(Text, &token, &id);  
	dfg_Match(DFG_ID, token);
	symbol = symbol_Lookup(id); 
	if (symbol == 0) { 
	  misc_StartUserErrorReport();
	  misc_UserErrorReport("\n Undefined symbol %s in precedence list.\n", id); 
	  misc_FinishUserErrorReport(); 
	} 
	misc_StringFree(id); 
	symbol_AddProperty(symbol, DOMPRED);
	Text = dfg_SettingsScanner(Text, &token, &id);  
      } while (token == ',');
      dfg_Match(')', token);
      Text = dfg_SettingsScanner(Text, &token, &id); 
      dfg_Match('.', token); 
    } else {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n Unknown identifier in settings: ");
      misc_UserErrorReport("%s.\n",id);
      misc_FinishUserErrorReport();
    }
    Text = dfg_SettingsScanner(Text, &token, &id);
  }

  *(Text-1) = '*'; 
}    





typedef struct {
  SYMBOL symbol;
  BOOL   valid;
  int    arity;
} DFG_SYMENTRY, *DFG_SYM;

static __inline__ DFG_SYM dfg_SymCreate(void)
{
  return (DFG_SYM) memory_Malloc(sizeof(DFG_SYMENTRY));
}

static __inline__ void dfg_SymFree(DFG_SYM Entry)
{
  memory_Free(Entry, sizeof(DFG_SYMENTRY));
}


static void dfg_SymAdd(SYMBOL Symbol)

{
  DFG_SYM newEntry = dfg_SymCreate();
  newEntry->symbol = Symbol;
  newEntry->valid  = FALSE;
  newEntry->arity  = 0;
  dfg_SYMBOLLIST = list_Cons(newEntry, dfg_SYMBOLLIST);
}


static void dfg_SymCheck(SYMBOL Symbol, NAT Arity)

{
  LIST scan = dfg_SYMBOLLIST;
  while (!list_Empty(scan)) {
    DFG_SYM actEntry = (DFG_SYM) list_Car(scan);
    if (actEntry->symbol == Symbol) {
      if (actEntry->valid) {
	if (actEntry->arity != Arity) {
	  misc_StartUserErrorReport();
	  misc_UserErrorReport("\n Line %u: The actual arity %u of symbol %s differs "
			       " from the previous arity %u.\n", dfg_LINENUMBER, Arity,
			       symbol_Name(Symbol), actEntry->arity);
	  misc_FinishUserErrorReport();
	}
      } else {
	
	actEntry->arity = Arity;
	actEntry->valid = TRUE;
      }
      return;
    }
    scan = list_Cdr(scan);
  }

  
  
  if (symbol_Arity(Symbol) != Arity) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Line %u: Symbol %s was declared with arity %u.\n",
			 dfg_LINENUMBER, symbol_Name(Symbol), symbol_Arity(Symbol));
    misc_FinishUserErrorReport();
  }
}


static void dfg_SymCleanUp(void)

{
  while (!list_Empty(dfg_SYMBOLLIST)) {
    DFG_SYM actEntry  = (DFG_SYM) list_Car(dfg_SYMBOLLIST);
    SYMBOL  actSymbol = actEntry->symbol;

    if (actEntry->arity != symbol_Arity(actSymbol))
      symbol_SetArity(actSymbol, actEntry->arity);

    dfg_SymFree(actEntry);
    dfg_SYMBOLLIST = list_Pop(dfg_SYMBOLLIST);
  }
}





  
typedef struct {
  char*  name;
  SYMBOL symbol;
} DFG_VARENTRY, *DFG_VAR;

static __inline__ char* dfg_VarName(DFG_VAR Entry)
{
  return Entry->name;
}

static __inline__ SYMBOL dfg_VarSymbol(DFG_VAR Entry)
{
  return Entry->symbol;
}

static __inline__ DFG_VAR dfg_VarCreate(void)
{
  return (DFG_VAR) memory_Malloc(sizeof(DFG_VARENTRY));
}

static void dfg_VarFree(DFG_VAR Entry)
{
  misc_StringFree(Entry->name);
  memory_Free(Entry, sizeof(DFG_VARENTRY));
}

static void dfg_VarStart(void)
{
  dfg_VARLIST = list_Push(list_Nil(), dfg_VARLIST);
  dfg_VARDECL = TRUE;
}

static void dfg_VarStop(void)
{
  dfg_VARDECL = FALSE;
}

static void dfg_VarBacktrack(void)
{
  list_DeleteWithElement(list_Top(dfg_VARLIST), dfg_VarFree);
  dfg_VARLIST = list_Pop(dfg_VARLIST);
}

static void dfg_VarCheck(void)

{
  if (!list_Empty(dfg_VARLIST)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In dfg_VarCheck: List of variables should be empty!\n");
    misc_FinishErrorReport();
  }
  symbol_ResetStandardVarCounter();
}

static SYMBOL dfg_VarLookup(char* Name)

{
  LIST   scan, scan2;
  SYMBOL symbol = symbol_Null();

  scan  = dfg_VARLIST;
  scan2 = list_Nil();
  while (!list_Empty(scan) && list_Empty(scan2)) {
    scan2 = list_Car(scan);
    while (!list_Empty(scan2) &&
	   (strcmp(dfg_VarName(list_Car(scan2)), Name) != 0))
      scan2 = list_Cdr(scan2);
    scan = list_Cdr(scan);
  }

  if (!list_Empty(scan2)) {
    
    misc_StringFree(Name);
    symbol = dfg_VarSymbol(list_Car(scan2));
  } else {
    
    if (dfg_VARDECL) {
      DFG_VAR newEntry = dfg_VarCreate();
      newEntry->name   = Name;
      newEntry->symbol = symbol_CreateStandardVariable();
      
      list_Rplaca(dfg_VARLIST, list_Cons(newEntry,list_Car(dfg_VARLIST)));
      symbol = dfg_VarSymbol(newEntry);
    } else {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n Line %u: Free Variable %s.\n", dfg_LINENUMBER, Name);
      misc_FinishUserErrorReport();
    }
  }
  return symbol;
}
