




 





















%{

#include <ctype.h>    
#include <errno.h>
#include "misc.h"
#include "memory.h"
#include "symbol.h"
#include "term.h"
#include "dfg.h"
#include "dfgparser.h"

extern NAT dfg_LINENUMBER;  

static NAT dfg_CountNewlines(char*);

static __inline__ char* dfg_StringCopy(void)
{
  char *copy;
  copy = (char*) memory_Malloc(yyleng+1);
  strcpy(copy, yytext);
  return copy;
}

%}

%option noyywrap

%option always-interactive

%option nounput

%%

and				return DFG_AND;
author				return DFG_AUTHOR;
axioms				return DFG_AXIOMS;
begin_problem			return DFG_BEGPROB;
by				return DFG_BY;
clause				return DFG_CLAUSE;
cnf				return DFG_CNF;
conjectures			return DFG_CONJECS;
date				return DFG_DATE;
description			return DFG_DESC;
dnf				return DFG_DNF;
end_of_list			return DFG_ENDLIST;
end_problem			return DFG_ENDPROB;
equal				return DFG_EQUAL;
equiv				return DFG_EQUIV;
exists				return DFG_EXISTS;
false				return DFG_FALSE;
forall				return DFG_FORALL;
formula				return DFG_FORMULA;
freely				return DFG_FREELY;
functions			return DFG_FUNC;
generated			return DFG_GENERATED;
hypothesis			return DFG_HYPOTH;
implied				return DFG_IMPLIED;
implies				return DFG_IMPLIES;
KIV				return DFG_KIV;
LEM				return DFG_LEM;
list_of_clauses			return DFG_CLSLIST;
list_of_declarations		return DFG_DECLLIST;
list_of_descriptions		return DFG_DESCLIST;
list_of_formulae		return DFG_FORMLIST;
list_of_general_settings	return DFG_GENSET;
list_of_proof			return DFG_PRFLIST;
list_of_settings		return DFG_SETTINGS;
list_of_symbols			return DFG_SYMLIST;
list_of_terms                   return DFG_TERMLIST;
logic				return DFG_LOGIC;
name				return DFG_NAME;
not				return DFG_NOT;
operators			return DFG_OPERAT;
or				return DFG_OR;
OTTER				return DFG_OTTER;
predicate			return DFG_PRED;
predicates			return DFG_PRDICAT;
PROTEIN				return DFG_PROTEIN;
quantifiers			return DFG_QUANTIF;
resolution			return DFG_RESOLU;
satisfiable			return DFG_SATIS;
SATURATE			return DFG_SATURATE;
SETHEO				return DFG_SETHEO;
sort				return DFG_SORT;
sorts				return DFG_SORTS;
SPASS				return DFG_SPASS;
status				return DFG_STATUS;
subsort				return DFG_SUBSORT;
true				return DFG_TRUE;
unknown				return DFG_UNKNOWN;
unsatisfiable			return DFG_UNSATIS;
version				return DFG_VERSION;
3TAP				return DFG_3TAP;
"-1"				return DFG_MINUS1;

"%"[^\n]*			
\{\*([^\*]|(\*[^\}]))*\*\}	{ dfg_lval.string = dfg_StringCopy();
				  dfg_LINENUMBER += dfg_CountNewlines(yytext);
                                  return DFG_TEXT;
                                }
[0-9]+				{ unsigned long n;
                                  errno = 0;
                                  n = strtoul(yytext, NULL, 10);
				  if (errno != 0 || n > INT_MAX) {
                                    misc_StartUserErrorReport();
                                    misc_UserErrorReport("\n Number too big in line %d.\n",
							 dfg_LINENUMBER);
                                    misc_FinishUserErrorReport();
                                  }
                                  dfg_lval.number = (int) n;
                                  return DFG_NUM;
                                }
[a-zA-Z0-9_]+			{ dfg_lval.string = dfg_StringCopy();
                                  return DFG_ID;
                                }
[\xd \t]+				
"\n"				dfg_LINENUMBER++;
[-()\[\],\.]			return yytext[0];
.				{ misc_StartUserErrorReport();
                                  misc_UserErrorReport("\n Illegal character '");
				  if (isprint((int)yytext[0]))
				    misc_UserErrorReport("%c",yytext[0]);
				  else
				    misc_UserErrorReport("\\x%x", (unsigned int) yytext[0]);
				  misc_UserErrorReport("' in line %d.\n", dfg_LINENUMBER);
                                  misc_FinishUserErrorReport();
                                }

%%

static NAT dfg_CountNewlines(char* Text)
{
  NAT result = 0;

  while (*Text != 0) {
    if (*Text++ == '\n')
      result++;
  }
  return result;
}
