




 

























#include "doc-proof.h"





int dp_DEPTH;






void dp_Init(void)
{
  dp_DEPTH = 0;
}


void dp_FPrintDFGProof(LIST Clauses, char *argv)

{
  FILE   *Output;
  CLAUSE Clause;
  LIST   AxClauses,ConClauses,ProofClauses,Scan;
  char   *name;

  AxClauses = ConClauses = ProofClauses = list_Nil();
  
  name = memory_Malloc(sizeof(char)*strlen(argv)+5);
  sprintf(name,"%s.prf",argv);

  Output = misc_OpenFile(name,"w");
  fprintf(Output, "begin_problem(Unknown).\n\n");
  fprintf(Output, "list_of_descriptions.\n");
  fprintf(Output, "name({*%s*}).\n",argv);
  fprintf(Output, "author({*SPASS*}).\n");
  fprintf(Output, "status(unsatisfiable).\n");
  fprintf(Output, "description({*File generated by SPASS containing a proof.*}).\n");
  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "list_of_symbols.\n");
  fol_FPrintDFGSignature(Output);
  fprintf(Output, "end_of_list.\n\n");

  for(Scan=Clauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    if (clause_IsFromInput(Clause)) {
      if (clause_GetFlag(Clause,CONCLAUSE))
	ConClauses = list_Cons(Clause,ConClauses);
      else
	AxClauses  = list_Cons(Clause,AxClauses);
    }
    else
      ProofClauses  = list_Cons(Clause,ProofClauses);
  }

  ConClauses   = list_NReverse(ConClauses);
  AxClauses    = list_NReverse(AxClauses);
  ProofClauses = list_NReverse(ProofClauses);
  
  clause_FPrintCnfDFG(Output, FALSE, AxClauses, ConClauses);
  fprintf(Output, "\nlist_of_proof(resolution).\n");
  for(Scan=ProofClauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    clause_FPrintDFG(Output,list_Car(Scan),TRUE);
  }
  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "end_problem.\n\n");
  misc_CloseFile(Output,name);
  printf("\nDFG Proof printed to: %s\n",name);

  list_Delete(ConClauses);
  list_Delete(AxClauses);
  list_Delete(ProofClauses);
  memory_Free(name,sizeof(char)*strlen(argv)+5);
}

LIST dp_PrintProof(PROOFSEARCH Search, LIST Clauses, char *argv)

{
  LIST   ProofClauses,Scan,EmptyClauses,AllClauses, ReducedProof;
  LIST   Missing;
  LIST   Incomplete;
  LIST   SplitClauses;

  Missing = pcheck_ConvertParentsInSPASSProof(Search, Clauses);
  
  if (!list_Empty(Missing)) {
    printf("\nNOTE: clauses with following numbers have not been found:\n");
    for (; !list_Empty(Missing); Missing = list_Cdr(Missing))
      printf("%d ", (int)list_Car(Missing)); 
    printf("\n"); }

  EmptyClauses = list_Copy(Clauses); 
  ProofClauses = list_Nil();
  AllClauses   = list_Nconc(list_Copy(prfs_DocProofClauses(Search)),
			    list_Nconc(list_Copy(prfs_UsableClauses(Search)), list_Copy(prfs_WorkedOffClauses(Search))));

  

  SplitClauses = list_Nil();
  for (Scan = AllClauses; !list_Empty(Scan); Scan = list_Cdr(Scan)) 
    if (clause_IsFromSplitting(list_Car(Scan))) 
      SplitClauses = list_Cons(list_Car(Scan), SplitClauses);

  
  pcheck_ClauseListRemoveFlag(EmptyClauses, MARKED);
  pcheck_ClauseListRemoveFlag(AllClauses, MARKED);
  pcheck_MarkRecursive(EmptyClauses);
  pcheck_MarkRecursive(SplitClauses);
  
  
  ProofClauses = list_Nil();
  for (Scan = AllClauses; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    if (clause_GetFlag(list_Car(Scan), MARKED))
      ProofClauses = list_Cons(list_Car(Scan), ProofClauses); 
  }

  
  ProofClauses = list_Nconc(ProofClauses, list_Copy(EmptyClauses));
  ProofClauses = pcheck_ClauseNumberMergeSort(ProofClauses);
  ReducedProof = pcheck_ReduceSPASSProof(ProofClauses); 

  dp_SetProofDepth(pcheck_SeqProofDepth(ReducedProof));
  
  pcheck_ParentPointersToParentNumbers(AllClauses);
  pcheck_ParentPointersToParentNumbers(Clauses);

  

  Incomplete = list_Nil();
  for (Scan = ReducedProof; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    if (clause_GetFlag(list_Car(Scan), HIDDEN))
      Incomplete = list_Cons(list_Car(Scan), Incomplete); }
  if (!list_Empty(Incomplete)) {
    printf("NOTE: Following clauses in reduced proof have incomplete parent sets:\n");
    for (Scan = Incomplete; !list_Empty(Scan); Scan = list_Cdr(Scan))
      printf("%d ", clause_Number(list_Car(Scan))); 
    printf("\n"); }

  printf("\n\nHere is a proof with depth %d, length %d :\n",
	 dp_ProofDepth(), list_Length(ReducedProof));
  clause_ListPrint(ReducedProof);
  if (flag_GetFlagValue(flag_FPDFGPROOF))
    dp_FPrintDFGProof(ReducedProof,argv);

  fflush(stdout);

  list_Delete(EmptyClauses);
  list_Delete(AllClauses);
  list_Delete(ProofClauses);
  list_Delete(SplitClauses);
  list_Delete(Missing);
  list_Delete(Incomplete); 

  return ReducedProof;
}




