




 





















#ifndef _HASHARRAY_
#define _HASHARRAY_





#include "list.h"





#define hsh__SIZE 29            



typedef LIST* HASH;

void hsh_Check(HASH H);





static __inline__ unsigned long hsh_Index(POINTER Key)

{
  return (unsigned long)Key % hsh__SIZE;
}

static __inline__ void hsh_Put(HASH H, POINTER Key, POINTER Value)

{
  LIST Scan, Pair;
  unsigned long HashKey;
  HashKey = hsh_Index(Key);
  for(Scan = H[HashKey]; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Pair = list_Car(Scan);
    if (list_PairFirst(Pair) == Key) {
      if (!list_PointerMember(list_PairSecond(Pair), Value))
	list_Rplacd(Pair, list_Cons(Value, list_PairSecond(Pair)));
#ifdef DEBUG
      hsh_Check(H);
#endif
      return;
    }
  }
  H[HashKey] = list_Cons(list_PairCreate(Key, list_List(Value)), H[HashKey]);
#ifdef DEBUG
  hsh_Check(H);
#endif
}

static __inline__ void hsh_PutList(HASH H, POINTER Key, LIST List)

{
  LIST Scan, Pair;
  unsigned long HashKey;
  HashKey = hsh_Index(Key);
  for(Scan = H[HashKey]; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Pair = list_Car(Scan);
    if (list_PairFirst(Pair) == Key) {
      list_Rplacd(Pair, list_Nconc(list_PairSecond(Pair), List));
#ifdef DEBUG
      hsh_Check(H);
#endif
      return;
    }
  }
  H[HashKey] = list_Cons(list_PairCreate(Key, List), H[HashKey]);
#ifdef DEBUG
  hsh_Check(H);
#endif
}

static __inline__ void hsh_PutListWithCompareFunc(HASH H, POINTER Key, LIST List, BOOL (*Test)(), 
						  unsigned long (*HashFunc)())

{
  LIST Scan, Pair;
  unsigned long HashKey;
  HashKey = (unsigned long) HashFunc(Key);
  for(Scan = H[HashKey]; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Pair = list_Car(Scan);
    if (Test(list_PairFirst(Pair), Key)) {
      list_Rplacd(Pair, list_Nconc(list_PairSecond(Pair), List));
#ifdef DEBUG
      hsh_Check(H);
#endif
      return;
    }
  }
  H[HashKey] = list_Cons(list_PairCreate(Key, List), H[HashKey]);
#ifdef DEBUG
  hsh_Check(H);
#endif
}

static __inline__ LIST hsh_Get(HASH H, POINTER Key)

{
  LIST Scan, Pair;
  for(Scan = H[hsh_Index(Key)]; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Pair = list_Car(Scan);
    if (list_PairFirst(Pair) == Key)
      return list_PairSecond(Pair);
  }
  return NULL;
}

static __inline__ void hsh_DelItem(HASH H, POINTER Key)

{
  LIST Scan, Pair;
  unsigned long k;
  k = hsh_Index(Key);
  for(Scan = H[k]; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Pair = list_Car(Scan);
    if (list_PairFirst(Pair) == Key) {
      list_Delete(list_PairSecond(Pair));
      list_PairFree(Pair);
      H[k] = list_PointerDeleteElement(H[k], Pair);
      return;
    }
  }
}

static __inline__ LIST hsh_GetWithCompareFunc(HASH H, POINTER Key, BOOL (*Test)(),
					      unsigned long (*HashFunc)())

{
  LIST Scan, Pair;
  for(Scan = H[HashFunc(Key)]; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Pair = list_Car(Scan);
    if (Test(list_PairFirst(Pair), Key))
      return list_PairSecond(Pair);
  }
  return NULL;
}


static __inline__ unsigned long hsh_StringHashKey(char* Label)
{
  unsigned long i, s;
  s=0;
  for(i=0; i <= strlen(Label); i++)
    s += (int) Label[i];
  s = s % hsh__SIZE;
  return s;
}





HASH hsh_Create(void);
void hsh_Reset(HASH H);
void hsh_Delete(HASH H);
LIST hsh_GetAllEntries(HASH H);

#endif


