




 





















%{

#include "flags.h"
#include "ia.h"
#include "symbol.h"
#include "term.h"
#include "foldfg.h"
#include "clause.h"

extern NAT dfg_LINENUMBER;    
LIST       ia_PROOFREQUEST;   

void yyerror(const char*);
int  yylex(void);		

static SYMBOL ia_Symbol(char*, NAT);
static TERM   ia_CreateQuantifier(SYMBOL, LIST, TERM);

static __inline__ void  ia_StringFree(char* String)
{
  memory_Free(String, sizeof(char)*(strlen(String)+1));
}

static __inline__ TERM ia_TermCreate(char* Name, LIST Arguments)

{
  return term_Create(ia_Symbol(Name,list_Length(Arguments)), Arguments);
}





static void ia_SymCheck(SYMBOL, NAT);











static LIST ia_VARLIST;
static BOOL ia_VARDECL;

static void   ia_VarStart(void);
static void   ia_VarStop(void);
static void   ia_VarBacktrack(void);
static void   ia_VarCheck(void);
static SYMBOL ia_VarLookup(char*);

#define YY_INPUT(buf,result,max_size) \
{ \
  int c = getc(ia_in); \
  result = (c == EOF) ? YY_NULL : (buf[0] = c, 1); \
}

#define YYERROR_VERBOSE

%}

%union {
  int       number;
  char*     string;
  SYMBOL    symbol;
  TERM      term;
  LIST      list;
}

%token IA_AND IA_EQUAL IA_EQUIV IA_EXISTS IA_FALSE IA_FORALL IA_IMPLIED
%token IA_IMPLIES IA_NOT IA_OR IA_PROVE IA_TRUE
%token <number> IA_NUM
%token <string> IA_ID
%type <string>	id
%type <symbol>  binsymbol nsymbol quantsymbol
%type <term>    qterm term
%type <list>    labellist labellistopt qtermlist termlist

%%





proofrequest:		{ YYABORT; }
		| IA_PROVE '(' term ',' labellistopt ',' IA_NUM ')' '.' {
					  ia_VarCheck();
					  ia_PROOFREQUEST = list_PairCreate($3,$5);
					  flag_SetFlagValue(flag_TIMELIMIT, $7);
					  YYACCEPT;
                                        }
;





termlist:	term			{ $$ = list_List($1); }
		| termlist ',' term	{ $$ = list_Nconc($1, list_List($3)); }
;

term:		id		{ $$ = ia_TermCreate($1, list_Nil()); }
		| IA_TRUE	{ $$ = term_Create(fol_True(),list_Nil()); }
		| IA_FALSE	{ $$ = term_Create(fol_False(),list_Nil()); }
		| IA_NOT '(' term ')'
				{ $$ = term_Create(fol_Not(),list_List($3)); }
		| binsymbol '(' term ',' term ')'
				{ $$ = term_Create($1, list_Cons($3, list_List($5))); }
		| nsymbol '(' termlist ')'
					{ $$ = term_Create($1, $3); }
		| id '(' termlist ')'
				{ $$ = ia_TermCreate($1, $3); }
		| quantsymbol '(' '[' 	{ ia_VarStart(); }
		  qtermlist 		{ ia_VarStop(); }
		  ']' ',' term ')'
			  	{ ia_VarBacktrack();
				  $$ = ia_CreateQuantifier($1,$5,$9);
				}
		| id '(' '['		{ ia_VarStart(); }
		  qtermlist		{ ia_VarStop(); }
		  ']' ',' term ')'
				{ misc_StartUserErrorReport();
				  misc_UserErrorReport("\n Line %d: SPASS can't handle the quantifier %s.\n", dfg_LINENUMBER, $1);
				  misc_FinishUserErrorReport();
				}
;

binsymbol:	IA_EQUAL		{ $$ = fol_Equality(); }
		| IA_EQUIV		{ $$ = fol_Equiv();    }
		| IA_IMPLIED		{ $$ = fol_Implied();  }
		| IA_IMPLIES		{ $$ = fol_Implies();  }
;

nsymbol:	IA_AND			{ $$ = fol_And(); }
		| IA_OR		{ $$ = fol_Or();  }
;

quantsymbol:	IA_EXISTS		{ $$ = fol_Exist(); }
		| IA_FORALL		{ $$ = fol_All(); }
;

id:		IA_ID			{ $$ = $1; }
		| IA_NUM		{ $$ = misc_IntToString($1); }
;

qtermlist:	qterm			{ $$ = list_List($1); }
		| qtermlist ',' qterm	{ $$ = list_Nconc($1, list_List($3)); }
;

qterm:		id			{ SYMBOL s = ia_Symbol($1,0);
					  if (!symbol_IsVariable(s)) {
					    misc_StartUserErrorReport();
					    misc_UserErrorReport("\n Line %d: %s",dfg_LINENUMBER,
								 symbol_Name(s));
					    misc_UserErrorReport(" is not a variable.\n");
					    misc_FinishUserErrorReport();
					  }
					  $$ = term_Create(s, list_Nil());
					}
		| id '(' id ')'		{ SYMBOL p, v;
					  p = ia_Symbol($1, 1);
					  if (!symbol_IsPredicate(p)) {
					    misc_StartUserErrorReport();
					    misc_UserErrorReport("\n Line %d: %s",dfg_LINENUMBER,
								 symbol_Name(p));
					    misc_UserErrorReport(" is not a predicate.\n");
					    misc_FinishUserErrorReport();
					  }
					  v = ia_Symbol($3, 0);
					  if (!symbol_IsVariable(v)) {
					    misc_StartUserErrorReport();
					    misc_UserErrorReport("\n Line %d: %s",dfg_LINENUMBER,
								 symbol_Name(v));
					    misc_UserErrorReport(" is not a variable.\n");
					    misc_FinishUserErrorReport();
					  }
					  $$ = term_Create(p, list_List(term_Create(v,list_Nil())));
					}
;

labellistopt:	'[' ']'			{ $$ = list_Nil(); }
		| '[' labellist ']'	{ $$ = $2; }
;

labellist:	id		        { $$ = list_List($1); }
		| labellist ',' id	{ $$ = list_Nconc($1, list_List($3)); }
;

%%


void yyerror(const char *s)
{
  misc_StartUserErrorReport();
  misc_UserErrorReport("\n Line %i: %s\n", dfg_LINENUMBER, s);
  misc_FinishUserErrorReport();

}

LIST ia_GetNextRequest(FILE* Input)

{
  extern FILE* ia_in;  

  ia_in           = Input;
  ia_PROOFREQUEST = list_Nil();
  ia_parse();
  
  return ia_PROOFREQUEST;
}






static SYMBOL ia_Symbol(char* Name, NAT Arity)

{
  SYMBOL symbol;

  symbol = symbol_Lookup(Name);
  if (symbol != 0) {
    ia_StringFree(Name);
    ia_SymCheck(symbol, Arity); 
  } else {
    
    if (Arity > 0) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n Line %d: Undefined symbol %s.\n",dfg_LINENUMBER,Name);
      misc_FinishUserErrorReport();
    }
    symbol = ia_VarLookup(Name);
  }
  return symbol;
}


static TERM ia_CreateQuantifier(SYMBOL Symbol, LIST VarTermList, TERM Term)

{
  LIST varlist, sortlist, scan;
  TERM helpterm;

  
  varlist = sortlist = list_Nil();
  for ( ; !list_Empty(VarTermList); VarTermList = list_Pop(VarTermList)) {
    helpterm = list_Car(VarTermList);
    if (term_IsVariable(helpterm)) {
      varlist = list_Nconc(varlist, list_List((POINTER)term_TopSymbol(helpterm)));
      term_Delete(helpterm);
    } else {
      SYMBOL var = term_TopSymbol(term_FirstArgument(helpterm));
      varlist  = list_Nconc(varlist, list_List((POINTER)var));
      sortlist = list_Nconc(sortlist, list_List(helpterm));
    }
  }

  varlist = list_PointerDeleteDuplicates(varlist);
  
  for (scan = varlist; !list_Empty(scan); scan = list_Cdr(scan))
    list_Rplaca(scan, term_Create((SYMBOL)list_Car(scan), list_Nil()));

  if (!list_Empty(sortlist)) {
    if (symbol_Equal(fol_All(), Symbol)) {
      
      if (symbol_Equal(fol_Or(), term_TopSymbol(Term))) {
	
	
	
	for (scan = sortlist; !list_Empty(scan); scan = list_Cdr(scan))
	  
	  list_Rplaca(scan, term_Create(fol_Not(), list_List(list_Car(scan))));
	sortlist = list_Nconc(sortlist, term_ArgumentList(Term));
	term_RplacArgumentList(Term, sortlist);
      } else {
	
	if (list_Empty(list_Cdr(sortlist))) {
	  
	  list_Rplacd(sortlist, list_List(Term));
	  Term = term_Create(fol_Implies(), sortlist);
	} else {
	  
	  helpterm = term_Create(fol_And(), sortlist);
	  Term = term_Create(fol_Implies(), list_Cons(helpterm, list_List(Term)));
	}
      }
    } else if (symbol_Equal(fol_Exist(), Symbol)) {
      
      if (symbol_Equal(fol_And(), term_TopSymbol(Term))) {
	
	
	sortlist = list_Nconc(sortlist, term_ArgumentList(Term));
	term_RplacArgumentList(Term, sortlist);
      } else {
	sortlist = list_Nconc(sortlist, list_List(Term));
	Term = term_Create(fol_And(), sortlist);
      }
    }
  }
  helpterm = fol_CreateQuantifier(Symbol, varlist, list_List(Term));
  return helpterm;
}






static void ia_SymCheck(SYMBOL Symbol, NAT Arity)

{
  
  if (symbol_Arity(Symbol) != symbol_ArbitraryArity() &&
      symbol_Arity(Symbol) != Arity) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Line %u: Symbol %s", dfg_LINENUMBER, symbol_Name(Symbol));
    misc_UserErrorReport(" was declared with arity %u.\n", symbol_Arity(Symbol));
    misc_FinishUserErrorReport();
  }
}





  
typedef struct {
  char*  name;
  SYMBOL symbol;
} IA_VARENTRY, *IA_VAR;

static __inline__ char* ia_VarName(IA_VAR Entry)
{
  return Entry->name;
}

static __inline__ SYMBOL ia_VarSymbol(IA_VAR Entry)
{
  return Entry->symbol;
}

static __inline__ IA_VAR ia_VarCreate(void)
{
  return (IA_VAR) memory_Malloc(sizeof(IA_VARENTRY));
}

static void ia_VarFree(IA_VAR Entry)
{
  ia_StringFree(Entry->name);
  memory_Free(Entry, sizeof(IA_VARENTRY));
}

static void ia_VarStart(void)
{
  ia_VARLIST = list_Push(list_Nil(), ia_VARLIST);
  ia_VARDECL = TRUE;
}

static void ia_VarStop(void)
{
  ia_VARDECL = FALSE;
}

static void ia_VarBacktrack(void)
{
  list_DeleteWithElement(list_Top(ia_VARLIST), ia_VarFree);
  ia_VARLIST = list_Pop(ia_VARLIST);
}

static void ia_VarCheck(void)

{
  if (!list_Empty(ia_VARLIST)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In ia_VarCheck: List of variables should be empty!\n");
    misc_FinishErrorReport();
  }
  symbol_ResetStandardVarCounter();
}

static SYMBOL ia_VarLookup(char* Name)

{
  LIST   scan, scan2;
  SYMBOL symbol;

  symbol = symbol_Null();

  scan  = ia_VARLIST;
  scan2 = list_Nil();
  while (!list_Empty(scan) && list_Empty(scan2)) {
    scan2 = list_Car(scan);
    while (!list_Empty(scan2) &&
	   (strcmp(ia_VarName(list_Car(scan2)), Name) != 0))
      scan2 = list_Cdr(scan2);
    scan = list_Cdr(scan);
  }

  if (!list_Empty(scan2)) {
    
    ia_StringFree(Name);
    symbol = ia_VarSymbol(list_Car(scan2));
  } else {
    
    if (ia_VARDECL) {
      IA_VAR newEntry = ia_VarCreate();
      newEntry->name   = Name;
      newEntry->symbol = symbol_CreateStandardVariable();
      
      list_Rplaca(ia_VARLIST, list_Cons(newEntry,list_Car(ia_VARLIST)));
      symbol = ia_VarSymbol(newEntry);
    } else {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n Line %u: Free Variable %s.\n", dfg_LINENUMBER, Name);
      misc_FinishUserErrorReport();
    }
  }
  return symbol;
}
