




 





















#include "list.h"









LIST list_Copy(const LIST List)

{
  LIST Copy;
  LIST Scan1,Scan2;

 
  if (list_Empty(List))
    return(list_Nil());

  Copy = list_List(list_Car(List));
  Scan1 = Copy;
  Scan2 = list_Cdr(List);

  while (!list_Empty(Scan2)) {
    list_Rplacd(Scan1, list_List(list_Car(Scan2)));
    Scan1 = list_Cdr(Scan1);
    Scan2 = list_Cdr(Scan2);
  }
  return Copy;
}


LIST list_CopyWithElement(const LIST List, POINTER (*CopyElement)())

{
  LIST Copy;
  LIST Scan1,Scan2;

  if (list_Empty(List))
    return(list_Nil());

  Copy  = list_List(CopyElement(list_Car(List)));
  Scan1 = Copy;
  Scan2 = list_Cdr(List);

  while (!list_Empty(Scan2)) {
    list_Rplacd(Scan1, list_List(CopyElement(list_Car(Scan2))));
    Scan1 = list_Cdr(Scan1);
    Scan2 = list_Cdr(Scan2);
  }
  return Copy;
}


void list_InsertNext(LIST List, POINTER Pointer)

{
#ifdef DEBUG
  if (Pointer == NULL) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In list_InsertNext: NULL Pointer. ");
    misc_FinishErrorReport();
  }
#endif

  list_Rplacd(List, list_Cons(Pointer, list_Cdr(List)));
}


void list_NMapCar(LIST List, POINTER (*ElementFunc)())

{
  LIST Scan;

  for (Scan = List; !list_Empty(Scan); Scan = list_Cdr(Scan))
    list_Rplaca(Scan, ElementFunc(list_Car(Scan)));
}


void list_Apply(void (*Function)(),  LIST List)

{
  while (!list_Empty(List)) {
    Function(list_Car(List));
    List = list_Cdr(List);
  }
}


LIST list_Reverse(const LIST List)

{
  LIST ReverseList;
  LIST Scan;
  
  ReverseList = list_Nil();

  for (Scan=List;!list_Empty(Scan);Scan=list_Cdr(Scan)) 
    ReverseList = list_Cons(list_Car(Scan), ReverseList);

  return ReverseList;
}


LIST list_NReverse(LIST List)

{
  LIST ReverseList;
  LIST Scan1;
  LIST Scan2;
  
  ReverseList = list_Nil();

  for (Scan1=List;!list_Empty(Scan1);Scan1=list_Cdr(Scan1)) 
    ReverseList = list_Cons(list_Car(Scan1),ReverseList);

  for (Scan1=List, Scan2=ReverseList;
       !list_Empty(Scan1);
       Scan1=list_Cdr(Scan1), Scan2=list_Cdr(Scan2)) 
    list_Rplaca(Scan1, list_Car(Scan2));

  list_Delete(ReverseList);
  return List;
}


LIST list_PointerSort(LIST List)

{
  LIST Scan1,Scan2,Min;
  POINTER Exchange;
  
  for (Scan1=List;!list_Empty(Scan1);Scan1=list_Cdr(Scan1)) {
    Min = Scan1;
    for (Scan2=list_Cdr(Scan1);!list_Empty(Scan2);Scan2=list_Cdr(Scan2)) {
      if ((NAT)list_Car(Scan2)<(NAT)list_Car(Min)) {
	Exchange = list_Car(Min);
	list_Rplaca(Min,list_Car(Scan2));
	list_Rplaca(Scan2, Exchange);
      }
    }
  }

  return(List);
}


LIST list_Sort(LIST List, BOOL (*Test)())

{
  LIST Scan1,Scan2,Min;
  POINTER Exchange;
  
  for (Scan1=List; !list_Empty(Scan1); Scan1=list_Cdr(Scan1)) {
    Min = Scan1;
    for (Scan2 = list_Cdr(Scan1);
	 !list_Empty(Scan2);
	 Scan2 = list_Cdr(Scan2))
      if (Test(list_Car(Scan2), list_Car(Min))) {
	Exchange = list_Car(Min);
	list_Rplaca(Min, list_Car(Scan2));
	list_Rplaca(Scan2, Exchange);
      }
  }

  return(List);
}

LIST list_NumberSort(LIST List, NAT (*Number)())

{
  LIST Scan1,Scan2,Min;
  POINTER Exchange;
  
  for (Scan1=List; !list_Empty(Scan1); Scan1=list_Cdr(Scan1)) {
    Min = Scan1;
    for (Scan2 = list_Cdr(Scan1);
	 !list_Empty(Scan2);
	 Scan2 = list_Cdr(Scan2))
      if (Number(list_Car(Scan2)) < Number(list_Car(Min))) {
	Exchange = list_Car(Min);
	list_Rplaca(Min, list_Car(Scan2));
	list_Rplaca(Scan2, Exchange);
      }
  }

  return(List);
}

LIST list_GreaterNumberSort(LIST List, NAT (*Number)())

{
  LIST Scan1,Scan2,Min;
  POINTER Exchange;
  
  for (Scan1=List; !list_Empty(Scan1); Scan1=list_Cdr(Scan1)) {
    Min = Scan1;
    for (Scan2 = list_Cdr(Scan1);
	 !list_Empty(Scan2);
	 Scan2 = list_Cdr(Scan2))
      if (Number(list_Car(Scan2)) > Number(list_Car(Min))) {
	Exchange = list_Car(Min);
	list_Rplaca(Min, list_Car(Scan2));
	list_Rplaca(Scan2, Exchange);
      }
  }

  return(List);
}

LIST list_NNumberMerge(LIST List1, LIST List2, NAT (*Number)())

{
  LIST    Scan1,Scan2,Result,ResultStart;

  if (list_Empty(List1))
    return List2;

  if (list_Empty(List2))
    return List1;

  Result      = list_List(NULL);
  ResultStart = Result;
  Scan1       = List1;
  Scan2       = List2;

  while (!list_Empty(Scan1) && !list_Empty(Scan2)) {
    if (Number(list_Car(Scan1)) <= Number(list_Car(Scan2))) {
      list_Rplacd(Result,Scan1);
      Scan1  = list_Cdr(Scan1);
    }
    else {
      list_Rplacd(Result,Scan2);
      Scan2  = list_Cdr(Scan2);
    }
    Result = list_Cdr(Result);
  }
  if (list_Empty(Scan1))
    list_Rplacd(Result,Scan2);
  else
    list_Rplacd(Result,Scan1);
  Result = list_Cdr(ResultStart);
  list_Free(ResultStart);
  return Result;
}
      


POINTER list_DequeueNext(LIST List)

{
  POINTER Pointer;
  LIST    Memo;

  if (list_Empty(List)) return (POINTER)NULL;

  Memo = list_Cdr(List);

  if (list_Empty(Memo)) return (POINTER)NULL;
      
  Pointer = list_Car(Memo);
  list_Rplacd(List, Memo->cdr);
  memory_Free(Memo, sizeof(LIST_NODE));
  return Pointer;
}


POINTER list_NthElement(LIST List, NAT Number)

{
  while (!list_Empty(List) && --Number > 0)
    List = list_Cdr(List);
  
  if (list_Empty(List))
    return NULL;
  else
    return (list_Car(List));
}


void list_DeleteWithElement(LIST List, void (*ElementDelete)())

{
  LIST Scan;

  while (!list_Empty(List)) {
    Scan = list_Cdr(List);
    ElementDelete(list_Car(List));
    list_Free(List);
    List = Scan;
    }
}


NAT list_DeleteWithElementCount(LIST List, void (*ElementDelete)())

{
  int  Result;
  LIST Scan;

  Result = 0;

  while (!list_Empty(List)) {
    Scan = list_Cdr(List);
    ElementDelete(list_Car(List));
    list_Free(List);
    List = Scan;
    Result++;
  }

  return Result;
}


LIST list_DeleteElement(LIST List, POINTER Element, BOOL (*Test)())

{
  LIST   Scan1,Scan2;

  while(!list_Empty(List) && Test(Element, list_Car(List))) {
    Scan1 = list_Cdr(List);
    list_Free(List);
    List = Scan1;
  }

  if (list_Empty(List)) return(list_Nil());
  
  Scan2 = List;
  Scan1 = list_Cdr(List);

  while(!list_Empty(Scan1)) {
    if (Test(Element, list_Car(Scan1))) {
      list_Rplacd(Scan2, list_Cdr(Scan1));
      memory_Free(Scan1, sizeof(LIST_NODE));
      Scan1 = list_Cdr(Scan2);
    }
    else {
      Scan2 = Scan1;
      Scan1 = list_Cdr(Scan1);
    }
  }
  return(List);     
}

LIST list_DeleteElementIf(LIST List, BOOL (*Test)())

{
  LIST   Scan1,Scan2;

  while(!list_Empty(List) && Test(list_Car(List))) {
    Scan1 = list_Cdr(List);
    list_Free(List);
    List = Scan1;
  }

  if (list_Empty(List)) 
    return(list_Nil());
  
  Scan2 = List;
  Scan1 = list_Cdr(List);

  while(!list_Empty(Scan1)) {
    if (Test(list_Car(Scan1))) {
      list_Rplacd(Scan2, list_Cdr(Scan1));
      memory_Free(Scan1, sizeof(LIST_NODE));
      Scan1 = list_Cdr(Scan2);
    }
    else {
      Scan2 = Scan1;
      Scan1 = list_Cdr(Scan1);
    }
  }
  return(List);     
}

LIST list_DeleteElementIfFree(LIST List, BOOL (*Test)(), void (*Delete)())

{
  LIST   Scan1,Scan2;

  while(!list_Empty(List) && Test(list_Car(List))) {
    Scan1 = list_Cdr(List);
    Delete(list_Car(List));
    list_Free(List);
    List = Scan1;
  }

  if (list_Empty(List)) return(list_Nil());
  
  Scan2 = List;
  Scan1 = list_Cdr(List);

  while(!list_Empty(Scan1)) {
    if (Test(list_Car(Scan1))) {
      Delete(list_Car(Scan1));
      list_Rplacd(Scan2, list_Cdr(Scan1));
      memory_Free(Scan1, sizeof(LIST_NODE));
      Scan1 = list_Cdr(Scan2);
    }
    else {
      Scan2 = Scan1;
      Scan1 = list_Cdr(Scan1);
    }
  }
  return(List);     
}



LIST list_DeleteElementFree(LIST List, POINTER Element, BOOL (*Test)(),
			    void (*Free)())

{
  LIST   Scan1,Scan2;

  while(!list_Empty(List) && Test(Element, list_Car(List))) {
    Scan1 = list_Cdr(List);
    Free(list_Car(List));
    list_Free(List);
    List = Scan1;
  }

  if (list_Empty(List)) return(list_Nil());
  
  Scan2 = List;
  Scan1 = list_Cdr(List);

  while(!list_Empty(Scan1)) {
    if (Test(Element, list_Car(Scan1))) {
      list_Rplacd(Scan2, list_Cdr(Scan1));
      Free(list_Car(Scan1));
      list_Free(Scan1);
      Scan1 = list_Cdr(Scan2);
    }
    else {
      Scan2 = Scan1;
      Scan1 = list_Cdr(Scan1);
    }
  }
  return(List);     
}


LIST list_DeleteOneElement(LIST List, POINTER Element, BOOL (*Test)())

{
  LIST scan1, scan2;

  if (list_Empty(List))
    return List;
  else {
    if (Test(Element, list_Car(List)))
      return list_Pop(List);
  }
  
  for (scan2 = List, scan1 = list_Cdr(List); !list_Empty(scan1);
       scan2 = scan1, scan1 = list_Cdr(scan1)) {
    if (Test(Element, list_Car(scan1))) {
      list_Rplacd(scan2, list_Cdr(scan1));
      memory_Free(scan1, sizeof(LIST_NODE));
      scan1 = list_Cdr(scan2);
      return List;
    }
  }
  return List;
}


LIST list_PointerDeleteElement(LIST List, POINTER Element)

{
  LIST   Scan1,Scan2;

  while(!list_Empty(List) && Element == list_Car(List)) {
    Scan1 = list_Cdr(List);
    memory_Free(List, sizeof(LIST_NODE));
    List = Scan1;
  }

  if (list_Empty(List)) return(list_Nil());
  
  Scan2 = List;
  Scan1 = list_Cdr(List);

  while(!list_Empty(Scan1)) {
    if (Element == list_Car(Scan1)) {
      list_Rplacd(Scan2, list_Cdr(Scan1));
      memory_Free(Scan1, sizeof(LIST_NODE));
      Scan1 = list_Cdr(Scan2);
    }
    else {
      Scan2 = Scan1;
      Scan1 = list_Cdr(Scan1);
    }
  }
  return(List);     
}

LIST list_PointerDeleteElementFree(LIST List, POINTER Element, void (*Free)())

{

  LIST   Scan1,Scan2;
  
  while(!list_Empty(List) && Element == list_Car(List)) {
    Scan1 = list_Cdr(List);
    Free(list_Car(List));
    memory_Free(List, sizeof(LIST_NODE));
    List = Scan1;
  }

  if (list_Empty(List)) return(list_Nil());
  
  Scan2 = List;
  Scan1 = list_Cdr(List);

  while(!list_Empty(Scan1)) {
    if (Element == list_Car(Scan1)) {
      list_Rplacd(Scan2, list_Cdr(Scan1));
      Free(list_Car(Scan1));
      memory_Free(Scan1, sizeof(LIST_NODE));
      Scan1 = list_Cdr(Scan2);
    }
    else {
      Scan2 = Scan1;
      Scan1 = list_Cdr(Scan1);
    }
  }
  return(List);     
}
 

LIST list_PointerDeleteOneElement(LIST List, POINTER Element)

{
  LIST   Scan1,Scan2;

  if (list_Empty(List))
    return List;
  else {
    if (Element == list_Car(List))
      return list_Pop(List);
  }
  
  Scan2 = List;
  Scan1 = list_Cdr(List);

  while(!list_Empty(Scan1)) {
    if (Element == list_Car(Scan1)) {
      list_Rplacd(Scan2, list_Cdr(Scan1));
      memory_Free(Scan1, sizeof(LIST_NODE));
      Scan1 = list_Cdr(Scan2);
      return List;
    }
    else {
      Scan2 = Scan1;
      Scan1 = list_Cdr(Scan1);
    }
  }
  return List;     
}


BOOL list_DeleteFromList(LIST* List, POINTER Element)

{
  BOOL Found;
  LIST Scan1;

  Found = FALSE;

  while(list_Exist(*List) && Element == list_Car(*List)) {
    Scan1 = list_Cdr(*List);
    memory_Free(*List, sizeof(LIST_NODE));
    *List = Scan1;
    Found = TRUE;
  }

  if (list_Exist(*List)) {
    LIST Scan2;

    Scan2 = *List;
    Scan1 = list_Cdr(*List);

    while(list_Exist(Scan1)) {
      if (Element == list_Car(Scan1)) {
	list_Rplacd(Scan2, list_Cdr(Scan1));
	memory_Free(Scan1, sizeof(LIST_NODE));
	Scan1 = list_Cdr(Scan2);
	Found = TRUE;
      } else {
	Scan2 = Scan1;
	Scan1 = list_Cdr(Scan1);
      }
    }
  }

  return Found;
}


BOOL list_DeleteOneFromList(LIST* List, POINTER Element)

{
  if (list_Exist(*List)) {
    LIST Scan1;

    if (Element == list_Car(*List)) {
      Scan1 = list_Cdr(*List);
      memory_Free(*List, sizeof(LIST_NODE));
      *List = Scan1;
      return TRUE;
    } else {
      LIST Scan2;

      for (Scan2 = *List, Scan1 = list_Cdr(*List);
	   list_Exist(Scan1);) {
	if (Element == list_Car(Scan1)) {
	  list_Rplacd(Scan2, list_Cdr(Scan1));
	  memory_Free(Scan1, sizeof(LIST_NODE));
	  Scan1 = list_Cdr(Scan2);
	  return TRUE;
	} else {
	  Scan2 = Scan1;
	  Scan1 = list_Cdr(Scan1);
	}
      }
    }
  }
  return FALSE;
}


BOOL list_IsSetOfPointers(LIST L)

{
  for ( ; !list_Empty(L); L = list_Cdr(L)) {
    if (list_PointerMember(list_Cdr(L), list_Car(L)))
      return FALSE;
  }
  return TRUE;
}


LIST list_DeleteDuplicates(LIST List, BOOL (*Test)())

{
  LIST Scan;
  
  Scan = List;

  while (!list_Empty(Scan)) {
    list_Rplacd(Scan,
		 list_DeleteElement(list_Cdr(Scan), list_Car(Scan), Test));
    Scan = list_Cdr(Scan);
  }
  return List;
}


LIST list_DeleteDuplicatesFree(LIST List, BOOL (*Test)(), void (*Free)())

{
  LIST Scan;
  
  Scan = List;

  while (!list_Empty(Scan)) {
    list_Rplacd(Scan, list_DeleteElementFree(list_Cdr(Scan), list_Car(Scan), Test, Free));
    Scan = list_Cdr(Scan);
  }
  return List;
}


LIST list_PointerDeleteDuplicates(LIST List)

{
  LIST Scan;
  
  Scan = List;

  while (!list_Empty(Scan)) {
    list_Rplacd(Scan, list_PointerDeleteElement(list_Cdr(Scan),
						list_Car(Scan)));
    Scan = list_Cdr(Scan);
  }
  return List;
}


LIST list_NPointerUnion(LIST List1, LIST List2)

{
  return(list_PointerDeleteDuplicates(list_Nconc(List1,List2)));
}


LIST list_NUnion(LIST List1, LIST List2, BOOL (*Test)())

{
  return(list_DeleteDuplicates(list_Nconc(List1,List2), Test));
}


LIST list_NListTimes(LIST List1, LIST List2)

{
  LIST Result, Scan1, Scan2;

  Result = list_Nil();

  for(Scan1=List1; !list_Empty(Scan1); Scan1=list_Cdr(Scan1))
    for(Scan2=List2; !list_Empty(Scan2); Scan2=list_Cdr(Scan2))
      Result = list_Cons(list_Append(((LIST)list_Car(Scan1)),
				     list_Copy((LIST)list_Car(Scan2))),
			  Result);

  list_Apply(list_Delete, List1);
  list_Apply(list_Delete, List2);
  list_Delete(List1);
  list_Delete(List2);

  return(Result);
}


LIST list_NIntersect(LIST List1, LIST List2, BOOL (*Test)())

{
  LIST Scan1, Scan2;
  
  while (!list_Empty(List1) && !list_Member(List2, list_Car(List1), Test)) {
    Scan1 = list_Cdr(List1);
    list_Free(List1);
    List1 = Scan1;
  }

  if (list_Empty(List1))
    return(List1);

  Scan1 = List1;
  Scan2 = list_Cdr(List1);

  while(!list_Empty(Scan2)) {
    if (list_Member(List2, list_Car(Scan2), Test)) {
      Scan2 = list_Cdr(Scan2);
      Scan1 = list_Cdr(Scan1);
    }
    else {
      list_Rplacd(Scan1, list_Cdr(Scan2));
      list_Free(Scan2);
      Scan2 = list_Cdr(Scan1);
    }
  }
  return(List1);
}


LIST list_NPointerIntersect(LIST List1, LIST List2)

{
  LIST Scan1, Scan2;
  
  while (!list_Empty(List1) && !list_PointerMember(List2, list_Car(List1))) {
    Scan1 = list_Cdr(List1);
    list_Free(List1);
    List1 = Scan1;
  }

  if (list_Empty(List1))
    return(List1);

  Scan1 = List1;
  Scan2 = list_Cdr(List1);

  while(!list_Empty(Scan2)) {
    if (list_PointerMember(List2, list_Car(Scan2))) {
      Scan2 = list_Cdr(Scan2);
      Scan1 = list_Cdr(Scan1);
    }
    else {
      list_Rplacd(Scan1, list_Cdr(Scan2));
      list_Free(Scan2);
      Scan2 = list_Cdr(Scan1);
    }
  }
  return(List1);
}

void list_NInsert(LIST List1, LIST List2)

{
  LIST Help;

#ifdef DEBUG
  if (list_Empty(List1)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In list_NInsert: Empty list argument.");
    misc_FinishErrorReport();
  }
#endif

  Help = list_Cdr(List1);
  list_Rplacd(List1,List2);
  List2 = List1;

  while (!list_Empty(list_Cdr(List2)))
    List2 = list_Cdr(List2);

  list_Rplacd(List2,Help);
}

      
BOOL list_HasIntersection(LIST List1, LIST List2)

{ 
  LIST Scan;

  for(Scan=List1; !list_Empty(Scan); Scan=list_Cdr(Scan))
    if (list_PointerMember(List2,list_Car(Scan)))
      return TRUE;

  return FALSE;
}


LIST list_NPointerDifference(LIST List1, LIST List2)

{ 
  LIST Scan;

  for(Scan=List2; !list_Empty(Scan); Scan=list_Cdr(Scan))
    List1 = list_PointerDeleteElement(List1, list_Car(Scan));

  return(List1);
}


LIST list_NDifference(LIST List1, LIST List2, BOOL (*Test)())

{ 
  LIST Scan;
  
  for(Scan=List2; !list_Empty(Scan); Scan=list_Cdr(Scan))
    List1 = list_DeleteElement(List1, list_Car(Scan), Test);

  return(List1);
}


LIST list_NMultisetDifference(LIST List1, LIST List2, BOOL (*Test)())

{
  LIST scan;
  
  for (scan = List2; !list_Empty(scan); scan = list_Cdr(scan))
    
    List1 = list_DeleteOneElement(List1, list_Car(scan), Test);

  return List1;
}

    
BOOL list_PointerReplaceMember(LIST List, POINTER Old, POINTER New)

{
  while (!list_Empty(List)) {
    if (Old == list_Car(List)) {
      list_Rplaca(List, New);
      return TRUE;
    }
    List = list_Cdr(List);
  }
  return FALSE;
}   
  
void list_DeleteAssocListWithValues(LIST List, void (*ValueDelete)())

{
  LIST Scan;

  for (Scan=List;!list_Empty(Scan);Scan = list_Cdr(Scan)) {
    ValueDelete(list_PairSecond(list_Car(Scan)));
    list_PairFree(list_Car(Scan));
  }

  list_Delete(List);
}

POINTER list_AssocListValue(LIST List, POINTER Key)

{
  LIST Scan;

  for (Scan=List;!list_Empty(Scan);Scan = list_Cdr(Scan))
    if (Key == list_PairFirst(list_Car(Scan)))
      return list_PairSecond(list_Car(Scan));

  return (POINTER)NULL;
}

LIST list_AssocListPair(LIST List, POINTER Key)

{
  LIST Scan;

  for (Scan=List;!list_Empty(Scan);Scan = list_Cdr(Scan))
    if (Key == list_PairFirst(list_Car(Scan)))
      return list_Car(Scan);

  return list_PairNull();
}


NAT list_Length(LIST List)

{
  NAT Result;

  Result = 0;

  while (!list_Empty(List)) {
    Result++;
    List = list_Cdr(List);
  }

  return(Result);
}


NAT list_Bytes(LIST List)

{
  return(list_Length(List)*sizeof(LIST_NODE));
}
