




 





















#include "memory.h"

unsigned int  memory_PAGESIZE;   
long          memory_MAXMEM;     
static int    memory__EOF = EOF; 
unsigned long memory_NEWBYTES;   
unsigned long memory_FREEDBYTES; 

const unsigned int memory_ALIGN = sizeof(POINTER); 


#ifdef DEBUG
unsigned int memory_LEFTTAG;     
unsigned int memory_OFFSET;      
unsigned int memory_MARKSIZE;    

BOOL         memory_MANAGEMENT_INITIALIZED = FALSE;

#else  
unsigned int memory_MARKSIZE = 0;
unsigned int memory_OFFSET   = 0;
#endif 

const unsigned int memory_MAGICMALLOC = 1; 
const unsigned int memory_MAGICFREE   = 2; 



static MEMORY_RESOURCE memory_PAGES[memory__DYNMAXSIZE/memory__SHAREDPAGES + 1];



MEMORY_RESOURCE * memory_ARRAY[memory__DYNMAXSIZE];

 
MEMORY_BIGBLOCKHEADER memory_BIGBLOCKS = NULL;









void memory_Init(long Maxmem)

{
  int i;
  int extra;             

  memory_FREEDBYTES = 0; 
  memory_NEWBYTES   = 0; 

  
  memory_PAGESIZE   = memory__DEFAULTPAGESIZE;

#ifdef DEBUG

  
 
  if (!memory_ManagementInitialized()) {
    
    memory_MANAGEMENT_INITIALIZED = TRUE;
  }
  else {
    
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_Init:");
    misc_UserErrorReport("\n Memory Error.");
    misc_UserErrorReport(" Trying to initialize memory management");
    misc_UserErrorReport(" for a second time.\n");
    misc_FinishUserErrorReport();
  }

  
  memory_LEFTTAG  = sizeof(MEMORY_INFONODE) + sizeof(unsigned int);

  if ((sizeof(MEMORY_INFONODE) + sizeof(unsigned int)) % memory_ALIGN == 0) {
    memory_OFFSET   = memory_LEFTTAG;
  }
  else {
    memory_OFFSET   = memory_LEFTTAG + memory_ALIGN 
      - (memory_LEFTTAG % memory_ALIGN);
  }

  if ((sizeof(unsigned int) % memory_ALIGN) == 0) {
    memory_MARKSIZE =  memory_OFFSET + sizeof(unsigned int);
  }
  else {
    memory_MARKSIZE =  memory_OFFSET + sizeof(unsigned int) + memory_ALIGN 
      - (sizeof(unsigned int) % memory_ALIGN);
  }
#endif

  
  
  extra = 2*sizeof(POINTER) + memory_MARKSIZE; 


  
  while (memory_PAGESIZE < (2*(memory__DYNMAXSIZE + memory_MARKSIZE) + extra)) {
    
    memory_PAGESIZE += memory__DEFAULTPAGESIZE/2;    
  }

  
  if (Maxmem <= 0) {
    
    memory_MAXMEM = -1;
  }
  else {
    
    memory_MAXMEM = Maxmem;
  }

  
  for (i=1; i<memory__DYNMAXSIZE; i++) {
    MEMORY_RESOURCE *CurrentResource;
    int             TotalSize;

    
    memory_ARRAY[i]               = &memory_PAGES[(i-1)/memory__SHAREDPAGES];

    CurrentResource = memory_ARRAY[i];

    CurrentResource->free            = &memory__EOF; 
    CurrentResource->next            = &memory__EOF; 
    CurrentResource->end_of_page     = &memory__EOF; 
    CurrentResource->page            = &memory__EOF; 

    
    CurrentResource->aligned_size    = memory_CalculateRealBlockSize(i);

    
    CurrentResource->total_size      = memory_MARKSIZE 
      + CurrentResource->aligned_size;

    TotalSize                        = CurrentResource->total_size;

    
    CurrentResource->offset          =
      ((memory_PAGESIZE-extra)/TotalSize)*TotalSize 
      + sizeof(POINTER) + memory_OFFSET;
  }
}


void memory_Restrict(long Maxmem)

{
   
  if (Maxmem <= 0) {
    
    memory_MAXMEM = -1;
  }
  else {
    
    memory_MAXMEM = Maxmem;
  }
}









#ifdef NO_MEMORY_MANAGEMENT

POINTER memory_Malloc(unsigned int Bytes)
{
  char *mem; 

  
  mem = (char *)malloc(Bytes);

  
  if (mem == NULL) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_Malloc:");
    misc_UserErrorReport("\n Memory Error. Out of memory.\n");
    misc_FinishUserErrorReport();
  }

  return mem;
}

#else

#ifdef DEBUG
POINTER memory_MallocIntern(unsigned int Bytes, 
			    const char * File, 
			    unsigned short int Line)
#else
POINTER memory_Malloc(unsigned int Bytes)
#endif

{
  char                *NewMemory;       

  MEMORY_RESOURCE     *Resource;        

#ifdef DEBUG
  MEMORY_INFO         NewInfo;          
#endif


#ifdef DEBUG
  
  if (!memory_ManagementInitialized()) {
    
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_Malloc:");
    misc_UserErrorReport("\n Memory Error.");
    misc_UserErrorReport(" Memory management is not initialized.");
    misc_UserErrorReport("\n You have to call memory_Init()");
    misc_UserErrorReport(" before you can use memory management functions.\n");
    misc_UserErrorReport("\n Error occured in memory_Malloc");
    misc_UserErrorReport(" called from file %s at line %d.\n",
			 File, Line);
    misc_FinishUserErrorReport();
  }

  
  if (Bytes == 0) {
    

    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_Malloc:");
    misc_UserErrorReport("\n Memory Error. Tried to allocate 0 Bytes!");
    misc_UserErrorReport("\n Error occured in memory_Malloc");
    misc_UserErrorReport(" called from file %s at line %d.\n",
			 File, Line);
    misc_FinishUserErrorReport();
  }
#endif
  
  
  if (Bytes >= memory__DYNMAXSIZE) {
    unsigned int        RealBigBlockSize; 

    


    

    RealBigBlockSize = sizeof(MEMORY_BIGBLOCKHEADERNODE) + 
      memory_MARKSIZE + memory_CalculateRealBlockSize(Bytes);

    
    if (memory_MAXMEM >= 0) {
      
      if ((unsigned int)memory_MAXMEM < RealBigBlockSize) {
	
	misc_StartUserErrorReport();
	misc_UserErrorReport("\n In memory_Malloc:");
	misc_UserErrorReport("\n Memory Error.");
	misc_UserErrorReport(" Terminated by user given memory restriction,\n");
	misc_UserErrorReport("\n while trying to allocate %lu bytes.\n", 
			     RealBigBlockSize);
	misc_UserErrorReport("\n Maximum amount of memory");
	misc_UserErrorReport(" left for allocation is %d.\n", 
			     memory_MAXMEM);
#ifdef DEBUG
	misc_UserErrorReport("\n Error occured in memory_Malloc");
	misc_UserErrorReport(" called from file %s at line %d.\n",
			     File, Line);
#endif
	misc_FinishUserErrorReport();
      } 
      else
	
	memory_MAXMEM -= RealBigBlockSize;
    }

    
    NewMemory = (char *)malloc(RealBigBlockSize);

    
    if(NewMemory != NULL) {

      

      MEMORY_BIGBLOCKHEADER NewBigBlock; 

      
      NewBigBlock = (MEMORY_BIGBLOCKHEADER) NewMemory;
      NewBigBlock->next = memory_BIGBLOCKS;
      NewBigBlock->previous = NULL;

      
      if (memory_BIGBLOCKS != NULL) {
	memory_BIGBLOCKS->previous = NewBigBlock;
      }

       
      memory_BIGBLOCKS = NewBigBlock;

      
      NewMemory += sizeof(MEMORY_BIGBLOCKHEADERNODE);

#ifdef DEBUG
      
      NewInfo = (MEMORY_INFO) NewMemory;

      
      NewMemory += memory_OFFSET;
#endif

      
      memory_NEWBYTES    += RealBigBlockSize;
    }
    else {
      
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n In memory_MallocIntern:");
      misc_UserErrorReport("\n Memory Error. Out of memory.");
      misc_UserErrorReport("\n Failed to allocate %d bytes.\n", 
			   RealBigBlockSize);
#ifdef DEBUG
      misc_UserErrorReport("\n Error occured in memory_Malloc");
      misc_UserErrorReport(" called from file %s at line %d.\n",
			   File, Line);
#endif
      misc_FinishUserErrorReport();
    }
  }
  else {
    

    
    Resource = memory_ARRAY[Bytes];


    
    if (*((int *)Resource->free) != EOF) {

      

      NewMemory                 = (char *) Resource->free;

      
      Resource->free            = *((POINTER *)(NewMemory));

      
      memory_FREEDBYTES        -= Resource->total_size;

#ifdef DEBUG
       
      NewInfo = (MEMORY_INFO) ((char*) NewMemory - memory_OFFSET);

      
      memory_CheckPointer(NewMemory, Bytes);  
#endif
    } 
    else { 
      
 
      
      if (Resource->next != Resource->end_of_page) {

	
	NewMemory                = (char *)Resource->next;

	
	Resource->next           = NewMemory + Resource->total_size;

	
	memory_NEWBYTES         += Resource->total_size;

#ifdef DEBUG
	
	if ((char *)NewMemory > (char *) Resource->end_of_page) {
	  
	  misc_StartErrorReport();
	  misc_ErrorReport("\n In memory_Malloc:");
	  misc_ErrorReport("\n Memory Error. Address overflow %d.",Bytes);
	  misc_ErrorReport("\n Error occured in memory_Malloc");
	  misc_ErrorReport(" called from file %s at line %d.\n", File, Line);
	  misc_FinishErrorReport();
	}

	
	NewInfo = (MEMORY_INFO)((char*) NewMemory - memory_OFFSET);
#endif

      } 
      else { 
	
	if (memory_MAXMEM >=0) {
	  
	  if ((unsigned int)memory_MAXMEM < memory_PAGESIZE) {
	    
	    misc_StartUserErrorReport();
	    misc_UserErrorReport("\n In memory_Malloc:");
	    misc_UserErrorReport("\n Memory Error.");
	    misc_UserErrorReport(" Terminated by user given memory restriction.\n");
#ifdef DEBUG
	    misc_UserErrorReport("\n Error occured in memory_Malloc");
	    misc_UserErrorReport(" called from file %s at line %d.\n",
				 File, Line);
#endif
	    misc_FinishUserErrorReport();
	  } 
	  else {
	    
	    memory_MAXMEM -= memory_PAGESIZE;
	  }
	}

	
	NewMemory=(char *)malloc(memory_PAGESIZE);

	
	if (NewMemory == NULL) {
	  
	  misc_StartUserErrorReport();
	  misc_UserErrorReport("\n In memory_Malloc:");
	  misc_UserErrorReport("\n Memory Error.");
	  misc_UserErrorReport(" Terminated, ran out of system memory.\n");
#ifdef DEBUG
	  misc_UserErrorReport("\n Error occured in memory_Malloc");
	  misc_UserErrorReport(" called from file %s at line %d.\n",
			       File, Line);
#endif
	  misc_FinishUserErrorReport();
	}

	
	*((POINTER *)NewMemory)     = Resource->page;
	Resource->page              = NewMemory;

	
	memory_NEWBYTES            += Resource->total_size;

	
	Resource->end_of_page       = (char *) NewMemory + Resource->offset;

	
	NewMemory += sizeof(POINTER);

#ifdef DEBUG
	
	NewInfo    = (MEMORY_INFO) NewMemory;

	
	NewMemory += memory_OFFSET;
#endif

		
	Resource->next              = NewMemory + Resource->total_size;
      }
    }
  }

#ifdef DEBUG
  
  memory_SetInfo(NewInfo,File, Line, NULL, 0);
  memory_SetBlockStatusAndSize(NewMemory,  
			       memory_MAGICMALLOC, Bytes);

  
  memset(NewMemory, memory__FREESHREDDER,
	 memory_LookupRealBlockSize(Bytes));
#endif

  return NewMemory;
}

#endif



#ifdef NO_MEMORY_MANAGEMENT

POINTER memory_Calloc(unsigned int Elements, unsigned int Bytes)
{
  char *mem; 

  
  mem = (char *)calloc(Elements, Bytes);
  
  
  if (mem == NULL) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_Calloc:");
    misc_UserErrorReport("\n Memory Error. Out of memory.\n");
    misc_FinishUserErrorReport();
  }

  return mem;
}

#else

#ifdef DEBUG
POINTER memory_CallocIntern(unsigned int Elements, unsigned int Bytes, 
			    const char * File, unsigned short int Line)
#else
POINTER memory_Calloc(unsigned int Elements, unsigned int Bytes)
#endif

{
  char       * mem; 

  
#ifdef DEBUG
  mem = (char *)memory_MallocIntern(Elements * Bytes, File, Line);
#else
  mem = (char *)memory_Malloc(Elements * Bytes);
#endif

   
  if (mem != NULL) {
    memset(mem,0, Elements * Bytes);
  }
  
  else {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_Calloc:");
    misc_UserErrorReport("\n Memory Error. Out of memory.\n");
#ifdef DEBUG
    misc_UserErrorReport("\n Error occured in memory_Calloc");
    misc_UserErrorReport(" called from file %s at line %d.\n",
			 File, Line);
#endif
    misc_FinishUserErrorReport();
  }

  return mem;
}
#endif

void memory_FreeAllMem(void) 

{
  int i;

  

  for (i = 1; i < memory__DYNMAXSIZE; i++) {
    POINTER thispage, nextpage;
    MEMORY_RESOURCE * Resource;

    Resource = memory_ARRAY[i];

    thispage = Resource->page;

    if (*((int *)thispage) != EOF) {
      do {
	nextpage = *((POINTER *)thispage);
	free(thispage);
	thispage = nextpage;
      } while  (*((int *)thispage) != EOF);

      
      Resource->page        = &memory__EOF;
      Resource->free        = &memory__EOF;
      Resource->next        = &memory__EOF;
      Resource->end_of_page = &memory__EOF;
    }
  }

  

  if (memory_BIGBLOCKS != NULL) {
    MEMORY_BIGBLOCKHEADER thisblock, nextblock;
    
    for (thisblock = memory_BIGBLOCKS; 
	 thisblock != NULL; 
	 thisblock = nextblock) {
      nextblock = thisblock->next;
      free(thisblock);
    }

    
    memory_BIGBLOCKS = NULL;
  }
}









void memory_Print(void)

{
#ifndef NO_MEMORY_MANAGEMENT
  
  memory_FPrint(stdout);
#endif
}

void memory_FPrint(FILE* File)

{
#ifndef NO_MEMORY_MANAGEMENT
  int     Pages;    
  int     i;
  POINTER ActPage;  

  
  Pages = 0;
  for(i = 1; i < memory__DYNMAXSIZE; i+=memory__SHAREDPAGES) {
    
    ActPage = memory_ARRAY[i]->page;

    
    while (*((int *)ActPage) != EOF) {
      Pages++;
      ActPage = *((POINTER *)ActPage);
    }
  }

  
  fprintf(File,"\n###\n");
  fprintf(File,"### Pagesize: %d\n",
	  memory_PAGESIZE);
  fprintf(File,"### Marksize: %d\n",
	  (int)memory_MARKSIZE);
  fprintf(File,"### Memory demanded:  %lu KBytes\n", 
	  memory_NEWBYTES/memory__KILOBYTE);
  fprintf(File,"### Memory freed:     %lu KBytes\n", 
	  memory_FREEDBYTES/memory__KILOBYTE);
  fprintf(File,"### Memory remaining: %lu Bytes\n", 
	  memory_NEWBYTES-memory_FREEDBYTES);
  fprintf(File,"### Pages allocated:  %d Pages\n", 
	  Pages);
  fprintf(File,"###\n");
#endif
}

void memory_PrintAllocatedBlocks(unsigned int Size)

{
#ifndef NO_MEMORY_MANAGEMENT
  MEMORY_RESOURCE *Resource;     

  POINTER          ActPage;      
  POINTER          ActNext;      
  POINTER          ActEndOfPage; 

  unsigned int     BlockSize;    

#ifdef DEBUG
  MEMORY_INFO      Info;         
#endif

  


  

  if (Size >= memory__DYNMAXSIZE) {
    
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_PrintAllocatedBlocks:");
    misc_UserErrorReport("\n Parameter size is too big: %d.",
			 Size);
    misc_UserErrorReport("\n Maximal allowed value is: %d.\n", 
			 memory__DYNMAXSIZE);
    misc_FinishUserErrorReport();
  }
  else {
    

    
    Resource     = memory_ARRAY[Size];
    ActPage      = Resource->page;
    ActNext      = Resource->next;
    ActEndOfPage = Resource->end_of_page;
    BlockSize    = Resource->total_size;

    
    if (*((int *)ActPage) == EOF) {
      
      if (*((int *)ActNext) == EOF) {
	
	printf("   No request so far\n");
      }
      else {
	
	 misc_StartErrorReport();
	 misc_ErrorReport("\n In memory_PrintAllocatedBlocks:");
	 misc_ErrorReport("\n Memory Error. No Page entry but Next entry.\n");
	 misc_FinishErrorReport();
      }
    }
    else {
       
#ifdef DEBUG

      POINTER ActData; 


      
      while(*((int *)ActPage) != EOF) {

	
	ActData      = (char *)ActPage + sizeof(POINTER) + memory_OFFSET;
	ActEndOfPage = (char *)ActPage + Resource->offset;

	
	while (ActData != ActNext 
	       && ActData != ActEndOfPage
	       && memory_GetBlockStatus(ActData) != memory_MAGICMALLOC) {
	  ActData = (char *)ActData + BlockSize;
	}

	
	if (ActData == ActNext || ActData == ActEndOfPage) {
	  
	  printf("\n\n   No memory allocated from page at address %p\n", ActPage);
	}
	else {
	  
	  printf("\n\n   Allocated but not freed: ");
	  do  {
	    Info = (MEMORY_INFO) ((char *) ActData - memory_OFFSET);
	    if (memory_GetBlockStatus(ActData) == memory_MAGICMALLOC 
		&& memory_GetBlockSize(ActData) == Size) {
	      printf("\n\t%p allocated in file %s at line %d ", 
		     ActData, Info->mallocInFile, Info->mallocAtLine);
	    }
	    ActData = (char *)ActData + BlockSize;
	  } while (ActData != ActNext && ActData != ActEndOfPage);
	}

	
	ActPage = *((POINTER *)ActPage);
      }
#endif
    }
  }
#endif
}

void memory_PrintFreedBlocks(unsigned int Size)

{
#ifndef NO_MEMORY_MANAGEMENT
  POINTER     ActFree; 

#ifdef DEBUG
  MEMORY_INFO Info;    
#endif

  

  
  if (Size >= memory__DYNMAXSIZE) {
    
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_PrintFreedBlocks.");
    misc_UserErrorReport("\n Parameter Size is too big: %d.", 
			 Size);
    misc_UserErrorReport("\n Maximal allowed value is: %d.\n", 
			 memory__DYNMAXSIZE);
    misc_FinishUserErrorReport();
  }
  else {
    

    
    ActFree = memory_ARRAY[Size]->free;

    
    if(*((int *)ActFree) == EOF) {
      
      printf("\n\n   No freed memory\n");
    }
    else {
      

      printf("\n\n   Free: ");
      while(*((int *)ActFree) != EOF) {
#ifdef DEBUG
	

	 
	if ( memory_GetBlockSize(ActFree) == Size) {
	  
	  Info = (MEMORY_INFO) ((char *) ActFree - memory_OFFSET);
	  printf("\n\t%p\tallocated in file %s at line %d",
		 ActFree,  Info->mallocInFile, Info->mallocAtLine);
	  printf("\n\t\tfreed in file %s at line %d",
		 Info->freeInFile, Info->freeAtLine);
	}
	else {
	  

	  
	  if(memory__SHAREDPAGES == 1) {
	    
	    misc_StartUserErrorReport();
	    misc_UserErrorReport("\n In memory_PrintFreedBlocks:");
	    misc_UserErrorReport("\n Memory Error. Memory block size mismatch.");
	    misc_UserErrorReport("\n Expected %d found %d for memory block at %p.\n",
				 Size, memory_GetBlockSize(ActFree), ActFree);
	    misc_UserErrorReport("\n Probably the memory block was corrupted.\n");	    
	    misc_FinishUserErrorReport();
	  }
	}

#endif

	
	ActFree = *((POINTER *)ActFree);
      }
    }
  }
#endif
}


void memory_PrintAllocatedBigBlocks(void)

{
#ifndef NO_MEMORY_MANAGEMENT
#ifdef DEBUG
  MEMORY_BIGBLOCKHEADER Ptr;         
  MEMORY_INFO           Info;        
  char                * BlockStart;  
  
  
  Ptr = memory_BIGBLOCKS;

  
  if (Ptr != NULL) {
    
    do {
      BlockStart = (char *)Ptr + memory_OFFSET 
	+ sizeof(MEMORY_BIGBLOCKHEADERNODE);
      
      Info = (MEMORY_INFO) (BlockStart - memory_OFFSET);
      printf("\n\t%p %d bytes allocated in file %s at line %d ", 
	     BlockStart, memory_GetBlockSize(BlockStart),
	     Info->mallocInFile, Info->mallocAtLine);
      Ptr = Ptr->next;
    } while (Ptr != NULL);
    puts("");
  }
  else {
    
    printf("   No request so far\n");
  }
#endif
#endif
}

void memory_PrintDetailed(void)

{
#ifndef NO_MEMORY_MANAGEMENT
  MEMORY_RESOURCE  *Resource;     

  POINTER           ActPage;      
  POINTER           ActData;      
  POINTER           ActEndOfPage; 
  unsigned int      BlockSize;    
  unsigned int      PageOffset;   

  unsigned int      i;


  
  printf("\n\nEOF Pointer: %p\n", &memory__EOF);

  
  for (i=1; i<memory__DYNMAXSIZE; i++) {
    
    Resource     = memory_ARRAY[i];
    ActPage      = Resource->page;
    ActData      = Resource->next;
    ActEndOfPage = Resource->end_of_page;
    PageOffset   = Resource->offset;
    BlockSize    = Resource->total_size;

    
    printf("\n\n Entry: %d aligned size: %d total size: %d\n", 
	   i , Resource->aligned_size, BlockSize);

    
    if (*((int *)ActPage) == EOF) {
      
      if (*((int *)ActData) == EOF) {
	
	printf("   No request so far\n");
      }
      else {
	
	misc_StartErrorReport();
	misc_ErrorReport("\n In memory_PrintDetailed:");
	misc_ErrorReport("\n Memory Error. No Page entry but Next entry.\n");
	misc_FinishErrorReport();
      }
    }
    else {
      

      
      while(*((int *)ActPage) != EOF) {
	
	printf("\n\n   Page: %p Next Page: %p\n",
	       ActPage, *((POINTER *)ActPage));

	
	ActData = ((char *)ActPage + sizeof(POINTER) + memory_OFFSET);
	ActEndOfPage = (char *)ActPage + PageOffset;

	
	printf("   Data: ");
	while (ActData != ActEndOfPage) {
	  int column;

	  printf("\n\t\t");
	  for (column = 0; column < 6; column++) {
	    printf("%p ", ActData);
	    ActData = (char *)ActData + BlockSize;
	    if (ActData == ActEndOfPage) {
	      break;
	    }
	  }
	}

	
	ActPage = *((POINTER *)ActPage);
      }

      
      memory_PrintAllocatedBlocks(i);
      memory_PrintFreedBlocks(i);
    }
  }

#ifdef DEBUG
  
  printf("\n\n Allocated blocks of size >= %d\n", 
	 memory__DYNMAXSIZE);
  memory_PrintAllocatedBigBlocks();
#endif
#endif
}


void memory_PrintLeaks(void)

{
#ifndef NO_MEMORY_MANAGEMENT
  POINTER           ActPage;       
  POINTER           ActNext;       
  POINTER           ActEndOfPage;  
  MEMORY_RESOURCE  *Resource;      
  unsigned int      Size;          
  unsigned int      BlockSize;     

   
  if (memory_UsedBytes() != 0UL) { 

    

    
    for(Size = 1; Size < memory__DYNMAXSIZE; Size++) {
      
      Resource     = memory_ARRAY[Size];
      ActPage      = Resource->page;
      ActNext      = Resource->next;
      ActEndOfPage = Resource->end_of_page;
      BlockSize    = Resource->total_size;

      
      if (*((int *)ActPage) != EOF) {
      
	
#ifdef DEBUG
	
	POINTER ActData;
	BOOL    LeakFound;
	
	LeakFound = FALSE;
	
	while(*((int *)ActPage) != EOF) {
	  
	  
	  
	  ActData      = (char *)ActPage + sizeof(POINTER) + memory_OFFSET;
	  ActEndOfPage = (char *)ActPage + Resource->offset;
	  
	  while (ActData != ActNext && ActData != ActEndOfPage) {
	    if (memory_GetBlockStatus(ActData) == memory_MAGICMALLOC) {
	      LeakFound = TRUE;
	      break;
	    }
	    ActData = (char *)ActData + BlockSize;
	  }
	  
	  if (LeakFound) {
	    
	    
	    
	    printf("\n\n  Leaked blocks of size %d:", Size);
	    
	    memory_PrintAllocatedBlocks(Size);
	    printf("\n");
	    
	    break;
	  }
	  else {
	    
	    ActPage = *((POINTER *)ActPage);
	  }
	}

#endif

      }      
    }

#ifdef DEBUG
    
    if(memory_BIGBLOCKS != NULL) {
      printf("\n\n  Leaked blocks of size >= %d\n", 
	     memory__DYNMAXSIZE);
      memory_PrintAllocatedBigBlocks();
      printf("\n");
    }
#endif

  }
#endif
}
