




 






















#ifndef _MEMORY_
#define _MEMORY_





#include "misc.h"





#ifndef memory__DYNMAXSIZE
#define memory__DYNMAXSIZE   1024 
#endif

#ifndef memory__SHAREDPAGES
#define memory__SHAREDPAGES    1  
#endif

#ifndef memory__FREESHREDDER
#define memory__FREESHREDDER  'S' 
#endif

#define memory__KILOBYTE        1024

#ifndef memory__DEFAULTPAGESIZE
#define memory__DEFAULTPAGESIZE (8 * memory__KILOBYTE)
                                  
#endif

typedef struct MEMORY_RESOURCEHELP {
  POINTER free;                    
  POINTER next;                    
  POINTER page;                    
  POINTER end_of_page;             
  int     total_size;              
  int     aligned_size;            
  int     offset;                  
} MEMORY_RESOURCE;

extern MEMORY_RESOURCE * memory_ARRAY[];

#if defined(DEBUG)
typedef struct MEMORY_INFOHELP {
  const char * mallocInFile;       
  const char * freeInFile;         
  unsigned short int mallocAtLine; 
  unsigned short int freeAtLine;   
} MEMORY_INFONODE, * MEMORY_INFO;

#endif

typedef struct MEMORY_BIGBLOCKHEADERHELP {
  struct MEMORY_BIGBLOCKHEADERHELP * previous, * next;
} MEMORY_BIGBLOCKHEADERNODE, * MEMORY_BIGBLOCKHEADER;

extern long                  memory_MAXMEM;

extern unsigned long         memory_NEWBYTES;
extern unsigned long         memory_FREEDBYTES;

extern const unsigned int    memory_MAGICMALLOC;
extern const unsigned int    memory_MAGICFREE;

extern const unsigned int    memory_ALIGN;

extern MEMORY_BIGBLOCKHEADER memory_BIGBLOCKS;





extern unsigned int memory_MARKSIZE;
extern unsigned int memory_OFFSET;

#ifdef DEBUG
extern BOOL         memory_MANAGEMENT_INITIALIZED;
#endif






static __inline__ unsigned int memory_CalculateRealBlockSize(unsigned int 
							     BlockSize)

{  
  unsigned int RealSize;

  RealSize = BlockSize;

  if (RealSize % memory__SHAREDPAGES) {
    RealSize += memory__SHAREDPAGES - (RealSize % memory__SHAREDPAGES);
  }

  if (RealSize % memory_ALIGN) {
    RealSize += memory_ALIGN - (RealSize % memory_ALIGN);
  }

  return RealSize;

}

static __inline__ unsigned int memory_LookupRealBlockSize(unsigned int 
							  BlockSize)

{  

  unsigned int RealSize;

  if (BlockSize < memory__DYNMAXSIZE) {
    RealSize = memory_ARRAY[BlockSize]->aligned_size;
  }
  else {
    RealSize = memory_CalculateRealBlockSize(BlockSize);
  }

  return RealSize;

}

#ifdef DEBUG
static __inline__ void memory_SetBlockStatusAndSize(POINTER Mem, 
						    unsigned int Status,
						    unsigned int Size)

{
  *((int *)Mem - 1) = Size; 
  *((int *)((char *)Mem + memory_LookupRealBlockSize(Size))) = Status;
}

static __inline__ unsigned int memory_GetBlockSize(POINTER Mem)

{
  return *((int *)Mem - 1);
}

static __inline__ unsigned int memory_GetRealBlockSize(POINTER Mem)

{
  return memory_LookupRealBlockSize(memory_GetBlockSize(Mem));
}

static __inline__ unsigned int memory_GetBlockStatus(POINTER Mem) 

{
  unsigned int Size;

  Size = memory_GetBlockSize(Mem); 

  return  *((int *)((char *)Mem + memory_LookupRealBlockSize(Size)));
}

static __inline__ void memory_SetInfo (MEMORY_INFO Info,
				       const char * MallocInFile,
				       unsigned short int MallocAtLine,
				       const char * FreeInFile,
				       unsigned short int FreeAtLine)

{

  if (!Info) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In memory_SetInfo:");
    misc_ErrorReport("\n Memory Error. Info is a NULL pointer.\n");
    misc_FinishErrorReport();
  }
 
  Info->mallocAtLine   = MallocAtLine;
  Info->mallocInFile   = MallocInFile;
  Info->freeAtLine     = FreeAtLine;
  Info->freeInFile     = FreeInFile;
}

static __inline__ void memory_CheckPointer(POINTER Pointer, unsigned int Size) 

{

  MEMORY_INFO Info;
  unsigned int BlockSize, RealBlockSize, BlockStatus;

  Info = (MEMORY_INFO) ((char *) Pointer - memory_OFFSET);

  RealBlockSize = memory_LookupRealBlockSize(Size);

  if (Pointer == NULL) {
    
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_CheckPointer:");
    misc_UserErrorReport("\n Memory Error. Pointer is a NULL pointer.\n");
    misc_FinishUserErrorReport();
  }


  if (Size == 0) {
    
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_CheckPointer:");
    misc_UserErrorReport("\n Memory Error.");
    misc_UserErrorReport(" Pointer %p points to a block of memory", Pointer);
    misc_UserErrorReport(" with size 0.\n"); 
    misc_FinishUserErrorReport();
  }

  if ((unsigned long)Pointer % (unsigned long)memory_ALIGN){
    
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_CheckPointer:");
    misc_UserErrorReport("\n Memory Error.");
    misc_UserErrorReport(" Pointer %p is not a legal pointer.\n", Pointer);
    misc_FinishUserErrorReport();
  }

  

  BlockStatus   = memory_GetBlockStatus(Pointer);
  BlockSize     = memory_GetBlockSize(Pointer);

  if (BlockStatus != memory_MAGICMALLOC 
      && BlockStatus != memory_MAGICFREE) {

    

    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_CheckPointer:");
    misc_UserErrorReport("\n Memory Error.");
    misc_UserErrorReport(" Pointer %p was not (de)allocated by the module,",
			 Pointer);
    misc_UserErrorReport("\n or the memory block was corrupted.\n");
    misc_FinishUserErrorReport();
  }

  if (BlockStatus == memory_MAGICMALLOC) {
    if (BlockSize != Size) {
      
      

      misc_StartUserErrorReport();
      misc_UserErrorReport("\n In memory_CheckPointer:");
      misc_UserErrorReport("\n Memory Error.");
      misc_UserErrorReport(" Pointer %p was apparently allocated for",
			   Pointer);
      misc_UserErrorReport(" a block of size %d,",
			   BlockSize);
      misc_UserErrorReport("\n but it is expected to be a block of size %d.",
			   Size);
      misc_UserErrorReport("\n Probably the memory block was corrupted.\n");
      misc_FinishUserErrorReport();

      
    }

    if ((Size % memory_ALIGN) || (Size % memory__SHAREDPAGES)) {
      
      char * ptr, * limit;
      
      limit    = (char *)Pointer + RealBlockSize;
      
      for (ptr = (char *)Pointer + Size; ptr < limit; ptr++) {
	if (*ptr != memory__FREESHREDDER) {
	  misc_StartUserErrorReport();
	  misc_UserErrorReport("\n In memory_CheckPointer:");
	  misc_UserErrorReport("\n Memory Error.");
	  misc_UserErrorReport(" Pointer %p was allocated in file %s at line %d,",
			       Pointer, Info->mallocInFile, Info->mallocAtLine);
	  misc_UserErrorReport("\n for a block of size %d.",
			       BlockSize);
	  misc_UserErrorReport("\n The memory block was corrupted.\n");
	  misc_FinishUserErrorReport();
	}
      }
    }
  }

  if (Size >= memory__DYNMAXSIZE) {
    
    MEMORY_BIGBLOCKHEADER BigBlockHeader;

    BigBlockHeader = (MEMORY_BIGBLOCKHEADER) ((char *) Pointer - memory_OFFSET 
					      - sizeof(MEMORY_BIGBLOCKHEADERNODE));

    
    if (((BigBlockHeader->previous != NULL) 
	 && (BigBlockHeader->previous->next != BigBlockHeader))
	|| ((BigBlockHeader->previous == NULL) 
	    && (memory_BIGBLOCKS != BigBlockHeader))
	|| ((BigBlockHeader->next != NULL) 
	    && (BigBlockHeader->next->previous != BigBlockHeader))) {
      
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n In memory_CheckPointer:");
      misc_UserErrorReport("\n Memory Error.");
      misc_UserErrorReport(" Pointer %p was not allocated by the module,",
			   Pointer);
      misc_UserErrorReport("\n or the memory block was corrupted.\n");
      misc_FinishUserErrorReport();
    }
  }
  
  if (BlockStatus == memory_MAGICFREE) {
    
    char * ptr, * limit;
    
    limit = (char *)Pointer + RealBlockSize;

    for (ptr = (char *)Pointer + sizeof(POINTER); ptr < limit ; ptr++){
      
      if (*ptr != memory__FREESHREDDER) {
	misc_StartUserErrorReport();
	misc_UserErrorReport("\n In memory_CheckPointer:");
	misc_UserErrorReport("\n Memory Error.");
	misc_UserErrorReport(" Pointer %p was allocated in file %s at line %d",
			     Pointer, Info->mallocInFile, Info->mallocAtLine);
	misc_UserErrorReport("\n for a block of size %d",BlockSize);
	misc_UserErrorReport("\n and freed in file %s at line %d.",
			     Info->freeInFile, Info->freeAtLine);
	misc_UserErrorReport("\n The memory block was used after deallocation.\n");
	misc_FinishUserErrorReport();
      }
    }
  }
}

static __inline__ BOOL memory_ManagementInitialized(void)

{
  return memory_MANAGEMENT_INITIALIZED;
}

#endif

static __inline__ unsigned long memory_DemandedBytes(void)

{
  return memory_NEWBYTES;
}


static __inline__ unsigned long memory_UsedBytes(void)

{
  return memory_NEWBYTES-memory_FREEDBYTES;
}

#ifdef NO_MEMORY_MANAGEMENT

static __inline__ void memory_Free(POINTER Freepointer, unsigned int Size)
{
  free((char*)Freepointer);
}

#else

#ifdef DEBUG
static __inline__ void memory_FreeIntern(POINTER  Freepointer, 
					 unsigned int Size, 
					 const char * File, 
					 unsigned short int Line)
#else
static __inline__ void memory_Free(POINTER Freepointer, unsigned int Size)
#endif

{ 
  unsigned int RealBlockSize; 

#ifdef DEBUG
  MEMORY_INFO  Info;          
#endif

  RealBlockSize = memory_LookupRealBlockSize(Size);

#ifdef DEBUG

  
  if (!memory_ManagementInitialized()) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_Free:");
    misc_UserErrorReport("\n Memory Error.");
    misc_UserErrorReport(" Memory management is not initialized.");
    misc_UserErrorReport("\n You have to call memory_Init()");
    misc_UserErrorReport(" before you can use memory management functions.\n");
    misc_UserErrorReport("\n Error occured in memory_Free");
    misc_UserErrorReport(" called from file %s at line %d.\n",
			 File, Line);
    misc_FinishUserErrorReport();
  }

  
  memory_CheckPointer(Freepointer, Size);

  
  Info = (MEMORY_INFO) ((char *) Freepointer - memory_OFFSET);

  
  if ( memory_GetBlockStatus(Freepointer) == memory_MAGICFREE) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In memory_Free:");
    misc_UserErrorReport("\n Memory Error.");
    misc_UserErrorReport(" Pointer %p was allocated in file %s at line %d.",
			 Freepointer, Info->mallocInFile, Info->mallocAtLine);
    misc_UserErrorReport("\n It has already been freed in file %s at line %d.",
			 Info->freeInFile, Info->freeAtLine);
    misc_UserErrorReport("\n Size of memory block is %d bytes.",
			 memory_GetBlockSize(Freepointer));
    misc_UserErrorReport("\n Error occured in memory_Free");
    misc_UserErrorReport(" called from file %s at line %d.\n",
			 File, Line);
    misc_FinishUserErrorReport();
  }

  

  memset (Freepointer,memory__FREESHREDDER, RealBlockSize);

  
  memory_SetInfo(Info,Info->mallocInFile, Info->mallocAtLine, File, Line);
  memory_SetBlockStatusAndSize(Freepointer, memory_MAGICFREE, Size);
#endif


   
  if (Size >= memory__DYNMAXSIZE) {
    
    MEMORY_BIGBLOCKHEADER BigBlockHeader;

    BigBlockHeader = 
      (MEMORY_BIGBLOCKHEADER) ((char *) Freepointer - memory_OFFSET
			       - sizeof(MEMORY_BIGBLOCKHEADERNODE));

    
    if (BigBlockHeader->previous != NULL) {
      
      BigBlockHeader->previous->next = BigBlockHeader->next;
    }
    else {
      
      memory_BIGBLOCKS = BigBlockHeader->next;
    }

    
    if (BigBlockHeader->next != NULL) {
      
      BigBlockHeader->next->previous = BigBlockHeader->previous;
    }

    
    memory_FREEDBYTES       += RealBlockSize + memory_MARKSIZE + 
      sizeof(MEMORY_BIGBLOCKHEADERNODE);

    if (memory_MAXMEM >= 0) {
      memory_MAXMEM           += RealBlockSize + memory_MARKSIZE + 
	sizeof(MEMORY_BIGBLOCKHEADERNODE);
    }

    
#ifdef DEBUG
    free((char *)Freepointer - memory_OFFSET 
	 - sizeof(MEMORY_BIGBLOCKHEADERNODE));
#else
    free((char*) Freepointer - sizeof(MEMORY_BIGBLOCKHEADERNODE));
#endif
  }
  else {
    
    
    memory_FREEDBYTES       += memory_ARRAY[Size]->total_size;
    *(POINTER *)Freepointer  = memory_ARRAY[Size]->free;
    memory_ARRAY[Size]->free = Freepointer;
  }

}

#endif





#ifdef __cplusplus
extern "C" {
#endif


void    memory_Init(long);
void    memory_Restrict(long);
  
void    memory_Print(void);
void    memory_FPrint(FILE*);
  
void    memory_PrintLeaks(void);
void    memory_PrintDetailed(void);
void    memory_PrintAllocatedBlocks(unsigned int Size);
void    memory_PrintFreedBlocks(unsigned int Size);
void    memory_PrintAllocatedBigBlocks(void);
  
void    memory_FreeAllMem(void);

#ifdef __cplusplus
}
#endif

#if defined(DEBUG) && !defined(NO_MEMORY_MANAGEMENT)

#ifdef __cplusplus
extern "C" {
#endif
POINTER memory_MallocIntern(unsigned int, const char *, unsigned short int);
POINTER memory_CallocIntern(unsigned int,
			    unsigned int, 
			    const char *, 
			    unsigned short int);
#ifdef __cplusplus
}
#endif
#define memory_Malloc(Size) memory_MallocIntern((Size), __FILE__, __LINE__)
#define memory_Calloc(Elements, Size) \
memory_CallocIntern((Elements), (Size), __FILE__, __LINE__)
#define memory_Free(Pointer, Size) \
memory_FreeIntern((Pointer), (Size), __FILE__, __LINE__)
#else
#ifdef __cplusplus
extern "C" {
#endif
POINTER memory_Malloc(unsigned int);
POINTER memory_Calloc(unsigned int, unsigned int);
#ifdef __cplusplus
}
#endif
#endif

#endif
