




 























#include "options.h"






static LIST opts_DECLARATIONS;


static LIST opts_PARAMETERS;

static OPTID opts_IdNextAvailable;






static int  opts_StrToInt(char*, int*);
static void opts_AddParam(OPTID, char*);
static BOOL opts_AddParamCheck(OPTID, char*);

static int  opts_GetOptLongOnly(int, char *const *, const char *,
				const struct OPTION *, int *);

OPTDECL*      opts_DeclGetById(OPTID);
static void   opts_PrintDeclarations(void);
static void   opts_PrintDeclarationList(LIST);
static int    opts_StrLength(char*);
static void   opts_IdIncAvailable(void);
static OPTID  opts_IdGetNextAvailable(void);
static OPTID  opts_IdNext(OPTID);
static OPTID  opts_IdEqual(OPTID, OPTID);
static OPTID  opts_IdCmp(OPTID, OPTID);






static enum
{
  REQUIRE_ORDER, PERMUTE, RETURN_IN_ORDER
} opts_Ordering;


static char *opts_PosixlyCorrect;

static int opts_FirstNonOpt;
static int opts_LastNonOpt;



static int opts_NonOptionFlagslen;




static char *opts_Arg = NULL;




static int opts_Ind = 1;



static int opts_GetOptInitialized = 0;



static char *opts_NextChar;



static int opts_Err = 1;



static int opts_Opt = '?';






static int opts_StrLength(char* s)

{
  return(strlen(s));
}

static void opts_StrFree(char *s)

{
  memory_Free(s, opts_StrLength(s)+1);
}

static char* opts_StrMalloc(int l)

{
  return((char*)memory_Malloc(l+1));
}

static BOOL opts_StrEqual(char* s1, char* s2)

{
  return (strcmp(s1,s2) == 0);
}


static char* opts_StrNConc(char* s1, char* s2)

{ 
  char* dst;

  dst = opts_StrMalloc(opts_StrLength(s1)+opts_StrLength(s2));
  strcpy(dst, s1);
  dst = strcat(dst,s2);
  
  opts_StrFree(s1);
  opts_StrFree(s2);

  return(dst);
}

static char* opts_StrDup(char* s)

{
  char* new;

  new = opts_StrMalloc(opts_StrLength(s));

  strcpy(new,s);

  return(new);
}

OPTID opts_IdFirst(void)

{
  return(opts_IDFIRST);
}

static OPTID opts_IdNull(void)

{
  return(-1);
}

BOOL opts_IdIsNull(OPTID Id)

{
  return(opts_IdEqual(opts_IdNull(), Id));
}

static void opts_IdIncAvailable(void)

{
  opts_IdNextAvailable = opts_IdNext(opts_IdNextAvailable);
}

static OPTID opts_IdGetNextAvailable(void)

{
  return(opts_IdNextAvailable);
}

static void  opts_DeclSetClName(OPTDECL* D, char* s)

{
  D->clname = s;
}

static char* opts_DeclGetClName(OPTDECL* D)

{
  return(D->clname);
}


static void  opts_DeclSetType(OPTDECL* D, OPTTYPE type)

{
  D->type = type;
}

static OPTTYPE opts_DeclGetType(OPTDECL* D)

{
  return(D->type);
}

static BOOL opts_DeclIsShortOpt(OPTDECL* D)

{
  return(opts_StrLength(opts_DeclGetClName(D)) == 1);
}

static BOOL opts_DeclHasOptArg(OPTDECL* D) 

{
  return(opts_DeclGetType(D) == opts_OPTARGTYPE);
}

static BOOL opts_DeclHasReqArg(OPTDECL* D) 

{
  return(opts_DeclGetType(D) == opts_REQARGTYPE);
}

static BOOL opts_DeclHasNoArg(OPTDECL* D) 

{
  return(opts_DeclGetType(D) == opts_NOARGTYPE );

}

static void opts_ErrorExit(void)

{
  exit(1);
}

OPTID opts_Declare(char* ClName, OPTTYPE OptType)


{
  OPTDECL* D;
  OPTID Id;

  if (!opts_IdIsNull(opts_Id(ClName))) {
    fprintf(stderr, "internal error: option with command line name '%s' redeclared.\n", ClName);
    opts_ErrorExit(); }

  D  = memory_Malloc(sizeof(OPTDECL));

  opts_DeclSetClName(D, opts_StrDup(ClName));
  opts_DeclSetType(D,OptType);
  
  opts_DECLARATIONS = list_Nconc(opts_DECLARATIONS, list_List(D));

  Id = opts_IdGetNextAvailable();
  opts_IdIncAvailable();

  return(Id);
}

OPTID opts_DeclareVector(OPTDECL Decls[]) 

{
  int i;

  i = 0;

  while (opts_StrLength(opts_DeclGetClName(&Decls[i])) != 0) {
    opts_Declare(opts_DeclGetClName(&Decls[i]), opts_DeclGetType(&Decls[i])); 
    i++;
  }
  return(opts_IdGetNextAvailable());
}

static char* opts_TranslateShortOptDeclarations(void)

{
  LIST  Scan;
  char* ShortDecl;
  OPTDECL* Decl;

  ShortDecl = opts_StrDup("\0");

  Scan = opts_DECLARATIONS;
  
  while (Scan) {
    
    Decl = (OPTDECL*)list_Car(Scan);
    
    if (opts_DeclIsShortOpt(Decl)) {
      ShortDecl = opts_StrNConc(ShortDecl, opts_StrDup(opts_DeclGetClName(Decl)));      

      
      if (opts_DeclHasReqArg(Decl)||opts_DeclHasOptArg(Decl)) 
	ShortDecl = opts_StrNConc(ShortDecl, opts_StrDup(":"));
    }
    Scan = list_Cdr(Scan);
  }

  
  if (opts_StrLength(ShortDecl) != 0) {
    ShortDecl = opts_StrNConc(opts_StrDup(":"),ShortDecl); }

  return(ShortDecl);
}

static LIST opts_GetLongOptDeclarations()

{
  LIST     Scan, Long;
  OPTDECL* Decl;

  Scan = opts_DECLARATIONS;
  Long = list_Nil();

  while (!list_Empty(Scan)) {
    Decl = list_Car(Scan);

    if (!opts_DeclIsShortOpt(Decl)) {
      Long = list_Cons(Decl, Long);
    }
    Scan = list_Cdr(Scan);
  }  
  return(Long);
}

static struct OPTION *opts_GetLongOptsArray(int OptNum)

{
  return((struct OPTION*)memory_Malloc(sizeof(struct OPTION)*(OptNum+1)));
}

static void opts_FreeLongOptsArray(struct OPTION *LongOpts)

{
  int i;

  for (i=0; LongOpts[i].name != 0; i++) ;

  memory_Free(LongOpts, (i+1)*sizeof(struct OPTION));
}


static struct OPTION* opts_TranslateLongOptDeclarations()

{
  LIST           Scan;
  LIST           LongDeclarations;
  int            OptNum;
  int            OptCnt;
  struct OPTION* LongOpts;

  OPTDECL* Decl;

  LongDeclarations = opts_GetLongOptDeclarations();
  OptNum           = list_Length(LongDeclarations);
  LongOpts         = opts_GetLongOptsArray(OptNum);
  OptCnt = 0;
  Scan   = LongDeclarations;

  while (!list_Empty(Scan)) {
    Decl = list_Car(Scan);
    
    LongOpts[OptCnt].name = opts_DeclGetClName(Decl);

    if (opts_DeclHasOptArg(Decl))
      LongOpts[OptCnt].has_arg = 2;
    else if (opts_DeclHasReqArg(Decl)) 
      LongOpts[OptCnt].has_arg = 1; 
    else 
      LongOpts[OptCnt].has_arg = 0; 
    LongOpts[OptCnt].flag    = 0;
    LongOpts[OptCnt].val     = 0; 
    
    Scan = list_Cdr(Scan);
    OptCnt++;
  }
  
  LongOpts[OptCnt].name    = (char*)NULL;
  LongOpts[OptCnt].has_arg = 0;
  LongOpts[OptCnt].flag = 0;
  LongOpts[OptCnt].val = 0;

  list_Delete(LongDeclarations);

  return(LongOpts);
}


static void opts_PrintLongOpts(struct OPTION *LongOpts)

{
  int i;

  if (LongOpts == NULL) {
    printf("\nPrintLongOpts gets NULL pointer.\n");
    return;
  }
  printf("\nLong options array:\n");
  
  i = 0;
  while (LongOpts[i].name != NULL) {
    printf("\nentry %d:\n",i);

    printf("Name:    %s\n", LongOpts[i].name);
    printf("has_arg: %d\n", LongOpts[i].has_arg);
    printf("flag;  : %d\n", (int)LongOpts[i].flag);
    printf("val    : %d\n", LongOpts[i].val);
    i++;
  }
}

static OPTID opts_IdCmp(OPTID Id1, OPTID Id2)

{
  return(Id1-Id2);
}

static OPTID opts_IdEqual(OPTID Id1, OPTID Id2)

{
  return(opts_IdCmp(Id1,Id2) == 0);
}

static OPTID opts_IdNext(OPTID Id)

{
  return(Id+1);
}


char* opts_ClName(OPTID Id)

{
  OPTDECL* Decl;

  Decl = opts_DeclGetById(Id);
  return(opts_DeclGetClName(Decl));
}

OPTID opts_Id(char* ClName) 

{
  LIST    Scan;
  BOOL    found;
  OPTID   Id;

  Scan  = opts_DECLARATIONS;
  Id    = opts_IdFirst();
  found = FALSE;

  while (!found && !list_Empty(Scan)) {
    if (opts_StrEqual(opts_DeclGetClName(list_Car(Scan)), ClName)) {
      found = TRUE;
    } else {
      Scan = list_Cdr(Scan);
      Id = opts_IdNext(Id);
    }
  }
  if (!found)
    Id = opts_IdNull();
  return(Id);
}

static OPTID opts_ShortOptId(char c)

{
  char Str[2];

  Str[0] = c;
  Str[1] = '\0';

  return(opts_Id(Str));
}

void opts_Init(void)

{
  opts_DECLARATIONS = list_Nil();
  opts_PARAMETERS   = list_Nil();
  opts_Err = 1; 
  opts_IdNextAvailable = opts_IdFirst();
}

void opts_DeclareSPASSFlagsAsOptions(void)

{
  int i;

  for (i=0; i < flag_MAXFLAG; i++) {
    opts_Declare(flag_Name(i), opts_OPTARGTYPE);
  }
}

void opts_FreeParameterPair(LIST Pair)

{
  opts_StrFree(list_PairSecond(Pair));

  list_PairFree(Pair);
}

void opts_FreeDecl(OPTDECL* D)

{
  opts_StrFree(opts_DeclGetClName(D));
  memory_Free(D, sizeof(OPTDECL));
}

void opts_Free(void)

{
  list_DeleteWithElement(opts_PARAMETERS,  opts_FreeParameterPair);
  list_DeleteWithElement(opts_DECLARATIONS,opts_FreeDecl);
}

static void opts_PrintDeclarationList(LIST Scan)

{
  OPTDECL* Decl;
  OPTID    Id;

  Id = opts_IdFirst();

  printf("\nDeclarations:\n");

  while (Scan) {
    Decl = (OPTDECL*)list_Car(Scan);
    printf("Id:%-6d Name:%-18s Type:%d\n", Id, opts_DeclGetClName(Decl),
	   opts_DeclGetType(Decl));
    Scan = list_Cdr(Scan);
    Id   = opts_IdNext(Id);
  }
  printf("\n End of list.");
}

static void opts_PrintDeclarations(void)

{
  opts_PrintDeclarationList(opts_DECLARATIONS);
}

static void opts_PrintParameters(void)

{
  LIST Scan;
  LIST Pair;

  printf("\nopts_PARAMETERS:\n");

  Scan = opts_PARAMETERS;

  while (!list_Empty(Scan)) {

    Pair = list_Car(Scan);
    printf("\nId: %d ", (OPTID)list_PairFirst(Pair));
    printf("Par: %s",  (char*)list_PairSecond(Pair));

    Scan = list_Cdr(Scan);
  }

  printf("\n end of list.\n");
}


void opts_PrintSPASSNames(void)

{
  int i,j;
  
  for (i=0; i < flag_MAXFLAG; i=i+4) { 
    for (j =0; j <=3; j++) {
      if (i+j < flag_MAXFLAG)
	printf("%-18s ", flag_Name(i+j)); } 
    printf("\n");
  }
}

OPTDECL* opts_DeclGetById(OPTID Id)

{
  OPTID ScanId;
  LIST  Scan;

  ScanId = opts_IdFirst();
  Scan   = opts_DECLARATIONS;
  
  while (!list_Empty(Scan)) {
    if (opts_IdEqual(Id, ScanId))
      return(list_Car(Scan));
    Scan = list_Cdr(Scan);
    ScanId = opts_IdNext(ScanId);
  }

  return((OPTDECL*)NULL);
}

OPTDECL* opts_DeclGetByClName(char* ClName)

{
  OPTID Id;
  
  Id = opts_Id(ClName);
  if (opts_IdIsNull(Id))
    return(NULL);
  return(opts_DeclGetById(Id));
}

BOOL opts_Read(int argc, char* argv[])

{
  int     OptIndex,c;
  char*   ShortOpts;
  BOOL    Ok;
  OPTID   OptId;
  OPTDECL* OptDecl;
  char*   OptName;
  struct  OPTION *LongOpts;

  Ok = TRUE;
  
  ShortOpts = opts_TranslateShortOptDeclarations();
  LongOpts  = opts_TranslateLongOptDeclarations();

  while (Ok && (c = opts_GetOptLongOnly(argc, argv, ShortOpts, 
					LongOpts, &OptIndex)) != -1) {
    
    if (c == '?') { 
       

      
      return(FALSE);
    } else if (c == 0) {
      
      
      OptName = LongOpts[OptIndex].name;
      OptId   = opts_Id(OptName);
      OptDecl = opts_DeclGetById(OptId);
      
      if (opts_Arg == NULL) {
	
	if (opts_DeclHasReqArg(OptDecl)) {
	  fprintf(stderr, "\nerror, option %s requires argument.\n", OptName); 
	  return(FALSE);
	}
	
	
	Ok = opts_AddParamCheck(OptId,opts_DEFAULTOPTARG);
      } else  
	Ok = opts_AddParamCheck(OptId,opts_Arg);
    } else {
      
      
      
      if (c == ':') 
	return(FALSE); 
      
      
      
      else {
	OptId = opts_ShortOptId(c);
	if (opts_IdIsNull(OptId)) {
	  fprintf(stderr, "\ninternal error: option %c not found.\n", c);
	  opts_ErrorExit(); }
	OptDecl = opts_DeclGetById(OptId);
	
	if (opts_DeclHasReqArg(OptDecl)) {
	  if(!opts_Arg) {
	    fprintf(stderr, "\nerror: option %c requires argument.\n",c); 
	    Ok = FALSE;
	  } else if (opts_StrEqual(opts_Arg, opts_ENDMARKER)) {
	    fprintf(stderr, "\nerror: option %c has delimiter -- as argument.\n",c);
	    Ok = FALSE;
	  } else 
	    Ok = opts_AddParamCheck(OptId,opts_Arg);
	}
	
	else if (opts_DeclHasOptArg(OptDecl)) {
	  
	  if (opts_Arg) {
	    if (opts_StrEqual(opts_Arg, opts_ENDMARKER)) 
	      Ok = opts_AddParamCheck(OptId,opts_DEFAULTOPTARG);
	    else 
	      Ok = opts_AddParamCheck(OptId,opts_Arg); }
	  else 
	    Ok = opts_AddParamCheck(OptId,opts_DEFAULTOPTARG);
	}
	
	else 
	  Ok = opts_AddParamCheck(OptId,opts_DEFAULTOPTARG);
      }
    }
  }
  
  opts_StrFree(ShortOpts);
  opts_FreeLongOptsArray(LongOpts);
  
  return(Ok);
}


BOOL opts_GetValueByName(char* Name, char** Value)

{
  LIST Scan;
  LIST Pair;
  BOOL found;

  found = FALSE;
  Pair  = list_Nil(); 

  for (Scan = opts_PARAMETERS;
       (!found && !list_Empty(Scan)); Scan = list_Cdr(Scan)) {
    Pair = list_Car(Scan);
    if (opts_StrEqual(Name, opts_ClName((OPTID)list_PairFirst(Pair))))
      found = TRUE;
  }

  if (found) 
    (*Value) = list_PairSecond(Pair); 

  return(found);
}

BOOL opts_GetValue(OPTID Id, char** s) 

{
  LIST Scan;
  LIST Pair;
  BOOL found;

  Pair  = list_Nil();
  found = FALSE;

  for (Scan = opts_PARAMETERS;
       (!found && !list_Empty(Scan)); Scan = list_Cdr(Scan)) {
    Pair = list_Car(Scan);
    if (opts_IdEqual(Id, (OPTID)list_PairFirst(Pair)))
      found = TRUE;
  }

  if (found) 
    (*s) = list_PairSecond(Pair);

  return(found);
}


BOOL opts_GetIntValueByName(char* Name, int* Val)

{
  char* ValStr ;

  if (!opts_GetValueByName(Name, &ValStr)) 
    return(FALSE);
  
  return(opts_StrToInt(ValStr, Val)); 
}

BOOL opts_GetIntValue(OPTID Id, int* i) 

{
  return(opts_GetIntValueByName(opts_ClName(Id), i));
}



BOOL opts_IsSet(OPTID Id)

{
  LIST Scan;
  LIST Pair;
  BOOL found;
  
  found = FALSE;

  for (Scan = opts_PARAMETERS;
       (!found && !list_Empty(Scan)); Scan = list_Cdr(Scan)) {
    Pair = list_Car(Scan);
    if (opts_IdEqual(Id, (OPTID)list_Car(Pair)))
      found = TRUE;
  }
  return(found);
}


BOOL opts_IsSetByName(char* Name)

{
  LIST Scan;
  LIST Pair;
  BOOL found;
  
  found = FALSE;

  for (Scan = opts_PARAMETERS;
       (!found && !list_Empty(Scan)); Scan = list_Cdr(Scan)) {
    Pair = list_Car(Scan);
    if (opts_StrEqual(Name, opts_ClName((OPTID)list_PairFirst(Pair))))
      found = TRUE;
  }
  return(found);
}

void opts_SetFlags(FLAGSTORE Store)

{
  int        IntValue;
  OPTID      Id;
  FLAG_FLAGS i;
  FLAGSTORE  OldStore;

  OldStore = flag_GetInstalledStore();
  flag_InstallStore(Store);

  for (i=0; i < flag_MAXFLAG; i++) {
    Id = opts_Id(flag_Name(i));
    if (opts_IsSet(Id)) {
      if (opts_GetIntValue(Id, &IntValue)) {
	flag_SetFlagValue(Id, IntValue);
      } else {
	fprintf(stderr, "\nerror: argument of option %s must be integer.\n",
		flag_Name(i));
      }
    }
  }
  flag_InstallStore(OldStore);
}

void opts_Transfer(FLAGSTORE Store)

{
  LIST      Scan;
  LIST      Pair;
  int       IntValue;
  char      *Name, *ValStr;
  OPTID     Id;
  BOOL      ok;
  FLAGSTORE OldStore;

  OldStore = flag_GetInstalledStore();
  flag_InstallStore(Store);

  Scan = opts_PARAMETERS;

  while (!list_Empty(Scan)) {
    Pair    = list_Car(Scan);
    Id      = (int)list_PairFirst(Pair);
    ValStr  = (char*)list_PairSecond(Pair);
    Name    = opts_ClName(Id);
    
    ok = opts_StrToInt(ValStr, &IntValue);
    if (!ok) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\nerror: argument '%s' of option '%s' must be integer.\n",
			   ValStr, Name);
      misc_FinishUserErrorReport();
    } else { 
      flag_SetFlagValue(flag_Id(Name), IntValue);
    }
    Scan = list_Cdr(Scan);
  }
  flag_InstallStore(OldStore);
}


static void opts_AddParam(OPTID Id, char* ValueString)

{ 
  LIST Pair;
  Pair = list_PairCreate((POINTER)Id, opts_StrDup(ValueString));
  opts_PARAMETERS = list_Cons(Pair, opts_PARAMETERS); 
}

static BOOL opts_AddParamCheck(OPTID Id, char* ValueString)

{
  char* Dummy;
  if (opts_GetValue(Id, &Dummy)) {
    fprintf(stderr, "error: option %s is multiply defined.\n", opts_ClName(Id));
    return(FALSE);
  }
  opts_AddParam(Id, ValueString);
  return(TRUE);
}


static int opts_StrToInt(char* s, int* np)

{
  int i, sign, n;
  
  i = 0;
  sign = 1;
  if (s[0] == '+' || s[0] == '-') {
    if (s[0] == '-')
      sign = -1;
    i = 1;
  }
  if (s[i] == '\0')
    return(0);
  else {
    n = 0;
    for( ; s[i] >= '0' && s[i] <= '9'; i++)
      n = n * 10 + s[i] - '0';
    *np = n * sign;
    return(s[i] == '\0');
  }
}

int opts_Indicator(void) 

{
  return(opts_Ind);
}


static void opts_Exchange (char **argv)



{
  int bottom = opts_FirstNonOpt;
  int middle = opts_LastNonOpt;
  int top = opts_Ind;
  char *tem;

  

  while (top > middle && middle > bottom) {
    if (top - middle > middle - bottom) {
      
      int len = middle - bottom;
      register int i;
      
      
      for (i = 0; i < len; i++) {
	tem = argv[bottom + i];
	argv[bottom + i] = argv[top - (middle - bottom) + i];
	argv[top - (middle - bottom) + i] = tem;
      }
      
      top -= len;
    }
    else {
      
      int len = top - middle;
      register int i;
      
      
      for (i = 0; i < len; i++) {
	tem = argv[bottom + i];
	argv[bottom + i] = argv[middle + i];
	argv[middle + i] = tem;
      }
      
      bottom += len;
    }
  }
  
  

  opts_FirstNonOpt += (opts_Ind - opts_LastNonOpt);
  opts_LastNonOpt = opts_Ind;
}

static const char * opts_GetOptInitialize (int argc, char *const *argv,
					   const char *optstring)


{
  

  opts_FirstNonOpt = opts_LastNonOpt = opts_Ind = 1;

  opts_NextChar = NULL;

  opts_PosixlyCorrect = getenv ("POSIXLY_CORRECT");

  

  if (optstring[0] == '-') {
    opts_Ordering = RETURN_IN_ORDER;
    ++optstring;
  }
  else if (optstring[0] == '+') {
    opts_Ordering = REQUIRE_ORDER;
    ++optstring;
  }
  else if (opts_PosixlyCorrect != NULL)
    opts_Ordering = REQUIRE_ORDER;
  else
    opts_Ordering = PERMUTE;
  
  opts_NonOptionFlagslen = 0;
  
  return optstring;
}

int opts_GetOptInternal (int argc, char *const *argv, const char *optstring,
			 const struct OPTION *longopts, int *longind,
			 int long_only)


{
  opts_Arg = NULL;

  if (!opts_GetOptInitialized || opts_Ind == 0) {
    optstring = opts_GetOptInitialize (argc, argv, optstring);
    opts_Ind = 1;		
    opts_GetOptInitialized = 1;
  }
  
  


  if (opts_NextChar == NULL || *opts_NextChar == '\0') {
    
    
    
    if (opts_LastNonOpt > opts_Ind)
      opts_LastNonOpt = opts_Ind;
    if (opts_FirstNonOpt > opts_Ind)
      opts_FirstNonOpt = opts_Ind;
    
    if (opts_Ordering == PERMUTE) {
      
      
      if (opts_FirstNonOpt != opts_LastNonOpt && opts_LastNonOpt != opts_Ind)
	opts_Exchange ((char **) argv);
      else if (opts_LastNonOpt != opts_Ind)
	opts_FirstNonOpt = opts_Ind;
      
      
      
      while (opts_Ind < argc &&  (argv[opts_Ind][0] != '-' || argv[opts_Ind][1] == '\0'))
	opts_Ind++;
      opts_LastNonOpt = opts_Ind;
    }
    
    
    
    if (opts_Ind != argc && !strcmp (argv[opts_Ind], "--")) {
      opts_Ind++;
      
      if (opts_FirstNonOpt != opts_LastNonOpt && opts_LastNonOpt != opts_Ind)
	opts_Exchange ((char **) argv);
      else if (opts_FirstNonOpt == opts_LastNonOpt)
	opts_FirstNonOpt = opts_Ind;
      opts_LastNonOpt = argc;
      
      opts_Ind = argc;
    }
    
    
    
    if (opts_Ind == argc) {
      
      if (opts_FirstNonOpt != opts_LastNonOpt)
	opts_Ind = opts_FirstNonOpt;
      return -1;
    }
    
    
    
    if ( (argv[opts_Ind][0] != '-' || argv[opts_Ind][1] == '\0')) {
      if (opts_Ordering == REQUIRE_ORDER)
	return -1;
      opts_Arg = argv[opts_Ind++];
      return 1;
    }
    
    
    
    opts_NextChar = (argv[opts_Ind] + 1
		     + (longopts != NULL && argv[opts_Ind][1] == '-'));
  }
  
  
  
  
  
  if (longopts != NULL
      && (argv[opts_Ind][1] == '-'
	  || (long_only && (argv[opts_Ind][2] || !strchr (optstring, argv[opts_Ind][1]))))) {
    char *nameend;
    const struct OPTION *p;
    const struct OPTION *pfound = NULL;
    int exact = 0;
    int ambig = 0;
    int indfound = -1;
    int option_index;
    
    for (nameend = opts_NextChar; *nameend && *nameend != '='; nameend++)
       ;
    
    
    for (p = longopts, option_index = 0; p->name; p++, option_index++)
      if (!strncmp (p->name, opts_NextChar, nameend - opts_NextChar)) {
	if ((unsigned int) (nameend - opts_NextChar)
	    == (unsigned int) strlen (p->name)) {
	  
	  pfound = p;
	  indfound = option_index;
	  exact = 1;
	  break;
	}
	else if (pfound == NULL) {
	  
	  pfound = p;
	  indfound = option_index;
	}
	else
	  
	  ambig = 1;
      }
    
    if (ambig && !exact) {
      if (opts_Err)
	fprintf (stderr, ("%s: option `%s' is ambiguous\n"),
		 argv[0], argv[opts_Ind]);
      opts_NextChar += strlen (opts_NextChar);
      opts_Ind++;
      opts_Opt = 0;
      return '?';
    }
    
    if (pfound != NULL) {
      option_index = indfound;
      opts_Ind++;
      if (*nameend) {
	
	if (pfound->has_arg)
	  opts_Arg = nameend + 1;
	else {
	  if (opts_Err) {
	    if (argv[opts_Ind - 1][1] == '-')
	      
	      fprintf (stderr,
		       ("%s: option `--%s' doesn't allow an argument\n"),
		       argv[0], pfound->name);
	    else
	      
	      fprintf (stderr,
		       ("%s: option `%c%s' doesn't allow an argument\n"),
		       argv[0], argv[opts_Ind - 1][0], pfound->name);
	  }
	  opts_NextChar += strlen (opts_NextChar);
	  
	  opts_Opt = pfound->val;
	  return '?';
	}
      }
      else if (pfound->has_arg == 1) {
	if (opts_Ind < argc)
	  opts_Arg = argv[opts_Ind++];
	else {
	  if (opts_Err)
	    fprintf (stderr,
		     ("%s: option `%s' requires an argument\n"),
		     argv[0], argv[opts_Ind - 1]);
	  opts_NextChar += strlen (opts_NextChar);
	  opts_Opt = pfound->val;
	  return optstring[0] == ':' ? ':' : '?';
	}
      }
      opts_NextChar += strlen (opts_NextChar);
      if (longind != NULL)
	*longind = option_index;
      if (pfound->flag) {
	*(pfound->flag) = pfound->val;
	return 0;
      }
      return pfound->val;
    }
    
    
    if (!long_only || argv[opts_Ind][1] == '-'
	|| strchr (optstring, *opts_NextChar) == NULL) {
      if (opts_Err) {
	if (argv[opts_Ind][1] == '-')
	  
	  fprintf (stderr, ("%s: unrecognized option `--%s'\n"),
		   argv[0], opts_NextChar);
	else
	  
	  fprintf (stderr, ("%s: unrecognized option `%c%s'\n"),
		   argv[0], argv[opts_Ind][0], opts_NextChar);
      }
      opts_NextChar = (char *) "";
      opts_Ind++;
      opts_Opt = 0;
      return '?';
    }
  }
  
  
  
  {
    char c = *opts_NextChar++;
    char *temp = strchr (optstring, c);

    
    if (*opts_NextChar == '\0')
      ++opts_Ind;

    if (temp == NULL || c == ':') {
      if (opts_Err) {
	if (opts_PosixlyCorrect)
	  
	  fprintf (stderr, ("%s: illegal option -- %c\n"),
		   argv[0], c);
	else
	  fprintf (stderr, ("%s: invalid option -- %c\n"),
		   argv[0], c);
      }
      opts_Opt = c;
      return '?';
    }
    
    if (temp[0] == 'W' && temp[1] == ';') {
      char *nameend;
      const struct OPTION *p;
      const struct OPTION *pfound = NULL;
      int exact = 0;
      int ambig = 0;
      int indfound = 0;
      int option_index;
      
      
      if (*opts_NextChar != '\0') {
	opts_Arg = opts_NextChar;
	
	opts_Ind++;
      }
      else if (opts_Ind == argc) {
	if (opts_Err) {
	  
	  fprintf (stderr, ("%s: option requires an argument -- %c\n"),
		   argv[0], c);
	}
	opts_Opt = c;
	if (optstring[0] == ':')
	  c = ':';
	else
	  c = '?';
	return c;
      }
      else
	
	opts_Arg = argv[opts_Ind++];
      
      
      
      for (opts_NextChar = nameend = opts_Arg; *nameend && *nameend != '='; nameend++)
	 ;
      
      
      for (p = longopts, option_index = 0; p->name; p++, option_index++)
	if (!strncmp (p->name, opts_NextChar, nameend - opts_NextChar)) {
	  if ((unsigned int) (nameend - opts_NextChar) == strlen (p->name)) {
	    
	    pfound = p;
	    indfound = option_index;
	    exact = 1;
	    break;
	  }
	  else if (pfound == NULL) {
	    
	    pfound = p;
	    indfound = option_index;
	  }
	  else
	    
	    ambig = 1;
	}
      if (ambig && !exact) {
	if (opts_Err)
	  fprintf (stderr, ("%s: option `-W %s' is ambiguous\n"),
		   argv[0], argv[opts_Ind]);
	opts_NextChar += strlen (opts_NextChar);
	opts_Ind++;
	return '?';
      }
      if (pfound != NULL) {
	option_index = indfound;
	if (*nameend) {
	  
	  if (pfound->has_arg)
	    opts_Arg = nameend + 1;
	  else {
	    if (opts_Err)
	      fprintf(stderr, "%s: option `-W %s' doesn't allow an argument\n",
		      argv[0], pfound->name);
	    
	    opts_NextChar += strlen (opts_NextChar);
	    return '?';
	  }
	}
	else if (pfound->has_arg == 1) {
	  if (opts_Ind < argc)
	    opts_Arg = argv[opts_Ind++];
	  else {
	    if (opts_Err)
	      fprintf(stderr, "%s: option `%s' requires an argument\n",
		      argv[0], argv[opts_Ind - 1]);
	    opts_NextChar += strlen (opts_NextChar);
	    return optstring[0] == ':' ? ':' : '?';
	  }
	}
	opts_NextChar += strlen (opts_NextChar);
	if (longind != NULL)
	  *longind = option_index;
	if (pfound->flag) {
	  *(pfound->flag) = pfound->val;
	  return 0;
	}
	return pfound->val;
      }
      opts_NextChar = NULL;
      return 'W';	
    }
    if (temp[1] == ':') {
      if (temp[2] == ':') {
	
	if (*opts_NextChar != '\0') {
	  opts_Arg = opts_NextChar;
	  opts_Ind++;
	}
	else
	  opts_Arg = NULL;
	opts_NextChar = NULL;
      }
      else {
	
	if (*opts_NextChar != '\0') {
	  opts_Arg = opts_NextChar;
	  
	  opts_Ind++;
	}
	else if (opts_Ind == argc) {
	  if (opts_Err) {
	    
	    fprintf (stderr,
		     ("%s: option requires an argument -- %c\n"),
		     argv[0], c);
	  }
	  opts_Opt = c;
	  if (optstring[0] == ':')
	    c = ':';
	  else
	    c = '?';
	}
	else
	  
	  opts_Arg = argv[opts_Ind++];
	opts_NextChar = NULL;
      }
    }
    return c;
  }
}




static int opts_GetOptLongOnly(int argc, char *const *argv,
			       const char *options,
			       const struct OPTION *long_options,
			       int *opt_index)

{
  return opts_GetOptInternal (argc, argv, options, long_options, opt_index, 1);
}


static void opts_Dummy(void)

{
  opts_PrintParameters();
  opts_PrintDeclarations();
  opts_DeclHasNoArg(NULL);
  opts_PrintLongOpts((struct OPTION*)NULL);
  opts_Dummy();
}
