




 





















#include "pqueue.h"
#include "memory.h"

static __inline__ NAT pq_Parent(NAT Index)
{
  return ((Index+1) / 2) - 1;
}

static __inline__ NAT pq_Left(NAT Index)
{
  return (2 * Index) + 1;
}

static __inline__ NAT pq_Right(NAT Index)
{
  return (2 * Index) + 2;
}


PRIORITY_QUEUE pq_Create(NAT MaxSize, BOOL (*SmallerFct)())
{
  PRIORITY_QUEUE newQueue;
  NAT            i;

#ifdef DEBUG
  if (MaxSize == 0) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In pq_Create: Illegal Input.");
    misc_FinishErrorReport();
  }
#endif

  newQueue = memory_Malloc(sizeof(PQ_NODE));
  newQueue->maxSize = MaxSize;
  newQueue->size    = 0;
  newQueue->items   = memory_Malloc(sizeof(POINTER)*MaxSize);
  newQueue->smaller = SmallerFct;

  for (i = 0; i < MaxSize; i++)
    newQueue->items[i] = NULL;

  return newQueue;
}

void pq_Delete(PRIORITY_QUEUE PQueue)
{
  memory_Free(PQueue->items, sizeof(POINTER)*PQueue->maxSize);
  memory_Free(PQueue, sizeof(PQ_NODE));
}

static __inline__ POINTER pq_AtIndex(PRIORITY_QUEUE PQueue, NAT Index)
{
  return PQueue->items[Index];
}

static void pq_Heapify(PRIORITY_QUEUE PQueue, NAT Index)
{
  NAT l, r, smallest;

  l = pq_Left(Index);
  r = pq_Right(Index);
  smallest = Index;

  
  if (l < pq_Size(PQueue) &&
      PQueue->smaller(pq_AtIndex(PQueue, l), pq_AtIndex(PQueue, Index)))
    smallest = l;
  if (r < pq_Size(PQueue) &&
      PQueue->smaller(pq_AtIndex(PQueue, r), pq_AtIndex(PQueue, smallest)))
    smallest = r;

  if (smallest != Index) {
    
    POINTER help;
    
    help                    = pq_AtIndex(PQueue, Index);
    PQueue->items[Index]    = pq_AtIndex(PQueue, smallest);
    PQueue->items[smallest] = help;

    pq_Heapify(PQueue, smallest);
  }
}
    

POINTER pq_GetMin(PRIORITY_QUEUE PQueue)
{
#ifdef DEBUG
  if (pq_Empty(PQueue)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In pq_GetMin: Heap underflow."); 
    misc_FinishErrorReport();
  }
#endif

  return pq_AtIndex(PQueue, 0);
}


POINTER pq_ExtractMin(PRIORITY_QUEUE PQueue)
{
  POINTER min;

#ifdef DEBUG
  if (pq_Empty(PQueue)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In pq_ExtractMin: Heap underflow.");
    misc_FinishErrorReport();
  }
#endif

  min = pq_AtIndex(PQueue, 0);
  
  
  PQueue->items[0] = pq_AtIndex(PQueue, pq_Size(PQueue)-1);
  PQueue->size--;

  
  pq_Heapify(PQueue, 0);

  return min;
}

void pq_Insert(PRIORITY_QUEUE PQueue, POINTER Item)
{
  NAT index;

#ifdef DEBUG
  if (pq_Size(PQueue) >= pq_MaxSize(PQueue)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In pq_Insert: Heap overflow.");
    misc_FinishErrorReport();
  }
#endif

  index = pq_Size(PQueue);  
  PQueue->size++;

  while (index > 0 &&
	 !PQueue->smaller(pq_AtIndex(PQueue, pq_Parent(index)), Item)) {
    PQueue->items[index] = pq_AtIndex(PQueue, pq_Parent(index));
    index = pq_Parent(index);
  }

  PQueue->items[index] = Item;
}


void pq_Print(PRIORITY_QUEUE PQueue, void (*ItemPrint)())
{
  NAT i;

  for (i = 0; i < pq_Size(PQueue); i++) {
    printf("%u ", i);
    ItemPrint(pq_AtIndex(PQueue, i));
    putchar('\n');
  }
}
