






















#include "proofcheck.h"



static void   PrintFormulaList(FILE* File, LIST Formulas, char* Name);
static void   pcheck_SaveAsDFG2(FILE* File, LIST Axioms, LIST Conjectures);
static string pcheck_StrConc(char* s1, char* s2);
static string pcheck_StrEmpty();
static BOOL   pcheck_IsRightSplitHalf(CLAUSE C);

static BOOL   pcheck_StrToInt(char*, int*);
static BOOL   pcheck_ClauseIsGround(CLAUSE C);


int    pcheck_Timelimit;
char*  pcheck_ProofFileSuffix;
BOOL   pcheck_Quiet;


BOOL        pcheck_ClauseCg;
BOOL        pcheck_GenNamedCg, pcheck_GenRedCg;
char        *pcheck_CgName, *pcheck_RedCgName;
GRAPHFORMAT pcheck_GraphFormat;


static int pcheck_MaxParentSplitLevel(CLAUSE Clause) 

{
  LIST Parents;
  int Max;
  int Act;

  Max = 0;
  Parents = clause_ParentClauses(Clause);

  while (!list_Empty(Parents)) {
    Act = clause_SplitLevel(list_Car(Parents));
    if (Act > Max) 
      Max = Act;
    Parents = list_Cdr(Parents);
  }
  return(Max);
}

static BOOL pcheck_ClauseIsFromLeftSplit(CLAUSE Clause)

{
  if (clause_SplitLevel(Clause) > pcheck_MaxParentSplitLevel(Clause)) {
    return(TRUE);
  }
  return(FALSE);
}

static BOOL pcheck_ClauseIsFromRightSplit(CLAUSE Clause)

{
  if (list_Empty(clause_ParentClauses(Clause)))
    return(FALSE);

  return(clause_IsEmptyClause(list_Car(clause_ParentClauses(Clause))));
}

static BOOL pcheck_ClauseIsFromSplit(CLAUSE Clause)

{
  return (pcheck_ClauseIsFromRightSplit(Clause) ||
	  pcheck_ClauseIsFromLeftSplit(Clause));
}


static int pcheck_LabelToNumber(char* Label)

{
  int Number;

  if (!pcheck_StrToInt(Label, &Number)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In pcheck_LabelToNumber: Could not convert clause label %s to number.\n", 
			 Label);
    misc_FinishUserErrorReport();
  }

  return(Number);
}


static int pcheck_CompareNumberAndClause(const void* Number, const void* ClausePtr)

{
  return (int)Number - clause_Number(*(CLAUSE*)ClausePtr);
}


static void pcheck_ParentNumbersToPointersInVector(CLAUSE* ClauseVector, int Size)

{
  int     Position;
  LIST    NewParents, OldParents;
  LIST    ScanParents;
  int     ParentNum;
  CLAUSE* Parent;
  
  for (Position = 0; Position < Size; Position++)  {
    OldParents = clause_ParentClauses(ClauseVector[Position]);
    NewParents = list_Copy(OldParents);
    
    for (ScanParents = NewParents; !list_Empty(ScanParents); ScanParents = list_Cdr(ScanParents)) {     
      ParentNum = (int)list_Car(ScanParents);
      
      
      
      Parent = bsearch((const void*)ParentNum, ClauseVector, Position,
		       sizeof(CLAUSE), pcheck_CompareNumberAndClause);
      if (Parent != NULL)
	list_Rplaca(ScanParents, *Parent);  
      else {
	misc_StartUserErrorReport();
	misc_UserErrorReport("\n Error: Missing parent clause %d of clause %d.\n", 
			     ParentNum, clause_Number(ClauseVector[Position]));
	misc_FinishUserErrorReport();
      }
    }    
    clause_SetParentClauses(ClauseVector[Position], NewParents);
    list_Delete(OldParents); 
  }
}

static LIST pcheck_ForceParentNumbersToPointersInVector(CLAUSE* ClauseVector,
							int Size)

{
  int     Position;
  LIST    NewParents, OldParents;
  LIST    ScanParents, ScanPLits;
  LIST    OldPLits, NewPLits;
  int     ParentNum, PLitNum;
  CLAUSE  *Parent, Clause;
  BOOL    Ok;
  LIST    Missing;

  Ok = TRUE;
  Missing = list_Nil();

  for (Position = 0; Position < Size; Position++) {
    clause_RemoveFlag(ClauseVector[Position], MARKED); 
    clause_RemoveFlag(ClauseVector[Position], HIDDEN);
  }

  for (Position = 0; Position < Size; Position++)  {
    Clause = ClauseVector[Position];
    if (!clause_GetFlag(Clause, MARKED)) { 
      clause_SetFlag(Clause, MARKED);
      OldParents = clause_ParentClauses(Clause);
      OldPLits   = clause_ParentLiterals(Clause);

      NewParents  = list_Nil();
      NewPLits    = list_Nil();
      ScanParents = OldParents;
      ScanPLits   = OldPLits;
      while  (!list_Empty(ScanParents)) {
	ParentNum = (int)list_Car(ScanParents);
	PLitNum   = (int)list_Car(ScanPLits);
	
	
	
	Parent = bsearch((const void*)ParentNum, ClauseVector, Position,
			 sizeof(CLAUSE), pcheck_CompareNumberAndClause);
	if (Parent == NULL) {
	  Missing = list_Cons((POINTER)ParentNum, Missing); 
	  clause_SetFlag(Clause, HIDDEN);
	} else {
	  if (clause_GetFlag(*Parent, HIDDEN))
	    clause_SetFlag(Clause, HIDDEN);
	  NewParents  = list_Cons((POINTER)*Parent, NewParents);
	  NewPLits    = list_Cons((POINTER)PLitNum,NewPLits);
	}
	ScanParents = list_Cdr(ScanParents); 
	ScanPLits   = list_Cdr(ScanPLits);
      }
      NewParents = list_NReverse(NewParents);
      NewPLits   = list_NReverse(NewPLits);
      clause_SetParentClauses(Clause, NewParents);
      clause_SetParentLiterals(Clause, NewPLits);
      list_Delete(OldPLits);
      list_Delete(OldParents);  
    } 
  } 
 
  return(Missing);
}

static int pcheck_CompareClauseNumber(const void* C1, const void* C2)

{
  return clause_Number(*((CLAUSE*)C1)) - clause_Number(*((CLAUSE*)C2));
}


static LIST pcheck_ConvertParentsInList(LIST List)

{
  int     Size, Index;
  CLAUSE* ClauseVector;
  LIST    Missing;

  Size = list_Length(List);
  if (Size == 0)
    return(list_Nil());

  
  ClauseVector = (CLAUSE*)memory_Malloc(Size * sizeof(CLAUSE));
  for (Index = 0; !list_Empty(List); List = list_Cdr(List), Index++)
    ClauseVector[Index] = list_Car(List);

  
  qsort(ClauseVector, Size, sizeof(CLAUSE), pcheck_CompareClauseNumber);

  
  Missing = pcheck_ForceParentNumbersToPointersInVector(ClauseVector, Size);
  
  memory_Free(ClauseVector, Size * sizeof(CLAUSE));
  return(Missing);
}

LIST pcheck_ConvertParentsInSPASSProof(PROOFSEARCH Search, LIST EmptyClauses) 

{
  LIST    AllLists; 
  LIST    Missing;

  AllLists = list_Nconc(list_Copy(prfs_WorkedOffClauses(Search)),
			list_Nconc(list_Copy(prfs_UsableClauses(Search)),
				   list_Nconc(list_Copy(prfs_DocProofClauses(Search)),
					      list_Copy(EmptyClauses))));
  
  AllLists = pcheck_ClauseNumberMergeSort(AllLists);
  Missing  = pcheck_ConvertParentsInList(AllLists);
  list_Delete(AllLists);

  return(Missing);
}


static LIST pcheck_ParentNumbersToParents(LIST Proof)

{
  LIST    ScanClauses;
  int     ProofLength, Position;
  CLAUSE* ClauseVector;

  if (list_Empty(Proof))
    return(list_Nil());

  
  ProofLength  = list_Length(Proof);
  ClauseVector = (CLAUSE*) memory_Malloc(ProofLength * sizeof(CLAUSE));

  for (ScanClauses = Proof, Position = 0; !list_Empty(ScanClauses);
       ScanClauses = list_Cdr(ScanClauses), Position++) {
    ClauseVector[Position] = list_Car(ScanClauses);
  }

  
  qsort(ClauseVector, ProofLength, sizeof(CLAUSE), pcheck_CompareClauseNumber);
 
  
  pcheck_ParentNumbersToPointersInVector(ClauseVector, ProofLength);
 
  memory_Free(ClauseVector, ProofLength * sizeof(CLAUSE));

  return(Proof);
}


LIST pcheck_ParentPointersToParentNumbers(LIST Clauses)

{
  LIST ScanClauses;
  LIST ScanParents;

  pcheck_ClauseListRemoveFlag(Clauses, MARKED);

  for (ScanClauses = Clauses; !list_Empty(ScanClauses); ScanClauses = list_Cdr(ScanClauses)) {
    if (!clause_GetFlag(list_Car(ScanClauses), MARKED)) {
      for (ScanParents = clause_ParentClauses(list_Car(ScanClauses)); !list_Empty(ScanParents);
	   ScanParents = list_Cdr(ScanParents))
	list_Rplaca(ScanParents, (POINTER)clause_Number(list_Car(ScanParents))); 
      clause_SetFlag(list_Car(ScanClauses), MARKED);
    }
  }
  return(Clauses);
}


LIST pcheck_ConvertTermListToClauseList(LIST ProofRest)

{
  LIST    Clauses;
  LIST    ProofLine;
  TERM    ClauseTerm;
  CLAUSE  Clause;
  int     Level;
  int     ClauseNumber;
  LIST    ParentLabels;
  LIST    ParentIds;     
  LIST    ParentLits;    
  RULE    Origin;
  char*   ClauseLabel;

  Clauses = list_Nil();  

  while (!list_Empty(ProofRest)) {    
    
    ProofLine     = list_Car(ProofRest);
    
    ClauseLabel   = list_First(ProofLine);
    ClauseTerm    = list_Second(ProofLine);    
    
    list_Rplaca(list_Cdr(ProofLine), clause_Null()); 
    ParentLabels  = (LIST)list_Third(ProofLine);
    Level         = (int)list_Fourth(ProofLine);
    Origin        = (RULE)list_Fifth(ProofLine);

    
    Clause       = dfg_CreateClauseFromTerm(ClauseTerm, TRUE);
    ClauseNumber = pcheck_LabelToNumber(ClauseLabel);
    ParentIds    = list_Nil();
    ParentLits   = list_Nil();

    while (!list_Empty(ParentLabels)) {
      ParentIds  = list_Cons((POINTER)pcheck_LabelToNumber(list_Car(ParentLabels)), ParentIds); 
      ParentLits = list_Cons(0, ParentLits);
      ParentLabels   = list_Cdr(ParentLabels); }
    
    
    clause_SetNumber(Clause, ClauseNumber);
    ParentIds = list_NReverse(ParentIds);
    clause_SetParentClauses(Clause, ParentIds);
    clause_SetParentLiterals(Clause, ParentLits);
    Clause->origin = Origin;

    clause_SetSplitLevel(Clause, Level); 
    if (Level > 0) {
      clause_ClearSplitField(Clause);
      clause_SetSplitFieldBit(Clause, Level);
    } else 
      clause_SetSplitField(Clause, (SPLITFIELD)NULL,0);

    clause_RemoveFlag(Clause, MARKED);
    Clauses   = list_Cons(Clause, Clauses);
    ProofRest = list_Cdr(ProofRest);  
  }
  Clauses = list_NReverse(Clauses);
  
  
  Clauses = pcheck_ParentNumbersToParents(Clauses);
  
  return(Clauses);
}

static int pcheck_MaxClauseLevel(LIST Clauses)

{
  int Max;
  int Act;
  LIST Parents;
  
  Max = 0;
  Parents = list_Nil();
  
  while (!list_Empty(Clauses)) {
    Act = clause_SplitLevel(list_Car(Clauses));
    if (Act > Max) 
      Max = Act;
    Clauses = list_Cdr(Clauses);
  }
  return(Max);
}


static BOOL pcheck_ClauseIsUnmarked(CLAUSE C)

{
  return(!clause_GetFlag(C, MARKED));
}


static void pcheck_RemoveUnmarkedFromTableau(TABLEAU T)

{
  if (tab_IsEmpty(T))
    return;
  
  tab_SetClauses(T, list_DeleteElementIf(tab_Clauses(T), pcheck_ClauseIsUnmarked));
  
  pcheck_RemoveUnmarkedFromTableau(tab_LeftBranch(T));
  pcheck_RemoveUnmarkedFromTableau(tab_RightBranch(T));
}

static void pcheck_CollectUnmarkedSplits(TABLEAU T, LIST* Splits)

{
  LIST Scan;

  if (tab_IsEmpty(T))
    return;

  for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    if (!clause_GetFlag(list_Car(Scan), MARKED) && clause_IsFromSplitting(list_Car(Scan))) 
      (*Splits) = list_Cons(list_Car(Scan), *Splits);
  }
  
  pcheck_CollectUnmarkedSplits(tab_LeftBranch(T), Splits);
  pcheck_CollectUnmarkedSplits(tab_RightBranch(T), Splits);
}


static void pcheck_TableauSplitsComplete(TABLEAU T)

{
  if (tab_IsEmpty(T))
    return;
  
  if (tab_RightBranchIsEmpty(T) && !tab_LeftBranchIsEmpty(T)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Error: Split of clause %d has no right branch.\n", 
			 clause_Number(tab_SplitClause(T)));
    misc_FinishUserErrorReport();
  }
 
  if (!tab_RightBranchIsEmpty(T) && tab_LeftBranchIsEmpty(T)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Error: Split of clause %d has no left branch.\n", 
			 clause_Number(tab_SplitClause(T)));
    misc_FinishUserErrorReport();
  }
  
  pcheck_TableauSplitsComplete(tab_LeftBranch(T));
  pcheck_TableauSplitsComplete(tab_RightBranch(T));
}

static void pcheck_RightSplitParents(CLAUSE SplitClause, CLAUSE RightSplitClause,
				     CLAUSE LeftSplitClause)

{
  LIST Scan;
  BOOL HasEmpty, ContainsSplitClause;

  HasEmpty = ContainsSplitClause = FALSE;

  for (Scan = clause_ParentClauses(RightSplitClause);
       !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    if (clause_IsEmptyClause(list_Car(Scan)))
      HasEmpty = TRUE;
    if (clause_Number(list_Car(Scan)) == clause_Number(SplitClause)) 
      ContainsSplitClause = TRUE; 
  }
  
  if (!HasEmpty) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Error: Right split clause %d has no empty clause as parent.\n", 
			 clause_Number(SplitClause));
    misc_FinishUserErrorReport();
  }

  if (!ContainsSplitClause) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Error: Right split clause %d does not have "
			 "its split parent as parent clause.\n",
			 clause_Number(SplitClause));
    misc_FinishUserErrorReport();
  }
  
}


static void pcheck_SplitFormats(TABLEAU T)

{
  LIST Scan;

  if (tab_IsEmpty(T))
    return;

  
  

  for (Scan = tab_RightSplitClauses(T); !list_Empty(Scan);
       Scan = list_Cdr(Scan)) {
    pcheck_RightSplitParents(tab_SplitClause(T), list_Car(Scan),
			     tab_LeftSplitClause(T));
  }
    
  pcheck_SplitFormats(tab_RightBranch(T));
  pcheck_SplitFormats(tab_LeftBranch(T));
}

static void pcheck_SplitLevels(TABLEAU T)

{
  LIST   Scan;
  CLAUSE Clause;
  int    CorrectLevel;

  if (tab_IsEmpty(T))
    return;

  
  for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Clause = list_Car(Scan);
    if (!list_Empty(clause_ParentClauses(Clause))
	&& !clause_IsFromSplitting(Clause)) {
 
     CorrectLevel = pcheck_MaxParentSplitLevel(Clause);
     if (clause_SplitLevel(Clause) != CorrectLevel) {
       misc_StartUserErrorReport();
       misc_UserErrorReport("\n Error: Split level of clause %d should be %d.\n",
			    clause_Number(Clause), CorrectLevel);
       misc_FinishUserErrorReport();
     }
    }
  }

  pcheck_SplitLevels(tab_RightBranch(T));
  pcheck_SplitLevels(tab_LeftBranch(T));
}

static void pcheck_SplitPrecheck(TABLEAU T)

{
  if (tab_IsEmpty(T))
    return;
  
  if (!subs_Subsumes(tab_LeftSplitClause(T), tab_SplitClause(T),-1,-1)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Error: Incorrect split of %d, "
			 "left half of split does not subsume splitted clause.\n",tab_SplitClause(T));
    misc_FinishUserErrorReport();
  }

  if (list_Length(tab_RightSplitClauses(T)) > 1 &&
	!pcheck_ClauseIsGround(tab_LeftSplitClause(T))) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Error: Incorrect split of %d, "
			 "non-ground split generated more than two clause.\n",tab_SplitClause(T));
    misc_FinishUserErrorReport();
  }

  pcheck_SplitPrecheck(tab_LeftBranch(T));
  pcheck_SplitPrecheck(tab_RightBranch(T));
}

BOOL pcheck_BuildTableauFromProof(LIST Proof, TABLEAU* Tableau)

{
  LIST     ProofRest;
  TABLEAU  SplitPos;
  TABPATH  Path;
  CLAUSE   Clause;
  int      ClauseLevel, SplitLevel;
  int      ProofDepth;

  if (list_Empty(Proof)) {
    *Tableau = tab_EmptyTableau();
    return(TRUE); }

  ProofDepth = pcheck_MaxClauseLevel(Proof);
  *Tableau   = tab_CreateNode();
  Path       = tab_PathCreate(ProofDepth, *Tableau);

  ProofRest = Proof;
  while (!list_Empty(ProofRest)) {

    SplitLevel  = tab_PathLength(Path);
    Clause      = list_Car(ProofRest);
    ClauseLevel = clause_SplitLevel(Clause); 

    if (pcheck_ClauseIsFromSplit(Clause)) {

      if (ClauseLevel == 0) {
	misc_StartUserErrorReport();
	misc_UserErrorReport("\n Error: Split level of split clause %d is 0.\n", clause_Number(Clause));
	misc_FinishUserErrorReport();
      }

      if (ClauseLevel > SplitLevel+1) {
	misc_StartUserErrorReport();
	misc_UserErrorReport("\n Error: Split level of split clause %d is not increment"
			     " of current split level.\n", clause_Number(Clause));
	misc_FinishUserErrorReport();
      }

      SplitPos = tab_PathNthNode(Path, ClauseLevel-1);

      if (pcheck_ClauseIsFromLeftSplit(Clause)) {
	if (!tab_LeftBranchIsEmpty(SplitPos)) {
	  misc_StartUserErrorReport();
	  misc_UserErrorReport("\n Error: Multiple left splits for clause %d.\n", 
			       clause_Number(tab_SplitClause(SplitPos)));
	  misc_FinishUserErrorReport();
	}
	else {
	  Path = tab_PathPrefix(ClauseLevel-1, Path);
	  tab_SetSplitClause(SplitPos, list_Car(clause_ParentClauses(Clause)));
	  tab_SetLeftSplitClause(SplitPos, Clause);
	  tab_AddSplitAtCursor(Path, TRUE); }
      }
      
      

      else {
	if (tab_RightBranchIsEmpty(SplitPos)) {
	  
	  if (tab_LeftBranchIsEmpty(SplitPos)) {
	    misc_StartUserErrorReport();
	    misc_UserErrorReport("\n Error: Right split with incorrect split level, clause %d.\n", 
				 clause_Number(Clause));
	    misc_FinishUserErrorReport();
	  }
	  else { 
	    Path = tab_PathPrefix(ClauseLevel-1, Path);
	    tab_AddSplitAtCursor(Path, FALSE); }
	}
	tab_AddRightSplitClause(SplitPos, Clause); 
      } 
    } 
      
    if (ClauseLevel > tab_PathLength(Path)) {
      misc_StartUserErrorReport();
      misc_UserErrorReport("\n Error: Split level of clause %d greater than current level.\n", 
			   clause_Number(Clause));
      misc_FinishUserErrorReport();
    }
    tab_AddClauseOnItsLevel(Clause, Path); 
    
    ProofRest = list_Cdr(ProofRest);   
  } 


  tab_PathDelete(Path);

  return(TRUE);
}


static BOOL pcheck_TableauJustificationsRec(TABLEAU T, TABPATH Path)

{
  CLAUSE  Clause;
  LIST    ScanClauses;
  LIST    ScanParents;
  LIST    Parents;
  CLAUSE  Parent;
  BOOL    Ok, RightSplit;

  if (tab_IsEmpty(T))
    return(TRUE);

  Ok = TRUE; 

  

  for (ScanClauses = tab_Clauses(tab_PathTop(Path));
       !list_Empty(ScanClauses); ScanClauses = list_Cdr(ScanClauses)) {

    Clause  = list_Car(ScanClauses);
    Parents = clause_ParentClauses(Clause);
    
    RightSplit = pcheck_ClauseIsFromRightSplit(Clause); 
    
    
    
    for (ScanParents = Parents; !list_Empty(ScanParents); 
	 ScanParents = list_Cdr(ScanParents)) {
      
      Parent = list_Car(ScanParents);
      
      if ((!(RightSplit && clause_IsEmptyClause(Parent)) &&
	  !(RightSplit && pcheck_ClauseIsFromLeftSplit(Parent))) ||
	  (clause_Number(Parent) > clause_Number(Clause)) ) {
	if (!tab_PathContainsClause(Path, Parent)) {
	  misc_StartUserErrorReport();
	  misc_UserErrorReport("\n Error: Parent clause with number %d is not yet justified.\n", 
			       clause_Number(Parent));
	  misc_FinishUserErrorReport();
	}
      }
    }
  }  
  
  
  
  
  if (!tab_LeftBranchIsEmpty(T)) {
    Path = tab_PathPush(tab_LeftBranch(T), Path);
    Ok   = Ok && pcheck_TableauJustificationsRec(tab_LeftBranch(T), Path);
    Path = tab_PathPop(Path); }
  
  if (!tab_RightBranchIsEmpty(T)) {
    Path = tab_PathPush(tab_RightBranch(T), Path);
    Ok   = Ok && pcheck_TableauJustificationsRec(tab_RightBranch(T), Path);
    Path = tab_PathPop(Path); }

  return(Ok);
}

static BOOL pcheck_TableauJustifications(TABLEAU T)

{
  TABPATH Path;
  BOOL    Ok;

  Path = tab_PathCreate(tab_Depth(T),T);
  Ok   = pcheck_TableauJustificationsRec(T,Path); 
  tab_PathDelete(Path);
  
  return(Ok);
}

BOOL pcheck_TableauProof(TABLEAU* Tableau, LIST Proof)

{
  LIST RedundantClauses;
  LIST EmptyClauses;
  LIST UnmarkedSplits;

  tab_LabelNodes(*Tableau);
  
  if (pcheck_GenNamedCg) 
    tab_WriteTableau(*Tableau, pcheck_CgName, pcheck_GraphFormat);
  
  RedundantClauses = list_Nil();
  if (!pcheck_Quiet) {
    printf("pruning closed branches...");
    fflush(stdout);
  }
  (*Tableau) = tab_PruneClosedBranches(*Tableau, &RedundantClauses);   
  if (!pcheck_Quiet) {
    printf("finished.\n");
  }

  if (!pcheck_Quiet) {
    printf("removing incomplete splits...");
    fflush(stdout);
  }
  (*Tableau) = tab_RemoveIncompleteSplits(*Tableau, &RedundantClauses);  
  if (!pcheck_Quiet) {
    printf("finished.\n");
  }
  
  list_Delete(RedundantClauses);

  
  

  EmptyClauses = list_Nil();
  tab_GetEarliestEmptyClauses(*Tableau, &EmptyClauses);
  pcheck_ClauseListRemoveFlag(Proof, MARKED);
  pcheck_MarkRecursive(EmptyClauses);
  UnmarkedSplits = list_Nil();
  pcheck_CollectUnmarkedSplits(*Tableau, &UnmarkedSplits);
  pcheck_MarkRecursive(UnmarkedSplits);
  pcheck_RemoveUnmarkedFromTableau(*Tableau); 
  list_Delete(UnmarkedSplits);
  list_Delete(EmptyClauses);

  
  if (pcheck_GenRedCg) 
    tab_WriteTableau(*Tableau, pcheck_RedCgName, pcheck_GraphFormat);

  tab_SetSplitLevels(*Tableau);
  pcheck_SplitLevels(*Tableau);
  tab_CheckEmpties(*Tableau);

 if (!tab_IsClosed(*Tableau)) {
    printf("\nerror: tableau is not closed.\n");  
    return(FALSE);
 }

  
  if (!pcheck_Quiet) {
    printf("checking justifications...");
    fflush(stdout);
  }
  if (!pcheck_TableauJustifications(*Tableau))
    return(FALSE);
  if (!pcheck_Quiet) {
    printf("finished.\n");
  }
 
  return(TRUE);
}


void pcheck_MarkRecursive(LIST Clauses)

{
  CLAUSE Clause;

  for (; !list_Empty(Clauses); Clauses = list_Cdr(Clauses)) {
    Clause = list_Car(Clauses);
    if (!clause_GetFlag(Clause, MARKED)) {
      pcheck_MarkRecursive(clause_ParentClauses(Clause)); 
      clause_SetFlag(Clause, MARKED);
    }
  }
}


static void pcheck_CollectTermVariables(TERM Term, LIST* Vars)

{
  LIST Scan;

  if (term_IsComplex(Term)) {
    for (Scan = term_ArgumentList(Term); !list_Empty(Scan);
	 Scan = list_Cdr(Scan))
      pcheck_CollectTermVariables(list_Car(Scan), Vars);
  }
  else if (term_IsVariable(Term)) {
    if (!list_Member(*Vars, Term, term_Equal))
      (*Vars) = list_Cons(term_Copy(Term), *Vars);
  }
}


static BOOL pcheck_IsRightSplitHalf(CLAUSE C)

{
  LIST Parents;
  BOOL Ok;

  Parents = list_Copy(clause_ParentClauses(C));
  Parents = list_PointerDeleteDuplicates(Parents);

  Ok = FALSE;
  if (list_Length(Parents) == 2) {
    if (clause_IsEmptyClause(list_First(Parents)))
      Ok = subs_Subsumes(C, list_Second(Parents),-1,-1);
  }
  list_Delete(Parents);

  return(Ok);
}


static BOOL pcheck_ClauseIsGround(CLAUSE C)

{
  int i;
  LIST Vars;

  Vars = list_Nil();
  for (i = clause_FirstLitIndex(); i <= clause_LastLitIndex(C); i++) {
    pcheck_CollectTermVariables(clause_LiteralAtom(clause_GetLiteral(C,i)), &Vars);
    if (!list_Empty(Vars)) {
      term_ListDelete(Vars);
      return(FALSE);
    }
  }
    
  return(TRUE);
}

static TERM pcheck_UnivClosure(TERM T)

{
  LIST FreeVars;

  FreeVars = list_Nil();
  pcheck_CollectTermVariables(T, &FreeVars);
  
  if (list_Empty(FreeVars))
    return(T);
  return(fol_CreateQuantifier(fol_All(), FreeVars, list_List(T)));
}

static TERM pcheck_ClauseToTerm(CLAUSE Clause)

{
  int  LitScan;
  LIST Args;
  TERM Lit;
  TERM ClauseTerm;

  Args = list_Nil();

  for (LitScan = clause_FirstLitIndex(); LitScan <= clause_LastLitIndex(Clause);
       LitScan++) {
    Lit = clause_LiteralSignedAtom(clause_GetLiteral(Clause, LitScan));
    Args = list_Cons(term_Copy(Lit), Args);
  }

  if (list_Empty(Args))
    Args = list_List(term_Create(fol_False(), list_Nil()));
  
  ClauseTerm = term_Create(fol_Or(), Args);
  ClauseTerm = pcheck_UnivClosure(ClauseTerm);

  return(ClauseTerm);
}

static LIST pcheck_ClauseListToTermList(LIST Clauses)

{
  LIST Terms;

  Terms = list_Nil();
  for (; !list_Empty(Clauses); Clauses = list_Cdr(Clauses))
    Terms = list_Cons(pcheck_ClauseToTerm(list_Car(Clauses)), Terms);
  
  return(Terms);
}

static void pcheck_SaveNumberedDFGProblem(int Number, LIST Axioms, LIST Conjectures,
					  string ProofFileName, string DestPrefix)

{
  char *Filename, *Tmp, *NumStr;
  FILE*     File;

  NumStr   = misc_IntToString(Number);
  Tmp      = pcheck_GenericFilename(ProofFileName, NumStr);
  Filename = pcheck_StrConc(DestPrefix, Tmp);
    
  File = misc_OpenFile(Filename, "w");
  pcheck_SaveAsDFG2(File, Axioms, Conjectures);
  misc_CloseFile(File, Filename);

  misc_StringFree(NumStr);
  misc_StringFree(Tmp);
  misc_StringFree(Filename);
}

static void pcheck_SplitToProblems(TABLEAU T, string ProofFileName,
				   string DestPrefix)

{
  TERM SplitClauseTerm;
  TERM LeftClauseTerm, RightClauseTerm;
  TERM Equiv, Disj, Tmp;
  
  LIST   Conj, Args, Negations;

  SplitClauseTerm = pcheck_ClauseToTerm(tab_SplitClause(T));
  LeftClauseTerm  = pcheck_ClauseToTerm(tab_LeftSplitClause(T));
  
  

  Negations = tab_RightSplitClauses(T);
  if (!list_Empty(Negations) && pcheck_IsRightSplitHalf(list_Car(Negations))) {
    Negations = list_Cdr(Negations);
    
    RightClauseTerm = pcheck_ClauseToTerm(list_Car(tab_RightSplitClauses(T)));
    Disj  = term_Create(fol_Or(), list_Cons(LeftClauseTerm, list_Cons(RightClauseTerm, list_Nil())));
    Equiv = term_Create(fol_Equiv(), list_Cons(SplitClauseTerm, list_Cons(Disj, list_Nil())));
    Conj  = list_List(Equiv);
    
    pcheck_SaveNumberedDFGProblem(clause_Number(tab_LeftSplitClause(T)), list_Nil(), 
				  Conj, ProofFileName, DestPrefix);
    term_ListDelete(Conj);
  }
  
  Args = list_Nil();

  
  if (!list_Empty(Negations)) {

    LeftClauseTerm  = pcheck_ClauseToTerm(tab_LeftSplitClause(T));
    Args  = pcheck_ClauseListToTermList(Negations);
    Tmp   = term_Create(fol_And(), Args);
    Tmp   = term_Create(fol_Not(), list_List(Tmp));
    Equiv = term_Create(fol_Implies(), list_Cons(Tmp, list_Cons(LeftClauseTerm, list_Nil()))); 
    Conj  = list_List(Equiv);
    
    
    pcheck_SaveNumberedDFGProblem(clause_Number(list_Car(tab_RightSplitClauses(T))), 
				  list_Nil(), Conj, ProofFileName, DestPrefix);
    term_ListDelete(Conj);
  }
}

void pcheck_TableauToProofTask(TABLEAU T, string ProofFileName,
			       string DestPrefix)

{
  LIST   Scan;
  LIST   Axioms, Conj, Help;
  CLAUSE Clause;

  if (tab_IsEmpty(T))
    return;

  
  if (!tab_IsLeaf(T))
    pcheck_SplitToProblems(T, ProofFileName, DestPrefix);
 
  
  for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) {

    Clause = list_Car(Scan);
    if (!clause_IsFromSplitting(Clause) &&
	!list_Empty(clause_ParentClauses(Clause))) {
      Axioms = list_Copy(clause_ParentClauses(Clause));
      Axioms = list_PointerDeleteDuplicates(Axioms);
      Help   = Axioms;
      Axioms = pcheck_ClauseListToTermList(Axioms);
      list_Delete(Help);
      Conj   = list_List(pcheck_ClauseToTerm(Clause));
      pcheck_SaveNumberedDFGProblem(clause_Number(Clause), Axioms, Conj,
				    ProofFileName, DestPrefix);
      term_ListDelete(Axioms);
      term_ListDelete(Conj);
    }
  }
  
  
  pcheck_TableauToProofTask(tab_RightBranch(T), ProofFileName, DestPrefix);
  pcheck_TableauToProofTask(tab_LeftBranch(T), ProofFileName, DestPrefix);
}


int pcheck_SeqProofDepth(LIST Proof)
 
{
  int  Max;

  Max = 0;
  
  for ( ; !list_Empty(Proof); Proof = list_Cdr(Proof)) 
    if (clause_Depth(list_Car(Proof)) > Max)
      Max = clause_Depth(list_Car(Proof));
  
  return(Max);
}


LIST pcheck_ReduceSPASSProof(LIST Proof)
  
{
  LIST    EmptyClauses, RedundantClauses;
  LIST    ReducedProof;
  TABLEAU Tableau;
  LIST    UnmarkedSplits;

  if (!pcheck_BuildTableauFromProof(Proof, &Tableau)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n Error: Proof could not be translated in a closed tableau.\n");
    misc_FinishUserErrorReport();
  }

  RedundantClauses = list_Nil();
  Tableau = tab_PruneClosedBranches(Tableau, &RedundantClauses);
  Tableau = tab_RemoveIncompleteSplits(Tableau, &RedundantClauses);
  list_Delete(RedundantClauses);
  
  tab_SetSplitLevels(Tableau);
  
  

  EmptyClauses = list_Nil();
  tab_GetEarliestEmptyClauses(Tableau, &EmptyClauses);
  pcheck_ClauseListRemoveFlag(Proof, MARKED);
  pcheck_MarkRecursive(EmptyClauses);
  UnmarkedSplits = list_Nil();
  pcheck_CollectUnmarkedSplits(Tableau, &UnmarkedSplits);
  pcheck_MarkRecursive(UnmarkedSplits);
  pcheck_RemoveUnmarkedFromTableau(Tableau); 
  list_Delete(UnmarkedSplits);

  ReducedProof = list_Nil(); 
  tab_ToClauseList(Tableau, &ReducedProof);
  ReducedProof = pcheck_ClauseNumberMergeSort(ReducedProof);

  tab_Delete(Tableau);
  list_Delete(EmptyClauses); 

  return(ReducedProof);
}


void pcheck_DeleteProof(LIST Proof)

{
  LIST Line, Scan2, Scan1;
  
  Scan1 = Proof;
  while (!list_Empty(Scan1)) {
    Line = list_Car(Scan1);
    
    misc_StringFree(list_Car(Line));
    if (list_Second(Line) != clause_Null())                
      term_Delete(list_Second(Line));

    

    for(Scan2 = list_Third(Line); !list_Empty(Scan2); Scan2 = list_Cdr(Scan2)) 
      misc_StringFree(list_Car(Scan2));
    list_Delete(list_Third(Line));     

    
    list_Delete(Line);     
    Scan1 = list_Cdr(Scan1);
  }
  list_Delete(Proof);
}


string pcheck_GenericFilename(string Filename, string Id)

{	
  string Help1, Help2, Help3, Res;
  int i;

  Help1 = pcheck_StrConc("_", Id); 
  Help2 = pcheck_StrConc(Help1, pcheck_ProofFileSuffix); 
  
  
  for (i = 0; Filename[i] != '.' && i < strlen(Filename); i++) ;
  Help3 = pcheck_StrPrefix(Filename, i);

  Res = pcheck_StrConc(Help3, Help2);
  misc_StringFree(Help1);
  misc_StringFree(Help2); 
  misc_StringFree(Help3); 

  return(Res);
}


static string pcheck_StrConc(char* s1, char* s2)

{ 
  string dst;

  dst = memory_Malloc(strlen(s1) + strlen(s2) + 1);
  strcpy(dst, s1);
  return(strcat(dst,s2));
}

static string pcheck_StrEmpty(void)

{
  string s ;

  s = memory_Malloc(1);
  s[0] = '\0';
  return(s);
}

string pcheck_StrPrefix(string s, int i)

{
  string dst;

  dst = memory_Malloc(i+1);
  strncpy(dst, s, i);
  dst[i] = '\0';
  return(dst);
}


string pcheck_StrSuffix(string s, int i)

{
  int l;
  string dst;
  
  l = strlen(s);
  if (i >= l)
    return(pcheck_StrEmpty());

  dst = memory_Malloc(l-i+1);
  strcpy(dst, s+i);
  return(dst);
}


void pcheck_ClauseListRemoveFlag(LIST Clauses, CLAUSE_FLAGS Flag)

{
  for (; !list_Empty(Clauses); Clauses = list_Cdr(Clauses)) 
    clause_RemoveFlag(list_Car(Clauses), Flag);
}


LIST pcheck_ClauseNumberMergeSort(LIST L)

{
  LIST SingleStep, DoubleStep, Prev;
  LIST L1, L2;

  if (list_Length(L) < 2)
    return(L);
  
  
  Prev = L;
  SingleStep = list_Cdr(L);
  DoubleStep = list_Cdr(SingleStep);
  
  while (!list_Empty(DoubleStep) && !list_Empty(list_Cdr(DoubleStep))) {
    Prev       = SingleStep;
    SingleStep = list_Cdr(SingleStep);
    DoubleStep = list_Cdr(list_Cdr(DoubleStep)); }
  
  L1 = L;
  L2 = SingleStep;
  list_Rplacd(Prev, list_Nil());

  L = list_NNumberMerge(pcheck_ClauseNumberMergeSort(L1),
			pcheck_ClauseNumberMergeSort(L2),
			(NAT (*)()) clause_Number);

  return(L);
}

static void PrintFormulaList(FILE* File, LIST Formulas, char* Name)

{
  LIST scan;

  fprintf(File, "list_of_formulae(%s).\n", Name);
  for (scan = Formulas; !list_Empty(scan); scan= list_Cdr(scan)) {
    fprintf(File, "  formula(");
    fol_FPrintDFG(File, list_Car(scan));
    fprintf(File, ").\n");
  }
  fprintf(File, "end_of_list.\n\n");
}

static void pcheck_SaveAsDFG2(FILE* File, LIST Axioms, LIST Conjectures)

{
  fprintf(File, "begin_problem(Unknown).\n\n"); 

  fprintf(File, "list_of_descriptions.\n");
  fprintf(File, "name({*Sub Proof*}).\n");
  fprintf(File, "author({* Proof Checker *}).\n");
  fprintf(File, "status(unsatisfiable).\n");
  fprintf(File, "description({* Checks correctness of proof line *}).\n");
  fprintf(File, "end_of_list.\n\n");
  
  fprintf(File, "list_of_symbols.\n");
  fol_FPrintDFGSignature(File);
  fprintf(File, "end_of_list.\n\n");

  PrintFormulaList(File, Axioms, "axioms");
  PrintFormulaList(File, Conjectures, "conjectures");

  fprintf(File, "end_problem.\n");
}


static BOOL pcheck_StrToInt(char* s, int* np)

{
  int i, sign, n;
  
  i = 0;
  sign = 1;
  if (s[0] == '+' || s[0] == '-') {
    if (s[0] == '-')
      sign = -1;
    i = 1;
  }
  if (s[i] == '\0')
    return TRUE;
  else {
    n = 0;
    for( ; s[i] >= '0' && s[i] <= '9'; i++)
      n = n * 10 + s[i] - '0';
    *np = n * sign;
    return (s[i] == '\0');
  }
}
