




 





















#include "renaming.h"

NAT ren_STAMPID;

static BOOL ren_AFactorOk(TERM,TERM);
static BOOL ren_BFactorOk(TERM,TERM);
static BOOL ren_AExtraFactorOk(TERM,TERM);
static BOOL ren_BExtraFactorOk(TERM,TERM);
static TERM ren_FormulaRename(TERM Term, LIST Termlist, LIST *SkolemSymbols);
static LIST ren_GetRenamings(TERM Term1, int Pol1, TERM Term2, int Pol2);
static BOOL ren_HasBenefit(TERM Term1, int Pol1, TERM Term2, int Pol2);
static BOOL ren_PFactorOk(TERM);
static BOOL ren_PExtraFactorOk(TERM);
static int  ren_Polarity(TERM Term);
static BOOL ren_NotPFactorOk(TERM);
static BOOL ren_NotPExtraFactorOk(TERM);
static void ren_ResetTermStamp(TERM Term);

void ren_Init(void)

{
  ren_STAMPID = term_GetStampID();
}

static void ren_ResetTermStamp(TERM Term)

{
  SYMBOL Top;
 
  term_ResetTermStamp(Term);
  Top = term_TopSymbol(Term);

  if (!symbol_IsPredicate(Top)) {
    if (fol_IsQuantifier(Top))
      ren_ResetTermStamp(term_SecondArgument(Term));
    else {
      LIST Scan;
      for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	ren_ResetTermStamp(list_Car(Scan)); 
    }
  }
}

static BOOL ren_HasNEquivFathers(TERM Term1, TERM Term2, NAT n)

{
  Term2 = term_Superterm(Term2);

  while (Term1 != Term2) {
    if (symbol_Equal(term_TopSymbol(Term2),fol_Equiv())) {
      n--;
      if (n == 0)
	return TRUE;
    }
    Term2 = term_Superterm(Term2);
  }

  return FALSE;
}


static BOOL ren_PExtraFactorOk(TERM Term)

{ 
  SYMBOL Top;

  if (term_HasTermStamp(Term)) 
    return FALSE;

  Top = term_TopSymbol(Term);
  
  if (symbol_Equal(Top,fol_Equiv())) {
    if (!fol_IsLiteral(term_FirstArgument(Term)) && !fol_IsLiteral(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;
  }
  if (symbol_Equal(Top,fol_And())) {
    if (list_Length(term_ArgumentList(Term))>2 ||
	ren_PFactorOk(term_FirstArgument(Term)) ||
	ren_PFactorOk(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;
  }
  if (symbol_Equal(Top,fol_Implies()))
    return (ren_NotPExtraFactorOk(term_FirstArgument(Term)) || 
	    ren_PExtraFactorOk(term_SecondArgument(Term)));

  if (symbol_Equal(Top,fol_Not()))
    return ren_NotPExtraFactorOk(term_FirstArgument(Term));

  if (fol_IsQuantifier(Top))
    return ren_PExtraFactorOk(term_SecondArgument(Term));

  if (symbol_Equal(Top,fol_Or())) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (ren_PExtraFactorOk(list_Car(Scan)))
	return TRUE;
  }

  return FALSE; 
}

static BOOL ren_PFactorOk(TERM Term)

{ 
  SYMBOL Top;

  if (term_HasTermStamp(Term)) 
    return FALSE;

  Top = term_TopSymbol(Term);
  
  if (symbol_Equal(Top,fol_Equiv()) || symbol_Equal(Top,fol_And()))
    return TRUE;

  if (symbol_Equal(Top,fol_Implies()))
    return (ren_NotPFactorOk(term_FirstArgument(Term)) || ren_PFactorOk(term_SecondArgument(Term)));

  if (symbol_Equal(Top,fol_Not()))
    return ren_NotPFactorOk(term_FirstArgument(Term));

  if (fol_IsQuantifier(Top))
    return ren_PFactorOk(term_SecondArgument(Term));

  if (symbol_Equal(Top,fol_Or())) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (ren_PFactorOk(list_Car(Scan)))
	return TRUE;
  }

  return FALSE; 
}

static BOOL ren_NotPExtraFactorOk(TERM Term)

{ 
  SYMBOL Top;

  if (term_HasTermStamp(Term)) 
    return FALSE;

  Top = term_TopSymbol(Term);
  
  if (symbol_Equal(Top,fol_Equiv())) {
    if (!fol_IsLiteral(term_FirstArgument(Term)) && !fol_IsLiteral(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;
  }
  if (symbol_Equal(Top,fol_Or())) {
    if (list_Length(term_ArgumentList(Term))>2 ||
	ren_NotPFactorOk(term_FirstArgument(Term)) ||
	ren_NotPFactorOk(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;
  }
  if (symbol_Equal(Top,fol_Implies())) {
    if (ren_PFactorOk(term_FirstArgument(Term)) ||
	ren_NotPFactorOk(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;
  }
  if (symbol_Equal(Top,fol_Not()))
    return ren_PExtraFactorOk(term_FirstArgument(Term));

  if (fol_IsQuantifier(Top))
    return ren_NotPExtraFactorOk(term_SecondArgument(Term));

  if (symbol_Equal(Top,fol_And())) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (ren_NotPExtraFactorOk(list_Car(Scan)))
	return TRUE;
  }

  return FALSE; 
}

static BOOL ren_NotPFactorOk(TERM Term)

{ 
  SYMBOL Top;

  if (term_HasTermStamp(Term)) 
    return FALSE;

  Top = term_TopSymbol(Term);
  
  if (symbol_Equal(Top,fol_Equiv()) || symbol_Equal(Top,fol_Or()) || symbol_Equal(Top,fol_Implies()))
    return TRUE;

  if (symbol_Equal(Top,fol_Not()))
    return ren_PFactorOk(term_FirstArgument(Term));

  if (fol_IsQuantifier(Top))
    return ren_NotPFactorOk(term_SecondArgument(Term));

  if (symbol_Equal(Top,fol_And())) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (ren_NotPFactorOk(list_Car(Scan)))
	return TRUE;
  }

  return FALSE; 
}


static BOOL ren_AFactorOk(TERM Term1, TERM Term2)

{ 
  SYMBOL Top;
  TERM   Super;

  if (Term1 == Term2)
    return FALSE;

  Super = term_Superterm(Term2);
  Top   = term_TopSymbol(Super);    
  
  if (symbol_Equal(Top,fol_And()) || fol_IsQuantifier(Top))
    return ren_AFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Not()))
    return ren_BFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Or())) {
    LIST Scan;
    TERM Sub;
    for (Scan=term_ArgumentList(Super);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Sub = (TERM)list_Car(Scan);
      if (Sub != Term2 && ren_PFactorOk(Sub))
	return TRUE;
    }
    return ren_AFactorOk(Term1, Super);
  }

  if (symbol_Equal(Top,fol_Implies())) {
    if (Term2 == term_FirstArgument(Super))
      return ren_BFactorOk(Term1, Super);
    else
      return (ren_NotPFactorOk(term_FirstArgument(Super)) || ren_AFactorOk(Term1, Super));
  }
  if (symbol_Equal(Top,fol_Equiv())) {
    int Pol;
    Pol = ren_Polarity(Super);
    if (Pol == 0)
      return TRUE;
    if (Term2 == term_FirstArgument(Super))
      Term2 = term_SecondArgument(Super);
    else
      Term2 = term_FirstArgument(Super);

    if (Pol == 1)
      return (ren_NotPFactorOk(Term2) || ren_AFactorOk(Term1,Super));
    else
      return (ren_PFactorOk(Term2) ||  ren_BFactorOk(Term1,Super));
  }
  misc_StartErrorReport();
  misc_ErrorReport("In ren_AFactorOk: Unknown first order operator.");
  misc_FinishErrorReport();
  return FALSE; 
}

static BOOL ren_AExtraFactorOk(TERM Term1, TERM Term2)

{ 
  SYMBOL Top;
  TERM   Super;

  if (Term1 == Term2)
    return FALSE;

  Super = term_Superterm(Term2);
  Top   = term_TopSymbol(Super);    
  
  if (symbol_Equal(Top,fol_And()) || fol_IsQuantifier(Top))
    return ren_AExtraFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Not()))
    return ren_BExtraFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Or())) {
    LIST Scan;
    TERM Sub;
    BOOL Extra;
    Extra = FALSE;
    for (Scan=term_ArgumentList(Super);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Sub = (TERM)list_Car(Scan);
      if (Sub != Term2) {
	if (ren_PFactorOk(Sub)) {
	  if (Extra || ren_PExtraFactorOk(Sub))
	    return TRUE;
	  else
	    Extra = TRUE;
	}
      }
    }
    return ren_AExtraFactorOk(Term1, Super);
  }

  if (symbol_Equal(Top,fol_Implies())) {
    if (Term2 == term_FirstArgument(Super))
      return ren_BExtraFactorOk(Term1, Super);
    else
      return (ren_NotPExtraFactorOk(term_FirstArgument(Super)) || ren_AExtraFactorOk(Term1, Super));
  }
  if (symbol_Equal(Top,fol_Equiv())) {
    int Pol;
    Pol = ren_Polarity(Super);
    if (Term2 == term_FirstArgument(Super))
      Term2 = term_SecondArgument(Super);
    else
      Term2 = term_FirstArgument(Super);

    if (Pol == 0)
      return (ren_PFactorOk(Term2) ||  ren_NotPFactorOk(Term2) ||
	      ren_AFactorOk(Term1,Super) ||  ren_BFactorOk(Term1,Super));

    if (Pol == 1)
      return (ren_NotPExtraFactorOk(Term2) || ren_AExtraFactorOk(Term1,Super));
    else
      return (ren_PExtraFactorOk(Term2) ||  ren_BExtraFactorOk(Term1,Super));
  }
  misc_StartErrorReport();
  misc_ErrorReport("In ren_AExtraFactorOk: Unknown first order operator.");
  misc_FinishErrorReport();
  return FALSE; 
}

static BOOL ren_BFactorOk(TERM Term1, TERM Term2)

{ 
  SYMBOL Top;
  TERM   Super;

  if (Term1 == Term2)
    return FALSE;

  Super = term_Superterm(Term2);
  Top   = term_TopSymbol(Super);    
  
  if (symbol_Equal(Top,fol_Or()) || fol_IsQuantifier(Top))
    return ren_BFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Not()))
    return ren_AFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_And())) {
    LIST Scan;
    TERM Sub;
    for (Scan=term_ArgumentList(Super);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Sub = (TERM)list_Car(Scan);
      if (Sub != Term2 && ren_NotPFactorOk(Sub))
	return TRUE;
    }
    return ren_BFactorOk(Term1, Super);
  }

  if (symbol_Equal(Top,fol_Implies())) {
    if (Term2 == term_FirstArgument(Super))
      return (ren_PFactorOk(term_SecondArgument(Super)) || ren_AFactorOk(Term1, Super));
    else
      return ren_BFactorOk(Term1, Super);
  }
  if (symbol_Equal(Top,fol_Equiv())) {
    int Pol;
    Pol = ren_Polarity(Super);
    if (Pol == 0)
      return TRUE;
    if (Term2 == term_FirstArgument(Super))
      Term2 = term_SecondArgument(Super);
    else
      Term2 = term_FirstArgument(Super);

    if (Pol == 1)
      return (ren_PFactorOk(Term2) || ren_AFactorOk(Term1,Super));
    else
      return (ren_NotPFactorOk(Term2) || ren_BFactorOk(Term1,Super));
  }
  misc_StartErrorReport();
  misc_ErrorReport("In ren_BFactorOk: Unknown first order operator.");
  misc_FinishErrorReport();
  return FALSE; 
}

static BOOL ren_BExtraFactorOk(TERM Term1, TERM Term2)

{ 
  SYMBOL Top;
  TERM   Super;

  if (Term1 == Term2)
    return FALSE;

  Super = term_Superterm(Term2);
  Top   = term_TopSymbol(Super);    
  
  if (symbol_Equal(Top,fol_Or()) || fol_IsQuantifier(Top))
    return ren_BExtraFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Not()))
    return ren_AExtraFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_And())) {
    LIST Scan;
    TERM Sub;
    BOOL Extra;
    Extra = FALSE;
    for (Scan=term_ArgumentList(Super);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Sub = (TERM)list_Car(Scan);
      if (Sub != Term2) {
	if (ren_NotPFactorOk(Sub)) {
	  if (Extra || ren_NotPExtraFactorOk(Sub))
	    return TRUE;
	  else
	    Extra = TRUE;
	}
      }
    }
    return ren_BExtraFactorOk(Term1, Super);
  }

  if (symbol_Equal(Top,fol_Implies())) {
    if (Term2 == term_FirstArgument(Super))
      return (ren_PExtraFactorOk(term_SecondArgument(Super)) || ren_AExtraFactorOk(Term1, Super));
    else
      return ren_BExtraFactorOk(Term1, Super);
  }
  if (symbol_Equal(Top,fol_Equiv())) {
    int Pol;
    Pol = ren_Polarity(Super);
    if (Term2 == term_FirstArgument(Super))
      Term2 = term_SecondArgument(Super);
    else
      Term2 = term_FirstArgument(Super);

    if (Pol == 0)
      return (ren_PFactorOk(Term2) ||  ren_NotPFactorOk(Term2) ||
	      ren_AFactorOk(Term1,Super) ||  ren_BFactorOk(Term1,Super));

    if (Pol == 1)
      return (ren_PExtraFactorOk(Term2) || ren_AExtraFactorOk(Term1,Super));
    else
      return (ren_NotPExtraFactorOk(Term2) ||  ren_BExtraFactorOk(Term1,Super));
    }
  misc_StartErrorReport();
  misc_ErrorReport("In ren_BExtraFactorOk: Unknown first order operator.");
  misc_FinishErrorReport();
  return FALSE; 
}

static BOOL ren_HasBenefit(TERM Term1, int Pol1, TERM Term2, int Pol2)

{ 
  BOOL  PFacOk, NotPFacOk, AFacOk, BFacOk;
  
  switch (Pol1*Pol2) {
    
  case 0:
    PFacOk    = ren_PFactorOk(Term2);
    NotPFacOk = ren_NotPFactorOk(Term2);
    AFacOk    = ren_AFactorOk(Term1,Term2);
    BFacOk    = ren_BFactorOk(Term1,Term2);
    

    return ((AFacOk && BFacOk && PFacOk && NotPFacOk) ||
	    (AFacOk && PFacOk && (ren_PExtraFactorOk(Term2) || ren_AExtraFactorOk(Term1,Term2))) ||
	    (BFacOk && NotPFacOk && (ren_NotPExtraFactorOk(Term2) || ren_BExtraFactorOk(Term1,Term2))));
    break;

  case 1:
    return (ren_PFactorOk(Term2) && ren_AFactorOk(Term1,Term2));
    break;

  case -1:
    return (ren_NotPFactorOk(Term2) && ren_BFactorOk(Term1,Term2));
    break;
  }
  misc_StartErrorReport();
  misc_ErrorReport("In ren_HasBenefit: Unknown polarity.");
  misc_FinishErrorReport();
  return FALSE;  
}

static BOOL ren_HasNonZeroBenefit(TERM Term1, int Pol1, TERM Term2, int Pol2)

{ 
  BOOL  PFacOk, NotPFacOk, AFacOk, BFacOk, PEFacOk, NotPEFacOk, AEFacOk, BEFacOk;
  
  switch (Pol1*Pol2) {
    
  case 0:
    PFacOk     = ren_PFactorOk(Term2);
    NotPFacOk  = ren_NotPFactorOk(Term2);
    AFacOk     = ren_AFactorOk(Term1,Term2);
    BFacOk     = ren_BFactorOk(Term1,Term2);
    PEFacOk    = ren_PExtraFactorOk(Term2);
    NotPEFacOk = ren_NotPExtraFactorOk(Term2);
    AEFacOk    = ren_AExtraFactorOk(Term1,Term2);
    BEFacOk    = ren_BExtraFactorOk(Term1,Term2);
    
    return ((AFacOk && BFacOk && PFacOk && NotPFacOk && (AEFacOk || BEFacOk || PEFacOk || NotPEFacOk)) || 
	    (AFacOk && PFacOk && (ren_PExtraFactorOk(Term2) && ren_AExtraFactorOk(Term1,Term2))) ||
	    (Pol1 == 0 && (symbol_Equal(term_TopSymbol(Term2),fol_Equiv()) ||
			   ren_HasNEquivFathers(Term1,Term2,1))) ||
	    ren_HasNEquivFathers(Term1,Term2,2) ||
	    (BFacOk && NotPFacOk && (ren_NotPExtraFactorOk(Term2) && ren_BExtraFactorOk(Term1,Term2))));
    break;

  case 1:
    return (ren_PExtraFactorOk(Term2) || ren_AExtraFactorOk(Term1,Term2) ||
	    ren_HasNEquivFathers(Term1,Term2,2));
    break;

  case -1:
    return (ren_NotPExtraFactorOk(Term2) || ren_BExtraFactorOk(Term1,Term2) ||
	    ren_HasNEquivFathers(Term1,Term2,2));
    break;
  }
  misc_StartErrorReport();
  misc_ErrorReport("In ren_HasNonZeroBenefit: Unknown polarity.");
  misc_FinishErrorReport();
  return FALSE;  
}


static LIST ren_GetRenamings(TERM Term1, int Pol1, TERM Term2, int Pol2) 

{ 
  SYMBOL Top;
  LIST   Result,Scan;

  Result = list_Nil();

  

  if (fol_IsLiteral(Term2))
    return Result;

  Top = term_TopSymbol(Term2);

  if (symbol_Equal(Top,fol_Not()))
    return ren_GetRenamings(Term1, Pol1, term_FirstArgument(Term2), Pol2*-1);
  
  if (term_Superterm(Term2) && !fol_IsQuantifier(term_TopSymbol(term_Superterm(Term2))) &&
      ren_HasBenefit(Term1, Pol1, Term2, Pol2)) {
    Result = list_Cons(Term2,Result);
    term_SetTermStamp(Term2);
    Term1 = Term2;
    Pol1  = Pol2;
  }

  if (fol_IsQuantifier(Top))
    return list_Nconc(Result,ren_GetRenamings(Term1, Pol1, term_SecondArgument(Term2), Pol2));

  if (symbol_Equal(Top,fol_And()) || symbol_Equal(Top,fol_Or()))
    for (Scan=term_ArgumentList(Term2);!list_Empty(Scan);Scan=list_Cdr(Scan))
      Result = list_Nconc(Result,ren_GetRenamings(Term1,Pol1,list_Car(Scan),Pol2));
  else
    if (symbol_Equal(Top,fol_Implies()))
      Result = list_Nconc(Result,
        list_Nconc(ren_GetRenamings(Term1,Pol1,term_FirstArgument(Term2),Pol2*-1),
		   ren_GetRenamings(Term1,Pol1,term_SecondArgument(Term2),Pol2)));
    else
      if (symbol_Equal(Top,fol_Equiv()))
	Result = list_Nconc(Result,
			    list_Nconc(ren_GetRenamings(Term1,Pol1,term_FirstArgument(Term2),0),
				       ren_GetRenamings(Term1,Pol1,term_SecondArgument(Term2),0)));
      else {
	misc_StartErrorReport();
	misc_ErrorReport("In ren_GetRenamings: Unknown first-order operator.");
	misc_FinishErrorReport();
      }

  return Result;
}

static int ren_Polarity(TERM Term)

{
  TERM SuperTerm;

  SuperTerm = term_Superterm(Term);

  if (SuperTerm) {
    SYMBOL Top;
    Top = term_TopSymbol(SuperTerm);
    if (symbol_Equal(Top,fol_And()) || symbol_Equal(Top,fol_Or()) || fol_IsQuantifier(Top))
      return ren_Polarity(SuperTerm);
    if (symbol_Equal(Top,fol_Not()))
      return (-1*ren_Polarity(SuperTerm));
    if (symbol_Equal(Top,fol_Equiv()))
      return 0;
    if (symbol_Equal(Top,fol_Implies())) {
      if (Term == term_FirstArgument(SuperTerm))
	return (-1*ren_Polarity(SuperTerm));
      else
	return ren_Polarity(SuperTerm);
    }
    misc_StartErrorReport();
    misc_ErrorReport("In ren_Polarity: Unknown first-order operator.");
    misc_FinishErrorReport();
  }

  return 1;
}


static LIST ren_RemoveTerm(TERM Term, LIST Renamings)

{
  LIST Scan;

  if (list_Empty(Renamings))
    return Renamings;

  while (!list_Empty(Renamings) && ren_Hit(list_Car(Renamings)) == Term) {
    Scan      = list_Cdr(Renamings);
    ren_Delete(list_Car(Renamings));
    list_Free(Renamings);
    Renamings = Scan;
  }

  if (list_Empty(Renamings))
    return Renamings;

  ren_SetMatches(list_Car(Renamings), 
		 list_PointerDeleteElement(ren_Matches(list_Car(Renamings)), Term));

  list_Rplacd(Renamings, ren_RemoveTerm(Term, list_Cdr(Renamings)));

  return Renamings;
}

static LIST ren_RemoveAllSubterms(TERM Term, LIST Renamings)

{
  Renamings = ren_RemoveTerm(Term, Renamings);
  
  if (!symbol_IsPredicate(term_TopSymbol(Term))) {
    if (fol_IsQuantifier(term_TopSymbol(Term)))
      Renamings = ren_RemoveAllSubterms(term_SecondArgument(Term), Renamings);
    else {
      LIST Scan;
      for(Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	Renamings = ren_RemoveAllSubterms(list_Car(Scan), Renamings);
    }
  }

  return Renamings;
}



static LIST ren_SolveDependencies(LIST Renamings)

{
  LIST     Scan;
  RENAMING Renaming;
  TERM     ActMatch;

  if (list_Empty(Renamings))
    return Renamings;

  Renaming = (RENAMING)list_Car(Renamings);
  for(Scan=ren_Matches(Renaming);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    ActMatch = (TERM)list_Car(Scan);
    list_Rplacd(Renamings, ren_RemoveAllSubterms(ActMatch, list_Cdr(Renamings)));
  }
  list_Rplacd(Renamings, ren_SolveDependencies(list_Cdr(Renamings)));

  return Renamings;
}


static TERM ren_FormulaRename(TERM Term, LIST Renamings, LIST *SkolemSymbols) 

{
  TERM     Result,ActTerm,Hit,DefTerm,Superterm,NewTerm;
  LIST     Scan,FreeVariables,Args;
  SYMBOL   ActSymbol;
  RENAMING Renaming;
 
  DefTerm = (TERM)NULL;

  if (!list_Empty(Renamings))
    Result = term_Create(fol_AND,list_List(Term));
  else
    return Term;

  ActSymbol = 0;

  while (!list_Empty(Renamings)) {

    Renaming       = (RENAMING)list_Car(Renamings); 
    Renamings      = list_Cdr(Renamings);
    Hit            = ren_Hit(Renaming);
    Superterm      = term_Superterm(Hit);
    FreeVariables  = fol_FreeVariables(Hit);
    ActSymbol      = symbol_CreateSkolemPredicate(list_Length(FreeVariables));
    *SkolemSymbols = list_Cons((POINTER)ActSymbol,*SkolemSymbols);

    
    NewTerm        = term_Create(ActSymbol, 
				 list_CopyWithElement(FreeVariables,(POINTER (*)())term_Copy));
    switch (ren_OverallPolarity(Renaming)) {
    case 0:
      DefTerm = term_Create(fol_Equiv(),list_Cons(term_Copy(NewTerm),list_List(Hit)));
      break;
	  
    case 1:
      DefTerm = term_Create(fol_Implies(),list_Cons(term_Copy(NewTerm),list_List(Hit)));
      break;
	  
    case -1:
      DefTerm = term_Create(fol_Implies(),list_Cons(Hit,list_List(term_Copy(NewTerm))));
      break;
    }
    term_RplacSuperterm(term_FirstArgument(DefTerm),DefTerm);
    term_RplacSuperterm(term_SecondArgument(DefTerm),DefTerm);
    if (!list_Empty(FreeVariables))
      DefTerm = fol_CreateQuantifier(fol_All(),
				     list_CopyWithElement(FreeVariables,
							  (POINTER (*)())term_Copy), 
				     list_List(DefTerm));
    term_RplacArgumentList(Result,list_Nconc(term_ArgumentList(Result),list_List(DefTerm)));

    
    term_RplacSuperterm(NewTerm, Superterm);
    for(Args=term_ArgumentList(Superterm);!list_Empty(Args); Args=list_Cdr(Args)) 
      if ((TERM)list_Car(Args) == Hit) {
	list_Rplaca(Args, NewTerm);
	break;
      }    

    for (Scan=ren_Matches(Renaming); !list_Empty(Scan); Scan=list_Cdr(Scan)) {
      
      ActTerm        = (TERM)list_Car(Scan);
      Superterm      = term_Superterm(ActTerm);

      
      NewTerm        = term_Create(ActSymbol, 
				   list_CopyWithElement(FreeVariables,
							(POINTER (*)())term_Copy));
      
      cont_StartBinding();
      if (unify_MatchFlexible(cont_LeftContext(), Hit, ActTerm))
	cont_ApplyBindingsModuloMatching(cont_LeftContext(), NewTerm);
      else {
	misc_StartErrorReport();
	misc_ErrorReport("\n In ren_FormulaRename: Further match is no instance of hit.\n");
	misc_FinishErrorReport();
      }
      cont_BackTrack();

      
      term_RplacSuperterm(NewTerm, Superterm);
      for(Args=term_ArgumentList(Superterm);!list_Empty(Args); Args=list_Cdr(Args)) 
	if ((TERM)list_Car(Args) == ActTerm) {
	  list_Rplaca(Args, NewTerm);
	  break;
	}  
      term_Delete(ActTerm);
    }
    list_Delete(ren_Matches(Renaming));
    ren_SetMatches(Renaming, list_Nil());
    list_Delete(FreeVariables);
  }
  return Result;			   
}

static LIST ren_FreeRenaming(LIST Renamings)

{
  LIST     Scan;
  TERM     Father, Term;
  RENAMING Candidate;

  for(Scan=Renamings;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    Candidate = (RENAMING)list_Car(Scan);
    if (list_Empty(ren_Matches(Candidate))) {
      Term   = ren_Hit(Candidate);
      Father = term_Superterm(Term);
      while (!term_HasTermStamp(Father) && term_Superterm(Father)) {
	Father = term_Superterm(Father);
      }
	
      if (!ren_HasNonZeroBenefit( Father,  (term_Superterm(Father) ? ren_Polarity(Father) : 1),  
				  Term,  ren_OverallPolarity(Candidate))) {
	ren_Delete(Candidate);
	list_Rplaca(Scan,NULL);
	term_ResetTermStamp(Term);
      }
    }
  }

  Renamings = list_PointerDeleteElement(Renamings,NULL);

  return Renamings;
}

LIST ren_FurtherMatches(TERM Formula, LIST Formulas)

{
  LIST Scan1, Scan2, Allmatches, Matchables, Renamings;
  TERM Hit;
  int  Polarity, NewPol;

  Allmatches = list_Nil();
  Renamings  = list_Nil();

  for(Scan1=Formulas; !list_Empty(Scan1); Scan1=list_Cdr(Scan1)) {
    Hit = (TERM)list_Car(Scan1);

    if (!list_PointerMember(Allmatches, Hit)) {
      Matchables = list_Cons(Hit, fol_Generalizations(Formula, Hit));
      Hit        = fol_MostGeneralFormula(Matchables); 
      list_Delete(Matchables);
      Matchables = fol_Instances(Formula, Hit);
      Polarity   = ren_Polarity(Hit);
	
      for(Scan2=Matchables; !list_Empty(Scan2); Scan2=list_Cdr(Scan2)) {
	if(list_PointerMember(Allmatches, list_Car(Scan2))) {
	  list_Rplaca(Scan2, NULL);
	}
	else {
	  NewPol = ren_Polarity(list_Car(Scan2));
	  if(Polarity != 0 && ((NewPol == 0) || (NewPol != Polarity))) {
	    Polarity = 0;
	  }
	}
      }
      Matchables = list_PointerDeleteElement(Matchables, NULL);
      Allmatches = list_Nconc(list_Copy(Matchables), Allmatches);
      Renamings  = list_Cons(ren_Create(Hit, Matchables, Polarity),Renamings);
    }
  }
  list_Delete(Allmatches);

  return Renamings;
}


TERM ren_Rename(TERM Term, LIST *SkolemSymbols, BOOL Document, BOOL Match)

{
  LIST Renamings, Scan, Formulas;

  Renamings = list_Nil();
  Formulas  = list_Nil();

  if (term_StampOverflow(ren_STAMPID))
    ren_ResetTermStamp(Term);

#ifdef DEBUG
  fol_CheckFatherLinks(Term);
#endif

  term_StartStamp();

  Formulas = ren_GetRenamings(Term,1,Term,1);

  Formulas = list_GreaterNumberSort(Formulas, fol_Depth);

  if (Match)
    Renamings = ren_FurtherMatches(Term, Formulas);
  else {
    for(Scan=Formulas;!list_Empty(Scan);Scan=list_Cdr(Scan))
      Renamings = list_Cons(ren_Create(list_Car(Scan),list_Nil(),ren_Polarity(list_Car(Scan))),Renamings);
  }

  Renamings = ren_FreeRenaming(Renamings);
  
  Renamings = ren_SolveDependencies(Renamings);

  if (!list_Empty(Renamings) && Document) {
    printf("\n\n\t Renaming term:\n ");
    fol_PrettyPrintDFG(Term);
    for(Scan=Renamings;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      LIST     Matches;
      RENAMING Candidate;
      Candidate = (RENAMING)list_Car(Scan);
      printf("\n\t Found:\n");
      fol_PrettyPrintDFG(ren_Hit(Candidate));
      printf("\n\t Instances:\n");
      for(Matches=ren_Matches(Candidate); !list_Empty(Matches); Matches=list_Cdr(Matches)) {
	fol_PrettyPrintDFG(list_Car(Matches));
	if (!list_Empty(list_Cdr(Matches)))
	  printf("\n");
      }
    }
  }

  Term = ren_FormulaRename(Term, Renamings, SkolemSymbols);
  
  if (!list_Empty(Renamings) && Document) {
    printf("\n\n\t Renamed term:\n");
    fol_PrettyPrintDFG(Term);
  }

  list_DeleteWithElement(Renamings, ren_Delete);
  list_Delete(Formulas);

  term_StopStamp();

  return Term;
}
