























     


#include "rpos.h"

const NAT RPOS_NONGROUND = 0;
const NAT RPOS_DONTKNOW  = 1;

static RPOS_TERM rpos_TermCreate(void)

{
  RPOS_TERM result   = memory_Malloc(sizeof(RPOS_TERMNODE));
  result->term       = NULL;
  result->parent     = NULL;
  result->children   = list_Nil();
  result->ord        = RPOS_DONTKNOW;
  result->size       = 1;   
  result->count      = 0;
  result->newSubList = list_Nil();
  result->smaller    = list_Nil();
  result->equal      = list_Nil();
  result->visited    = FALSE;
  return result;
}

static void rpos_TermDelete(RPOS_TERM RTerm)

{
  LIST scan;

  for (scan = RTerm->children; !list_Empty(scan); scan = list_Pop(scan))
    rpos_TermDelete(list_Car(scan));
  for (scan = RTerm->newSubList; !list_Empty(scan); scan = list_Pop(scan))
    rpos_TermDelete(list_Car(scan));
  list_Delete(RTerm->smaller);
  list_Delete(RTerm->equal);
  memory_Free(RTerm, sizeof(RPOS_TERMNODE));
}


static void rpos_CountVariables(TERM Term, BOOL *Ground, NAT Index)

{
  int  stack;

#ifdef DEBUG
  if (!term_IsTerm(Term) || Index > 1) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In rpos_CountVariables: Illegal input."); 
    misc_FinishErrorReport();
  }
#endif

  stack   = stack_Bottom();
  *Ground = TRUE;             

  do {
    if (term_IsStandardVariable(Term)) {
      ord_VARCOUNT[term_TopSymbol(Term)][Index]++;
      *Ground = FALSE;
    } else {
      if (term_IsComplex(Term))
	stack_Push(term_ArgumentList(Term));
    }

    while (!stack_Empty(stack) && list_Empty(stack_Top()))
      stack_Pop();

    if (!stack_Empty(stack)) {
      Term = list_Car(stack_Top());
      stack_RplacTop(list_Cdr(stack_Top()));
    }
  } while (!stack_Empty(stack));
}


static void rpos_VarCondition(TERM Term1, BOOL *VarCond1, BOOL *Ground1,
			      TERM Term2, BOOL *VarCond2, BOOL *Ground2)

{
  SYMBOL maxVar1, maxVar2;
  NAT    i;

  *VarCond1 = *VarCond2 = TRUE;
  maxVar1   = term_MaxVar(Term1);
  maxVar2   = term_MaxVar(Term2);

  if (symbol_GreaterVariable(maxVar2, maxVar1))
    maxVar1 = maxVar2;

  for (i = 0; i <= maxVar1; i++) {
    ord_VARCOUNT[i][0] = 0;
    ord_VARCOUNT[i][1] = 0;
  }

  rpos_CountVariables(Term1, Ground1, 0);
  rpos_CountVariables(Term2, Ground2, 1);

  for (i = 0; i <= maxVar1; i++) {
    if (ord_VARCOUNT[i][0] < ord_VARCOUNT[i][1]) {
      *VarCond1 = FALSE;
      if (!*VarCond2)
	return;
    }
    else if (ord_VARCOUNT[i][0] > ord_VARCOUNT[i][1]) {
      *VarCond2 = FALSE;
      if (!*VarCond1)
	return;
    }
  }
}


static RPOS_TERM rpos_InitComparison(TERM Term, PRIORITY_QUEUE PQueue)

{
  RPOS_TERM rTerm;
  LIST      scan;

  rTerm = rpos_TermCreate();
  rTerm->term = Term;

  if (symbol_IsVariable(term_TopSymbol(Term)))
    rTerm->ord = RPOS_NONGROUND;
  else if (!term_IsComplex(Term))
    pq_Insert(PQueue, rTerm);   
  else {  
    if (symbol_HasProperty(term_TopSymbol(Term), ORDRIGHT))
      
      rTerm->children = list_Reverse(term_ArgumentList(Term));
    else
      
      rTerm->children = list_Copy(term_ArgumentList(Term));

    for (scan = rTerm->children; !list_Empty(scan); scan = list_Cdr(scan)) {
      RPOS_TERM child = rpos_InitComparison(list_Car(scan), PQueue);
      list_Rplaca(scan, child);
      child->parent = rTerm;
      if (child->ord == RPOS_NONGROUND)
	rTerm->ord = RPOS_NONGROUND;
      rTerm->size += child->size;
    }
  }
  return rTerm;
}


static RPOS_TERM rpos_ContInitComparison(CONTEXT Context, TERM Term, PRIORITY_QUEUE PQueue)

{
  RPOS_TERM rTerm;
  LIST      scan;

  Term  = cont_Deref(&Context, Term);
  rTerm = rpos_TermCreate();
  rTerm->term = Term;

  if (symbol_IsVariable(term_TopSymbol(Term)))
    rTerm->ord = RPOS_NONGROUND;
  else if (!term_IsComplex(Term))
    pq_Insert(PQueue, rTerm);   
  else {  
    if (symbol_HasProperty(term_TopSymbol(Term), ORDRIGHT))
      
      rTerm->children = list_Reverse(term_ArgumentList(Term));
    else
      
      rTerm->children = list_Copy(term_ArgumentList(Term));

    for (scan = rTerm->children; !list_Empty(scan); scan = list_Cdr(scan)) {
      RPOS_TERM child = rpos_ContInitComparison(Context, list_Car(scan), PQueue);
      list_Rplaca(scan, child);
      child->parent = rTerm;
      if (child->ord == RPOS_NONGROUND)
	rTerm->ord = RPOS_NONGROUND;
      rTerm->size += child->size;
    }
  }
  return rTerm;
}


static ord_RESULT rpos_GroundTermCompare(RPOS_TERM RT1, RPOS_TERM RT2)

{
  SYMBOL s1, s2;

  if (RT1->term == RT2->term)
    return ord_EQUAL;     

  s1 = term_TopSymbol(RT1->term);
  s2 = term_TopSymbol(RT2->term);

  if (symbol_PrecedenceGreater(s1, s2))
    return ord_GREATER_THAN;
  else if (symbol_PrecedenceGreater(s2, s1))
    return ord_SMALLER_THAN;
  else {
    
    LIST scan1, scan2;
    
    for (scan1 = RT1->children, scan2 = RT2->children;
	 !list_Empty(scan1) && !list_Empty(scan2);
	 scan1 = list_Cdr(scan1), scan2 = list_Cdr(scan2)) {
      RT1 = list_Car(scan1);
      RT2 = list_Car(scan2);

      if (RT1->ord > RT2->ord)
	return ord_GREATER_THAN;
      else if (RT1->ord < RT2->ord)
	return ord_SMALLER_THAN;
    }

    if (list_Empty(scan1) && list_Empty(scan2))
      return ord_EQUAL;
    else if (list_Empty(scan1))
      
      return ord_SMALLER_THAN;
    else
      
      return ord_GREATER_THAN;
  }
}


static BOOL rpos_GroundTermEqual(RPOS_TERM RT1, RPOS_TERM RT2)

{
  return rpos_GroundTermCompare(RT1, RT2) == ord_EQUAL;
}


static BOOL rpos_GroundTermSmaller(RPOS_TERM RT1, RPOS_TERM RT2)

{
  return rpos_GroundTermCompare(RT1, RT2) == ord_SMALLER_THAN;
}


static void rpos_UpdateMUL(RPOS_TERM RTerm)
{
  
  BOOL mul;
  LIST scan, nground;

  if (term_IsComplex(RTerm->term)) {
    nground = list_Nil();
    mul = symbol_HasProperty(term_TopSymbol(RTerm->term), ORDMUL);

    for (scan = RTerm->children; !list_Empty(scan); scan = list_Cdr(scan)) {
      rpos_UpdateMUL(list_Car(scan));
      if (mul && ((RPOS_TERM)list_Car(scan))->ord == RPOS_NONGROUND)
	nground = list_Cons(list_Car(scan), nground);
    }
    
    if (mul) {
      if (RTerm->ord == RPOS_NONGROUND) {
	
	
	list_Delete(RTerm->children);
	RTerm->children   = RTerm->newSubList;
	RTerm->newSubList = nground;
      } else {
	
	
	
	
	
	
	if (RTerm->ord < symbol_Arity(term_TopSymbol(RTerm->term)))
	  list_Delete(RTerm->newSubList);
      }
    }
  }
}

      
static LIST rpos_UpdateParent(RPOS_TERM RTerm)
{
  RPOS_TERM parent;
  LIST      result;

  parent = RTerm->parent;
  result = list_Nil();

  if (parent != NULL) {
    if (symbol_HasProperty(term_TopSymbol(parent->term), ORDMUL))
      parent->newSubList = list_Cons(RTerm, parent->newSubList);
    
    parent->count++;
    
    if (parent->count == symbol_Arity(term_TopSymbol(parent->term))) {
      if (symbol_HasProperty(term_TopSymbol(parent->term), ORDMUL)) {
	
	
	 
	list_Delete(parent->children);
	parent->children = parent->newSubList;
	parent->newSubList = list_Nil();
      }
      result = list_Cons(parent, result); 
    }
  }
  return result;
}


static ord_RESULT rpos_ShortCompareGround(CONTEXT Context1, TERM T1,
					  CONTEXT Context2, TERM T2)
{
  ord_RESULT     result;
  PRIORITY_QUEUE pqueue;  
  RPOS_TERM      rTerm1, rTerm2;
  LIST           newCandidates;
  BOOL           T1finished, T2finished;
  NAT            order;

#ifdef DEBUG
  if ((Context1 == NULL && Context2 != NULL) ||
      (Context1 != NULL && Context2 == NULL)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In rpos_ShortCompareGround: Illegal input.");
    misc_FinishErrorReport();
  }
#endif

  newCandidates = list_Nil();
  T1finished    = T2finished = FALSE;
  order         = 1;  

  
  if (Context1 == NULL) {
    pqueue = pq_Create(term_ComputeSize(T1)+term_ComputeSize(T2), rpos_GroundTermSmaller);
    rTerm1 = rpos_InitComparison(T1, pqueue);
    rTerm2 = rpos_InitComparison(T2, pqueue);
  } else {
    
    pqueue = pq_Create(cont_TermSize(Context1,T1)+cont_TermSize(Context2,T2),
		       rpos_GroundTermSmaller);
    rTerm1 = rpos_ContInitComparison(Context1, T1, pqueue);
    rTerm2 = rpos_ContInitComparison(Context2, T2, pqueue);
  }

  while (!pq_Empty(pqueue) && !T1finished && !T2finished) {
    RPOS_TERM s = pq_ExtractMin(pqueue);

    s->ord = ++order;
    newCandidates = rpos_UpdateParent(s);

    if (s == rTerm1)
      T1finished = TRUE;
    else if (s == rTerm2)
      T2finished = TRUE;

    while (!pq_Empty(pqueue) &&
	   rpos_GroundTermEqual(s, pq_GetMin(pqueue))) {
      
      s = pq_ExtractMin(pqueue);
      s->ord = order;
      newCandidates = list_Nconc(rpos_UpdateParent(s), newCandidates);

      if (s == rTerm1)
	T1finished = TRUE;
      else if (s == rTerm2)
	T2finished = TRUE;
    }

    while (!list_Empty(newCandidates)) {
      pq_Insert(pqueue, list_Car(newCandidates));
      newCandidates = list_Pop(newCandidates);
    }
  }

  result = ord_UNCOMPARABLE;
  if (T1finished) {
    if (T2finished)
      result = ord_EQUAL;
    else
      result = ord_SMALLER_THAN;
  }
  else if (T2finished)
    result = ord_GREATER_THAN;
  else {
    misc_StartErrorReport();
    misc_ErrorReport("\n In rpos_ShortCompareGround: ground terms are uncomparable?\n");
    misc_FinishErrorReport();
  }

  rpos_TermDelete(rTerm1);
  rpos_TermDelete(rTerm2);
  pq_Delete(pqueue);

  return result;
}

static void rpos_FullCompareGround(CONTEXT Context1, TERM T1,
				   CONTEXT Context2, TERM T2,
				   RPOS_TERM* rt1, RPOS_TERM* rt2)
{
  PRIORITY_QUEUE pqueue;  
  RPOS_TERM      rTerm1, rTerm2;
  LIST           newCandidates;
  NAT            order;

#ifdef DEBUG
  if ((Context1 == NULL && Context2 != NULL) &&
      (Context1 != NULL && Context2 == NULL)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In rpos_FullCompareGround: Illegal input.");
    misc_FinishErrorReport();  
  }
#endif

  newCandidates = list_Nil();
  order         = 1;  
 
  
  if (Context1 == NULL) {
    pqueue = pq_Create(term_ComputeSize(T1)+term_ComputeSize(T2), rpos_GroundTermSmaller);
    rTerm1 = rpos_InitComparison(T1, pqueue);
    rTerm2 = rpos_InitComparison(T2, pqueue);
  } else {
    
    pqueue = pq_Create(cont_TermSize(Context1,T1)+cont_TermSize(Context2,T2),
		       rpos_GroundTermSmaller);
    rTerm1 = rpos_ContInitComparison(Context1, T1, pqueue);
    rTerm2 = rpos_ContInitComparison(Context2, T2, pqueue);
  }

  while (!pq_Empty(pqueue)) {
    RPOS_TERM s = pq_ExtractMin(pqueue);

    s->ord = ++order;
    newCandidates = rpos_UpdateParent(s);

    while (!pq_Empty(pqueue) &&
	   rpos_GroundTermEqual(s, pq_GetMin(pqueue))) {
      
      s = pq_ExtractMin(pqueue);
      s->ord = order;
      newCandidates = list_Nconc(rpos_UpdateParent(s), newCandidates);
    }

    while (!list_Empty(newCandidates)) {
      pq_Insert(pqueue, list_Car(newCandidates));
      newCandidates = list_Pop(newCandidates);
    }
  }

  pq_Delete(pqueue);
  rpos_UpdateMUL(rTerm1);
  rpos_UpdateMUL(rTerm2);

  *rt1 = rTerm1;
  *rt2 = rTerm2;
}


static LIST rpos_SortByLength(RPOS_TERM RTerm)
{
  RPOS_TERM actTerm;
  LIST      *A, scan, result;        
  NAT       i, stack, size;

  A = (LIST*) memory_Malloc(sizeof(LIST) * RTerm->size);

  for (i = 0; i < RTerm->size; i++)
    A[i] = list_Nil();
  
  stack = stack_Bottom();
  A[RTerm->size - 1] = list_List(RTerm);

  if (!list_Empty(RTerm->children))    
    stack_Push(RTerm->children);
  if (!list_Empty(RTerm->newSubList))  
    stack_Push(RTerm->newSubList);

  while (!stack_Empty(stack)) {
    scan = stack_Top();
    actTerm = (RPOS_TERM) list_Car(scan);
    stack_RplacTop(list_Cdr(scan));

    size = actTerm->size;
    A[size - 1] = list_Cons(actTerm, A[size - 1]);

    if (!list_Empty(actTerm->children))    
      stack_Push(actTerm->children);
    if (!list_Empty(actTerm->newSubList))  
      stack_Push(actTerm->newSubList);
    
    while (!stack_Empty(stack) && list_Empty(stack_Top()))
      stack_Pop();
  }

  
  
  
  
  
  
  
  
  
  
  
  
  result = list_Nil();
  for (i = RTerm->size; i-- > 0; )
    result = list_Nconc(A[i], result);

  memory_Free(A, sizeof(LIST) * RTerm->size); 
  return result;
}


static ord_RESULT rpos_InternNGCompare(RPOS_TERM RT1, RPOS_TERM RT2)

{
  if (RT1->ord != RPOS_NONGROUND && RT2->ord != RPOS_NONGROUND) {
    
    if (RT1->ord > RT2->ord)
      return ord_GREATER_THAN;
    else if (RT1->ord < RT2->ord)
      return ord_SMALLER_THAN;
    else
      return ord_EQUAL;    
  }

  if (list_PointerMember(RT1->smaller, RT2))
    return ord_GREATER_THAN;
  if (list_PointerMember(RT2->smaller, RT1))
    return ord_SMALLER_THAN;
  if (list_PointerMember(RT1->equal, RT2))
    return ord_EQUAL;
  return ord_UNCOMPARABLE;
}


static ord_RESULT rpos_NGCompare(RPOS_TERM RT1, RPOS_TERM RT2)
{
  RPOS_TERM rTerm1, rTerm2, parent;
  LIST      list1, list2, scan1, scan2, argScan1, argScan2;
  BOOL      is_greater, is_smaller, equal;

  
  
  

  list1 = rpos_SortByLength(RT1);
  list2 = rpos_SortByLength(RT2);
  
  
  

  
  
  
  
  for (scan1 = list1; !list_Empty(scan1); scan1 = list_Cdr(scan1)) {
    rTerm1 = (RPOS_TERM) list_Car(scan1);
    if (rTerm1->size > 1)
      break;
    if (!term_IsStandardVariable(rTerm1->term))
      continue;

    for (scan2 = list2; !list_Empty(scan2); scan2 = list_Cdr(scan2)) {
      rTerm2 = (RPOS_TERM) list_Car(scan2);
      if (rTerm2->size > 1)
	break;
      if (term_EqualTopSymbols(rTerm1->term, rTerm2->term)) {
	
	rTerm1->equal = list_Cons(rTerm2, rTerm1->equal);
	rTerm2->equal = list_Cons(rTerm1, rTerm2->equal);
	for (parent = rTerm1->parent; parent != NULL; parent = parent->parent)
	  parent->smaller = list_Cons(rTerm2, parent->smaller);
	for (parent = rTerm2->parent; parent != NULL; parent = parent->parent)
	  parent->smaller = list_Cons(rTerm1, parent->smaller);
      }
    }
  }

  
  for (scan1 = list1; !list_Empty(scan1); scan1 = list_Cdr(scan1)) {
    rTerm1 = (RPOS_TERM) list_Car(scan1);
    if (term_IsStandardVariable(rTerm1->term))
      continue;

    for (scan2 = list2; !list_Empty(scan2); scan2 = list_Cdr(scan2)) {
      LIST children1, children2, ngChildren1, ngChildren2;

      rTerm2 = (RPOS_TERM) list_Car(scan2);

      if (term_IsStandardVariable(rTerm2->term) ||
	  (rTerm1->ord != RPOS_NONGROUND && rTerm2->ord != RPOS_NONGROUND))
	
	
	continue;

      if (rTerm2->visited) {
	rTerm2->visited = FALSE;
	continue;
      }

      
      
      
      
      is_greater = FALSE;
      is_smaller = TRUE;
      for (argScan1 = rTerm1->children;
	   !list_Empty(argScan1) && !is_greater;
	   argScan1 = list_Cdr(argScan1)) {
	switch (rpos_InternNGCompare(list_Car(argScan1), rTerm2)) {
	case ord_EQUAL:
	case ord_GREATER_THAN:
	  is_greater = TRUE;
	case ord_UNCOMPARABLE:
	  is_smaller = FALSE;
	  break;
	default:
	  break;
	}
      }
      for (argScan1 = rTerm1->newSubList;
	   !list_Empty(argScan1) && !is_greater;
	   argScan1 = list_Cdr(argScan1)) {
	switch (rpos_InternNGCompare(list_Car(argScan1), rTerm2)) {
	case ord_EQUAL:
	case ord_GREATER_THAN:
	  is_greater = TRUE;
	case ord_UNCOMPARABLE:
	  is_smaller = FALSE;
	  break;
	default:
	  break;
	}
      }
      if (is_greater) {
	
	rTerm1->smaller = list_Cons(rTerm2, rTerm1->smaller);
	
	
	continue;
      }
      if (is_smaller &&
	  symbol_PrecedenceGreater(term_TopSymbol(rTerm2->term),
				   term_TopSymbol(rTerm1->term))) {
	
	
	rTerm2->smaller = list_Cons(rTerm1, rTerm2->smaller);
	
	
	for (parent = rTerm2->parent; parent != NULL; parent = parent->parent) {
	  parent->smaller = list_Cons(rTerm1, parent->smaller);
	  parent->visited = TRUE;
	}
	continue;
      }
      
      
      
      
      
      is_greater = TRUE;
      is_smaller = FALSE;
	
      for (argScan2 = rTerm2->children;
	   !list_Empty(argScan2) && !is_smaller;
	   argScan2 = list_Cdr(argScan2)) {
 	switch (rpos_InternNGCompare(rTerm1, list_Car(argScan2))) {
	case ord_EQUAL:
	case ord_SMALLER_THAN:
	  is_smaller = TRUE;
	case ord_UNCOMPARABLE:
	  is_greater = FALSE;
	  break;
 	default:
	  break;
	}
      }
      for (argScan2 = rTerm2->newSubList;
	   !list_Empty(argScan2) && !is_smaller;
	   argScan2 = list_Cdr(argScan2)) {
	switch (rpos_InternNGCompare(rTerm1, list_Car(argScan2))) {
	case ord_EQUAL:
	case ord_SMALLER_THAN:
	  is_smaller = TRUE;
	case ord_UNCOMPARABLE:
	  is_greater = FALSE;
	  break;
	default:
	  break;
	}
      }
      if (is_smaller) {
	
	rTerm2->smaller = list_Cons(rTerm1, rTerm2->smaller);
	
	
	for (parent = rTerm2->parent; parent != NULL; parent = parent->parent) {
	  parent->smaller = list_Cons(rTerm1, parent->smaller);
	  parent->visited = TRUE;
	}
	continue; 
      }
      if (is_greater &&
	  symbol_PrecedenceGreater(term_TopSymbol(rTerm1->term),
		 		   term_TopSymbol(rTerm2->term))) {
	
	
	rTerm1->smaller = list_Cons(rTerm2, rTerm1->smaller);
	continue;
      }


      if (!term_EqualTopSymbols(rTerm1->term, rTerm2->term))
	
	
	continue;

      
      
      if (!symbol_HasProperty(term_TopSymbol(rTerm1->term),ORDMUL)) {
	is_greater = FALSE;
	is_smaller = FALSE;
	equal      = TRUE;

	
	for (argScan1 = rTerm1->children, argScan2 = rTerm2->children;
	     !list_Empty(argScan1) && equal;
	     argScan1 = list_Cdr(argScan1), argScan2 = list_Cdr(argScan2)) {
	  switch (rpos_InternNGCompare(list_Car(argScan1), list_Car(argScan2))) {
	  case ord_UNCOMPARABLE:
	    equal = FALSE;      
	    break;
	  case ord_SMALLER_THAN:
	    is_smaller = TRUE;
	    equal      = FALSE;
	    break;
	  case ord_GREATER_THAN:
	    is_greater = TRUE;
	    equal      = FALSE;
	    break;
	  default:
	    break;
	  }
	}
	  
	if (is_greater) {
	  
	  for ( ; !list_Empty(argScan2) && is_greater; argScan2 = list_Cdr(argScan2)) {
	    if (rpos_InternNGCompare(rTerm1, list_Car(argScan2)) != ord_GREATER_THAN)
	      is_greater = FALSE; 
	    
	    
	  }
	}
	else if (is_smaller) {
	  
	  for ( ; !list_Empty(argScan1) && is_smaller; argScan1 = list_Cdr(argScan1)) {
	    if (rpos_InternNGCompare(list_Car(argScan1), rTerm2) != ord_SMALLER_THAN)
	      is_smaller = FALSE;
	    
	    
	  }
	}
	if (is_greater)
	  rTerm1->smaller = list_Cons(rTerm2, rTerm1->smaller);
	else if (is_smaller)
	  rTerm2->smaller = list_Cons(rTerm1, rTerm2->smaller);
	else if (equal) {
 	  rTerm1->equal = list_Cons(rTerm2, rTerm1->equal);
	  rTerm2->equal = list_Cons(rTerm1, rTerm2->equal);
	}
	if (is_smaller || equal) {
	  
	  
	  for (parent = rTerm2->parent; parent != NULL; parent = parent->parent) {
	    parent->smaller = list_Cons(rTerm1, parent->smaller);
	    parent->visited = TRUE;
	  }
	}
	  
	continue;
      }


      
      equal     = TRUE;
      children1 = children2 = list_Nil();
      
       
      for (argScan1 = rTerm1->children, argScan2 = rTerm2->children;
	   !list_Empty(argScan1) && equal;
	   argScan1 = list_Cdr(argScan1), argScan2 = list_Cdr(argScan2)) {
	if (((RPOS_TERM) list_Car(argScan1))->ord >
	    ((RPOS_TERM) list_Car(argScan2))->ord) {
	  equal     = FALSE;
	  children1 = list_Cons(list_Car(argScan1), children1);
	}
	else if (((RPOS_TERM) list_Car(argScan1))->ord <
		 ((RPOS_TERM) list_Car(argScan2))->ord) {
	  equal     = FALSE;
	  children2 = list_Cons(list_Car(argScan2), children2);
	}
      }
      
      ngChildren1 = list_Copy(rTerm1->newSubList);
      ngChildren2 = list_Copy(rTerm2->newSubList);
      for (argScan1 = ngChildren1; !list_Empty(argScan1); argScan1 = list_Cdr(argScan1))
	for (argScan2 = ngChildren2; !list_Empty(argScan2); argScan2 = list_Cdr(argScan2))
	  if (list_PointerMember(((RPOS_TERM)list_Car(argScan1))->equal,
				 list_Car(argScan2))) {
	    list_Rplaca(argScan1, NULL);
	    list_Rplaca(argScan2, NULL);
	    break;
	  }
      ngChildren1 = list_PointerDeleteElement(ngChildren1, NULL);
      ngChildren2 = list_PointerDeleteElement(ngChildren2, NULL);
      
      if (equal && list_Empty(ngChildren1)) {
	
	rTerm1->equal = list_Cons(rTerm2, rTerm1->equal);
	rTerm2->equal = list_Cons(rTerm1, rTerm2->equal);
	
	
	for (parent = rTerm2->parent; parent != NULL; parent = parent->parent) {
	  parent->smaller = list_Cons(rTerm1, parent->smaller);
	  parent->visited = TRUE;
	}
	continue;
      }
	
      children1 = list_Nconc(children1, ngChildren1);
      children2 = list_Nconc(children2, ngChildren2);
      
      

      
      ngChildren2 = list_Copy(children2);
      for (argScan1 = children1; !list_Empty(argScan1); argScan1 = list_Cdr(argScan1)) {
	for (argScan2 = ngChildren2; !list_Empty(argScan2);
	     argScan2 = list_Cdr(argScan2)) {
	  if (rpos_InternNGCompare(list_Car(argScan1), list_Car(argScan2))
	      == ord_GREATER_THAN) {
	    list_Rplaca(argScan2, NULL);
	  }
	}
	ngChildren2 = list_PointerDeleteElement(ngChildren2, NULL);
	if (list_Empty(ngChildren2))
	  break;
      }
      if (list_Empty(ngChildren2)) {
	
	rTerm1->smaller = list_Cons(rTerm2, rTerm1->smaller);
	list_Delete(children1);
	list_Delete(children2);
	continue;
      }
      list_Delete(ngChildren2);   

      
      ngChildren1 = list_Copy(children1);
      for (argScan2 = children2; !list_Empty(argScan2); argScan2 = list_Cdr(argScan2)) {
	for (argScan1 = ngChildren1; !list_Empty(argScan1);
	     argScan1 = list_Cdr(argScan1)) {
	  if (rpos_InternNGCompare(list_Car(argScan2), list_Car(argScan1))
	      == ord_GREATER_THAN) {
	    list_Rplaca(argScan1, NULL);
	  }
	}
	ngChildren1 = list_PointerDeleteElement(ngChildren1, NULL);
	if (list_Empty(ngChildren1))
	  break;
      }
      if (list_Empty(ngChildren1)) {
	
	rTerm2->smaller = list_Cons(rTerm1, rTerm2->smaller);
	
	
	for (parent = rTerm2->parent; parent != NULL; parent = parent->parent) {
	  parent->smaller = list_Cons(rTerm1, parent->smaller);
	  parent->visited = TRUE;
	}
      } else
	list_Delete(ngChildren1);
      list_Delete(children1);
      list_Delete(children2);
    }
  }

  

  list_Delete(list1);
  list_Delete(list2);

  return rpos_InternNGCompare(RT1, RT2);
}


ord_RESULT rpos_BorromUpCompare(TERM T1, TERM T2)

{
  BOOL       varCond1, varCond2, ground1, ground2;
  ord_RESULT result;

  
  
  

  if (term_IsVariable(T1)) {
    if (term_EqualTopSymbols(T1, T2))
      result = ord_EQUAL;
    else if (term_ContainsSymbol(T2, term_TopSymbol(T1)))
      result = ord_SMALLER_THAN;
    else
      result = ord_UNCOMPARABLE;

    
    
    
    return result;
  }

  if (term_IsVariable(T2)) {
    if (term_EqualTopSymbols(T1, T2))
      result = ord_EQUAL;
    else if (term_ContainsSymbol(T1, term_TopSymbol(T2)))
      result = ord_GREATER_THAN;
    else
      result = ord_UNCOMPARABLE;
    
    
    
    
    return result;
  }

  rpos_VarCondition(T1, &varCond1, &ground1, T2, &varCond2, &ground2);

  if (ground1 && ground2)
    
    result = rpos_ShortCompareGround(NULL, T1, NULL, T2); 
  else {
        
    if (varCond1 || varCond2) {
      RPOS_TERM  rt1, rt2;

      rpos_FullCompareGround(NULL, T1, NULL, T2, &rt1, &rt2); 
      result = rpos_NGCompare(rt1, rt2);
      
      rpos_TermDelete(rt1);
      rpos_TermDelete(rt2);
    } else
      result = ord_UNCOMPARABLE;
    
    
    
  }
  
  
  
  
  
  return result;
}


static void rpos_ContCountVariables(CONTEXT Context, TERM Term, BOOL *Ground, NAT Index)

{
  LIST scan;

#ifdef DEBUG
  if (!term_IsTerm(Term) || Index > 1) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In rpos_ContCountVariables: Illegal input.");
    misc_FinishErrorReport();
  }
#endif

  Term = cont_Deref(&Context, Term);

  if (term_IsStandardVariable(Term)) {
    ord_VARCOUNT[term_TopSymbol(Term)][Index]++;
    *Ground = FALSE;
  }
  else if (term_IsComplex(Term)) {
    for (scan = term_ArgumentList(Term); !list_Empty(scan); scan = list_Cdr(scan))
      rpos_ContCountVariables(Context, list_Car(scan), Ground, Index);
  }
}

static void rpos_ContVarCondition(CONTEXT Context1, TERM Term1,
				  BOOL *VarCond1, BOOL *Ground1,
				  CONTEXT Context2, TERM Term2,
				  BOOL *VarCond2, BOOL *Ground2)

{
  SYMBOL maxVar1, maxVar2;
  NAT    i;

  *VarCond1 = *VarCond2 = TRUE;
  maxVar1   = cont_TermMaxVar(Context1, Term1);
  maxVar2   = cont_TermMaxVar(Context2, Term2);

  if (symbol_GreaterVariable(maxVar2, maxVar1))
    maxVar1 = maxVar2;

  for (i = 0; i <= maxVar1; i++) {
    ord_VARCOUNT[i][0] = 0;
    ord_VARCOUNT[i][1] = 0;
  }

  *Ground1 = *Ground2 = TRUE;
  rpos_ContCountVariables(Context1, Term1, Ground1, 0);
  rpos_ContCountVariables(Context2, Term2, Ground2, 1);

  for (i = 0; i <= maxVar1; i++) {
    if (ord_VARCOUNT[i][0] < ord_VARCOUNT[i][1]) {
      *VarCond1 = FALSE;
      if (!*VarCond2)  
	return;
    }
    else if (ord_VARCOUNT[i][0] > ord_VARCOUNT[i][1]) {
      *VarCond2 = FALSE;
      if (!*VarCond1)
	return;
    }
  }
}


ord_RESULT rpos_BottomUpContCompare(CONTEXT Context1, TERM T1,
			    CONTEXT Context2, TERM T2)

{
  BOOL       varCond1, varCond2, ground1, ground2;
  ord_RESULT result;

  

  T1 = cont_Deref(&Context1, T1);
  T2 = cont_Deref(&Context2, T2);

  if (term_IsVariable(T1)) {
    if (term_EqualTopSymbols(T1, T2))
      result = ord_EQUAL;
    else if (cont_TermContainsSymbol(Context2, T2, term_TopSymbol(T1)))
      result = ord_SMALLER_THAN;
    else
      result = ord_UNCOMPARABLE;
  
    
    
    
    return result;
  }
  
  if (term_IsVariable(T2)) {
    if (term_EqualTopSymbols(T1, T2))
      result = ord_EQUAL;
    else if (cont_TermContainsSymbol(Context1, T1, term_TopSymbol(T2)))
      result = ord_GREATER_THAN;
    else
      result = ord_UNCOMPARABLE;
    
    
    
    
    return result;
  }

  rpos_ContVarCondition(Context1, T1, &varCond1, &ground1,
			Context2, T2, &varCond2, &ground2);

  if (ground1 && ground2)
    
    result = rpos_ShortCompareGround(Context1, T1, Context2, T2);
  else {
        
    if (varCond1 || varCond2) {
      RPOS_TERM rt1, rt2;

      rpos_FullCompareGround(Context1, T1, Context2, T2, &rt1, &rt2);
      result = rpos_NGCompare(rt1, rt2);

      rpos_TermDelete(rt1);
      rpos_TermDelete(rt2);
    } else {
      result = ord_UNCOMPARABLE;
      
      
      
    }
  }

  
  
  
  
  return result;
}
  
BOOL rpos_BottomUpContGreater(CONTEXT Context1, TERM T1, CONTEXT Context2, TERM T2)

{
  BOOL       varCond1, varCond2, ground1, ground2;
  ord_RESULT result;

  

  T1 = cont_Deref(&Context1, T1);
  T2 = cont_Deref(&Context2, T2);

  if (term_IsVariable(T1)) {
    

    
    return FALSE;
  }
  
  if (term_IsVariable(T2)) {
    if (cont_TermContainsSymbol(Context1, T1, term_TopSymbol(T2)))
      result = TRUE;
    else
      result = FALSE;  
    
    
    return result;
  }

  rpos_ContVarCondition(Context1, T1, &varCond1, &ground1,
			Context2, T2, &varCond2, &ground2);

  if (!varCond1) {
    
    

    
    return FALSE;
  }

  result = FALSE;
  if (ground1 && ground2) {
    
    if (ord_IsGreaterThan(rpos_ShortCompareGround(Context1, T1, Context2, T2)))
      result = TRUE;
  } else {
    
    RPOS_TERM rt1, rt2;
    
    rpos_FullCompareGround(Context1, T1, Context2, T2, &rt1, &rt2);
    if (ord_IsGreaterThan(rpos_NGCompare(rt1, rt2)))
      result = TRUE;
    
    rpos_TermDelete(rt1);
    rpos_TermDelete(rt2);
  }

  
  
  return result;
}

void rpos_TermPrint(RPOS_TERM RTerm)
{
  LIST scan;

  symbol_Print(term_TopSymbol(RTerm->term));
  if (!list_Empty(RTerm->children) || !list_Empty(RTerm->newSubList)) {
    putchar('(');
    for (scan = RTerm->children; !list_Empty(scan); scan = list_Cdr(scan)) {
      rpos_TermPrint(list_Car(scan));
      if (!list_Empty(list_Cdr(scan)))
	putchar(',');
    }
    if (!list_Empty(RTerm->newSubList)) {
      putchar(':');
      for (scan = RTerm->newSubList; !list_Empty(scan); scan = list_Cdr(scan)) {
	rpos_TermPrint(list_Car(scan));
	if (!list_Empty(list_Cdr(scan)))
	  putchar(',');
      }
    }
    putchar(')');
  }
}


void rpos_ListPrint(LIST NGList)
{
  LIST scan;

  for ( ; !list_Empty(NGList); NGList = list_Cdr(NGList)) {
    RPOS_TERM rTerm;
    rTerm = (RPOS_TERM)list_Car(NGList);
    putchar('\n');
    rpos_TermPrint(rTerm);
    if (!list_Empty(rTerm->equal)) {
      fputs(" == ", stdout);
      for (scan = rTerm->equal; !list_Empty(scan); scan = list_Cdr(scan)) {
	rpos_TermPrint(list_Car(scan));
	putchar(' ');
      }
    }
    fputs(" > ", stdout);
    for (scan = rTerm->smaller; !list_Empty(scan); scan = list_Cdr(scan)) {
      rpos_TermPrint(list_Car(scan));
      putchar(' ');
    }
  }
  fflush(stdout);
}





static LIST rpos_MultisetDifference(TERM T1, TERM T2)

{
  LIST result;

  result = list_Copy(term_ArgumentList(T1));
  result = list_NMultisetDifference(result, term_ArgumentList(T2), rpos_Equal);
  return result;
}
  
  
static ord_RESULT rpos_MulGreaterEqual(TERM T1, TERM T2)

{
  LIST l1, l2;

  l1 = rpos_MultisetDifference(T1, T2);
  if (list_Empty(l1))
     
    return ord_Equal();   
  else {
    LIST scan;
    BOOL greater;

    l2 = rpos_MultisetDifference(T2, T1);

    for (greater = TRUE; !list_Empty(l2) && greater; l2 = list_Pop(l2)) {
      for (scan = l1, greater = FALSE; !list_Empty(scan) && !greater; scan = list_Cdr(scan))
	greater = rpos_Greater(list_Car(scan), list_Car(l2));
    }
    list_Delete(l1); 
    if (greater)
      return ord_GreaterThan();
    else
      return ord_Uncomparable();
  }
}

static ord_RESULT rpos_LexGreaterEqual(TERM T1, TERM T2)

{
  ord_RESULT result;
  LIST       l1, l2, scan1, scan2;

  if (symbol_HasProperty(term_TopSymbol(T1), ORDRIGHT)) {
    l1 = list_Reverse(term_ArgumentList(T1)); 
    l2 = list_Reverse(term_ArgumentList(T2));
  } else {
    l1 = term_ArgumentList(T1);
    l2 = term_ArgumentList(T2);
  }
  
  result = ord_Equal();
  for (scan1 = l1, scan2 = l2; !list_Empty(scan1);
       scan1 = list_Cdr(scan1), scan2 = list_Cdr(scan2)) {
    result = rpos_GreaterEqual(list_Car(scan1), list_Car(scan2));
    if (!ord_IsEqual(result))
      break;
  }

  if (ord_IsEqual(result))  
    ;            
  else if (ord_IsGreaterThan(result)) {
    
    for (scan2 = list_Cdr(scan2); !list_Empty(scan2) && rpos_Greater(T1, list_Car(scan2));
	 scan2 = list_Cdr(scan2)); 
    if (list_Empty(scan2))
      result = ord_GreaterThan();
    else
      result = ord_Uncomparable();
  }
  else {
    

    
    for (scan1 = list_Cdr(scan1), result = ord_Uncomparable();
	 !list_Empty(scan1) && !ord_IsGreaterThan(result);
	 scan1 = list_Cdr(scan1)) {
      if (!ord_IsUncomparable(rpos_GreaterEqual(list_Car(scan1), T2)))
	result = ord_GreaterThan();
    }
  }

  if (symbol_HasProperty(term_TopSymbol(T1), ORDRIGHT)) {
    list_Delete(l1);  
    list_Delete(l2);
  }
  return result;
}


BOOL rpos_Equal(TERM T1, TERM T2)

{
  LIST l1, l2;

  if (!term_EqualTopSymbols(T1, T2))
    return FALSE;
  else if (!term_IsComplex(T1))  
    return TRUE;
  else {
    if (symbol_HasProperty(term_TopSymbol(T1), ORDMUL)) {  
      l1 = rpos_MultisetDifference(T1, T2);
      if (list_Empty(l1))
	return TRUE;
      else {
	list_Delete(l1);
	return FALSE;
      }
    } else {   
      for (l1 = term_ArgumentList(T1), l2 = term_ArgumentList(T2);
	   !list_Empty(l1) &&  rpos_Equal(list_Car(l1), list_Car(l2));
	   l1 = list_Cdr(l1), l2 = list_Cdr(l2)); 
      return list_Empty(l1);  
    }
  }
}


ord_RESULT rpos_GreaterEqual(TERM T1, TERM T2)

{
  LIST scan;

  if (term_IsVariable(T1)) {
    if (term_EqualTopSymbols(T1, T2))
      return ord_Equal();   
    else
      
      return ord_Uncomparable();
  } else if (term_IsVariable(T2)) {   
    if (term_ContainsSymbol(T1, term_TopSymbol(T2)))
      return ord_GreaterThan();
    else
      return ord_Uncomparable();
  } else if (term_EqualTopSymbols(T1, T2)) {
    if (symbol_HasProperty(term_TopSymbol(T1), ORDMUL))
      return rpos_MulGreaterEqual(T1, T2);
    else
      return rpos_LexGreaterEqual(T1, T2);
  } else {
    if (symbol_PrecedenceGreater(term_TopSymbol(T1), term_TopSymbol(T2))) {
      
      
      for (scan = term_ArgumentList(T2); !list_Empty(scan); scan = list_Cdr(scan))
	if (!rpos_Greater(T1,  list_Car(scan)))
	  return ord_Uncomparable();
      return ord_GreaterThan();
    } else {
      
      for (scan = term_ArgumentList(T1); !list_Empty(scan); scan = list_Cdr(scan))
	if (!ord_IsUncomparable(rpos_GreaterEqual(list_Car(scan), T2)))
	  return ord_GreaterThan();    
      return ord_Uncomparable();
    }
  }
}

ord_RESULT rpos_Compare(TERM T1, TERM T2)

{
  ord_RESULT result;

  result = rpos_GreaterEqual(T1, T2);
  if (!ord_IsUncomparable(result))
    return result;
  else if (rpos_Greater(T2, T1))
    return ord_SmallerThan();
  else
    return ord_UNCOMPARABLE;
}





static LIST rpos_ContMultisetDifference(CONTEXT C1, TERM T1, CONTEXT C2, TERM T2)

{
  LIST result, scan1, scan2;

  
  

  
  
  
  
  result = list_Copy(term_ArgumentList(T1));
  for (scan2 = term_ArgumentList(T2); !list_Empty(scan2);
       scan2 = list_Cdr(scan2)) {
    
    
    for (scan1 = result; !list_Empty(scan1); scan1 = list_Cdr(scan1)) {
      if (list_Car(scan1) != NULL &&
	  rpos_ContEqual(C1, list_Car(scan1), C2, list_Car(scan2))) {
	
	list_Rplaca(scan1, NULL);  
	break;
      }
    }
  }
  return list_PointerDeleteElement(result, NULL); 
}
  
  
static ord_RESULT rpos_ContMulGreaterEqual(CONTEXT C1, TERM T1,
					   CONTEXT C2, TERM T2)

{
  LIST l1, l2;

  
  

  l1 = rpos_ContMultisetDifference(C1, T1, C2, T2);
  if (list_Empty(l1))
     
    return ord_Equal();   
  else {
    LIST scan;
    BOOL greater;

    l2 = rpos_ContMultisetDifference(C2, T2, C1, T1);

    for (greater = TRUE; !list_Empty(l2) && greater; l2 = list_Pop(l2)) {
      for (scan = l1, greater = FALSE; !list_Empty(scan) && !greater;
	   scan = list_Cdr(scan))
	greater = rpos_ContGreater(C1, list_Car(scan), C2, list_Car(l2));
    }
    list_Delete(l1); 
    if (greater)
      return ord_GreaterThan();
    else
      return ord_Uncomparable();
  }
}

static ord_RESULT rpos_ContLexGreaterEqual(CONTEXT C1, TERM T1,
					   CONTEXT C2, TERM T2)

{
  ord_RESULT result;
  LIST       l1, l2, scan1, scan2;

  
  

  if (symbol_HasProperty(term_TopSymbol(T1), ORDRIGHT)) {
    l1 = list_Reverse(term_ArgumentList(T1)); 
    l2 = list_Reverse(term_ArgumentList(T2));
  } else {
    l1 = term_ArgumentList(T1);
    l2 = term_ArgumentList(T2);
  }
  
  result = ord_Equal();
  for (scan1 = l1, scan2 = l2; !list_Empty(scan1);
       scan1 = list_Cdr(scan1), scan2 = list_Cdr(scan2)) {
    result = rpos_ContGreaterEqual(C1, list_Car(scan1), C2, list_Car(scan2));
    if (!ord_IsEqual(result))
      break;
  }

  if (ord_IsEqual(result))  
    ;            
  else if (ord_IsGreaterThan(result)) {
    
    for (scan2 = list_Cdr(scan2);
	 !list_Empty(scan2) && rpos_ContGreater(C1, T1, C2, list_Car(scan2));
	 scan2 = list_Cdr(scan2)); 
    if (list_Empty(scan2))
      result = ord_GreaterThan();
    else
      result = ord_Uncomparable();
  }
  else {
    
    
    for (scan1 = list_Cdr(scan1), result = ord_Uncomparable();
	 !list_Empty(scan1) && !ord_IsGreaterThan(result);
	 scan1 = list_Cdr(scan1)) {
      if (!ord_IsUncomparable(rpos_ContGreaterEqual(C1,list_Car(scan1),C2,T2)))
	result = ord_GreaterThan();
    }
  }

  if (symbol_HasProperty(term_TopSymbol(T1), ORDRIGHT)) {
    list_Delete(l1);  
    list_Delete(l2);
  }
  return result;
}


BOOL rpos_ContEqual(CONTEXT C1, TERM T1, CONTEXT C2, TERM T2)

{
  LIST l1, l2;

  T1 = cont_Deref(&C1, T1);
  T2 = cont_Deref(&C2, T2);

  if (!term_EqualTopSymbols(T1, T2))
    return FALSE;
  else if (!term_IsComplex(T1))
    return TRUE;
  else {
    if (symbol_HasProperty(term_TopSymbol(T1), ORDMUL)) {
      l1 = rpos_ContMultisetDifference(C1, T1, C2, T2);
      if (list_Empty(l1))
	return TRUE;
      else {
	list_Delete(l1);
	return FALSE;
      }
    } else {   
      for (l1 = term_ArgumentList(T1), l2 = term_ArgumentList(T2);
	   !list_Empty(l1) &&  rpos_ContEqual(C1,list_Car(l1),C2,list_Car(l2));
	   l1 = list_Cdr(l1), l2 = list_Cdr(l2)); 
      return list_Empty(l1);  
    }
  }
}


ord_RESULT rpos_ContGreaterEqual(CONTEXT C1, TERM T1, CONTEXT C2, TERM T2)

{
  LIST scan;

  T1 = cont_Deref(&C1, T1);
  T2 = cont_Deref(&C2, T2);

  if (term_IsVariable(T1)) {
    if (term_EqualTopSymbols(T1, T2))
      return ord_Equal();   
    else
      
      return ord_Uncomparable();
  } else if (term_IsVariable(T2)) {   
    if (cont_TermContainsSymbol(C1, T1, term_TopSymbol(T2)))
      return ord_GreaterThan();
    else
      return ord_Uncomparable();
  } else if (term_EqualTopSymbols(T1, T2)) {
    if (symbol_HasProperty(term_TopSymbol(T1), ORDMUL))
      return rpos_ContMulGreaterEqual(C1, T1, C2, T2);
    else
      return rpos_ContLexGreaterEqual(C1, T1, C2, T2);
  } else {
    if (symbol_PrecedenceGreater(term_TopSymbol(T1), term_TopSymbol(T2))) {
      
      
      for (scan = term_ArgumentList(T2); !list_Empty(scan); scan = list_Cdr(scan))
	if (!rpos_ContGreater(C1, T1, C2, list_Car(scan)))
	  return ord_Uncomparable();
      return ord_GreaterThan();
    } else {
      
      for (scan = term_ArgumentList(T1); !list_Empty(scan); scan = list_Cdr(scan))
	if (!ord_IsUncomparable(rpos_ContGreaterEqual(C1,list_Car(scan),C2,T2)))
	  return ord_GreaterThan();    
      return ord_Uncomparable();
    }
  }
}

ord_RESULT rpos_ContCompare(CONTEXT C1, TERM T1, CONTEXT C2, TERM T2)

{
  ord_RESULT result;

  T1 = cont_Deref(&C1, T1);
  T2 = cont_Deref(&C2, T2);

  result = rpos_ContGreaterEqual(C1, T1, C2, T2);
  if (!ord_IsUncomparable(result))
    return result;
  else if (rpos_ContGreater(C2, T2, C1, T1))
    return ord_SmallerThan();
  else
    return ord_UNCOMPARABLE;
}
  
