




 





















#include "rules-red.h"










static NAT red_STAMPID;
 
const NAT red_USABLE    = 1;
const NAT red_WORKEDOFF = 2;
const NAT red_ALL       = 3;














static void red_HandleRedundantIndexedClauses(PROOFSEARCH Search, LIST Blocked, 
					      CLAUSE RedClause, BOOL Document)

{
  CLAUSE Clause;
  LIST   Scan;
  for(Scan = Blocked; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    if (prfs_SplitLevelCondition(clause_SplitLevel(RedClause),clause_SplitLevel(Clause),
				 prfs_LastBacktrackLevel(Search)))
      split_DeleteClauseAtLevel(Search, Clause, clause_SplitLevel(RedClause));
    else {
      if (clause_GetFlag(Clause, WORKEDOFF)) {
	if (Document)
	  prfs_MoveWorkedOffDocProof(Search, Clause);
	else
	  prfs_DeleteWorkedOff(Search, Clause);
      }
      else
	if (Document)
	  prfs_MoveUsableDocProof(Search, Clause);
	else
	  prfs_DeleteUsable(Search, Clause);
    }
  }
}

static void red_HandleRedundantDerivedClauses(PROOFSEARCH Search, LIST Blocked, 
					      CLAUSE RedClause, BOOL Document)

{
  CLAUSE Clause;
  LIST   Scan;
  for(Scan = Blocked; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    if (prfs_SplitLevelCondition(clause_SplitLevel(RedClause),clause_SplitLevel(Clause),
				 prfs_LastBacktrackLevel(Search))) {
      split_KeepClauseAtLevel(Search, Clause, clause_SplitLevel(RedClause));
    }
    else {
	if (Document)
	  prfs_InsertDocProofClause(Search, Clause);
	else
	  clause_Delete(Clause);
    }
  }
}

void red_Init()

{
  red_STAMPID = term_GetStampID();
}

void red_DocumentObviousReductions(CLAUSE Clause, LIST Indexes)

{

  list_Delete(clause_ParentClauses(Clause));
  list_Delete(clause_ParentLiterals(Clause));
  clause_SetParentClauses(Clause, list_Nil());
  clause_SetParentLiterals(Clause, Indexes);

  clause_AddParentClause(Clause,clause_Number(Clause));  

  clause_SetNumber(Clause, clause_IncreaseCounter());
  clause_SetFromObviousReductions(Clause);
}

static BOOL red_ObviousReductions(CLAUSE Clause, CLAUSE *Changed, BOOL Document)

{
  int     i,j,end;
  LIST    Indexes;
  TERM    Atom, PartnerAtom;

#ifdef DEBUG
  clause_Check(Clause);
#endif

  Indexes = list_Nil();
  end     = clause_LastAntecedentLitIndex(Clause);
    
  for (i = clause_FirstConstraintLitIndex(Clause); i <= end; i++) {
    Atom = clause_LiteralAtom(clause_GetLiteral(Clause,i));
    if (fol_IsEquality(Atom) &&
	term_Equal(term_FirstArgument(Atom),term_SecondArgument(Atom))) {
      Indexes = list_Cons((POINTER)i,Indexes);
    } 
    else
      for (j = i+1; j <= end; j++) {
	PartnerAtom = clause_LiteralAtom(clause_GetLiteral(Clause,j));
	if (term_Equal(PartnerAtom, Atom)
	    || (fol_IsEquality(Atom) 
		&& fol_IsEquality(PartnerAtom)
		&& term_Equal(term_FirstArgument(Atom),term_SecondArgument(PartnerAtom))
		&& term_Equal(term_FirstArgument(PartnerAtom),term_SecondArgument(Atom)))) {
	  Indexes = list_Cons((POINTER)i,Indexes);
	  j       = end;
	}
      }
  }

  end    = clause_LastSuccedentLitIndex(Clause);
    
  for (i = clause_FirstSuccedentLitIndex(Clause); i <= end; i++) {
    Atom = clause_LiteralAtom(clause_GetLiteral(Clause,i));
    for (j = i+1; j <= end; j++) {
      PartnerAtom = clause_LiteralAtom(clause_GetLiteral(Clause,j));
      if (term_Equal(PartnerAtom,Atom)
	    || (fol_IsEquality(Atom) 
		&& fol_IsEquality(PartnerAtom)
		&& term_Equal(term_FirstArgument(Atom),term_SecondArgument(PartnerAtom))
		&& term_Equal(term_FirstArgument(PartnerAtom),term_SecondArgument(Atom)))) {
	Indexes = list_Cons((POINTER)i,Indexes);
	j       = end;
      }
    }
  }

  if (clause_Length(Clause) == 1 &&
      clause_NumOfAnteLits(Clause) == 1 &&
      !list_PointerMember(Indexes,(POINTER)clause_FirstAntecedentLitIndex(Clause)) &&
      fol_IsEquality(clause_LiteralAtom(clause_GetLiteral(Clause,clause_FirstAntecedentLitIndex(Clause))))) {
    cont_StartBinding();
    if (unify_UnifyCom(cont_LeftContext(),
		    term_FirstArgument(clause_LiteralAtom(clause_GetLiteral(Clause,clause_FirstAntecedentLitIndex(Clause)))),
		    term_SecondArgument(clause_LiteralAtom(clause_GetLiteral(Clause,clause_FirstAntecedentLitIndex(Clause))))))
      Indexes = list_Cons((POINTER)clause_FirstAntecedentLitIndex(Clause),Indexes);
    cont_BackTrack();
  }

  if (!list_Empty(Indexes)) {
    if (flag_GetFlagValue(flag_POBV)) {
      printf("\nObvious: ");
      clause_Print(Clause);
      printf(" ==> ");
    }      
    if (Document) {
      CLAUSE Copy;
      Copy = clause_Copy(Clause);
      clause_DeleteLiterals(Copy,Indexes);
      red_DocumentObviousReductions(Copy,Indexes); 
      if (flag_GetFlagValue(flag_POBV))
	clause_Print(Copy);
      *Changed = Copy;
    }
    else {
      clause_DeleteLiterals(Clause,Indexes);
      list_Delete(Indexes);
      if (flag_GetFlagValue(flag_POBV))
	clause_Print(Clause);
    }
    return TRUE;
  }

  return FALSE;
}

void red_DocumentCondensing(CLAUSE Clause, LIST Indexes)

{

  list_Delete(clause_ParentClauses(Clause));
  list_Delete(clause_ParentLiterals(Clause));
  clause_SetParentClauses(Clause, list_Nil());
  clause_SetParentLiterals(Clause, Indexes);

  clause_AddParentClause(Clause,clause_Number(Clause));  

  clause_SetNumber(Clause, clause_IncreaseCounter());
  clause_SetFromCondensing(Clause);
}

static BOOL red_Condensing(CLAUSE Clause, CLAUSE *Changed, BOOL Document)

{
  LIST Indexes;

#ifdef DEBUG
  if (!clause_IsClause(Clause) || (*Changed != (CLAUSE)NULL)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_Condensing : ");
    misc_ErrorReport("Illegal input.\n");
    misc_FinishErrorReport();    
  }
  clause_Check(Clause);
#endif

  Indexes = cond_CondFast(Clause);

  if (!list_Empty(Indexes)) {
    if (flag_GetFlagValue(flag_PCON)) {
      printf("\nCondensing: "); 
      clause_Print(Clause);
      printf(" ==> ");
    }
    if (Document) {
      CLAUSE Copy;
      Copy = clause_Copy(Clause);
      clause_DeleteLiterals(Copy, Indexes);
      red_DocumentCondensing(Copy, Indexes);
      if (flag_GetFlagValue(flag_PCON))
	clause_Print(Copy);
      *Changed = Copy;
    }
    else {
      clause_DeleteLiterals(Clause, Indexes);
      list_Delete(Indexes);
      if (flag_GetFlagValue(flag_PCON))
	clause_Print(Clause);
    }
    return TRUE;
  }
  return FALSE;
}


void red_DocumentRemoveRedundantEquations(CLAUSE Clause, LIST Indexes, NAT NonTrivClauseNumber)

{

  list_Delete(clause_ParentClauses(Clause));
  list_Delete(clause_ParentLiterals(Clause));
  clause_SetParentClauses(Clause, list_Nil());
  clause_SetParentLiterals(Clause, Indexes);

  clause_AddParentClause(Clause,clause_Number(Clause));  

  clause_SetNumber(Clause, clause_IncreaseCounter());
  clause_SetFromRemoveRedundantEquations(Clause);

  clause_AddParentClause(Clause, NonTrivClauseNumber);
  clause_AddParentLiteral(Clause, 0); 
}


static BOOL red_RemoveRedundantEquations(CLAUSE Clause, CLAUSE *Changed, BOOL Document, 
					 NAT NonTrivClauseNumber, BOOL NonTrivDomain)

{
  LIST Indexes;                                 
  NAT  i;
  TERM LeftArg, RightArg;

#ifdef DEBUG
  if (!clause_IsClause(Clause) || (*Changed != (CLAUSE)NULL) ||
      (NonTrivDomain && NonTrivClauseNumber == 0) ||
      (!NonTrivDomain && NonTrivClauseNumber > 0)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_RemoveRedundantEquations: ");
    misc_ErrorReport("Illegal input.\n");
    misc_FinishErrorReport();    
  }
  clause_Check(Clause);
#endif

  Indexes = list_Nil();

  if (clause_ContainsNegativeEquations(Clause)) {
    for (i = clause_FirstAntecedentLitIndex(Clause); i <= clause_LastAntecedentLitIndex(Clause); i++) {
      if (clause_LiteralIsEquality(clause_GetLiteral(Clause,i))) {
	LeftArg  = term_FirstArgument(clause_GetLiteralAtom(Clause,i));
	RightArg = term_SecondArgument(clause_GetLiteralAtom(Clause,i));
	if ((term_IsVariable(LeftArg) &&
	    clause_NumberOfSymbolOccurrences(Clause, term_TopSymbol(LeftArg)) == 1) ||
	    (term_IsVariable(RightArg) &&
	     clause_NumberOfSymbolOccurrences(Clause, term_TopSymbol(RightArg)) == 1))
	  Indexes = list_Cons((POINTER)i, Indexes);
      }
    }
  }
  else
    if (NonTrivDomain && clause_ContainsPositiveEquations(Clause)) {
      for (i = clause_FirstSuccedentLitIndex(Clause); i <= clause_LastSuccedentLitIndex(Clause); i++) {
	if (clause_LiteralIsEquality(clause_GetLiteral(Clause,i))) {
	  LeftArg  = term_FirstArgument(clause_GetLiteralAtom(Clause,i));
	  RightArg = term_SecondArgument(clause_GetLiteralAtom(Clause,i));
	  if ((term_IsVariable(LeftArg) &&
	       clause_NumberOfSymbolOccurrences(Clause, term_TopSymbol(LeftArg)) == 1) ||
	      (term_IsVariable(RightArg) &&
	       clause_NumberOfSymbolOccurrences(Clause, term_TopSymbol(RightArg)) == 1))
	    Indexes = list_Cons((POINTER)i, Indexes);
	}
      }
    }

  if (!list_Empty(Indexes)) {
    if (flag_GetFlagValue(flag_PRRE)) {
      printf("\nRRE: "); 
      clause_Print(Clause);
      printf(" ==> ");
    }
    if (Document) {
      CLAUSE Copy;
      Copy = clause_Copy(Clause);
      clause_DeleteLiterals(Copy, Indexes);
      red_DocumentRemoveRedundantEquations(Copy, Indexes, NonTrivClauseNumber);
      if (flag_GetFlagValue(flag_PRRE))
	clause_Print(Copy);
      *Changed = Copy;
    }
    else {
      clause_DeleteLiterals(Clause, Indexes);
      list_Delete(Indexes);
      if (flag_GetFlagValue(flag_PRRE))
	clause_Print(Clause);
    }
    return TRUE;
  }

  return FALSE;
}

BOOL red_Tautology(CLAUSE Clause)

{
  TERM  Atom;
  int   i,j, la,n;
  BOOL   Result;

#ifdef DEBUG
  if (!clause_IsClause(Clause)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_Tautology :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(Clause);
#endif
  
  la     = clause_LastAntecedentLitIndex(Clause);
  n      = clause_Length(Clause);
  Result = FALSE;

  for (j = clause_FirstSuccedentLitIndex(Clause); j < n && !Result;  j++) {

    Atom = clause_LiteralAtom(clause_GetLiteral(Clause, j));

    if (fol_IsEquality(Atom) &&
	term_Equal(term_FirstArgument(Atom),term_SecondArgument(Atom)))
      Result = TRUE;
    
    for (i = clause_FirstLitIndex(); i <= la && !Result; i++)
      if (term_Equal(Atom, clause_LiteralAtom(clause_GetLiteral(Clause, i))))
	Result = TRUE;
  }

  if (Result && flag_GetFlagValue(flag_PTAUT)) {
    printf("\nTautology: "); 
    clause_Print(Clause);
  }
  return Result;
}

static LITERAL red_GetClRedLit(LITERAL ActLit, SHARED_INDEX ShIndex)

{
  LITERAL NextLit;
  int     i;
  CLAUSE  ActClause;
  TERM    CandTerm;
  LIST    LitScan;

  NextLit     = (LITERAL)NULL;
  ActClause   = clause_LiteralOwningClause(ActLit);
  i           = clause_LiteralGetIndex(ActLit);
  CandTerm    = st_ExistGen(cont_LeftContext(),
			    sharing_Index(ShIndex),
			    clause_LiteralAtom(ActLit));

  while (CandTerm) { 
    if (!term_IsVariable(CandTerm)) { 
      LitScan   = sharing_NAtomDataList(CandTerm); 
      while (!list_Empty(LitScan)) {
	NextLit = list_Car(LitScan);
	if (clause_LiteralsAreComplementary(ActLit,NextLit))
	  if (clause_Length(clause_LiteralOwningClause(NextLit)) == 1 ||
	      subs_SubsumesBasic(clause_LiteralOwningClause(NextLit),ActClause,
				 clause_LiteralGetIndex(NextLit),i)) {
	    st_CancelExistRetrieval();
	    return NextLit;
	  }
	LitScan = list_Cdr(LitScan);
      }
    }
    CandTerm = st_NextCandidate();
  }
  return (LITERAL)NULL;
}

static void red_DocumentClauseReduction(CLAUSE Clause, LIST LitInds, LIST ClauseNums, 
					LIST PLitInds)

{
  LIST   Scan,Help;

  Help  = list_Nil();

  for (Scan=LitInds; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Help = list_Cons((POINTER)clause_Number(Clause), Help);    
  }
  list_Delete(clause_ParentClauses(Clause));
  list_Delete(clause_ParentLiterals(Clause));
  clause_SetParentClauses(Clause, list_Nconc(Help,ClauseNums));
  clause_SetParentLiterals(Clause, list_Nconc(LitInds,PLitInds));
  
  clause_SetNumber(Clause, clause_IncreaseCounter());
  clause_SetFromClauseReduction(Clause);
}

static BOOL red_ClauseReduceRedClause(CLAUSE Clause, SHARED_INDEX ShIndex, CLAUSE *Changed, int Level,
				      BOOL Document)

{
  CLAUSE  PClause,Copy;
  LITERAL ActLit,PLit;
  int     i, j, length;
  LIST    ReducedBy,ReducedLits,PLits,Scan1,Scan2;

#ifdef DEBUG
  if (!clause_IsClause(Clause) || (*Changed != (CLAUSE)NULL)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_ClauseReduceRedClause :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(Clause);
#endif
  
  Copy          = Clause;
  length        = clause_Length(Clause);
  ReducedBy     = list_Nil();
  ReducedLits   = list_Nil();
  PLits         = list_Nil();
  i             = clause_FirstLitIndex();
  j             = 0;
  
  while (i < length) {
    ActLit  = clause_GetLiteral(Copy, i);
    
    if (!fol_IsEquality(clause_LiteralAtom(ActLit)) ||   
	clause_LiteralIsPositive(ActLit)) {
      PLit    = red_GetClRedLit(ActLit, ShIndex);
      if (clause_LiteralExists(PLit)) {
	if (list_Empty(PLits) && flag_GetFlagValue(flag_PCLR)) {
	  printf("\nFClauseReduction: "); 
	  clause_Print(Copy);
	}
	PClause     = clause_LiteralOwningClause(PLit);
	ReducedBy   = list_Cons((POINTER)clause_Number(PClause), ReducedBy);
	PLits       = list_Cons((POINTER)clause_LiteralGetIndex(PLit),PLits);
	ReducedLits = list_Cons((POINTER)(i+j), ReducedLits);
	if (Copy == Clause &&
	    (Document || prfs_SplitLevelCondition(clause_SplitLevel(PClause),clause_SplitLevel(Copy),Level)))
	  Copy = clause_Copy(Clause);
	clause_UpdateSplitDataFromPartner(Copy, PClause);	  
	clause_DeleteLiteral(Copy,i);
	length--;
	j++;
      }
      else
	i++;
    }
    else
      i++;
  }
  
  if (!list_Empty(ReducedBy)) {
    if (Document) {
      ReducedBy   = list_NReverse(ReducedBy);
      ReducedLits = list_NReverse(ReducedLits);
      PLits       = list_NReverse(PLits);
      red_DocumentClauseReduction(Copy, ReducedLits, ReducedBy, PLits); 
      if (flag_GetFlagValue(flag_PCLR)) {
	printf(" ==> [ ");
	for(Scan1=ReducedBy,Scan2=ReducedLits;!list_Empty(Scan1);
	    Scan1=list_Cdr(Scan1),Scan2=list_Cdr(Scan2))
	  printf("%d.%d ",(NAT)list_Car(Scan1),(NAT)list_Car(Scan2));
	printf("] ");
	clause_Print(Copy);
      }
    }
    else {
      if (flag_GetFlagValue(flag_PCLR)) {
	printf(" ==> [ ");
	for(Scan1=ReducedBy,Scan2=ReducedLits;!list_Empty(Scan1);
	    Scan1=list_Cdr(Scan1),Scan2=list_Cdr(Scan2))
	  printf("%d.%d ",(NAT)list_Car(Scan1),(NAT)list_Car(Scan2));
	printf("] ");
	clause_Print(Copy);
      }
      list_Delete(ReducedBy);
      list_Delete(ReducedLits);
      list_Delete(PLits);
    }
    if (Copy != Clause)
      *Changed = Copy;
    return TRUE;
  }
  return FALSE;
}

static void red_DocumentUnitConflict(CLAUSE Clause, LIST LitInds, LIST ClauseNums, LIST PLitInds)

{
  list_Delete(clause_ParentClauses(Clause));
  list_Delete(clause_ParentLiterals(Clause));
  clause_SetParentClauses(Clause, list_Nconc(list_List((POINTER)clause_Number(Clause)),ClauseNums));
  clause_SetParentLiterals(Clause, list_Nconc(LitInds,PLitInds));
  
  clause_SetNumber(Clause, clause_IncreaseCounter());
  clause_SetFromUnitConflict(Clause);
}


static BOOL red_UnitConflict(CLAUSE Clause, SHARED_INDEX ShIndex, CLAUSE *Changed, int Level,
			     BOOL Document)

{
  CLAUSE  PClause,Copy;
  LITERAL ActLit,PLit;
  LIST    Scan;
  TERM    CandTerm;

#ifdef DEBUG
  if (!clause_IsClause(Clause) || (*Changed != (CLAUSE)NULL)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_ForwardUnitConflict :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(Clause);
#endif
  
  Copy          = Clause;
  
  if (clause_Length(Clause) == 1) {
    ActLit   = clause_GetLiteral(Copy, clause_FirstLitIndex());
    PLit     = (LITERAL)NULL;
    CandTerm = st_ExistUnifier(cont_LeftContext(), sharing_Index(ShIndex), cont_RightContext(),
			       clause_LiteralAtom(ActLit));
    while (PLit == (LITERAL)NULL && CandTerm) {
      if (!term_IsVariable(CandTerm)) {
	Scan   = sharing_NAtomDataList(CandTerm); 
	while (!list_Empty(Scan)) {
	  PLit = list_Car(Scan);
	  if (clause_LiteralsAreComplementary(ActLit,PLit) &&
	      clause_Length(clause_LiteralOwningClause(PLit)) == 1) {
	    st_CancelExistRetrieval();
	    Scan = list_Nil();
	  }
	  else {
	    PLit  = (LITERAL)NULL;
	    Scan = list_Cdr(Scan);
	  }
	}
      }
      if (PLit == (LITERAL)NULL)
	CandTerm = st_NextCandidate();
    }

    if (PLit == (LITERAL)NULL && fol_IsEquality(clause_LiteralAtom(ActLit))) {
      TERM Atom;
      Atom = term_Create(fol_Equality(),list_Reverse(term_ArgumentList(clause_LiteralAtom(ActLit))));
      CandTerm = st_ExistUnifier(cont_LeftContext(), sharing_Index(ShIndex), cont_RightContext(), Atom);
      while (PLit == (LITERAL)NULL && CandTerm) {
	if (!term_IsVariable(CandTerm)) {
	  Scan   = sharing_NAtomDataList(CandTerm); 
	  while (!list_Empty(Scan)) {
	    PLit = list_Car(Scan);
	    if (clause_LiteralsAreComplementary(ActLit,PLit) &&
		clause_Length(clause_LiteralOwningClause(PLit)) == 1) {
	      st_CancelExistRetrieval();
	      Scan = list_Nil();
	    }
	    else {
	      PLit  = (LITERAL)NULL;
	      Scan = list_Cdr(Scan);
	    }
	  }
	}
	if (PLit == (LITERAL)NULL)
	  CandTerm = st_NextCandidate();
      }
      list_Delete(term_ArgumentList(Atom));
      term_Free(Atom);
    }
      
    if (clause_LiteralExists(PLit)) {
	if (flag_GetFlagValue(flag_PUNC)) {
	  printf("\nUnitConflict: "); 
	  clause_Print(Copy);
	}
	PClause     = clause_LiteralOwningClause(PLit);
	if (Copy == Clause &&
	    (Document || prfs_SplitLevelCondition(clause_SplitLevel(PClause),clause_SplitLevel(Copy),Level)))
	  Copy = clause_Copy(Clause);
	clause_UpdateSplitDataFromPartner(Copy, PClause);	  
	clause_DeleteLiteral(Copy,clause_FirstLitIndex());
	if (Document) 
	  red_DocumentUnitConflict(Copy, list_List((POINTER)clause_FirstLitIndex()), 
				      list_List((POINTER)clause_Number(PClause)), 
				      list_List((POINTER)clause_FirstLitIndex()));
	if (flag_GetFlagValue(flag_PUNC)) {
	  printf(" ==> [ %d.%d ]", clause_Number(PClause), clause_FirstLitIndex());	
	  clause_Print(Copy);
	}
	if (Copy != Clause)
	  *Changed = Copy;
	return TRUE;
    }
  }
  return FALSE;
}


static CLAUSE red_ForwardSubsumer(CLAUSE RedCl, SHARED_INDEX ShIndex)

{
  TERM   Atom,AtomGen;
  CLAUSE CandCl;
  LIST   LitScan;
  int    i,lits;

#ifdef DEBUG
  if (!clause_IsClause(RedCl)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_ForwardSubsumer :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(RedCl);
#endif

  lits     = clause_Length(RedCl);

  for (i = 0; i < lits; i++) {
    Atom    = clause_GetLiteralAtom(RedCl, i);
    AtomGen = st_ExistGen(cont_LeftContext(), sharing_Index(ShIndex), Atom);

    while (AtomGen) {
      if (!term_IsVariable(AtomGen)) {
	for (LitScan = sharing_NAtomDataList(AtomGen);
	     !list_Empty(LitScan);
	     LitScan = list_Cdr(LitScan)) {
	  CandCl = clause_LiteralOwningClause(list_Car(LitScan));

	  if (CandCl != RedCl &&
	      clause_GetLiteral(CandCl,clause_FirstLitIndex()) == (LITERAL)list_Car(LitScan) &&
	      term_TopSymbol(clause_GetLiteralTerm(RedCl,i)) == 
	      term_TopSymbol(clause_GetLiteralTerm(CandCl,
						   clause_FirstLitIndex())) &&
	      subs_SubsumesBasic(CandCl, RedCl, clause_FirstLitIndex(), i)) {
	    st_CancelExistRetrieval();
	    return (CandCl);
	  }
	}
      }
      AtomGen = st_NextCandidate();
    }

    if (fol_IsEquality(Atom) && 
	clause_LiteralIsNotOrientedEquality(clause_GetLiteral(RedCl,i))) {
      Atom = term_Create(fol_Equality(),list_Reverse(term_ArgumentList(Atom)));
      AtomGen = st_ExistGen(cont_LeftContext(), sharing_Index(ShIndex), Atom);
      while (AtomGen) {
	if (!term_IsVariable(AtomGen)) {
	  for (LitScan = sharing_NAtomDataList(AtomGen);
	       !list_Empty(LitScan);
	       LitScan = list_Cdr(LitScan)) {
	    CandCl = clause_LiteralOwningClause(list_Car(LitScan));
	    if ((CandCl != RedCl) && 
		clause_GetLiteral(CandCl,clause_FirstLitIndex())
		== (LITERAL)list_Car(LitScan) &&
		term_TopSymbol(clause_GetLiteralTerm(RedCl,i)) == 
		term_TopSymbol(clause_GetLiteralTerm(CandCl,
						 clause_FirstLitIndex())) &&
		subs_SubsumesBasic(CandCl, RedCl, clause_FirstLitIndex(), i)) {
	      st_CancelExistRetrieval();
	      list_Delete(term_ArgumentList(Atom));
	      term_Free(Atom);
	      return (CandCl);
	    }
	  }
	}
	AtomGen = st_NextCandidate();
      }
      list_Delete(term_ArgumentList(Atom));
      term_Free(Atom);
    }
  }

  return((CLAUSE)NULL);
}


static CLAUSE red_ForwardSubsumption(CLAUSE RedClause, SHARED_INDEX ShIndex)

{ 
  CLAUSE Subsumer;

#ifdef DEBUG
  if (!clause_IsClause(RedClause)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_ForwardSubSumer :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(RedClause);
#endif

  Subsumer = red_ForwardSubsumer(RedClause, ShIndex);

  if (flag_GetFlagValue(flag_PSUB) && Subsumer) {
    printf("\nFSubsumption:"); clause_Print(RedClause);
    printf(" by %d %d ",clause_Number(Subsumer),clause_SplitLevel(Subsumer));
  }

  return Subsumer;
}

static void red_DocumentRewriting(CLAUSE Clause, int i, CLAUSE Rule, int ri)

{
  list_Delete(clause_ParentClauses(Clause));
  list_Delete(clause_ParentLiterals(Clause));
  clause_SetParentClauses(Clause,
			  list_List((POINTER)clause_Number(Clause)));
  

  clause_SetParentLiterals(Clause, list_List((POINTER)i));
  clause_SetNumber(Clause, clause_IncreaseCounter());
  clause_SetFromRewriting(Clause);

  clause_AddParentClause(Clause,clause_Number(Rule));
  clause_AddParentLiteral(Clause,ri);
}

static void red_DocumentFurtherRewriting(CLAUSE Clause, int i, CLAUSE Rule, int ri)

{
  clause_AddParentClause(Clause,
    (int) list_Car(list_Cdr(clause_ParentClauses(Clause))));
  clause_AddParentLiteral(Clause, i);
  clause_AddParentClause(Clause, clause_Number(Rule));
  clause_AddParentLiteral(Clause, ri);
}

static BOOL red_RewriteRedUnitClause(CLAUSE RedClause, SHARED_INDEX ShIndex, CLAUSE *Changed, int Level,
				     BOOL Document)

{
  TERM   RedAtom, RedTermS;
  int    B_Stack;
  BOOL   Rewritten,Result,Oriented,Renamed;
  TERM   TermS,PartnerEq;
  LIST   EqList,EqScan,LitScan;
  CLAUSE Copy;

#ifdef DEBUG
  if (!clause_IsClause(RedClause) || (*Changed != (CLAUSE)NULL)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_RewriteUnitClause :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(RedClause);
#endif

  Result      = FALSE;
  Renamed     = FALSE;
  Copy        = RedClause;
  RedAtom     = clause_GetLiteralAtom(Copy, clause_FirstLitIndex());
  Rewritten   = TRUE;

  if (!list_Empty(term_ArgumentList(RedAtom))) { 
    if (term_StampOverflow(red_STAMPID))
      term_ResetTermStamp(clause_LiteralSignedAtom(clause_GetLiteral(RedClause, clause_FirstLitIndex())));
    term_StartStamp();
    while (Rewritten) {
      Rewritten = FALSE;
      B_Stack = stack_Bottom();
      sharing_PushListOnStackNoStamps(term_ArgumentList(RedAtom));

      while (!stack_Empty(B_Stack)) {
	RedTermS = (TERM)stack_PopResult();
	TermS    = st_ExistGen(cont_LeftContext(), sharing_Index(ShIndex), RedTermS);
	while (TermS && !Rewritten) {
	  EqList = term_SupertermList(TermS);
	  for (EqScan = EqList; !list_Empty(EqScan) && !Rewritten;
	       EqScan = list_Cdr(EqScan)) {
	    PartnerEq = list_Car(EqScan);
	    if (fol_IsEquality(PartnerEq)) {
	      CLAUSE  RewriteClause;
	      LITERAL RewriteLit;
	      TERM    Right;
	      if (TermS == term_FirstArgument(PartnerEq))
		Right = term_SecondArgument(PartnerEq);
	      else
		Right = term_FirstArgument(PartnerEq);
	      for (LitScan = sharing_NAtomDataList(PartnerEq); 
		   !list_Empty(LitScan) && !Rewritten;
		   LitScan = list_Cdr(LitScan)) {
		RewriteLit    = list_Car(LitScan);
		RewriteClause = clause_LiteralOwningClause(RewriteLit);
		if (clause_LiteralIsPositive(RewriteLit) &&
		    clause_Length(RewriteClause) == 1) {
		  Oriented = (clause_LiteralIsOrientedEquality(RewriteLit) && 
			      TermS == term_FirstArgument(PartnerEq));
		  if (!Oriented && !clause_LiteralIsOrientedEquality(RewriteLit)) { 
		    Renamed = TRUE;                            
		    term_StartMaxRenaming(clause_MaxVar(RewriteClause));
		    term_Rename(RedAtom); 
		    Oriented = ord_ContGreater(cont_LeftContext(),TermS,cont_LeftContext(),Right);
		      
		    
		  }
		  if (Oriented) {
		    TERM   TermT;

		    if (RedClause == Copy &&
			(Document || 
			 prfs_SplitLevelCondition(clause_SplitLevel(RewriteClause),
						  clause_SplitLevel(RedClause),Level))) {
		      Copy    = clause_Copy(RedClause);
		      RedAtom = clause_GetLiteralAtom(Copy, clause_FirstLitIndex());
		    }

		    if (!Result)
		      if (flag_GetFlagValue(flag_PREW)) {
			printf("\nFRewriting: ");
			clause_Print(Copy);
			printf(" ==>[ ");
		      }

		    if (Document) {
		      if (!Result)
			red_DocumentRewriting(Copy, clause_FirstLitIndex(),
					      RewriteClause, clause_FirstLitIndex());
		      else
			red_DocumentFurtherRewriting(Copy, clause_FirstLitIndex(), 
						     RewriteClause, clause_FirstLitIndex());
		    }
		    Result = TRUE;
		    TermT  = cont_ApplyBindingsModuloMatching(cont_LeftContext(), term_Copy(Right));
		    if (cont_BindingsAreRenamingModuloMatching(cont_RightContext()))
		      term_SetTermSubtermStamp(TermT);
		    term_ReplaceSubtermBy(RedAtom, RedTermS, TermT);
		    Rewritten = TRUE;
		    clause_UpdateSplitDataFromPartner(Copy, RewriteClause);
		    term_Delete(TermT);
		    stack_SetBottom(B_Stack);
		  
		    if (flag_GetFlagValue(flag_PREW))
		      printf("%d.%d ",clause_Number(RewriteClause), clause_FirstLitIndex());
		    clause_SetWeight(Copy, clause_ComputeWeight(Copy));
		  }
		}
	      }
	    }
	  }
	  if (!Rewritten)
	    TermS = st_NextCandidate();
	}
	st_CancelExistRetrieval();
	if (!Rewritten)
	  term_SetTermStamp(RedTermS);
      }
    }
    term_StopStamp(); 
  }
  if (Result) {
    clause_OrientAndReInit(Copy);
    if (Copy != RedClause)
      clause_OrientAndReInit(RedClause);
    if (flag_GetFlagValue(flag_PREW)) {
      printf("] "); 
      clause_Print(Copy);
    }
    if (Copy != RedClause)
      *Changed = Copy;
  }
  else
    if (Renamed)
      clause_OrientAndReInit(Copy);
	
      
#ifdef DEBUG
  clause_Check(Copy);
  clause_Check(RedClause);
#endif
  
  return Result;
}


static BOOL red_RewriteRedClause(CLAUSE RedClause, SHARED_INDEX ShIndex, CLAUSE *Changed, int Level,
				 BOOL Document)

{
  TERM   RedAtom, RedTermS;
  int    B_Stack;
  int    ci, length;
  BOOL   Rewritten,Result;
  TERM   TermS,PartnerEq;
  LIST   EqList,EqScan,LitScan;
  CLAUSE Copy;

#ifdef DEBUG
  if (!clause_IsClause(RedClause)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_RewriteRedClause :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(RedClause);
#endif
  
  length      = clause_Length(RedClause);

  if (length == 1)
    return red_RewriteRedUnitClause(RedClause, ShIndex, Changed, Level, Document);

  Result      = FALSE;
  Copy        = RedClause;

  for (ci = clause_FirstLitIndex(); ci < length; ci++) {
    Rewritten  = TRUE;
    if (!list_Empty(term_ArgumentList(clause_GetLiteralAtom(Copy, ci)))) {
      while (Rewritten) {
	Rewritten = FALSE;
	RedAtom   = clause_GetLiteralAtom(Copy, ci);

	B_Stack = stack_Bottom();
	sharing_PushListReverseOnStack(term_ArgumentList(RedAtom));

	while (!stack_Empty(B_Stack))  {
	  RedTermS = (TERM)stack_PopResult();
	  TermS    = st_ExistGen(cont_LeftContext(), sharing_Index(ShIndex), RedTermS);

	  while (TermS && !Rewritten) {
	    EqList = term_SupertermList(TermS);

	    for (EqScan = EqList; !list_Empty(EqScan) && !Rewritten;
		 EqScan = list_Cdr(EqScan)) {
	      PartnerEq = list_Car(EqScan);
	      if (fol_IsEquality(PartnerEq) &&
		  (term_FirstArgument(PartnerEq) == TermS)) {
		CLAUSE  RewriteClause;
		LITERAL RewriteLit;
		int     ri;

		for (LitScan = sharing_NAtomDataList(PartnerEq); 
		     !list_Empty(LitScan) && !Rewritten;
		     LitScan = list_Cdr(LitScan)) {
		  RewriteLit    = list_Car(LitScan);
		  RewriteClause = clause_LiteralOwningClause(RewriteLit);
		  ri            = clause_LiteralGetIndex(RewriteLit);

		  if (clause_LiteralIsPositive(RewriteLit) &&
		      clause_LiteralGetFlag(RewriteLit,STRICTMAXIMAL) &&
		      clause_LiteralIsOrientedEquality(RewriteLit) &&
		      subs_SubsumesBasic(RewriteClause, Copy, ri, ci)) {
		    TERM   TermT;

		    if (RedClause == Copy &&
			(Document || 
			 prfs_SplitLevelCondition(clause_SplitLevel(RewriteClause),
						  clause_SplitLevel(RedClause),Level))) {
		      Copy    = clause_Copy(RedClause);
		      RedAtom = clause_GetLiteralAtom(Copy, ci);
		    }

		    if (!Result) {
		      if (flag_GetFlagValue(flag_PREW)) {
			printf("\nFRewriting: ");
			clause_Print(Copy);
			printf(" ==>[ ");
		      }
		    }

		    if (Document) {
		      if (!Result)
			red_DocumentRewriting(Copy,ci, RewriteClause,ri);
		      else
			red_DocumentFurtherRewriting(Copy,ci,RewriteClause,ri);
		    }
		    Result = TRUE;
		    TermT  = cont_ApplyBindingsModuloMatching(cont_LeftContext(),
							      term_Copy(term_SecondArgument(PartnerEq)));
		    term_ReplaceSubtermBy(RedAtom, RedTermS, TermT);
		    Rewritten = TRUE;
		    clause_UpdateSplitDataFromPartner(Copy,RewriteClause);
		    term_Delete(TermT);
		    stack_SetBottom(B_Stack);

		    if (flag_GetFlagValue(flag_PREW))
		      printf("%d.%d ",clause_Number(RewriteClause), ri);
		    clause_SetWeight(Copy, clause_ComputeWeight(Copy));
		  }
		}
	      }
	    }
	    if (!Rewritten)
	      TermS = st_NextCandidate();
	  }
	  st_CancelExistRetrieval();
	} 
      }
    }
  }
  if (Result) {
    clause_OrientAndReInit(Copy);
    if (flag_GetFlagValue(flag_PREW)) {
      printf("] "); 
      clause_Print(Copy);
    }
    if (Copy != RedClause) {
      clause_OrientAndReInit(RedClause);
      *Changed = Copy;
    }
  }

#ifdef DEBUG
  clause_Check(Copy);
  clause_Check(RedClause);
#endif

  return Result;
}

LIST red_BackSubsumption(CLAUSE RedCl, SHARED_INDEX ShIndex)

{
  TERM   Atom,CandTerm;
  CLAUSE SubsumedCl;
  LIST   CandLits, Scan, SubsumedList;
  int    i,j;
  
#ifdef DEBUG
  if (!clause_IsClause(RedCl)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_BackSubsumption :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(RedCl);
#endif
  
  SubsumedList = list_Nil();
  
  if (!clause_IsEmptyClause(RedCl)) {
    
    i          = clause_FirstLitIndex();
    Atom       = clause_GetLiteralAtom(RedCl, i);
    CandTerm   = st_ExistInstance(cont_LeftContext(), sharing_Index(ShIndex), Atom);
    
    while (CandTerm) {
      CandLits = sharing_NAtomDataList(CandTerm);
      
      for (Scan = CandLits; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	SubsumedCl = clause_LiteralOwningClause(list_Car(Scan));
	j          = clause_LiteralGetIndex(list_Car(Scan));
	
	if ((RedCl != SubsumedCl) && 
	    term_TopSymbol(clause_GetLiteralTerm(RedCl,i)) == 
	    term_TopSymbol(clause_GetLiteralTerm(SubsumedCl,j)) &&
	    !list_PointerMember(SubsumedList, SubsumedCl) &&
	    subs_SubsumesBasic(RedCl, SubsumedCl, i, j))
	  SubsumedList = list_Cons(SubsumedCl, SubsumedList);
      }
      
      CandTerm = st_NextCandidate();
    }
    
    if (fol_IsEquality(Atom) && 
	clause_LiteralIsNotOrientedEquality(clause_GetLiteral(RedCl, i))) {
      Atom      = term_Create(fol_Equality(),
			      list_Reverse(term_ArgumentList(Atom)));
      CandTerm  = st_ExistInstance(cont_LeftContext(), sharing_Index(ShIndex), Atom);
      
      while (CandTerm) {
	CandLits = sharing_NAtomDataList(CandTerm);
	
	for (Scan = CandLits; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	  SubsumedCl = clause_LiteralOwningClause(list_Car(Scan));
	  
	  j          = clause_LiteralGetIndex(list_Car(Scan));
	  
	  if ((RedCl != SubsumedCl) && 
	      term_TopSymbol(clause_GetLiteralTerm(RedCl,i)) == 
	      term_TopSymbol(clause_GetLiteralTerm(SubsumedCl,j)) &&
	      !list_PointerMember(SubsumedList, SubsumedCl) &&
	      subs_SubsumesBasic(RedCl, SubsumedCl, i, j))
	    SubsumedList = list_Cons(SubsumedCl, SubsumedList);
	  
	}
	
	CandTerm = st_NextCandidate();
      }
      
      list_Delete(term_ArgumentList(Atom));
      term_Free(Atom);
    }
      
    if (flag_GetFlagValue(flag_PSUB)) {
	for (Scan = SubsumedList; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	    SubsumedCl = list_Car(Scan);
	    printf("\nBSubsumption: "); 
	    clause_Print(SubsumedCl);
	    printf(" by %d ",clause_Number(RedCl));
	}
    }
  }
  return SubsumedList;
}

static LIST red_GetBackClRedLits(CLAUSE Clause, LITERAL ActLit, SHARED_INDEX ShIndex)

{
  CLAUSE  PClause;
  LITERAL PLit;
  LIST    LitScan, PClLits;
  TERM    CandTerm;
  int     i;

  PClLits   = list_Nil();
  i         = clause_LiteralGetIndex(ActLit);

  CandTerm  = st_ExistInstance(cont_LeftContext(),
			       sharing_Index(ShIndex),
			       clause_LiteralAtom(ActLit));

  while (CandTerm) {

    LitScan = sharing_NAtomDataList(CandTerm); 

    for ( ; !list_Empty(LitScan); LitScan = list_Cdr(LitScan)) {

      PLit    = list_Car(LitScan);
      PClause = clause_LiteralOwningClause(PLit);

      if (PClause != Clause &&
	  clause_LiteralsAreComplementary(ActLit,PLit) &&
	  subs_SubsumesBasic(Clause,PClause,i,clause_LiteralGetIndex(PLit)))
	PClLits = list_Cons(PLit, PClLits);
    }
    
    CandTerm = st_NextCandidate();
  }
  return PClLits;
}


LIST red_BackClauseReduction(CLAUSE RedClause, SHARED_INDEX ShIndex, LIST* Result, BOOL Document)

{
    LIST Blocked;
    CLAUSE Copy;

#ifdef DEBUG
  if (!clause_IsClause(RedClause)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_BackClauseReduction :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(RedClause);
#endif

  Blocked = list_Nil();

  if (clause_Length(RedClause) == 1) {
    LITERAL ActLit, PLit;
    LIST    LitList, Scan, Iter;
    TERM    CandTerm;
    int     RedClNum;
    
    ActLit      = clause_GetLiteral(RedClause, clause_FirstLitIndex());

    if (!fol_IsEquality(clause_LiteralAtom(ActLit)) ||  
	clause_LiteralIsNegative(ActLit)) {
      CLAUSE PClause;
      int    PInd;
      LIST   PIndL;

      CandTerm = st_ExistInstance(cont_LeftContext(), sharing_Index(ShIndex), clause_LiteralAtom(ActLit));
      RedClNum = clause_Number(RedClause);
      LitList  = list_Nil();

      while (CandTerm) {
	for (Iter = sharing_NAtomDataList(CandTerm); !list_Empty(Iter); Iter = list_Cdr(Iter))
	  if (clause_LiteralsAreComplementary(ActLit,list_Car(Iter)))
	    LitList = list_Cons(list_Car(Iter),LitList);
	CandTerm = st_NextCandidate();
      }

      
      
      while (!list_Empty(LitList)) {
	PLit    = list_Car(LitList);
	PInd    = clause_LiteralGetIndex(PLit);
	PIndL   = list_List(PLit);
	PClause = clause_LiteralOwningClause(PLit);
	Blocked = list_Cons(PClause, Blocked);

	if (flag_GetFlagValue(flag_PCLR)) {
	  printf("\nBClauseReduction: ");
	  clause_Print(PClause);
	  printf(" ==>[ %d.%d ] ",clause_Number(RedClause),clause_FirstLitIndex());
	}

	Iter = LitList;
	for (Scan=list_Cdr(LitList);!list_Empty(Scan);Scan=list_Cdr(Scan)) 
	  if (PClause == clause_LiteralOwningClause(list_Car(Scan))) {
	    list_Rplacd(Iter,list_Cdr(Scan));
	    list_Rplacd(Scan,PIndL);
	    PIndL = Scan;
	    Scan  = Iter;
	  }
	  else
	    Iter = Scan;
	Iter    = LitList;
	LitList = list_Cdr(LitList);
	list_Free(Iter);
	Copy    = clause_Copy(PClause);
	clause_RemoveFlag(Copy,WORKEDOFF);
	clause_UpdateSplitDataFromPartner(Copy, RedClause);
	for(Scan=PIndL;!list_Empty(Scan);Scan=list_Cdr(Scan))       
	  list_Rplaca(Scan,(POINTER)clause_LiteralGetIndex(list_Car(Scan)));
	clause_DeleteLiterals(Copy, PIndL);
	
	if (Document)
	  red_DocumentClauseReduction(Copy, PIndL, list_List((POINTER)RedClNum),  
				      list_List((POINTER)clause_FirstLitIndex()));
	else
	  list_Delete(PIndL);	
	
	if (flag_GetFlagValue(flag_PCLR))
	  clause_Print(Copy);
	*Result = list_Cons(Copy, *Result);
      }
    }
    return Blocked;
  } 
  else {
    CLAUSE  PClause;
    LITERAL ActLit, PLit;
    LIST    LitScan,LitList;
    int     i,length,RedClNum,PInd;

    RedClNum    = clause_Number(RedClause);
    length      = clause_Length(RedClause);

    for (i = clause_FirstLitIndex(); i < length; i++) {
      ActLit = clause_GetLiteral(RedClause, i);
      
      if (!fol_IsEquality(clause_LiteralAtom(ActLit))) {
	LitList = red_GetBackClRedLits(RedClause,ActLit, ShIndex);
	
	for (LitScan = LitList;!list_Empty(LitScan);LitScan = list_Cdr(LitScan)) {
	  PLit    = list_Car(LitScan);
	  PClause = clause_LiteralOwningClause(PLit);
	  PInd    = clause_LiteralGetIndex(PLit);
	  Copy    = clause_Copy(PClause);
	  if (list_PointerMember(Blocked,PClause)) {
	    if (!flag_GetFlagValue(flag_DOCPROOF))
	      clause_NewNumber(Copy);
	  }
	  else
	    Blocked = list_Cons(PClause, Blocked);
	  clause_RemoveFlag(Copy,WORKEDOFF);
	  clause_UpdateSplitDataFromPartner(Copy, RedClause);
	  clause_DeleteLiteral(Copy, PInd);

	  if (Document)
	    red_DocumentClauseReduction(Copy, list_List((POINTER)PInd), 
				      list_List((POINTER)RedClNum), 
				      list_List((POINTER)i));
	  
	  if (flag_GetFlagValue(flag_PCLR)) {
	    printf("\nBClauseReduction: "); 
	    clause_Print(PClause);
	    printf(" ==>[ %d.%d ] ",clause_Number(RedClause),i);
	    clause_Print(Copy);
	  }
	  *Result = list_Cons(Copy, *Result);
	}
	list_Delete(LitList);
      }
    }
    return Blocked;
  }
}

static void red_ApplyRewriting(CLAUSE RuleCl, int ri, CLAUSE PartnerClause, int pli, TERM PartnerTermS,
			       BOOL Document)

{
  LITERAL PartnerLit;
  TERM    ReplaceTermT,NewAtom;

#ifdef DEBUG
  clause_Check(PartnerClause);
  clause_Check(RuleCl);
#endif

  if (Document)
    red_DocumentRewriting(PartnerClause,pli, RuleCl,ri);

  if (flag_GetFlagValue(flag_PREW)) {
    printf("\nBRewriting: ");
    clause_Print(PartnerClause);
    printf(" ==>[ %d.%d ] ",clause_Number(RuleCl), ri); 
  }

  PartnerLit   = clause_GetLiteral(PartnerClause, pli);

  ReplaceTermT =
    cont_ApplyBindingsModuloMatchingReverse(cont_LeftContext(),
	    term_Copy(term_SecondArgument(clause_LiteralSignedAtom(clause_GetLiteral(RuleCl, ri)))));

  NewAtom      = clause_LiteralSignedAtom(PartnerLit);
  term_ReplaceSubtermBy(NewAtom, PartnerTermS, ReplaceTermT);
  term_Delete(ReplaceTermT);

  clause_OrientAndReInit(PartnerClause);
  clause_UpdateSplitDataFromPartner(PartnerClause, RuleCl);

  if (flag_GetFlagValue(flag_PREW))
    clause_Print(PartnerClause);
}


static LIST red_LiteralRewriting(CLAUSE RedClause, LITERAL ActLit, int ri,
				 SHARED_INDEX ShIndex, LIST* Result, BOOL Document)

{
  TERM TermS, CandTerm;
  LIST Blocked;

#ifdef DEBUG
  if (!clause_LiteralIsLiteral(ActLit)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_LiteralRewriting :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(RedClause);
#endif

  Blocked    = list_Nil();
  TermS      = term_FirstArgument(clause_LiteralSignedAtom(ActLit));  

  CandTerm = st_ExistInstance(cont_LeftContext(), sharing_Index(ShIndex), TermS);

  while (CandTerm) {
    LIST LitList,LitListDel;

    if (!term_IsVariable(CandTerm) &&
	!symbol_IsPredicate(term_TopSymbol(CandTerm))) {
      LitListDel = sharing_GetDataList(CandTerm, ShIndex);

      for(LitList =  LitListDel; !list_Empty(LitList); LitList = list_Cdr(LitList)){
	LITERAL PartnerLit;
	CLAUSE PartnerClause;
	int pli;

	PartnerLit    = list_Car(LitList);
	pli           = clause_LiteralGetIndex(PartnerLit);
	PartnerClause = clause_LiteralOwningClause(PartnerLit);

	if (clause_Number(RedClause) != clause_Number(PartnerClause) &&
	    !list_PointerMember(Blocked, PartnerClause) &&
	    subs_SubsumesBasic(RedClause, PartnerClause, ri, pli)) {
	  CLAUSE Copy;
	  Blocked = list_Cons(PartnerClause,Blocked);
	  Copy    = clause_Copy(PartnerClause);
	  clause_RemoveFlag(Copy,WORKEDOFF);
	  red_ApplyRewriting(RedClause, ri, Copy, pli, CandTerm, Document);
	  *Result = list_Cons(Copy, *Result);
	}
      }
      list_Delete(LitListDel);
    }
    CandTerm = st_NextCandidate();
  }
  return Blocked;
}


LIST red_BackRewriting(CLAUSE RedClause, SHARED_INDEX ShIndex, LIST* Result, BOOL Document)

{
  int     i,length;
  LITERAL ActLit;
  LIST    Blocked;

#ifdef DEBUG
  if (!(clause_IsClause(RedClause))) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_BackRewriting :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(RedClause);
#endif
  
  Blocked    = list_Nil();
  length     = clause_Length(RedClause);
  
  if (!clause_HasEmptySuccedent(RedClause)) {
    
    for (i=clause_FirstSuccedentLitIndex(RedClause); i < length; i++) {
      ActLit = clause_GetLiteral(RedClause, i);
      if (clause_LiteralIsOrientedEquality(ActLit) && 
	  clause_LiteralGetFlag(ActLit,STRICTMAXIMAL)) {
	Blocked = list_Nconc(red_LiteralRewriting(RedClause, ActLit, i, ShIndex, Result, Document),
			     Blocked);
      }
      
#ifdef DEBUG
      if (fol_IsEquality(clause_LiteralSignedAtom(ActLit))) {
	ord_RESULT HelpRes;
	
	HelpRes =
	  ord_Compare(term_FirstArgument(clause_LiteralSignedAtom(ActLit)), 
		      term_SecondArgument(clause_LiteralSignedAtom(ActLit)));
	
	if (ord_IsSmallerThan(HelpRes)){ 
	  misc_StartErrorReport();
	  misc_ErrorReport("\n In red_BackRewriting :");
	  misc_ErrorReport(" Error. First Argument smaller than second in RedClause.\n");
	  misc_FinishErrorReport();
	  misc_DumpCore();
	}
      } 
#endif
    } 
  } 
  Blocked = list_PointerDeleteDuplicates(Blocked);
  return Blocked;
}


void red_DocumentSortSimplification(CLAUSE Clause, LIST Indexes, LIST Clauses)

{
  LIST Scan,Declarations,Self;

  Declarations = list_Nil();
  Self         = list_Nil();

  list_Delete(clause_ParentClauses(Clause));
  list_Delete(clause_ParentLiterals(Clause));  
  
  for(Scan=Indexes;!list_Empty(Scan);Scan=list_Cdr(Scan))
    Self = list_Cons((POINTER)clause_Number(Clause),Self);
  
  for(Scan=Clauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
   Declarations = list_Cons((POINTER)clause_FirstSuccedentLitIndex(list_Car(Scan)),Declarations);
   list_Rplaca(Scan,(POINTER)clause_Number(list_Car(Scan)));
  }

  clause_SetParentLiterals(Clause, list_Nconc(Indexes,Declarations));
  clause_SetParentClauses(Clause, list_Nconc(Self,Clauses));

  clause_SetNumber(Clause, clause_IncreaseCounter());
  clause_SetFromSortSimplification(Clause);
}


BOOL red_SortSimplification(SORTTHEORY Theory, CLAUSE Clause, CLAUSE *Changed, NAT Level, BOOL Document)

{ 
  if (Theory != (SORTTHEORY)NULL) {
    TERM      Atom,Term;
    SOJU      SortPair;
    SORT      TermSort,LitSort;
    LIST      Indexes,NewClauses,Clauses,Scan;
    int       i,lc,j,OldSplitLevel;
    CLAUSE    Copy;
    CONDITION Cond;

#ifdef DEBUG
    if (!clause_IsClause(Clause)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_SortSimplification :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
    }
    clause_Check(Clause);
#endif

    lc            = clause_LastConstraintLitIndex(Clause);
    i             = clause_FirstLitIndex();
    j             = 0;
    OldSplitLevel = clause_SplitLevel(Clause);
    Copy          = Clause;
    Indexes       = list_Nil();
    Clauses       = list_Nil();
  
    while (i <= lc) {

      Atom        = clause_LiteralAtom(clause_GetLiteral(Copy, i));
      Term        = term_FirstArgument(Atom);
      SortPair    = sort_ComputeSortNoResidues(Theory, Term, Copy, i);
      TermSort    = sort_PairSort(SortPair);
      NewClauses  = sort_ConditionClauses(sort_PairCondition(SortPair));
      sort_ConditionPutClauses(sort_PairCondition(SortPair),list_Nil());
      LitSort     = sort_TheorySortOfSymbol(Theory,term_TopSymbol(Atom));

      if ((Cond = sort_TheoryIsSubsortOfNoResidues(Theory, TermSort, LitSort)) != (CONDITION)NULL) {

	if (j == 0 && flag_GetFlagValue(flag_PSSI)) {
	  printf("\nSortSimplification: ");
	  clause_Print(Copy);
	  printf(" ==>[ ");
	}

	NewClauses = list_Nconc(NewClauses,sort_ConditionClauses(Cond));
	sort_ConditionPutClauses(Cond,list_Nil());
	sort_ConditionDelete(Cond);

	for (Scan = NewClauses; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	  if (Clause == Copy && 
	      (Document || 
	       prfs_SplitLevelCondition(clause_SplitLevel(list_Car(Scan)),OldSplitLevel,Level)))
	    Copy   = clause_Copy(Clause);
	  clause_UpdateSplitDataFromPartner(Copy, list_Car(Scan));
	  if (flag_GetFlagValue(flag_PSSI))
	    printf("%d ",clause_Number(list_Car(Scan)));
	}

	if (Document)
	  Indexes = list_Cons((POINTER)(i+j), Indexes);

	clause_DeleteLiteral(Copy, i);
	Clauses = list_Nconc(NewClauses,Clauses);
	j++;
	lc--;
      }	   
      else {
	list_Delete(NewClauses);
	i++;
      }
      sort_DeleteSortPair(SortPair);
      sort_Delete(LitSort);
    }

#ifdef DEBUG
    clause_Check(Copy);
#endif

    if (j > 0) {
      if (Document)
	red_DocumentSortSimplification(Copy,Indexes,Clauses);
      else
	list_Delete(Clauses);
      clause_ReInit(Copy);
      if (flag_GetFlagValue(flag_PSSI)) {
	printf("] ");
	clause_Print(Copy);
      }
      if (Copy != Clause)
	*Changed = Copy;
      return TRUE;
    }
  }

  return FALSE;
}

void red_ExchangeClauses(CLAUSE *RedClause, CLAUSE *Copy, LIST *Result)
{
  if (*Copy) {
    *Result    = list_Cons(*RedClause,*Result);
    *RedClause = *Copy;
    *Copy      = (CLAUSE)NULL;
  }
}
     


BOOL red_SimpleStaticReductions(CLAUSE *RedClause, LIST* Result, BOOL Document)

{
  CLAUSE Copy;

#ifdef DEBUG
  if (!clause_IsClause(*RedClause)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_SimpleStaticReductions :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(*RedClause);
#endif

  Copy = (CLAUSE)NULL;

  if (flag_GetFlagValue(flag_RTAUT) && red_Tautology(*RedClause))
    return TRUE;

  if (flag_GetFlagValue(flag_ROBV)) {
    red_ObviousReductions(*RedClause,&Copy,Document);
    red_ExchangeClauses(RedClause,&Copy,Result);
  }

  if (flag_GetFlagValue(flag_RCON)) {
    red_Condensing(*RedClause, &Copy, Document);
    red_ExchangeClauses(RedClause,&Copy,Result);
  }

  return FALSE;    
}

  


BOOL red_StaticReductions(PROOFSEARCH Search, CLAUSE *Clause, CLAUSE *Subsumer,
			  LIST* Result, BOOL Document, NAT Mode)

{ 
  CLAUSE       Copy;
  BOOL         Redundant;
  SHARED_INDEX Index;

#ifdef DEBUG
  if (!clause_IsClause(*Clause)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_Staticreductions :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(*Clause);
#endif

  
  Index     = (red_OnlyWorkedOffMode(Mode) ? 
	       prfs_WorkedOffSharingIndex(Search) : prfs_UsableSharingIndex(Search));
  Copy      = (CLAUSE)NULL;
  Redundant = red_SimpleStaticReductions(Clause, Result, Document);

  if (Redundant)
    return Redundant;

  if (flag_GetFlagValue(flag_RRRE) &&
      red_RemoveRedundantEquations(*Clause, &Copy, Document, prfs_NonTrivClauseNumber(Search), 
				   (flag_GetFlagValue(flag_RRRE) == 2))) {
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE;
  }

  if (flag_GetFlagValue(flag_RFSUB)) {
    *Subsumer = red_ForwardSubsumption(*Clause, Index);
    if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
      return Redundant;
  }

  if (flag_GetFlagValue(flag_RFREW) && 
      red_RewriteRedClause(*Clause, Index, &Copy, prfs_LastBacktrackLevel(Search), Document)) {
    red_ExchangeClauses(Clause,&Copy,Result);
    Redundant = red_SimpleStaticReductions(Clause, Result, Document);
    if (Redundant)
      return Redundant;
    if (clause_IsEmptyClause(*Clause))
      return FALSE;
    if (flag_GetFlagValue(flag_RFSUB)) {
      *Subsumer = red_ForwardSubsumption(*Clause, Index);
      if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
	return Redundant;
    }
  }

  if (red_OnlyWorkedOffMode(Mode) && flag_GetFlagValue(flag_RSSI)) { 
    red_SortSimplification(prfs_DynamicSortTheory(Search), *Clause, &Copy, 
			   prfs_LastBacktrackLevel(Search), Document);
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE;    
  }

  if (flag_GetFlagValue(flag_RFCLR)) {
    red_ClauseReduceRedClause(*Clause, Index, &Copy, prfs_LastBacktrackLevel(Search), Document);
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE;
  }

  if (flag_GetFlagValue(flag_RUNC)) {
    red_UnitConflict(*Clause, Index, &Copy, prfs_LastBacktrackLevel(Search), Document);
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE;
  }

  if (red_OnlyWorkedOffMode(Mode) && flag_GetFlagValue(flag_RSST))
    Redundant = red_ClauseDeletion(prfs_StaticSortTheory(Search),*Clause);
  
#ifdef DEBUG
  clause_Check(*Clause);
#endif

  return Redundant;
}

BOOL red_SelectedStaticReductions(PROOFSEARCH Search, CLAUSE *Clause, 
				  CLAUSE *Subsumer, LIST* Result, BOOL Document, NAT Mode)

{ 
  CLAUSE       Copy;
  BOOL         Redundant,Rewritten,Tried;
  SHARED_INDEX WoIndex,UsIndex;

#ifdef DEBUG
  if (!clause_IsClause(*Clause)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_SelectedStaticReductions:");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(*Clause);
#endif

  WoIndex = (SHARED_INDEX)NULL;
  UsIndex = (SHARED_INDEX)NULL;
  if (red_WorkedOffMode(Mode))
    WoIndex   = prfs_WorkedOffSharingIndex(Search);
  if (red_UsableMode(Mode))
    UsIndex   = prfs_UsableSharingIndex(Search);
  Copy      = (CLAUSE)NULL;
  Redundant = red_SimpleStaticReductions(Clause, Result, Document);

  if (Redundant)
    return Redundant;

  if (flag_GetFlagValue(flag_RRRE) &&
      red_RemoveRedundantEquations(*Clause, &Copy, Document, prfs_NonTrivClauseNumber(Search), 
				   (flag_GetFlagValue(flag_RRRE) == 2))) {
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE;
  }

  if (flag_GetFlagValue(flag_RFSUB)) {
    *Subsumer = (CLAUSE)NULL;
    if (WoIndex)
      *Subsumer = red_ForwardSubsumption(*Clause, WoIndex);
    if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
      return Redundant;
    if (UsIndex)
      *Subsumer = red_ForwardSubsumption(*Clause, UsIndex);
    if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
      return Redundant;
  }

  Rewritten = flag_GetFlagValue(flag_RFREW);
  Tried     = FALSE;
  while (Rewritten) {
    if (WoIndex &&
	red_RewriteRedClause(*Clause, WoIndex, &Copy, 
			     prfs_LastBacktrackLevel(Search), Document)) {
      Rewritten = TRUE;
      red_ExchangeClauses(Clause,&Copy,Result);
      Redundant = red_SimpleStaticReductions(Clause, Result, Document);
      if (Redundant)
	return Redundant;
      if (clause_IsEmptyClause(*Clause))
	return FALSE;
      if (flag_GetFlagValue(flag_RFSUB)) {
	*Subsumer = (CLAUSE)NULL;
	*Subsumer = red_ForwardSubsumption(*Clause, WoIndex);
	if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
	  return Redundant;
	if (UsIndex)
	  *Subsumer = red_ForwardSubsumption(*Clause, UsIndex);
	if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
	  return Redundant;
      }
    }
    else
      Rewritten = FALSE;

    if ((!Tried || Rewritten) && 
	UsIndex &&
	red_RewriteRedClause(*Clause, UsIndex, &Copy, prfs_LastBacktrackLevel(Search), Document)) {
      Rewritten = TRUE;
      red_ExchangeClauses(Clause,&Copy,Result);
      Redundant = red_SimpleStaticReductions(Clause, Result, Document);
      if (Redundant)
	return Redundant;
      if (clause_IsEmptyClause(*Clause))
	return FALSE;
      if (flag_GetFlagValue(flag_RFSUB)) {
	*Subsumer = (CLAUSE)NULL;
	if (WoIndex)
	  *Subsumer = red_ForwardSubsumption(*Clause, WoIndex);
	if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
	  return Redundant;
	*Subsumer = red_ForwardSubsumption(*Clause, UsIndex);
	if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
	  return Redundant;
      }
    }
    else
      Rewritten = FALSE;
    Tried = TRUE;
  }

  if (flag_GetFlagValue(flag_RSSI)) { 
    red_SortSimplification(prfs_DynamicSortTheory(Search), *Clause, &Copy, 
			   prfs_LastBacktrackLevel(Search), Document);
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE; 
  }
     
  if (flag_GetFlagValue(flag_RFCLR)) {
    if (WoIndex)
      red_ClauseReduceRedClause(*Clause, WoIndex, &Copy, prfs_LastBacktrackLevel(Search), Document);
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE;
    if (UsIndex)
      red_ClauseReduceRedClause(*Clause, UsIndex, &Copy, prfs_LastBacktrackLevel(Search), Document);
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE;
  }

  if (flag_GetFlagValue(flag_RUNC)) {
    if (WoIndex)
      red_UnitConflict(*Clause, WoIndex, &Copy, prfs_LastBacktrackLevel(Search), Document);
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE;
    if (UsIndex)
      red_UnitConflict(*Clause, UsIndex, &Copy, prfs_LastBacktrackLevel(Search), Document);
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE;
  }

  if (flag_GetFlagValue(flag_RSST))
    Redundant = red_ClauseDeletion(prfs_StaticSortTheory(Search),*Clause);
  
#ifdef DEBUG
  clause_Check(*Clause);
#endif

  return Redundant;
}

CLAUSE red_ReductionOnDerivedClause(PROOFSEARCH Search, CLAUSE Clause, BOOL Document, NAT Mode)

{
  CLAUSE RedClause;
  LIST   Redundant;

#ifdef DEBUG
  cont_SaveState();
#endif

  Redundant = list_Nil();
  RedClause = (CLAUSE)NULL;

  if (red_StaticReductions(Search,&Clause,&RedClause,&Redundant,Document,Mode)) { 
    red_HandleRedundantDerivedClauses(Search,Redundant,Clause,Document);
    list_Delete(Redundant);
    if (RedClause &&
	prfs_SplitLevelCondition(clause_SplitLevel(RedClause),clause_SplitLevel(Clause),
				 prfs_LastBacktrackLevel(Search))) {
      split_KeepClauseAtLevel(Search,Clause,clause_SplitLevel(RedClause));
    }
    else
      if (Document)
	prfs_InsertDocProofClause(Search,Clause);
      else
	clause_Delete(Clause);
    Clause = (CLAUSE)NULL;
  }
  else {
    red_HandleRedundantDerivedClauses(Search,Redundant,Clause,Document);
    list_Delete(Redundant);
  }

#ifdef DEBUG
  cont_CheckState();
#endif

  return Clause;
}

CLAUSE red_CompleteReductionOnDerivedClause(PROOFSEARCH Search, CLAUSE Clause, BOOL Document, NAT Mode)

{
  CLAUSE RedClause;
  LIST   Redundant;

#ifdef DEBUG
  cont_SaveState();
#endif

  Redundant = list_Nil();
  RedClause = (CLAUSE)NULL;

  if (red_SelectedStaticReductions(Search,&Clause,&RedClause,&Redundant,Document,Mode)) { 
    red_HandleRedundantDerivedClauses(Search,Redundant,Clause,Document);
    list_Delete(Redundant);
    if (RedClause && 
	prfs_SplitLevelCondition(clause_SplitLevel(RedClause),clause_SplitLevel(Clause),
				 prfs_LastBacktrackLevel(Search))) {
      split_KeepClauseAtLevel(Search,Clause,clause_SplitLevel(RedClause));
    }
    else
      if (Document)
	prfs_InsertDocProofClause(Search,Clause);
      else
	clause_Delete(Clause);
    Clause = (CLAUSE)NULL;
  }
  else {
    red_HandleRedundantDerivedClauses(Search,Redundant,Clause,Document);
    list_Delete(Redundant);
  }

#ifdef DEBUG
  cont_CheckState();
#endif

  return Clause;
}


LIST red_BackReduction(PROOFSEARCH Search, CLAUSE Clause, BOOL Document, NAT Mode)

{
  LIST Result, Redundant;

#ifdef DEBUG
  cont_SaveState();
#endif

  Result    = list_Nil();

  if (flag_GetFlagValue(flag_RBSUB)) {
    Redundant = list_Nil();
    if (red_WorkedOffMode(Mode))
      Redundant = red_BackSubsumption(Clause,prfs_WorkedOffSharingIndex(Search));
    if (red_UsableMode(Mode))
      Redundant = list_Nconc(Redundant,red_BackSubsumption(Clause,prfs_UsableSharingIndex(Search)));
    red_HandleRedundantIndexedClauses(Search, Redundant, Clause, Document);
    list_Delete(Redundant);
  }

  if (flag_GetFlagValue(flag_RBCLR)) {
    Redundant = list_Nil();
    if (red_WorkedOffMode(Mode))
      Redundant = red_BackClauseReduction(Clause,prfs_WorkedOffSharingIndex(Search),&Result,Document);
    if (red_UsableMode(Mode))
      Redundant = list_Nconc(Redundant,
			     red_BackClauseReduction(Clause,prfs_UsableSharingIndex(Search),
						     &Result,Document));
    red_HandleRedundantIndexedClauses(Search, Redundant, Clause, Document);
    list_Delete(Redundant);
  }

  if (flag_GetFlagValue(flag_RBREW)) {
    Redundant = list_Nil();
    if (red_WorkedOffMode(Mode))
      Redundant = red_BackRewriting(Clause,prfs_WorkedOffSharingIndex(Search),&Result,Document);
    if (red_UsableMode(Mode))
      Redundant = list_Nconc(Redundant,
			     red_BackRewriting(Clause,prfs_UsableSharingIndex(Search),&Result,Document));
    red_HandleRedundantIndexedClauses(Search, Redundant, Clause, Document);
    list_Delete(Redundant);
  }

#ifdef DEBUG
  cont_CheckState();
#endif

  return Result;
}

LIST red_CompleteReductionOnDerivedClauses(PROOFSEARCH Search, LIST DerivedClauses, BOOL Document, 
					   NAT Mode, int Bound, NAT BoundMode, int *BoundApplied)

{
  LIST   EmptyClauses,NewClauses,Scan;
  NAT    ClauseBound;
  CLAUSE Clause;

#ifdef DEBUG
  cont_SaveState();
#endif

  EmptyClauses   = list_Nil();
  DerivedClauses = clause_ListSortWeighed(DerivedClauses);
  ClauseBound    = 0;

  while (!list_Empty(DerivedClauses)) {
    Clause     = list_NCar(&DerivedClauses);
    if (prfs_SplitStackEmpty(Search))        
      Clause = red_CompleteReductionOnDerivedClause(Search, Clause, Document, Mode);

    if (Clause && BoundMode != 0 && Bound != -1 && 
	!clause_IsFromInput(Clause) && !clause_IsFromSplitting(Clause)) {
      switch (BoundMode) {
      case 1:
	ClauseBound = clause_Weight(Clause);
	break;
      case 2:
	ClauseBound = clause_ComputeTermDepth(Clause);
	break;
      default:
	misc_StartUserErrorReport();
	misc_UserErrorReport("\n\t Error while appying bound restrictions:");
	misc_UserErrorReport("\n You selected an unknown bound mode.\n");
	misc_FinishUserErrorReport();
      }
      if (ClauseBound > Bound) {
	if (flag_GetFlagValue(flag_PBDC)) {
	  printf("\nDeleted by bound: "); 
	  clause_Print(Clause);
	}
	clause_Delete(Clause);
	if (*BoundApplied == -1 || ClauseBound < *BoundApplied)
	  *BoundApplied = ClauseBound;
	Clause = (CLAUSE)NULL;
      }
    }

    if (Clause != (CLAUSE)NULL                             
	&& !prfs_SplitStackEmpty(Search))                  
      Clause = red_CompleteReductionOnDerivedClause(Search, Clause, Document, Mode);
	
    if (Clause) {	
      prfs_IncKeptClauses(Search);
      if (flag_GetFlagValue(flag_PKEPT)) {
	printf("\nKept: "); 
	clause_Print(Clause);
      }
      if (clause_IsEmptyClause(Clause))
	EmptyClauses = list_Cons(Clause,EmptyClauses);
      else {
	NewClauses = red_BackReduction(Search,Clause,Document,Mode);
	prfs_IncDerivedClauses(Search, list_Length(NewClauses));
	if (flag_GetFlagValue(flag_PDER))
	  for(Scan=NewClauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	    printf("\nDerived: "); 
	    clause_Print(list_Car(Scan));
	  }
	NewClauses = split_ExtractEmptyClauses(NewClauses,&EmptyClauses);

	prfs_InsertUsableClause(Search,Clause);
	DerivedClauses = list_NNumberMerge(DerivedClauses,NewClauses,clause_Weight);
      }
    }
  }

#ifdef DEBUG
  cont_CheckState();
#endif

  return EmptyClauses;
}



static CLAUSE red_CDForwardSubsumer(CLAUSE RedCl, st_INDEX Index)

{
  TERM   Atom,AtomGen;
  LIST   LitScan;
  int    i,length;
  CLAUSE CandCl;

#ifdef DEBUG
  if (!clause_IsClause(RedCl)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_CDForwardSubsumer :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(RedCl);
#endif

  length  = clause_Length(RedCl);

  for (i = 0; i < length; i++) {
    Atom     = clause_GetLiteralAtom(RedCl, i);
    AtomGen  = st_ExistGen(cont_LeftContext(), Index, Atom);

    while (AtomGen) {
      for (LitScan = term_SupertermList(AtomGen); 
	   !list_Empty(LitScan); LitScan = list_Cdr(LitScan)) {
	CandCl = clause_LiteralOwningClause(list_Car(LitScan));

	if (clause_GetLiteral(CandCl,clause_FirstLitIndex()) == (LITERAL)list_Car(LitScan) &&
	    subs_Subsumes(CandCl, RedCl, clause_FirstLitIndex(), i)) {
	  st_CancelExistRetrieval();
	  return CandCl;
	}
      }
      AtomGen = st_NextCandidate();  
    }
  }
  return (CLAUSE)NULL;
}


static BOOL red_CDForwardSubsumption(CLAUSE RedClause, st_INDEX Index)

{ 
  BOOL   IsSubsumed;
  CLAUSE Subsumer;

#ifdef DEBUG
  if (!clause_IsClause(RedClause)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_CDForwardSubSumption :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(RedClause);
#endif
  IsSubsumed = FALSE;
  Subsumer   = red_CDForwardSubsumer(RedClause, Index);

  if (clause_Exists(Subsumer)) {
    IsSubsumed = TRUE;

    if (flag_GetFlagValue(flag_DOCSST) &&
	flag_GetFlagValue(flag_PSUB)) {
      printf("\nFSubsumption:"); 
      clause_Print(RedClause);
      printf(" by %d ",clause_Number(Subsumer));
    }
  }
  return IsSubsumed;
}


static void red_CDBackSubsumption(CLAUSE RedCl, LIST* UsListPt, LIST* WOListPt, st_INDEX Index)

{
  TERM   Atom,AtomInst;
  CLAUSE SubsumedCl;
  LIST   Scan, SubsumedList;

#ifdef DEBUG
  if (!clause_IsClause(RedCl)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_CDBackupSubSumption :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(RedCl);
#endif

  SubsumedList = list_Nil();

  if (!clause_IsEmptyClause(RedCl)) {
    Atom     = clause_GetLiteralAtom(RedCl, clause_FirstLitIndex());
    AtomInst = st_ExistInstance(cont_LeftContext(), Index, Atom);
    
    while(AtomInst) {
      for (Scan = term_SupertermList(AtomInst); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	SubsumedCl = clause_LiteralOwningClause(list_Car(Scan));
	if ((RedCl != SubsumedCl) &&
	    subs_Subsumes(RedCl, SubsumedCl, clause_FirstLitIndex(), 
			  clause_LiteralGetIndex(list_Car(Scan))) &&
	    !list_PointerMember(SubsumedList, SubsumedCl))
	  SubsumedList = list_Cons(SubsumedCl, SubsumedList);
      }
      AtomInst = st_NextCandidate();
    }
    
    for (Scan = SubsumedList; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      SubsumedCl = list_Car(Scan);

      if (flag_GetFlagValue(flag_DOCSST) &&  flag_GetFlagValue(flag_PSUB)) {
	printf("\nBSubsumption: "); 
	clause_Print(SubsumedCl);printf(" by %d ",clause_Number(RedCl));
      }
	     

      if (clause_GetFlag(SubsumedCl,WORKEDOFF)) {
	*WOListPt = list_PointerDeleteOneElement(*WOListPt, SubsumedCl);
      }else {
	*UsListPt = list_PointerDeleteOneElement(*UsListPt, SubsumedCl);
      }
      clause_DeleteFlatFromIndex(SubsumedCl, Index);
    }
    list_Delete(SubsumedList);
  }
}


static LIST red_CDDerivables(SORTTHEORY Theory, CLAUSE GivenClause)

{
  LIST ListOfDerivedClauses;

#ifdef DEBUG
  if (!(clause_IsClause(GivenClause))) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_CDDeriveables :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
  clause_Check(GivenClause);
#endif
  
  if (clause_HasTermSortConstraintLits(GivenClause))
    ListOfDerivedClauses = inf_ForwardSortResolution(GivenClause, sort_TheoryIndex(Theory), Theory, TRUE);
  else
    ListOfDerivedClauses = inf_ForwardEmptySort(GivenClause, sort_TheoryIndex(Theory), Theory, TRUE);
  
  return ListOfDerivedClauses;
}


static BOOL red_CDReduce(SORTTHEORY Theory, CLAUSE RedClause,  LIST *ApproxUsListPt, LIST *ApproxWOListPt, 
			 st_INDEX Index)

{
  CLAUSE Copy;

#ifdef DEBUG
  clause_Check(RedClause);
#endif

  Copy = (CLAUSE)NULL; 
  
  red_ObviousReductions(RedClause,&Copy,FALSE);
  red_SortSimplification(Theory, RedClause, &Copy, NAT_MAX, FALSE);
  
  if (clause_IsEmptyClause(RedClause))
    return FALSE;

  red_Condensing(RedClause, &Copy, FALSE);

  if (red_CDForwardSubsumption(RedClause, Index))
    return TRUE;
  else {			
    red_CDBackSubsumption(RedClause, ApproxUsListPt, ApproxWOListPt, Index);
    clause_InsertFlatIntoIndex(RedClause, Index);
    *ApproxUsListPt = list_Cons(RedClause, *ApproxUsListPt);
  }

#ifdef DEBUG
  clause_Check(RedClause);
  if (Copy != (CLAUSE)NULL) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_CDReduce :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
    
#endif

  return FALSE;
}


BOOL red_ClauseDeletion(SORTTHEORY Theory, CLAUSE RedClause)

{
  if (Theory != (SORTTHEORY)NULL) {
    CLAUSE   ConstraintClause, GivenClause;
    LIST     ApproxUsableList, ApproxWOList, EmptyClauses, ApproxDerivables, Scan;
    int      i,nc, Count, OldClauseCounter;
    st_INDEX Index;

#ifdef DEBUG
    if (!(clause_IsClause(RedClause))) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In red_ClauseDeletion :");
      misc_ErrorReport(" Illegal input.\n");
      misc_FinishErrorReport();
    }
    clause_Check(RedClause);
#endif

    if (clause_HasEmptyConstraint(RedClause) || !flag_GetFlagValue(flag_RSST))
      return FALSE;
 
    if (flag_GetFlagValue(flag_DOCSST)) {
      printf("\n\nClauseDeletion tried on: "); 
      clause_Print(RedClause);
    }

    Index             = st_IndexCreate();
    Scan              = list_Nil();
    nc                = clause_NumOfConsLits(RedClause);
    OldClauseCounter  = clause_Counter();

    

    for (i = clause_FirstLitIndex(); i < nc; i++)
      Scan = list_Cons(term_Copy(clause_LiteralAtom(
						    clause_GetLiteral(RedClause, i))), Scan);

    Scan             = list_NReverse(Scan);
    ConstraintClause = clause_Create(Scan, list_Nil(), list_Nil());
    list_Delete(Scan);
    clause_InitSplitData(ConstraintClause);
    clause_AddParentClause(ConstraintClause, clause_Number(RedClause));
    clause_AddParentLiteral(ConstraintClause, clause_FirstLitIndex());
    clause_SetFromClauseDeletion(ConstraintClause);
    clause_InsertFlatIntoIndex(ConstraintClause, Index);
    ApproxUsableList = list_List(ConstraintClause);
    ApproxWOList     = list_Nil();

    

    

    EmptyClauses = list_Nil();
    Count        = 0;

    if (flag_GetFlagValue(flag_DOCSST)) {
      printf("\n*************** ClauseDeletion subproof: ***************");
      printf("\nThe usable list:\n"); clause_ListPrint(ApproxUsableList);
      printf("\nThe worked-off list:\n"); clause_ListPrint(ApproxWOList);
      
    }
    while (!list_Empty(ApproxUsableList) && list_Empty(EmptyClauses)) {
      GivenClause      = list_Car(ApproxUsableList); 
      clause_SetFlag(GivenClause,WORKEDOFF);

      if (flag_GetFlagValue(flag_DOCSST)) {
	printf("\n\tSubproof Given clause: "); 
	clause_Print(GivenClause); fflush(stdout);
      }
      ApproxWOList     = list_Cons(GivenClause, ApproxWOList);
      ApproxUsableList = list_PointerDeleteOneElement(ApproxUsableList,GivenClause);
      ApproxDerivables = red_CDDerivables(Theory,GivenClause);
      ApproxDerivables = split_ExtractEmptyClauses(ApproxDerivables, &EmptyClauses);
    
      if (!list_Empty(EmptyClauses)) { 
	if (flag_GetFlagValue(flag_DOCSST)) {
	  printf("\nClauseDeletion not possible: "); 
	  clause_Print(list_Car(EmptyClauses));
	}
	clause_DeleteClauseList(ApproxDerivables); 
	ApproxDerivables = list_Nil();
      }
      else {
	CLAUSE DerClause;
	for (Scan = ApproxDerivables; !list_Empty(Scan) && list_Empty(EmptyClauses);
	     Scan = list_Cdr(Scan)) {
	  DerClause = (CLAUSE)list_Car(Scan);
	  if (red_CDReduce(Theory, DerClause, &ApproxUsableList, &ApproxWOList, Index)) 
	    clause_Delete(DerClause);
	  else{
	    Count++;
	    if (flag_GetFlagValue(flag_DOCSST)) {
	      printf("\n"); clause_Print(DerClause);
	    }
	    if (clause_IsEmptyClause(DerClause))
	      EmptyClauses = list_Cons(DerClause,EmptyClauses);
	  }
	  list_Rplaca(Scan,(CLAUSE)NULL);
	}

	if (!list_Empty(EmptyClauses)) {
	  if (flag_GetFlagValue(flag_DOCSST)) {
	    printf(" ClauseDeletion not possible!");
	    clause_Print(list_Car(EmptyClauses));
	  }
	  clause_DeleteClauseList(ApproxDerivables);    
	  ApproxDerivables = list_Nil();
	}
	else {
	  list_Delete(ApproxDerivables); 
	  ApproxDerivables = list_Nil();
	}
      }
    }

    if (!list_Empty(EmptyClauses)) {
      if (flag_GetFlagValue(flag_DOCSST)) {
	printf("\nClauseDeletion failed, constraint solvable.");
	printf("\n************ ClauseDeletion subproof finished. ************\n");
      }
    }
    else
      if (flag_GetFlagValue(flag_PSST)) {
	printf("\nClauseDeletion: "); 
	clause_Print(RedClause);
      }

    
    clause_DeleteClauseListFlatFromIndex(ApproxUsableList, Index);
    clause_DeleteClauseListFlatFromIndex(ApproxWOList, Index);
    st_IndexDelete(Index);
    clause_SetCounter(OldClauseCounter);

    if (!list_Empty(EmptyClauses)) {
      list_DeleteWithElement(EmptyClauses,clause_Delete);
      return FALSE;
    } 

    return TRUE;

#ifdef DEBUG
    clause_Check(RedClause);
#endif
  }
  return FALSE;
}

LIST red_SatUnit(PROOFSEARCH Search, LIST ClauseList, BOOL Document) 

{
  CLAUSE Given,Clause;
  LIST   Scan, Derivables, EmptyClauses, BackReduced;
  NAT    n, Derived;

  Derived      = flag_GetFlagValue(flag_CNFPROOFSTEPS);
  EmptyClauses = list_Nil();

  ClauseList = clause_ListSortWeighed(ClauseList);
  
  while (!list_Empty(ClauseList) && list_Empty(EmptyClauses)) {
    Given = (CLAUSE)list_NCar(&ClauseList);
    Given = red_ReductionOnDerivedClause(Search, Given, Document, red_USABLE);
    if (Given) {
      if (clause_IsEmptyClause(Given))
	EmptyClauses = list_List(Given);
      else {
	BackReduced = red_BackReduction(Search, Given, Document, red_USABLE);

	if (Derived != 0) {
	  Derivables = inf_BoundedDepthUnitResolution(Given, prfs_UsableSharingIndex(Search),FALSE);
	  n          = list_Length(Derivables);
	  if (n > Derived)
	    Derived = 0;
	  else
	    Derived = Derived - n;
	}
	else 
	  Derivables = list_Nil();

	Derivables  = list_Nconc(BackReduced,Derivables);
	Derivables  = split_ExtractEmptyClauses(Derivables, &EmptyClauses);
    
	prfs_InsertUsableClause(Search, Given);

	for(Scan = Derivables; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	  Clause     = (CLAUSE)list_Car(Scan);
	  clause_SetDepth(Clause,0);
	}
	ClauseList = list_Nconc(ClauseList,Derivables);
	Derivables = list_Nil();
      }
    }
  }
  for(Scan = ClauseList; !list_Empty(Scan); Scan = list_Cdr(Scan))
      prfs_InsertUsableClause(Search, list_Car(Scan));
  list_Delete(ClauseList);
  return EmptyClauses;
}

CLAUSE red_SpecialInputReductions(CLAUSE Clause)

{
  int     i,end;
  LIST    Indexes;
  TERM    Atom;

#ifdef DEBUG
  clause_Check(Clause);
#endif

  Indexes = list_Nil();
  end     = clause_LastAntecedentLitIndex(Clause);
    
  for (i = clause_FirstConstraintLitIndex(Clause); i <= end; i++) {
    Atom = clause_LiteralAtom(clause_GetLiteral(Clause,i));
    if (fol_IsTrue(Atom)) 
      Indexes = list_Cons((POINTER)i,Indexes);
  }

  end    = clause_LastSuccedentLitIndex(Clause);
    
  for (i = clause_FirstSuccedentLitIndex(Clause); i <= end; i++) {
    Atom = clause_LiteralAtom(clause_GetLiteral(Clause,i));
    if (fol_IsFalse(Atom)) 
      Indexes = list_Cons((POINTER)i,Indexes);
  }
  
  clause_DeleteLiterals(Clause,Indexes);
  list_Delete(Indexes);

  return Clause;
}


LIST red_ReduceInput(PROOFSEARCH Search, LIST ClauseList, BOOL Document)

{
  CLAUSE Given;
  LIST   Scan, EmptyClauses, BackReduced;

  EmptyClauses = list_Nil();
  ClauseList   = clause_ListSortWeighed(list_Copy(ClauseList));
  ClauseList   = split_ExtractEmptyClauses(ClauseList, &EmptyClauses);
  
  while (!list_Empty(ClauseList) && list_Empty(EmptyClauses) &&
	 (flag_GetFlagValue(flag_TIMELIMIT) < 0 || 
	  flag_GetFlagValue(flag_TIMELIMIT) > clock_GetSeconds(clock_OVERALL))) {
    Given = (CLAUSE)list_NCar(&ClauseList);
#ifdef DEBUG
  if (!clause_IsClause(Given)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In red_ReduceInput :");
    misc_ErrorReport(" Illegal input.\n");
    misc_FinishErrorReport();
  }
#endif 
    Given = red_SpecialInputReductions(Given);
    Given = red_ReductionOnDerivedClause(Search, Given, Document, red_USABLE);
    if (Given) {
      prfs_IncKeptClauses(Search);
      if (clause_IsEmptyClause(Given))
	EmptyClauses = list_List(Given);
      else {
	BackReduced = red_BackReduction(Search, Given, Document, red_USABLE);
	prfs_IncDerivedClauses(Search, list_Length(BackReduced));
	BackReduced = split_ExtractEmptyClauses(BackReduced, &EmptyClauses);
	prfs_InsertUsableClause(Search, Given);
	BackReduced = clause_ListSortWeighed(BackReduced);
	ClauseList = list_NNumberMerge(ClauseList,BackReduced,clause_Weight);
      }
    }
  }
  for(Scan = ClauseList; !list_Empty(Scan); Scan = list_Cdr(Scan))
    prfs_InsertUsableClause(Search, list_Car(Scan));
  list_Delete(ClauseList);
  return EmptyClauses;
}


LIST red_SatInput(PROOFSEARCH Search, BOOL Document) 

{
  CLAUSE Given;
  LIST   Scan, ClauseList, Derivables, EmptyClauses;
  int    n;

  EmptyClauses = list_Nil();
  ClauseList   = list_Nil();
  Scan         = prfs_UsableClauses(Search);
  n            = list_Length(Scan);

  while(!list_Empty(Scan) &&
	n > 0 &&
	(flag_GetFlagValue(flag_TIMELIMIT) < 0 || 
	 flag_GetFlagValue(flag_TIMELIMIT) > clock_GetSeconds(clock_OVERALL))) {
    Given = (CLAUSE)list_Car(Scan);
    if (clause_GetFlag(Given,CONCLAUSE)) {
      Derivables = inf_BoundedDepthUnitResolution(Given, prfs_UsableSharingIndex(Search), FALSE);
      n         -= list_Length(Derivables);
      ClauseList = list_Nconc(Derivables,ClauseList);
    }
    Scan = list_Cdr(Scan);
  }
  
  prfs_IncDerivedClauses(Search, list_Length(ClauseList));
  EmptyClauses = red_ReduceInput(Search,ClauseList,Document);
  list_Delete(ClauseList);
  ClauseList   = list_Nil();
  if (list_Empty(EmptyClauses)) {
    Scan=prfs_UsableClauses(Search);
    while (!list_Empty(Scan) &&
	   n > 0 &&
	   (flag_GetFlagValue(flag_TIMELIMIT) < 0 || 
	    flag_GetFlagValue(flag_TIMELIMIT) > clock_GetSeconds(clock_OVERALL))) {
      Given = (CLAUSE)list_Car(Scan);
      if (clause_GetFlag(Given,CONCLAUSE) &&  clause_IsFromInput(Given)) {
	Derivables = inf_BoundedDepthUnitResolution(Given, prfs_UsableSharingIndex(Search), TRUE);
	n         -= list_Length(Derivables);
	ClauseList = list_Nconc(Derivables,ClauseList);
      }
      Scan = list_Cdr(Scan);
    }
    prfs_IncDerivedClauses(Search, list_Length(ClauseList));
    EmptyClauses = red_ReduceInput(Search,ClauseList,Document);
    list_Delete(ClauseList);
  }
  return EmptyClauses;
}

void red_CheckSplitSubsumptionCondition(PROOFSEARCH Search)

{
  LIST   Scan1,Scan2;
  CLAUSE Clause;

  for (Scan1=prfs_SplitStack(Search);!list_Empty(Scan1);Scan1=list_Cdr(Scan1))
    for (Scan2=prfs_SplitDeletedClauses(list_Car(Scan1));!list_Empty(Scan2);Scan2=list_Cdr(Scan2)) {
      Clause = (CLAUSE)list_Car(Scan2);
      if (!red_ForwardSubsumer(Clause, prfs_WorkedOffSharingIndex(Search)) &&
	  !red_ForwardSubsumer(Clause, prfs_UsableSharingIndex(Search)) &&
	  !red_ClauseDeletion(prfs_StaticSortTheory(Search),Clause)) {
	printf("\n No clause found implying ");
	clause_Print(Clause);
	printf("\n Current Split: ");
	prfs_PrintSplit(list_Car(Scan1));
	misc_DumpCore();
      }
    }
}
