




 





















#include "symbol.h"

SIGNATURE *symbol_SIGNATURE;
PRECEDENCE symbol_PRECEDENCE,symbol_DEFAULTPRECEDENCE;

SYMBOL symbol_STANDARDVARCOUNTER;
SYMBOL symbol_INDEXVARCOUNTER;

const int symbol_MASK         = 3;  
const int symbol_TYPEMASK     = 3;  
const int symbol_STATMASK     = 4;  
const int symbol_TYPESTATMASK = 7;  


const int symbol_TYPEBITS      = 2;
const int symbol_STATBITS      = 1;
const int symbol_TYPESTATBITS  = 3;

const int symbol_SIGTYPES = 4;

const char symbol_SKFNAME[]="skf";
const char symbol_SKCNAME[]="skc";
const char symbol_SKPNAME[]="SkP";
const char symbol_SKANAME[]="SkC";
const int  symbol_SKLENGTH = 3;

BOOL symbol_HASSIGNATURE;

int  symbol_ACTINDEX;
int  symbol_ACTSKOLEMFINDEX;
int  symbol_ACTSKOLEMCINDEX;
int  symbol_ACTSKOLEMAINDEX;
int  symbol_ACTSKOLEMPINDEX;
int  symbol_ORDERING;
LIST symbol_FREEDSYMBOLS;
char *symbol_VARSTRING;

SYMBOL symbol_CONTEXT[symbol__MAXSIGNATURE];









int symbol_GetIncreasedOrderingCounter(void)
{
  return symbol_ORDERING++;
}

NAT symbol_MaxStringLength()

{
  int       Index;
  NAT       Result,Length;
  SIGNATURE Entry;

  Result = 0;
  
  for(Index=1;Index<symbol_ACTINDEX;Index++)
    if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
      Entry  = symbol_Signature(Index);
      Length = Entry->length;
      if (Length > Result)
	Result = Length;
    }

  return Result;
}


static SYMBOL symbol_SignatureCreate(char* String, int Type,
				     int Arity, int Status)

{
  SIGNATURE Entry;
  
#ifdef DEBUG
  if (!symbol_HASSIGNATURE) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In symbol_SignatureCreate: Module was initialized"
		     "with no signature.\n");
    misc_FinishErrorReport();
  } 
  if (Type < 0 || Type >= symbol_SIGTYPES) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In symbol_SignatureCreate: Illegal input.\n");
    misc_FinishErrorReport();
  }
#endif

  if ((int)symbol_ACTINDEX >= symbol__MAXSIGNATURE &&
      list_Empty(symbol_FREEDSYMBOLS)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In symbol_SignatureCreate: No more symbols available.\n");
    misc_FinishUserErrorReport();
  }
  
  if (strlen(String)>=symbol__SYMBOLMAXLEN) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\n In symbol_SignatureCreate: String too long.\n");
    misc_FinishUserErrorReport();
  }
    
  Entry              = symbol_GetSignature();
  Entry->weight      = 1;
  symbol_PRECEDENCE[symbol_ACTINDEX] = symbol_ORDERING++;
  Entry->props       = 0;
  Entry->name        = String;
  Entry->length      = strlen(String);
  Entry->arity       = Arity;
  Entry->generatedBy = list_Nil();
  
  if (list_Empty(symbol_FREEDSYMBOLS)) {
    Entry->info = symbol_SignatureSymbol(symbol_ACTINDEX, Type, Status);
    symbol_SetSignature(symbol_ACTINDEX++, Entry);
  }
  else {
    int Index;
    
    Index               = (int)list_Car(symbol_FREEDSYMBOLS);
    symbol_FREEDSYMBOLS = list_PointerDeleteElement(symbol_FREEDSYMBOLS,
						    (POINTER)Index);
    Entry->info = symbol_SignatureSymbol(Index,  Type, Status);
    symbol_SetSignature(Index, Entry);
  }
  
  return Entry->info;
}


SYMBOL symbol_CreateFunction(const char* String, int Arity, int Status)

{
  if (Arity==0)
    return symbol_SignatureCreate(
	  strcpy(memory_Malloc(symbol__SYMBOLMAXLEN),String), 
	  symbol_CONSTANT, 
          Arity,
	  Status);
  else
    return symbol_SignatureCreate(
	  strcpy(memory_Malloc(symbol__SYMBOLMAXLEN),String), 
	  symbol_FUNCTION, 
          Arity,
	  Status);
}

SYMBOL symbol_CreateSkolemFunction(int Arity)

{
  char   newname[10];
  SYMBOL Result;

  Result = (SYMBOL)NULL;

  while (Result == (SYMBOL)NULL) {
    if (Arity==0)
      sprintf(newname,"%s%d",symbol_SKCNAME,symbol_ACTSKOLEMCINDEX++);
    else
      sprintf(newname,"%s%d",symbol_SKFNAME,symbol_ACTSKOLEMFINDEX++);
    if (symbol_Lookup(newname) == 0)
      Result = symbol_CreateFunction(newname, Arity,symbol_STATLEX);
  }

  symbol_AddProperty(Result,SKOLEM);

  return Result;
}


SYMBOL symbol_CreatePredicate(const char* String, int Arity, int Status)

{ 
  return symbol_SignatureCreate(
	    strcpy(memory_Malloc(symbol__SYMBOLMAXLEN), String),
	    symbol_PREDICATE, 
            Arity,
	    Status);
}

SYMBOL symbol_CreateSkolemPredicate(int Arity)

{ 
  char   newname[10];
  SYMBOL Result;

  Result = (SYMBOL)NULL;

  while (Result == (SYMBOL)NULL) {
    if (Arity==0)
      sprintf(newname,"%s%d",symbol_SKANAME,symbol_ACTSKOLEMAINDEX++);
    else
      sprintf(newname,"%s%d",symbol_SKPNAME,symbol_ACTSKOLEMPINDEX++);
    if (symbol_Lookup(newname) == 0)
      Result = symbol_CreatePredicate(newname, Arity, symbol_STATLEX);
  }

  return Result;
}


SYMBOL symbol_CreateJunctor(const char* String, int Arity, int Status)

{ 
  return symbol_SignatureCreate(
            strcpy(memory_Malloc(symbol__SYMBOLMAXLEN), String),
	    symbol_JUNCTOR, 
            Arity,
	    Status);
}


BOOL symbol_IsSymbol(SYMBOL Symbol)

{
  return 
    (!symbol_HASSIGNATURE ||
     ((!symbol_Equal(Symbol, symbol__NULL)) &&
      ((symbol_IsVariable(Symbol) && Symbol<symbol_MaxVars()) || 
       (symbol_IsSignature(Symbol) && symbol_Index(Symbol)<symbol_ACTINDEX))));
}


void symbol_Delete(SYMBOL Symbol)

{
#ifdef DEBUG
  if (!symbol_HASSIGNATURE) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In symbol_Delete: Module was initialized without signature.\n");
    misc_FinishErrorReport();
  }
  if (!symbol_IsSymbol(Symbol)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In symbol_Delete: Illegal input.\n");
    misc_FinishErrorReport();
  }
#endif

  if (!symbol_IsVariable(Symbol)) {
    int       Index;
    SIGNATURE Entry;
    
    Index               = symbol_Index(Symbol);
    symbol_FREEDSYMBOLS = list_Cons((POINTER)Index,symbol_FREEDSYMBOLS);
    Entry               = symbol_Signature(Index);
    symbol_FreeSignature(Entry);
  }
}

LIST symbol_GetAllSymbols(void)

{
  LIST Result;

  Result = list_Nil();

  if (symbol_HASSIGNATURE) {
    int Index;
    
    for(Index=1;Index<symbol_ACTINDEX;Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
	Result = list_Cons((POINTER)symbol_GetSigSymbol(Index), Result);
      }
  }
  
  return Result;
}

    
void symbol_FreeAllSymbols(void)

{
  if (symbol_HASSIGNATURE) {
    int Index;
    
    for(Index=1;Index<symbol_ACTINDEX;Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
	symbol_FreeSignature(symbol_Signature(Index));
      }
    
    memory_Free(symbol_SIGNATURE, sizeof(SIGNATURE) * symbol__MAXSIGNATURE);
    memory_Free(symbol_DEFAULTPRECEDENCE, sizeof(int) * symbol__MAXSIGNATURE);
  }
  
  memory_Free(symbol_VARSTRING, symbol__SYMBOLMAXLEN);
  list_Delete(symbol_FREEDSYMBOLS);
}

void symbol_Init(BOOL Signature)

{ 

  symbol_ResetSkolemIndex();
  symbol_ContextClean();

  if (Signature) {
    symbol_SIGNATURE          = (SIGNATURE*)memory_Malloc(sizeof(SIGNATURE)   * symbol__MAXSIGNATURE);
    symbol_DEFAULTPRECEDENCE  = (PRECEDENCE)memory_Malloc(sizeof(int) * symbol__MAXSIGNATURE);
    symbol_PRECEDENCE         = symbol_DEFAULTPRECEDENCE;
  }

  symbol_STANDARDVARCOUNTER  = symbol_GetInitialStandardVarCounter();
  symbol_INDEXVARCOUNTER     = symbol_GetInitialIndexVarCounter();
  
  symbol_ACTINDEX            = 1;
  symbol_ORDERING            = 1;
  symbol_VARSTRING           = memory_Malloc(symbol__SYMBOLMAXLEN);
  
  symbol_HASSIGNATURE        = Signature;
  symbol_FREEDSYMBOLS        = list_Nil();
}

void symbol_ReinitGenericNameCounters()

{
  if (symbol_HASSIGNATURE) {
    int          Index,Num;
    SYMBOL       Info;
    SIGNATURE    Entry;
    char         *Name,*Subname;
  
    for(Index=1;Index<symbol_ACTINDEX;Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
	Entry  = symbol_Signature(Index);
	Info   = Entry->info;
	Name   = Entry->name;
	if (strlen(Name)>symbol_SKLENGTH) {
	  Subname = Name + symbol_SKLENGTH;
	
	  switch (symbol_Type(Info)) {
	  
	  case symbol_CONSTANT:
	    if (strncmp(Name,symbol_SKCNAME,symbol_SKLENGTH) == 0 && 
		misc_StringIsNumber(Subname)) {
	      Num = atoi(Subname);
	      if (Num >= symbol_ACTSKOLEMCINDEX)
		symbol_ACTSKOLEMCINDEX = Num + 1;
	    }
	    break;
	  case symbol_FUNCTION:
	    if (strncmp(Name,symbol_SKFNAME,symbol_SKLENGTH) == 0  &&
		misc_StringIsNumber(Subname)) {
	      Num = atoi(Subname);
	      if (Num >= symbol_ACTSKOLEMFINDEX)
		symbol_ACTSKOLEMFINDEX = Num + 1;
	    }
	    break;
	  case symbol_PREDICATE:
	    if (Entry->arity == 0) {
	      if (strncmp(Name,symbol_SKANAME,symbol_SKLENGTH) == 0  &&
		  misc_StringIsNumber(Subname)) {
		Num = atoi(Subname);
		if (Num >= symbol_ACTSKOLEMAINDEX)
		  symbol_ACTSKOLEMAINDEX = Num + 1;
	      }
	    }
	    else {
	      if (strncmp(Name,symbol_SKPNAME,symbol_SKLENGTH) == 0  &&
		  misc_StringIsNumber(Subname)) {
		Num = atoi(Subname);
		if (Num >= symbol_ACTSKOLEMPINDEX)
		  symbol_ACTSKOLEMPINDEX = Num + 1;
	      }
	    }

	  }
	}
      }
  }
}

  








SYMBOL symbol_Lookup(const char* String)

{
  if (symbol_HASSIGNATURE) {
    int Index;
    for(Index = 1; Index < symbol_ACTINDEX; Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS, (POINTER)Index)) {
	if (strcmp(String, symbol_Signature(Index)->name) == 0)
	  return symbol_Signature(Index)->info;
      }
  }
  
  return 0;
}










void symbol_LowerSignature()

{
  int       Index;
  SIGNATURE Entry;
  SYMBOL    Info;
  char*     String;

  for(Index=1;Index<symbol_ACTINDEX;Index++)
    if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
      Entry  = symbol_Signature(Index);
      Info   = Entry->info;
	
      if (symbol_IsPredicate(Info) || symbol_IsFunction(Info)) {
	String = Entry->name;
	if ('A' < String[0] && String[0] < 'Z') {
	  char* New;
	  New = (char *)memory_Malloc(symbol__SYMBOLMAXLEN);
	  strcpy(&(New[2]),String);
	  New[0] = 115;                   
	  New[1] = 115;
	  Entry->name = New;
	  memory_Free(String,symbol__SYMBOLMAXLEN);
	}
      }
    }
}

void symbol_Dump(void)

{
  if (symbol_HASSIGNATURE) {
    int Index;
    fputs("Dump:", stdout);
    for(Index = 1; Index < symbol_ACTINDEX; Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
	printf("\n\t %4d:%4d:%4d:%4d:%4d:%s:%d",
	       Index,
	       symbol_Signature(Index)->info,
	       symbol_Signature(Index)->weight,
	       symbol_PRECEDENCE[Index],
	       symbol_Signature(Index)->props,
	       symbol_Signature(Index)->name,
	       symbol_Signature(Index)->length);
      }
  }
}


void symbol_PrintPrecedence(void)

{
  SIGNATURE ActSig;

  if (symbol_HASSIGNATURE) {
    LIST Symbols,Scan;
    int Index;
    Symbols = list_Nil();
    for(Index = 1; Index < symbol_ACTINDEX; Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index) &&
	  (symbol_IsPredicate(symbol_Signature(Index)->info) ||
	   symbol_IsFunction(symbol_Signature(Index)->info))) 
	Symbols = list_Cons((POINTER)(symbol_Signature(Index)->info),Symbols);
    Symbols = list_Sort(Symbols, symbol_PrecedenceGreater);
    for(Scan = Symbols; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      ActSig = symbol_Signature(symbol_Index((SYMBOL)list_Car(Scan)));
      if (list_Empty(list_Cdr(Scan)))
	printf(" %s", ActSig->name);
      else
	printf(" %s >", ActSig->name);
    }
    list_Delete(Symbols);
  }
}

void symbol_FPrintPrecedence(FILE *File)

{
  SIGNATURE ActSig;

  if (symbol_HASSIGNATURE) {
    LIST Symbols,Scan;
    int Index;
    Symbols = list_Nil();
    for(Index = 1; Index < symbol_ACTINDEX; Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index) &&
	  (symbol_IsPredicate(symbol_Signature(Index)->info) ||
	   symbol_IsFunction(symbol_Signature(Index)->info))) 
	Symbols = list_Cons((POINTER)(symbol_Signature(Index)->info),Symbols);
    Symbols = list_Sort(Symbols, symbol_PrecedenceGreater);
    Index   = 0;
    fputs("set_precedence(", File);
    for(Scan = Symbols; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      ActSig = symbol_Signature(symbol_Index((SYMBOL)list_Car(Scan)));
      fputs(ActSig->name, File);
      if (!list_Empty(list_Cdr(Scan)))
	putc(',', File);
      if (Index>15) {
	Index = 0;
	fputs("\n\t", File);
      }
      else
	Index++;
    }
    fputs(").", File);
    list_Delete(Symbols);
  }
}










void symbol_Print(SYMBOL Symbol)

{ 
#ifdef DEBUG
  if (!symbol_IsSymbol(Symbol) ||
      list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)symbol_Index(Symbol))) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In symbol_Print: Illegal input.\n");
    misc_FinishErrorReport();
  }
#endif

  if (symbol_Equal(symbol_Null(),Symbol))
    fputs("NULL", stdout);
  else if (symbol_IsVariable(Symbol)) {
    SYMBOL NormSymbol;

    NormSymbol = symbol_NormVar(Symbol);

    if (symbol_IsStandardVariable(Symbol)) {
      if (Symbol <= 6)
	
	sprintf(symbol_VARSTRING,"%c", 84 + NormSymbol);
      else
	
	sprintf(symbol_VARSTRING,"X%d", NormSymbol - 6);
    }
    else if (symbol_IsIndexVariable(Symbol))
      
      sprintf(symbol_VARSTRING,"I%d", NormSymbol);

    fputs(symbol_VARSTRING, stdout);
  }
  else if (symbol_HASSIGNATURE)
    fputs(symbol_Name(Symbol), stdout);
  else
    printf("%d", Symbol);
}


void symbol_FPrint(FILE* File, SYMBOL Symbol)

{ 
#ifdef DEBUG
  if (!symbol_IsSymbol(Symbol)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In symbol_FPrint: Illegal input.\n");
    misc_FinishErrorReport();
  }
#endif

  if (symbol_Equal(symbol_Null(),Symbol))
    fputs("NULL", File);
  else if (symbol_IsVariable(Symbol)) {
    SYMBOL NormSymbol;

    NormSymbol = symbol_NormVar(Symbol);

    if (symbol_IsStandardVariable(Symbol)) {
      if (Symbol <= 6)
	
	sprintf(symbol_VARSTRING,"%c", 84 + NormSymbol);
      else
	
	sprintf(symbol_VARSTRING,"X%d", NormSymbol - 6);
    }
    else if (symbol_IsIndexVariable(Symbol))
      
      sprintf(symbol_VARSTRING,"I%d", NormSymbol);

    fputs(symbol_VARSTRING, File);
  }
  else if (symbol_HASSIGNATURE)
    fputs(symbol_Name(Symbol), File);
  else
    fprintf(File, "%d", Symbol);
}


void symbol_FPrintOtter(FILE* File, SYMBOL Symbol)

{ 
#ifdef DEBUG
  if (!symbol_IsSymbol(Symbol)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In symbol_FPrintOtter: Illegal input.\n");
    misc_FinishErrorReport();
  }
#endif

  if (symbol_IsVariable(Symbol)) {
    SYMBOL NormSymbol;

    NormSymbol = symbol_NormVar(Symbol);

    if (symbol_IsStandardVariable(Symbol)) {
      if (Symbol <= 6)
	
	sprintf(symbol_VARSTRING,"%c", 116 + NormSymbol);
      else
	
	sprintf(symbol_VARSTRING,"x%d", NormSymbol - 6);
    }
    else if (symbol_IsIndexVariable(Symbol))
      
      sprintf(symbol_VARSTRING,"I%d", NormSymbol);

    fputs(symbol_VARSTRING, File);
  }
  else 
    if (symbol_HASSIGNATURE) {
      if (symbol_IsConstant(Symbol))
	fprintf(File, "c%s", symbol_Name(Symbol));
      else
	if (symbol_IsFunction(Symbol))
	  fprintf(File, "f%s", symbol_Name(Symbol));
	else
	  if (symbol_IsPredicate(Symbol))
	    fprintf(File, "P%s", symbol_Name(Symbol));
	  else
	    fprintf(File, "%s", symbol_Name(Symbol));
    }
    else
      fprintf(File, "%d", Symbol);
}


void symbol_PrintLn(SYMBOL Symbol)

{ 
  symbol_Print(Symbol);
  putchar('\n');
  
}


void symbol_PrintAll(void)

{
  const int symbol_TABLEBLANKS = 2;
  const int symbol_TABLETYPE   = 6;  
  const int symbol_TABLEARITY  = 7;
  const int symbol_TABLESTAT   = 6;
  const int symbol_TABLEPROP   = 8;  
  const int symbol_TABLESTRING = 36;

  if (symbol_HASSIGNATURE) {
    int          Index;
    SYMBOL       Info;
    SIGNATURE    Entry;
    unsigned int TypePos, ArityPos, StatPos, PropPos, StringPos, EndPos, ActPos;

    TypePos   = symbol_TABLEBLANKS;
    ArityPos  = TypePos   + symbol_TABLETYPE   + 1;
    StatPos   = ArityPos  + symbol_TABLEARITY  + 1;
    PropPos   = StatPos   + symbol_TABLESTAT   + 1;
    StringPos = PropPos   + symbol_TABLEPROP   + 1;
    EndPos    = StringPos + symbol_TABLESTRING + 1;

    fputs("\n\n", stdout);
    misc_PrintChar(symbol_TABLEBLANKS, ' '); putchar('+');
    misc_PrintChar(symbol_TABLETYPE,   '-'); putchar('+');
    misc_PrintChar(symbol_TABLEARITY,  '-'); putchar('+');
    misc_PrintChar(symbol_TABLESTAT,   '-'); putchar('+');
    misc_PrintChar(symbol_TABLEPROP,   '-'); putchar('+');
    misc_PrintChar(symbol_TABLESTRING, '-'); putchar('+'); putchar('\n');
    misc_PrintChar(symbol_TABLEBLANKS, ' '); putchar('|');
    ActPos = TypePos   + 1 + fputs(" Type", stdout);
    misc_PrintChar((ArityPos  - ActPos), ' '); putchar('|');
    ActPos = ArityPos  + 1 + fputs(" Arity", stdout);
    misc_PrintChar((StatPos   - ActPos), ' '); putchar('|');
    ActPos = StatPos   + 1 + fputs(" Stat", stdout);
    misc_PrintChar((PropPos   - ActPos), ' '); putchar('|');
    ActPos = PropPos   + 1 + fputs(" Prop", stdout);
    misc_PrintChar((StringPos - ActPos), ' '); putchar('|');
    ActPos = StringPos + 1 + fputs(" String", stdout);
    misc_PrintChar((EndPos    - ActPos), ' '); putchar('|'); putchar('\n');
    misc_PrintChar(symbol_TABLEBLANKS,   ' '); putchar('+');
    misc_PrintChar(symbol_TABLETYPE,     '-'); putchar('+');
    misc_PrintChar(symbol_TABLEARITY,    '-'); putchar('+');
    misc_PrintChar(symbol_TABLESTAT,     '-'); putchar('+');
    misc_PrintChar(symbol_TABLEPROP,     '-'); putchar('+');
    misc_PrintChar(symbol_TABLESTRING,   '-'); putchar('+'); putchar('\n');
  
    for(Index = 1; Index < symbol_ACTINDEX; Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
	Entry  = symbol_Signature(Index);
	Info   = Entry->info;
	
	misc_PrintChar(symbol_TABLEBLANKS,   ' '); putchar('|');
	ActPos = TypePos + 1;

	switch (symbol_Type(Info)) {
	case symbol_CONSTANT:
	  ActPos += fputs(" Con", stdout); break;
	case symbol_FUNCTION:
	  ActPos += fputs(" Fun", stdout); break;
	case symbol_PREDICATE:
	  ActPos += fputs(" Pre", stdout); break;
	case symbol_JUNCTOR:
	  ActPos += fputs(" Jun", stdout); break;
	}
	misc_PrintChar((ArityPos  - ActPos), ' '); putchar('|');
	ActPos = ArityPos + 1 + printf(" %3d", Entry->arity);
	misc_PrintChar((StatPos   - ActPos), ' '); putchar('|');
	ActPos = StatPos + 1;
	if (symbol_Type(Info) == symbol_FUNCTION) {
	  switch (symbol_Stat(Info)) {
	  case symbol_STATLEX:
	    ActPos += fputs(" Lex", stdout); break;
	  case symbol_STATMUL:
	    ActPos += fputs(" Mul", stdout); break;
	  }
	}
	misc_PrintChar((PropPos   - ActPos), ' '); putchar('|');
	ActPos = PropPos + 1 + printf(" %u", Entry->props);
	misc_PrintChar((StringPos - ActPos), ' '); putchar('|');
	ActPos = StringPos + 1 +  printf(" %s", Entry->name);
	misc_PrintChar((EndPos    - ActPos), ' '); putchar('|'); putchar('\n');
      }
    misc_PrintChar(symbol_TABLEBLANKS, ' '); putchar('+');
    misc_PrintChar(symbol_TABLETYPE,   '-'); putchar('+');
    misc_PrintChar(symbol_TABLEARITY,  '-'); putchar('+');
    misc_PrintChar(symbol_TABLESTAT,   '-'); putchar('+');
    misc_PrintChar(symbol_TABLEPROP,   '-'); putchar('|');
    misc_PrintChar(symbol_TABLESTRING, '-'); putchar('+'); putchar('\n');
  }
}

