




 






















#ifndef _SYMBOL_
#define _SYMBOL_





#include "list.h"











#define symbol__SYMBOLMAXLEN   64

#define symbol__NOOFSTANDARDVAR 2000
#define symbol__NOOFINDEXVAR    1000

#define symbol__MAXSTANDARDVAR  symbol__NOOFSTANDARDVAR
#define symbol__MAXINDEXVAR     (symbol__NOOFSTANDARDVAR + symbol__NOOFINDEXVAR)

#define symbol__NOOFVARIABLES (symbol__NOOFSTANDARDVAR + symbol__NOOFINDEXVAR)
#define symbol__MAXVARIABLES  (symbol__NOOFVARIABLES + 1)








extern const int symbol_MASK;
extern const int symbol_TYPEMASK;
extern const int symbol_STATMASK;
extern const int symbol_TYPESTATMASK;

extern const int symbol_ARBITRARYARITY;

extern const int symbol_TYPEBITS;
extern const int symbol_STATBITS;
extern const int symbol_TYPESTATBITS;

extern const int symbol_SIGTYPES;

#define symbol_CONSTANT  0 
#define symbol_FUNCTION  1
#define symbol_PREDICATE 2
#define symbol_JUNCTOR   3

#define symbol_STATLEX   0 
#define symbol_STATMUL   1





extern BOOL symbol_HASSIGNATURE;
extern LIST symbol_FREEDSYMBOLS;

typedef int SYMBOL;
typedef int *PRECEDENCE;

typedef struct signature {
  char   *name;         
  NAT    length;        
  int    weight;        
  int    arity;         
  NAT    props;         
  SYMBOL info;          
  LIST   generatedBy;
} SIGNATURE_NODE, *SIGNATURE;

typedef enum {SKOLEM=1,    CUMMUTATIVE=2, ASSOCIATIVE=4, ORDRIGHT=8, ORDMUL=16,
	      DECLSORT=32, DOMPRED=64, ISDEF=128, FREELY=256, GENERATED=512
} SPROPERTY;


#define symbol__MAXSIGNATURE 4000

extern SIGNATURE *symbol_SIGNATURE;
extern PRECEDENCE symbol_PRECEDENCE;
extern PRECEDENCE symbol_DEFAULTPRECEDENCE;

extern SYMBOL symbol_STANDARDVARCOUNTER;
extern SYMBOL symbol_INDEXVARCOUNTER;

extern int symbol_ACTINDEX;
extern int symbol_ACTSKOLEMFINDEX;
extern int symbol_ACTSKOLEMCINDEX;
extern int symbol_ACTSKOLEMPINDEX;
extern int symbol_ACTSKOLEMAINDEX;

extern SYMBOL symbol_CONTEXT[];                   








static __inline__ BOOL symbol_Equal(SYMBOL A, SYMBOL B)
{
  return A==B;
}

static __inline__ BOOL symbol_IsSignature(SYMBOL S)
{
  return S < 0;
}

static __inline__ int symbol_Type(SYMBOL ActSymbol)
{
  return (-ActSymbol) & symbol_TYPEMASK;
}

static __inline__ BOOL symbol_IsJunctor(SYMBOL S)
{
  return (S < 0) && (symbol_Type(S) == symbol_JUNCTOR);
}

static __inline__ BOOL symbol_IsFunction(SYMBOL S)
{
  return (S < 0)
    && ((symbol_Type(S) == symbol_FUNCTION)
	|| (symbol_Type(S) == symbol_CONSTANT));
}

static __inline__ BOOL symbol_IsConstant(SYMBOL S)
{
  return (S < 0) && (symbol_Type(S) == symbol_CONSTANT);
}

static __inline__ BOOL symbol_IsPredicate(SYMBOL S)
{
  return (S < 0) && (symbol_Type(S) == symbol_PREDICATE);
}

static __inline__ BOOL symbol_IsVariable(SYMBOL S)
{
  return S > 0;
}

static __inline__ BOOL symbol_IsStandardVariable(SYMBOL S)
{
  return symbol_IsVariable(S) && (S <= symbol__MAXSTANDARDVAR);
}

static __inline__ BOOL symbol_IsIndexVariable(SYMBOL S)
{
  return (S > symbol__MAXSTANDARDVAR) && (S <= symbol__MAXINDEXVAR);
}

static __inline__ BOOL symbol_IsComplex(SYMBOL S)
{
  return (!symbol_IsVariable(S)) && (!symbol_IsConstant(S));
}

static __inline__ BOOL symbol_IsSuccessor(SYMBOL S, SYMBOL P)
{
  return S > P;
}







static __inline__ int symbol_GetInitialStandardVarCounter(void)
{
  return 0;
}

static __inline__ int symbol_GetInitialIndexVarCounter(void)
{
  return symbol__MAXSTANDARDVAR;
}

static __inline__ int symbol_FirstIndexVariable(void)
{
  return symbol__MAXSTANDARDVAR + 1;
}

static __inline__ int symbol_LastIndexVariable(void)
{
  return symbol_INDEXVARCOUNTER;
}



#define symbol__NULL 0

static __inline__ int symbol_MaxVars(void)
{
  return symbol__MAXVARIABLES;
}

static __inline__ int symbol_MaxConsts(void)
{
  return symbol__MAXSIGNATURE;
}

static __inline__ int symbol_MaxBaseSorts(void)
{
  return symbol__MAXSIGNATURE;
}

static __inline__ int symbol_TypeBits(void)
{
  return symbol_TYPEBITS;
}

static __inline__ int symbol_Null(void)
{
  return 0;
}

static __inline__ int symbol_ActIndex(void)
{
  return symbol_ACTINDEX;
}

static __inline__ void symbol_ResetSkolemIndex(void)
{
  symbol_ACTSKOLEMFINDEX = 0;
  symbol_ACTSKOLEMCINDEX = 0;
  symbol_ACTSKOLEMPINDEX = 0;
  symbol_ACTSKOLEMAINDEX = 0;
}










static __inline__ void symbol_FreeSignature(SIGNATURE Sig)

{
  memory_Free(Sig->name, symbol__SYMBOLMAXLEN);
  list_Delete(Sig->generatedBy);
  memory_Free(Sig, sizeof(SIGNATURE_NODE));
}

static __inline__ SIGNATURE symbol_GetSignature(void)
{
  return (SIGNATURE) memory_Malloc(sizeof(SIGNATURE_NODE));
} 






static __inline__ SYMBOL symbol_CreateStandardVariable(void)

{
#ifdef DEBUG
  if (symbol_STANDARDVARCOUNTER >= symbol__MAXSTANDARDVAR) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In symbol_CreateStandardVariable: Number of standard variables exceeded.\n");
    misc_FinishErrorReport();
   }
#endif

  return (++symbol_STANDARDVARCOUNTER);
}


static __inline__ SYMBOL symbol_CreateIndexVariable(void)

{
#ifdef DEBUG
  if (symbol_INDEXVARCOUNTER >= symbol__MAXINDEXVAR) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In symbol_CreateIndexVariable: Number of index variables exceeded.\n");
    misc_FinishErrorReport();
  }
#endif

  return (++symbol_INDEXVARCOUNTER);
}


static __inline__ SYMBOL symbol_NextIndexVariable(SYMBOL Variable)
{
#ifdef DEBUG
  if (((Variable != symbol_GetInitialIndexVarCounter()) &&
       (!symbol_IsIndexVariable(Variable))) ||
      (Variable == symbol__MAXINDEXVAR)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In symbol_NextVariable: Illegal input.\n");
    misc_FinishErrorReport();
  }
#endif

  return (Variable + 1);
}


static __inline__ void symbol_SetStandardVarCounter(SYMBOL Variable)
{
#ifdef DEBUG
  if ((Variable != symbol_GetInitialStandardVarCounter()) &&
      !symbol_IsStandardVariable(Variable)) {
    misc_StartErrorReport();
    misc_ErrorReport("\n In symbol_SetStandardVarCounter: Illegal input.\n");
    misc_FinishErrorReport();
  } 
  else 
    if (Variable >= symbol__MAXSTANDARDVAR) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In symbol_SetStandardVarCounter: Number of standard variables exceeded.\n");
      misc_FinishErrorReport();
    }
#endif

  symbol_STANDARDVARCOUNTER = Variable;
}

static __inline__ SYMBOL symbol_FirstVariable(void)
{
  return 1;
}

static __inline__ BOOL symbol_GreaterVariable(SYMBOL Var1, SYMBOL Var2)
{
  return Var1 > Var2;
}

static __inline__ void symbol_ResetStandardVarCounter(void)
{
  symbol_STANDARDVARCOUNTER = symbol_GetInitialStandardVarCounter();
}

void   symbol_Init(BOOL);
void   symbol_FreeAllSymbols(void);
SYMBOL symbol_CreateFunction(const char*, int, int);
SYMBOL symbol_CreateSkolemFunction(int);
SYMBOL symbol_CreateSkolemPredicate(int);
SYMBOL symbol_CreatePredicate(const char*, int, int);
SYMBOL symbol_CreateJunctor(const char*, int, int);





SYMBOL symbol_Lookup(const char*);

static __inline__ int symbol_VarIndex(SYMBOL ActSymbol)
{
  return ActSymbol;
}

static __inline__ int symbol_NormVar(SYMBOL ActSymbol)
{
  
  return (ActSymbol <= symbol__MAXSTANDARDVAR) ? ActSymbol : (ActSymbol - symbol__MAXSTANDARDVAR);
}





static __inline__ int symbol_Index(SYMBOL ActSymbol)
{
  return (-ActSymbol) >> symbol_TYPESTATBITS;
}

static __inline__ SYMBOL symbol_SignatureSymbol(int ActIndex, int Type, int Status)
{
  return -((ActIndex << symbol_TYPESTATBITS)
	   | (Status << symbol_TYPEBITS)
	   | Type);
}

static __inline__ SIGNATURE symbol_Signature(int ActIndex)
{
  return symbol_SIGNATURE[ActIndex];
}

static __inline__ void symbol_SetSignature(int ActIndex, SIGNATURE Sig)
{
  symbol_SIGNATURE[ActIndex] = Sig;
}

static __inline__ SYMBOL symbol_GetSigSymbol(int Index)
{
  SIGNATURE S = symbol_Signature(Index);
  return S->info;
}

static __inline__ int symbol_Stat(SYMBOL ActSymbol)
{
  return ((-ActSymbol) & symbol_STATMASK) >> symbol_TYPEBITS;
}

static __inline__ SYMBOL symbol_ChangeType(SYMBOL S, int Type)

{
  SIGNATURE Sig;
  Sig = symbol_Signature(symbol_Index(S));
  S   = symbol_SignatureSymbol(symbol_Index(S), Type, symbol_Stat(S));
  Sig->info = S;
  return S;
}

static __inline__ int symbol_Arity(SYMBOL ActSymbol)
{
  SIGNATURE S = symbol_Signature(symbol_Index(ActSymbol)); 
  return S->arity;
}

static __inline__ NAT symbol_PositiveArity(SYMBOL ActSymbol)
{
  SIGNATURE S = symbol_Signature(symbol_Index(ActSymbol)); 
  if (S->arity<0)
    return NAT_MAX;
  else
    return S->arity;
}

static __inline__ void symbol_SetArity(SYMBOL ActSymbol, int Arity)
{
  SIGNATURE S = symbol_Signature(symbol_Index(ActSymbol)); 
  S->arity = Arity;
}
  
static __inline__ int symbol_ArbitraryArity()
{
  return -1;
}

static __inline__ char* symbol_Name(SYMBOL ActSymbol)
{
  SIGNATURE S = symbol_Signature(symbol_Index(ActSymbol));
  return S->name;
}

static __inline__ NAT symbol_NameLength(SYMBOL ActSymbol)
{
  SIGNATURE S = symbol_Signature(symbol_Index(ActSymbol));
  return S->length;
}

static __inline__ int symbol_Info(SYMBOL ActSymbol)
{
  SIGNATURE S = symbol_Signature(symbol_Index(ActSymbol));  
  return  S->info;
}

static __inline__ int symbol_Weight(SYMBOL ActSymbol)
{
  SIGNATURE S = symbol_Signature(symbol_Index(ActSymbol));
  return S->weight;
}

static __inline__ int symbol_Ordering(SYMBOL ActSymbol)
{
  return symbol_PRECEDENCE[symbol_Index(ActSymbol)];
}

static __inline__ void symbol_SetWeight(SYMBOL ActSymbol, int Weight)
{
  SIGNATURE S = symbol_Signature(symbol_Index(ActSymbol));
  S->weight = Weight;
}

static __inline__ void symbol_SetName(SYMBOL ActSymbol, char* Name)
{
  SIGNATURE S = symbol_Signature(symbol_Index(ActSymbol));
  S->name = Name;
}

static __inline__ LIST symbol_GeneratedBy(SYMBOL S)
{
  return symbol_Signature(symbol_Index(S))->generatedBy;
}

static __inline__ BOOL symbol_IsGeneratedBy(SYMBOL S1, SYMBOL S2)
{
  return list_PointerMember(symbol_GeneratedBy(S1), (POINTER)S2);
}

static __inline__ void symbol_SetGeneratedBy(SYMBOL S, LIST SymbolList)
{
  symbol_Signature(symbol_Index(S))->generatedBy = SymbolList;
}

static __inline__ void symbol_SetOrdering(SYMBOL ActSymbol, int Ordering)
{
  symbol_PRECEDENCE[symbol_Index(ActSymbol)] =  Ordering;
}

static __inline__ BOOL symbol_PrecedenceGreater(SYMBOL S1, SYMBOL S2)
{
  return symbol_Ordering(S1) < symbol_Ordering(S2);
}

static __inline__ BOOL symbol_HasProperty(SYMBOL ActSymbol, SPROPERTY Property)
{
  SIGNATURE S = symbol_Signature(symbol_Index(ActSymbol));
  return (S->props & Property);
}

static __inline__ void symbol_AddProperty(SYMBOL ActSymbol, SPROPERTY Property)
{
  SIGNATURE S = symbol_Signature(symbol_Index(ActSymbol));
  S->props    = S->props | Property;
}

static __inline__ void symbol_RemoveProperty(SYMBOL ActSymbol, SPROPERTY Property)
{
  SIGNATURE S = symbol_Signature(symbol_Index(ActSymbol));
  if(S->props & Property)
    S->props = S->props - Property;
}

static __inline__ BOOL symbol_IsBaseSort(SYMBOL Symbol)
{
  return (symbol_Arity(Symbol) == 1);
}

static __inline__ PRECEDENCE symbol_CreatePrecedence()
{
  return memory_Malloc(sizeof(int) * symbol__MAXSIGNATURE);
}

static __inline__ void symbol_DeletePrecedence(PRECEDENCE precedence)
{
  memory_Free(precedence, sizeof(int) * symbol__MAXSIGNATURE);
}

static __inline__ void symbol_InstallPrecedence(PRECEDENCE precedence)
{
  symbol_PRECEDENCE = precedence;
}

static __inline__ void symbol_InstallDefaultPrecedence()
{
  symbol_PRECEDENCE = symbol_DEFAULTPRECEDENCE;
}

static __inline__ BOOL symbol_HasInstalledDefaultPrecedence()
{
  return (symbol_PRECEDENCE == symbol_DEFAULTPRECEDENCE);
}





static __inline__ BOOL symbol_ContextIsClean()
{
  int i;
  for(i=0;i<symbol__MAXSIGNATURE;i++)
    if (symbol_CONTEXT[i] != (SYMBOL)0)
      return FALSE;
  return TRUE;
}

static __inline__ void symbol_ContextClean()
{
  int i;
  for(i=0;i<symbol__MAXSIGNATURE;i++)
    symbol_CONTEXT[i] = (SYMBOL)0;
}

static __inline__ BOOL symbol_ContextIsMapped(SYMBOL Symbol)
{
  int i;
  for(i=0;i<symbol__MAXSIGNATURE;i++)
    if (symbol_Equal(symbol_CONTEXT[i],Symbol))
      return TRUE;
  return FALSE;
}

static __inline__ SYMBOL symbol_ContextGetValue(SYMBOL Symbol)
{
  return symbol_CONTEXT[symbol_Index(Symbol)];
}

static __inline__ void symbol_ContextSetValue(SYMBOL Symbol, SYMBOL Value)
{
  symbol_CONTEXT[symbol_Index(Symbol)] = Value;
}

static __inline__ void symbol_ContextClearValue(SYMBOL Symbol)
{
  symbol_CONTEXT[symbol_Index(Symbol)] = (SYMBOL)0;
}

static __inline__ BOOL symbol_ContextIsBound(SYMBOL Symbol)
{
  return (symbol_CONTEXT[symbol_Index(Symbol)] != (SYMBOL)0);
}







NAT     symbol_MaxStringLength(void);

void    symbol_ReinitGenericNameCounters(void);

int     symbol_GetIncreasedOrderingCounter(void);

void    symbol_Delete(SYMBOL);
BOOL    symbol_IsSymbol(SYMBOL);
BOOL    symbol_Equal(SYMBOL, SYMBOL);
void    symbol_Dump(void);

void    symbol_LowerSignature(void);

LIST    symbol_GetAllSymbols(void);





void   symbol_Print(SYMBOL);
void   symbol_PrintPrecedence(void);
void   symbol_FPrintPrecedence(FILE*);
void   symbol_FPrint(FILE*, SYMBOL);
void   symbol_FPrintOtter(FILE*, SYMBOL);
void   symbol_PrintLn(SYMBOL);
void   symbol_PrintAll(void);

#endif
