






















#include "tableau.h" 


extern BOOL        pcheck_ClauseCg;
extern GRAPHFORMAT pcheck_GraphFormat;

TABPATH tab_PathCreate(int MaxLevel, TABLEAU Tab)

{
 TABPATH TabPath;

 TabPath            = (TABPATH)memory_Malloc(sizeof(TABPATH_NODE));
 TabPath->Path      = (TABLEAU*)memory_Malloc(sizeof(TABLEAU)*(MaxLevel+1));
 TabPath->Path[0]   = Tab;
 TabPath->MaxLength = MaxLevel;
 TabPath->Length    = 0;

 return(TabPath); 
}

void tab_PathDelete(TABPATH TabPath)

{
  memory_Free(TabPath->Path, (TabPath->MaxLength+1)*sizeof(TABLEAU));
  memory_Free(TabPath, sizeof(TABPATH_NODE));
}


BOOL tab_PathContainsClause(TABPATH Path, CLAUSE Clause)

{
  LIST Scan;

  if (clause_SplitLevel(Clause) > tab_PathLength(Path))
    return(FALSE);

  for(Scan = tab_Clauses(tab_PathNthNode(Path, clause_SplitLevel(Clause))); 
	!list_Empty(Scan); Scan = list_Cdr(Scan)) {
    if (list_Car(Scan) == Clause)
      return(TRUE);  }
  return(FALSE);
}

BOOL tab_PathContainsClauseSoft(TABPATH Path, CLAUSE Clause)

{
  LIST Scan;
  int Level;

  if (clause_SplitLevel(Clause) > tab_PathLength(Path))
    return(FALSE);
  
  for (Level = 0; Level <= tab_PathLength(Path); Level++) {
    for(Scan = tab_Clauses(tab_PathNthNode(Path, Level));  
	!list_Empty(Scan); Scan = list_Cdr(Scan)) {
      if (list_Car(Scan) == Clause)
	return(TRUE);  } } 

  return(FALSE);
}


BOOL tab_PathContainsClauseRobust(TABPATH P, CLAUSE C)

{
  if (tab_PathContainsClause(P,C)) 
    return(TRUE);

  if (tab_PathContainsClauseSoft(P,C)) {
    fprintf(stderr, "NOTE: Clause is found on path, but not indexed by level.\n");
    clause_PParentsFPrint(stderr,C); fflush(stderr);
    return(TRUE);
  }
  return(FALSE);
}

void tab_AddSplitAtCursor(TABPATH Path, BOOL LeftSide)

{
  TABLEAU Tab, NewBranch;

  Tab = tab_PathTop(Path);
  NewBranch = tab_CreateNode();
  if (LeftSide) {

#ifdef DEBUG
    if (!tab_LeftBranchIsEmpty(Tab)) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In tab_AddSplitAtCursor: Recreating existing"
		       " left branch in tableau.\n");
      misc_FinishErrorReport();
    }
#endif

    tab_SetLeftBranch(Tab,NewBranch);
  } else {

#ifdef DEBUG    
    if (!tab_RightBranchIsEmpty(Tab)) {
      misc_StartErrorReport();
      misc_ErrorReport("\n In tab_AddSplitAtCursor: Recreating existing"
		       " right branch in tableau.\n");
      misc_FinishErrorReport();
    }
#endif
    tab_SetRightBranch(Tab, NewBranch);
  }
  tab_PathPush(NewBranch, Path);
}

void tab_AddClauseOnItsLevel(CLAUSE C, TABPATH Path)

{
  int Level; 

  Level = clause_SplitLevel(C);
  if (Level > tab_PathLength(Path)) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\nError: Split level of some clause ");
    misc_UserErrorReport("\nis higher than existing level."); 
    misc_UserErrorReport("\nThis may be a bug in the proof file."); 
    misc_FinishUserErrorReport();  
  }
  
  tab_AddClause(C, tab_PathNthNode(Path, clause_SplitLevel(C)));
}


int tab_Depth(TABLEAU T)

{
  if (tab_IsEmpty(T))
    return(0);
  if (tab_IsLeaf(T))
    return(0);
  else 
    return(misc_Max(tab_Depth(tab_RightBranch(T))+1, tab_Depth(tab_LeftBranch(T)))+1);
}

BOOL tab_HasEmptyClause(TABLEAU T)

{
  LIST Scan;
  
  for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) 
    if (clause_IsEmptyClause(list_Car(Scan)))
      return(TRUE);

  return(FALSE);
}


BOOL tab_IsClosed(TABLEAU T)

{
  if (tab_IsEmpty(T))
    return(FALSE);

  if (tab_HasEmptyClause(T))
      return(TRUE);
  
  
  if (tab_RightBranchIsEmpty(T) || tab_LeftBranchIsEmpty(T)) {
    printf("\nopen node label: %d", T->Label);
    fflush(stdout);

    return(FALSE);
 }
  return(tab_IsClosed(tab_RightBranch(T)) && tab_IsClosed(tab_LeftBranch(T)));
}

LIST tab_DeleteFlat(TABLEAU T)

{
  LIST Clauses;

  Clauses = tab_Clauses(T);
  memory_Free(T, sizeof(TABLEAU_NODE));
  
  return(Clauses);
}


void tab_DeleteGen(TABLEAU T, LIST* Clauses, BOOL DeleteClauses)

{
  if (tab_IsEmpty(T))
    return;
  
  tab_DeleteGen(tab_RightBranch(T), Clauses, DeleteClauses);
  tab_DeleteGen(tab_LeftBranch(T),  Clauses, DeleteClauses);

  list_Delete(tab_RightSplitClauses(T));
  if (DeleteClauses) 
    list_Delete(tab_Clauses(T)); 
  else
    *Clauses = list_Nconc(tab_Clauses(T), *Clauses);
  
  tab_DeleteFlat(T);
  
}

void tab_DeleteCollectClauses(TABLEAU T, LIST* Clauses)

{
  tab_DeleteGen(T, Clauses, FALSE);
}

void tab_Delete(TABLEAU T)

{
  LIST Redundant;
  
  Redundant = list_Nil();
  tab_DeleteGen(T, &Redundant, TRUE);
}

static void tab_SetSplitLevelsRec(TABLEAU T, int Level)

{
  LIST Scan;

  if (tab_IsEmpty(T))
    return;
  
  for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    clause_SetSplitLevel(list_Car(Scan), Level); 
    if (Level >0) {
      clause_ClearSplitField(list_Car(Scan));
      clause_SetSplitFieldBit(list_Car(Scan), Level); }
    else 
      clause_SetSplitField(list_Car(Scan), (SPLITFIELD)NULL,0);
  }
  
  tab_SetSplitLevelsRec(tab_RightBranch(T), Level+1);
  tab_SetSplitLevelsRec(tab_LeftBranch(T), Level+1);
}

void tab_SetSplitLevels(TABLEAU T)

{
  tab_SetSplitLevelsRec(T,0);
}

TABLEAU tab_PruneClosedBranches(TABLEAU T, LIST* Clauses) 

{
  if (tab_IsEmpty(T))
    return(T);

  

  if (tab_HasEmptyClause(T)) {

    tab_DeleteCollectClauses(tab_RightBranch(T), Clauses); 
    tab_DeleteCollectClauses(tab_LeftBranch(T), Clauses); 
    tab_SetRightBranch(T, tab_EmptyTableau());
    tab_SetLeftBranch(T, tab_EmptyTableau()); 
    list_Delete(tab_RightSplitClauses(T));
    tab_SetRightSplitClauses(T, list_Nil());
    tab_SetSplitClause(T,clause_Null());
    tab_SetLeftSplitClause(T, clause_Null()); }

  
  
  else {
    tab_SetRightBranch(T, tab_PruneClosedBranches(tab_RightBranch(T), Clauses));
    tab_SetLeftBranch(T, tab_PruneClosedBranches(tab_LeftBranch(T), Clauses));  }
  
  return(T);
}


TABLEAU tab_RemoveIncompleteSplits(TABLEAU T, LIST* Clauses)

{
  LIST    NewClauses;
  TABLEAU Child;

  if (tab_IsEmpty(T))
    return(T);

  if (tab_IsLeaf(T))
    return(T);

  if (!tab_IsEmpty(tab_RightBranch(T)) &&
      !tab_IsEmpty(tab_LeftBranch(T))) {	   
    tab_SetRightBranch(T, tab_RemoveIncompleteSplits(tab_RightBranch(T), Clauses));
    tab_SetLeftBranch(T, tab_RemoveIncompleteSplits(tab_LeftBranch(T), Clauses));
    return(T);
  }
  if (tab_IsEmpty(tab_RightBranch(T))) 
    Child = tab_LeftBranch(T);
  else 
    Child = tab_RightBranch(T);

  Child = tab_RemoveIncompleteSplits(Child, Clauses);

  tab_SetLeftBranch(T, tab_LeftBranch(Child));   
  tab_SetRightBranch(T, tab_RightBranch(Child));

  

  tab_SetSplitClause(T, tab_SplitClause(Child));
  tab_SetLeftSplitClause(T, tab_LeftSplitClause(Child));
  tab_SetRightSplitClauses(T, tab_RightSplitClauses(Child));

  

  NewClauses = tab_DeleteFlat(Child);
  (*Clauses) = list_Nconc(NewClauses, *Clauses);

  return(T);
}

void tab_CheckEmpties(TABLEAU T)

{
  LIST Scan, Empties;
  BOOL Printem;

  if (tab_IsEmpty(T))
    return;

   
  Empties = list_Nil();
  for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    if (clause_IsEmptyClause(list_Car(Scan)))
      Empties = list_Cons(list_Car(Scan), Empties);
  }
  Printem = FALSE;
  if (!list_Empty(Empties) && !tab_IsLeaf(T)) {
    printf("\nNOTE: non-leaf node contains empty clauses.\n");
    Printem = TRUE;
  }
  
  if (tab_IsLeaf(T) && list_Length(Empties) > 1) {
    printf("\nNOTE: Leaf contains more than one empty clauses.\n");
    Printem =TRUE;
  }
  if (Printem) {
    printf("Clauses:\n");
    clause_PParentsListPrint(tab_Clauses(T));
  }
  list_Delete(Empties);
  tab_CheckEmpties(tab_LeftBranch(T));
  tab_CheckEmpties(tab_RightBranch(T));
}

void tab_GetAllEmptyClauses(TABLEAU T, LIST* L)

{
  if (tab_IsEmpty(T))
    return;
  
  tab_GetAllEmptyClauses(tab_LeftBranch(T), L);
  tab_GetAllEmptyClauses(tab_RightBranch(T), L);
}

void tab_GetEarliestEmptyClauses(TABLEAU T, LIST* L)

{
  CLAUSE FirstEmpty;
  LIST   Scan;

  if (tab_IsEmpty(T))
    return;

  if (tab_IsLeaf(T)) {  
    FirstEmpty = clause_Null();
  
    for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      if (clause_IsEmptyClause(list_Car(Scan))) {
	if (FirstEmpty == clause_Null()) 
	  FirstEmpty = list_Car(Scan);
	else if (clause_Number(FirstEmpty) > clause_Number(list_Car(Scan)))
	  FirstEmpty = list_Car(Scan);
      }
    }
    
    if (FirstEmpty != clause_Null())
      (*L) = list_Cons(FirstEmpty, *L);
  }
  tab_GetEarliestEmptyClauses(tab_LeftBranch(T), L);
  tab_GetEarliestEmptyClauses(tab_RightBranch(T), L);

}

void tab_ToClauseList(TABLEAU T, LIST* Proof)

{
  if (tab_IsEmpty(T))
    return;
  
  (*Proof) = list_Nconc(list_Copy(tab_Clauses(T)), *Proof);

  tab_ToClauseList(tab_LeftBranch(T),Proof);
  tab_ToClauseList(tab_RightBranch(T),Proof);
}


void tab_ToSeqProofOrdered(TABLEAU T, LIST* Proof)

{
  LIST Scan;
  BOOL RightSplitRead, LeftSplitRead;
  
  if (tab_IsEmpty(T))
    return;

  Scan = tab_Clauses(T); 
  RightSplitRead = LeftSplitRead = FALSE;

  while (!list_Empty(Scan)) {
    

    if (!RightSplitRead && !tab_RightBranchIsEmpty(T) &&
	clause_Number(list_Car(Scan)) <
	clause_Number(list_Car(tab_RightSplitClauses(T)))) {
      tab_ToSeqProofOrdered(tab_RightBranch(T), Proof);
      RightSplitRead = TRUE;
    }
    if (!LeftSplitRead && !tab_LeftBranchIsEmpty(T) &&
	clause_Number(list_Car(Scan)) < 
	clause_Number(tab_LeftSplitClause(T))) {
      tab_ToSeqProofOrdered(tab_LeftBranch(T), Proof);
      LeftSplitRead  = TRUE;
    }
    (*Proof) = list_Cons(list_Car(Scan), *Proof); 
    Scan  = list_Cdr(Scan);
  }
  

  if (!RightSplitRead)
    tab_ToSeqProofOrdered(tab_RightBranch(T), Proof);
  if (!LeftSplitRead)
    tab_ToSeqProofOrdered(tab_LeftBranch(T), Proof);
}






void tab_LabelNodesRec(TABLEAU T, int* Num)

{
  if (tab_IsEmpty(T))
    return;
  T->Label = *Num;
  (*Num)++;
  tab_LabelNodesRec(tab_LeftBranch(T), Num);
  tab_LabelNodesRec(tab_RightBranch(T), Num);
}

void tab_LabelNodes(TABLEAU T)

{
  int Num;

  Num = 0;
  tab_LabelNodesRec(T, &Num);
}


void tab_FPrintNodeLabel(FILE *File, TABLEAU T)

{
  LIST Scan;

  
  fprintf(File, "\""); 
  
  fprintf(File, "label: %d\\n", T->Label);
  
  
  fprintf(File, "SplitClause : "); 
  clause_PParentsFPrint(File, tab_SplitClause(T));
  fprintf(File, "\\nLeft Clause : ");
  clause_PParentsFPrint(File, tab_LeftSplitClause(T));
  fprintf(File, "\\n"); 
  fprintf(File, "RightClauses: ");
  if (list_Empty(tab_RightSplitClauses(T)))
    fprintf(File,"[]\\n");
  else {
    clause_PParentsFPrint(File, list_Car(tab_RightSplitClauses(T)));
    fprintf(File, "\\n");
    for (Scan = list_Cdr(tab_RightSplitClauses(T)); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      fprintf(File,"              ");
      clause_PParentsFPrint(File, list_Car(Scan));
      fprintf(File, "\\n");
    }
  }
  
  if (pcheck_ClauseCg) {
    if (list_Empty(tab_Clauses(T)))
      fprintf(File, "[]");
    else {
      for (Scan = tab_Clauses(T); !list_Empty(Scan); Scan = list_Cdr(Scan)){
	clause_PParentsFPrint(File, list_Car(Scan)); 
	fprintf(File, "\\");
	fprintf(File, "n");
      }
    }
  }
  fprintf(File, "\""); 
}

void tab_FPrintEdgeCgFormat(FILE* File, int Source, int Target, BOOL Left)

{
  fprintf(File, "\nedge: {");
  fprintf(File, "\nsourcename: \"%d\"", Source);
  fprintf(File, "\ntargetname: \"%d\"\n", Target);
  fprintf(File, "\nlabel: \"");
  if (Left) 
    fprintf(File, "0");
  else
    fprintf(File, "1");
  fprintf(File, "\"");

  fprintf(File, "  }\n");
} 

void tab_FPrintEdgesCgFormat(FILE* File, TABLEAU T)

{
  if (tab_IsEmpty(T))
    return;

  if (!tab_LeftBranchIsEmpty(T))
    tab_FPrintEdgeCgFormat(File, T->Label, tab_LeftBranch(T)->Label, TRUE);
  if (!tab_RightBranchIsEmpty(T))
    tab_FPrintEdgeCgFormat(File, T->Label, tab_RightBranch(T)->Label, FALSE);
  
  tab_FPrintEdgesCgFormat(File, tab_LeftBranch(T));
  tab_FPrintEdgesCgFormat(File, tab_RightBranch(T));
}

void tab_FPrintNodesCgFormat(FILE* File, TABLEAU T)

{
  if (tab_IsEmpty(T))
    return;

  fprintf(File, "\nnode: {\n");
  fprintf(File, "\nlabel: ");
  tab_FPrintNodeLabel(File, T);
  fprintf(File, "\n"); 
  
  fprintf(File, "title: \"%d\"\n", T->Label);
  fprintf(File, "  }\n");
  
  
  tab_FPrintNodesCgFormat(File, tab_LeftBranch(T));
  tab_FPrintNodesCgFormat(File, tab_RightBranch(T));
  
} 

void tab_FPrintCgFormat(FILE* File, TABLEAU T)

{
  fprintf(File, "graph: \n");
  fprintf(File,"{\n");
  fprintf(File, "display_edge_labels: yes\n");
  
  tab_FPrintNodesCgFormat(File, T);
  tab_FPrintEdgesCgFormat(File, T); 
  fprintf(File, "}\n");
}

void tab_PrintCgFormat(TABLEAU T)

{
  tab_FPrintCgFormat(stdout, T);
}





void tab_FPrintDaVinciEdge(FILE* File, int L1, int L2)

{
  fprintf(File, "l(\"%d->%d\","   ,L1,L2);
  fprintf(File, "e(\"\",[],\n");
  
  fprintf(File, "r(\"%d\")))\n", L2);
}

void tab_FPrintDaVinciFormatRec(FILE* File, TABLEAU T)

{
  
  fprintf(File, "l(\"%d\",", T->Label);
  
  fprintf(File, "n(\"\", [a(\"OBJECT\","); 
  tab_FPrintNodeLabel(File, T);
  fprintf(File, ")],\n");
  
  
  fprintf(File, "[");
  if (!tab_LeftBranchIsEmpty(T)) 
    tab_FPrintDaVinciEdge(File, T->Label, tab_LeftBranch(T)->Label); 

  if (!tab_RightBranchIsEmpty(T)) {
    if (!tab_LeftBranchIsEmpty(T))
      fprintf(File, ",");
    tab_FPrintDaVinciEdge(File, T->Label, tab_RightBranch(T)->Label);
  }
  fprintf(File, "]");

  fprintf(File, "))"); 

  if (!tab_LeftBranchIsEmpty(T)) {
    fprintf(File, ",");
    tab_FPrintDaVinciFormatRec(File, tab_LeftBranch(T));
  }
  if (!tab_RightBranchIsEmpty(T)) {
    fprintf(File, ",");
    tab_FPrintDaVinciFormatRec(File, tab_RightBranch(T));
  }
}

void tab_FPrintDaVinciFormat(FILE* File, TABLEAU T)

{
  fprintf(File, "[\n");
  tab_FPrintDaVinciFormatRec(File,T);
  fprintf(File, "]\n");
}


void tab_WriteTableau(TABLEAU T, char* Filename, GRAPHFORMAT Format)

{
  FILE*  File;
  
  if (Format != DAVINCI && Format != XVCG) {
    misc_StartUserErrorReport();
    misc_UserErrorReport("\nError: unknown output format for tableau.\n");
    misc_FinishUserErrorReport();
  }

  File = misc_OpenFile(Filename, "w");

  if (Format == DAVINCI)
    tab_FPrintDaVinciFormat(File, T);
  else 
    if (Format == XVCG)
      tab_FPrintCgFormat(File, T);

  misc_CloseFile(File, Filename);
}
