




 




























#include "dfg.h"
#include "defs.h"
#include "ia.h"
#include "rules-inf.h"
#include "rules-sort.h"
#include "rules-split.h"
#include "terminator.h"
#include "rules-red.h"
#include "analyze.h"
#include "clock.h"
#include "options.h"
#include "context.h"
#include "cnf.h"
#include "search.h"
#include "hasharray.h"
#include <errno.h>
#include <stdlib.h>


extern FILE *popen(const char *, const char *);
extern int  pclose (FILE *);






char *top_InputFile;

static OPTID top_RemoveFileOptId;

typedef enum {top_PROOF, top_COMPLETION, top_RESOURCE} top_SEARCHRESULT;





#if (defined(SPASS_SIGNALS))
#include <signal.h>

static PROOFSEARCH *top_PROOFSEARCH;

void top_SigHandler(int Signal)
{
  if (Signal == SIGSEGV || Signal == SIGBUS) {
    printf("\nSPASS %s ", misc_VERSION);
    printf("\nSPASS beiseite: Ran out of time. SPASS hit resource bound.");
    printf("\nProblem: %s ", 
	   (top_InputFile != (char*)NULL ? top_InputFile : "Read from stdin."));
    printf("\n");
    exit(EXIT_FAILURE);
  }

  if (Signal == SIGINT || Signal == SIGTERM) {
    clock_StopPassedTime(clock_OVERALL);
    printf("\nSPASS %s ", misc_VERSION);
    printf("\nSPASS beiseite: Ran out of time. SPASS was killed.");
    printf("\nProblem: %s ", 
	   (top_InputFile != (char*)NULL ? top_InputFile : "Read from stdin."));
    printf("\nSPASS derived %d clauses, backtracked %d clauses "
	   "and kept %d clauses.",
	   (*top_PROOFSEARCH == (PROOFSEARCH)NULL ? 0 : prfs_DerivedClauses(*top_PROOFSEARCH)),
	   (*top_PROOFSEARCH == (PROOFSEARCH)NULL ? 0 : prfs_BacktrackedClauses(*top_PROOFSEARCH)),
	   (*top_PROOFSEARCH == (PROOFSEARCH)NULL ? 0 : prfs_KeptClauses(*top_PROOFSEARCH)));
    printf("\nSPASS allocated %lu KBytes.", memory_DemandedBytes()/1024);
    printf("\nSPASS spent\t"); clock_PrintTime(clock_OVERALL);
    printf(" on the problem.");
    printf("\n\t\t"); clock_PrintTime(clock_INPUT);
    printf(" for the input.");
    printf("\n\t\t"); clock_PrintTime(clock_CNF);
    printf(" for the FLOTTER CNF translation.");
    printf("\n\t\t"); clock_PrintTime(clock_INFERENCE);
    printf(" for inferences.");
    printf("\n\t\t"); clock_PrintTime(clock_BACKTRACK);
    printf(" for the backtracking.");
    printf("\n\t\t"); clock_PrintTime(clock_REDUCTION);
    printf(" for the reduction.\n");
  }

  if (opts_IsSet(top_RemoveFileOptId))
    remove(top_InputFile);

  exit(EXIT_FAILURE);
}

#endif





NAT top_UltraPower()
{
  FILE *ultrapipe;
  char buffer[256];
  char *pos;
  NAT  MHz;

  MHz = 0;

  ultrapipe = popen("/usr/sbin/prtconf -pv | /usr/bin/grep MHz ", "r");
  if (ultrapipe == NULL) {
    perror(NULL);
    misc_Error();
  }
  if (fgets(buffer, 256, ultrapipe) != NULL) { 
    pclose(ultrapipe);
    pos = strstr(buffer, "MHz");
    *pos = '\0';
    while (isdigit((int)*(--pos))) ;
    MHz = atoi(pos+1);
  }
  else
    pclose(ultrapipe);

  if (MHz != 333 && MHz != 300) {
    printf("\n\n\t SPASS was not compiled and designed for this machine !! \n\n");
    printf("\t Please read the README/LICENSE file and contact weidenb@mpi-sb.mpg.de\n\n");
    exit(EXIT_FAILURE);
  }

  return MHz;
}


static CLAUSE top_SelectClauseDepth(LIST List)
{
  CLAUSE Result;
  NAT    Vars,NewVars,Weight,Depth,NewDepth;

  Result = (CLAUSE)list_Car(List);
  Depth  = clause_Depth(Result);
  Weight = clause_Weight(Result);
  Vars   = clause_NumberOfVarOccs(Result);
  List   = list_Cdr(List);

  while (!list_Empty(List)) {
    NewDepth =  clause_Depth(list_Car(List));
    if (NewDepth <= Depth) {
      if (NewDepth < Depth || clause_Weight(list_Car(List)) < Weight) {
	Depth  = NewDepth;
	Result = (CLAUSE)list_Car(List);
	Weight = clause_Weight(Result);
	Vars   = clause_NumberOfVarOccs(list_Car(List));
      }
      else
	if (clause_Weight(list_Car(List)) == Weight) {
	  NewVars = clause_NumberOfVarOccs(list_Car(List));
	  if (flag_GetFlagValue(flag_PREFVAR)) {
	    if (Vars < NewVars) {
	      Depth  = NewDepth;
	      Result = (CLAUSE)list_Car(List);
	      Weight = clause_Weight(Result);
	      Vars   = NewVars;
	    }
	  }
	  else
	    if (Vars > NewVars) {
	      Depth  = NewDepth;
	      Result = (CLAUSE)list_Car(List);
	      Weight = clause_Weight(Result);
	      Vars   = NewVars;
	    }
	}
    }
    List = list_Cdr(List);
  }

  return Result;
}

static CLAUSE top_SelectMinimalWeightClause(LIST List)
{
  CLAUSE Result;
  NAT    Vars,NewVars,Weight;

  Result = (CLAUSE)list_Car(List);
  Weight = clause_Weight(Result);
  Vars   = clause_NumberOfVarOccs(Result);
  List   = list_Cdr(List);

  while (!list_Empty(List)) {
    if (clause_Weight(list_Car(List)) == Weight) {
      NewVars = clause_NumberOfVarOccs(list_Car(List));
      if (flag_GetFlagValue(flag_PREFVAR)) {
	if (Vars < NewVars) {
	  Result = (CLAUSE)list_Car(List);
	  Weight = clause_Weight(Result);
	  Vars   = NewVars;
	}
      }
      else
	if (Vars > NewVars) {
	  Result = (CLAUSE)list_Car(List);
	  Weight = clause_Weight(Result);
	  Vars   = NewVars;
	}
    }
    else
      return Result;
    List = list_Cdr(List);
  }
  return Result;
}

static CLAUSE top_SelectMinimalConWeightClause(LIST List, NAT Factor)
{
  CLAUSE Result;
  NAT    NewWeight,Weight, NewVars, Vars;

  Result = (CLAUSE)list_Car(List);
  Weight = clause_Weight(Result);
  if (clause_GetFlag(Result,CONCLAUSE))
    Weight = Weight / Factor;
  Vars   = clause_NumberOfVarOccs(list_Car(List));
  List   = list_Cdr(List);

  while (!list_Empty(List)) {
    NewWeight = clause_Weight(list_Car(List));
    if (clause_GetFlag(list_Car(List),CONCLAUSE))
      NewWeight = NewWeight / Factor;
    if (NewWeight < Weight) {
      Weight = NewWeight;
      Result = list_Car(List);
      Vars   = clause_NumberOfVarOccs(list_Car(List));
    }
    else {
      if (NewWeight == Weight) {
	NewVars = clause_NumberOfVarOccs(list_Car(List));
	if (flag_GetFlagValue(flag_PREFVAR)) {
	  if (Vars < NewVars) {
	    Result = (CLAUSE)list_Car(List);
	    Weight = NewWeight;
	    Vars   = NewVars;
	  }
	}
	else
	  if (Vars > NewVars) {
	    Result = (CLAUSE)list_Car(List);
	    Weight = NewWeight;
	    Vars   = NewVars;
	  }
      }
    }

    List = list_Cdr(List);
  }
  return Result;
}




LIST top_FullReductionSelectGivenComputeDerivables(PROOFSEARCH Search, CLAUSE *SplitClause, int *Counter)
{
  CLAUSE GivenClause, TerminatorClause;
  LIST   Derivables;

  GivenClause = (CLAUSE)NULL;
  Derivables  = list_Nil();

  if (*SplitClause != (CLAUSE)NULL &&
      list_PointerMember(prfs_UsableClauses(Search),*SplitClause))
    GivenClause = *SplitClause;
  else {
    if ((*Counter)%flag_GetFlagValue(flag_WDRATIO) == 0)
      GivenClause = top_SelectClauseDepth(prfs_UsableClauses(Search));
    else {
      if (flag_GetFlagValue(flag_PREFCON) > 1)
	GivenClause = top_SelectMinimalConWeightClause(prfs_UsableClauses(Search), flag_GetFlagValue(flag_PREFCON));
      else
	GivenClause = top_SelectMinimalWeightClause(prfs_UsableClauses(Search));
    }
    (*Counter)++;
  }
  prfs_ExtractUsable(Search, GivenClause);
  *SplitClause = (CLAUSE)NULL;

  if (flag_GetFlagValue(flag_PGIVEN)) {
    printf("\n\tGiven clause: ");
    clause_Print(GivenClause); 
    fflush(stdout);
  }    
      
  if (prfs_SplitCounter(Search) != 0 && 
      (*SplitClause = prfs_PerformSplitting(Search,GivenClause)) != (CLAUSE)NULL)
    Derivables = list_Cons(*SplitClause,Derivables);
  else { 
    if (flag_GetFlagValue(flag_RTER) > 0) {
      TerminatorClause = red_Terminator(GivenClause,
					flag_GetFlagValue(flag_RTER),
					prfs_WorkedOffSharingIndex(Search),
					prfs_UsableSharingIndex(Search));
      if (TerminatorClause != NULL) {
	Derivables = list_Cons(TerminatorClause, Derivables);
	prfs_InsertUsableClause(Search,GivenClause);
      }
    }
    if (list_Empty(Derivables)) {
      clause_SelectLiteral(GivenClause);
      
      prfs_InsertWorkedOffClause(Search,GivenClause);
      clock_StartCounter(clock_INFERENCE);
      Derivables = list_Nconc(Derivables,inf_DerivableClauses(Search,GivenClause));
      clock_StopAddPassedTime(clock_INFERENCE);
    }
  }

  prfs_IncDerivedClauses(Search,list_Length(Derivables));

  return Derivables;
}

LIST top_LazyReductionSelectGivenComputeDerivables(PROOFSEARCH Search, CLAUSE *SplitClause, int *Counter)
{
  CLAUSE GivenClause, TerminatorClause;
  LIST   Derivables;

  GivenClause = (CLAUSE)NULL;
  Derivables  = list_Nil();

  while (GivenClause == (CLAUSE)NULL &&
	 !list_Empty(prfs_UsableClauses(Search))) { 
    if (*SplitClause != (CLAUSE)NULL &&
	list_PointerMember(prfs_UsableClauses(Search),*SplitClause))
      GivenClause = *SplitClause;
    else {
      if ((*Counter)%flag_GetFlagValue(flag_WDRATIO) == 0)
	GivenClause = top_SelectClauseDepth(prfs_UsableClauses(Search));
      else {
	if (flag_GetFlagValue(flag_PREFCON) > 1)
	GivenClause = top_SelectMinimalConWeightClause(prfs_UsableClauses(Search), flag_GetFlagValue(flag_PREFCON));
      else
	GivenClause = top_SelectMinimalWeightClause(prfs_UsableClauses(Search));
      }
      (*Counter)++;
    }
    prfs_ExtractUsable(Search, GivenClause);
    GivenClause = red_CompleteReductionOnDerivedClause(Search, GivenClause, 
						       flag_GetFlagValue(flag_DOCPROOF), red_WORKEDOFF);
  }
  *SplitClause = (CLAUSE)NULL;
  
  if (GivenClause != (CLAUSE)NULL) {  
    if (clause_IsEmptyClause(GivenClause)) {
      Derivables = list_List(GivenClause);
      return Derivables;
    }
    else {
      clock_StartCounter(clock_REDUCTION);
      Derivables = red_BackReduction(Search,GivenClause,
				     flag_GetFlagValue(flag_DOCPROOF),red_WORKEDOFF);
      clock_StopAddPassedTime(clock_REDUCTION);
    }

    if (list_Empty(Derivables)) {  
      if (flag_GetFlagValue(flag_PGIVEN)) {
	printf("\n\tGiven clause: ");
	clause_Print(GivenClause); 
	fflush(stdout);
      }     
      if (prfs_SplitCounter(Search) != 0 && 
	  (*SplitClause = prfs_PerformSplitting(Search,GivenClause)) != (CLAUSE)NULL)
	Derivables = list_Cons(*SplitClause,Derivables);
      else { 
	if (flag_GetFlagValue(flag_RTER) > 0) {
	  TerminatorClause = red_Terminator(GivenClause,
					    flag_GetFlagValue(flag_RTER),
					    prfs_WorkedOffSharingIndex(Search),
					    prfs_UsableSharingIndex(Search));
	  if (TerminatorClause != NULL) {
	    Derivables = list_Cons(TerminatorClause, Derivables);
	    prfs_InsertUsableClause(Search,GivenClause);
	  }
	}
	if (list_Empty(Derivables)) {   
	  clause_SelectLiteral(GivenClause);
	  
	  prfs_InsertWorkedOffClause(Search,GivenClause);
	  clock_StartCounter(clock_INFERENCE);
	  Derivables = list_Nconc(Derivables,inf_DerivableClauses(Search,GivenClause));
	  clock_StopAddPassedTime(clock_INFERENCE);
	}
      }
    }
    else
      prfs_InsertUsableClause(Search,GivenClause);

    prfs_IncDerivedClauses(Search,list_Length(Derivables));
  }

  return Derivables;
}

PROOFSEARCH top_ProofSearch(PROOFSEARCH Search, LIST ProblemClauses, FLAGSTORE InputStore, 
			    int *BoundApplied)
{
  LIST   Scan,Usables,WorkedOffs,SubsortClauses,EmptyClauses,Derivables,UsedEmptyClauses;
  CLAUSE SplitClause,HighestLevelEmptyClause;
  int    Counter, ActBound, BoundMode, BoundLoops;

  HighestLevelEmptyClause = (CLAUSE)NULL;
  UsedEmptyClauses        = list_Nil();
  EmptyClauses            = list_Nil();
  Usables                 = list_Nil();
  WorkedOffs              = list_Nil();
  SubsortClauses          = list_Nil();
  Derivables              = list_Nil();
  Counter                 = 1;

  clock_InitCounter(clock_REDUCTION);
  clock_InitCounter(clock_BACKTRACK);
  clock_InitCounter(clock_INFERENCE);

  ana_AnalyzeProblem(Search, ProblemClauses); 

  if (flag_GetFlagValue(flag_AUTO)) {
    prfs_InstallFiniteMonadicPredicates(Search, ProblemClauses, ana_FinMonPredList());
    symbol_InstallPrecedence(prfs_Precedence(Search));
    flag_InstallStore(prfs_Store(Search));
    ana_AutoConfiguration(ProblemClauses);
    flag_TransferSetFlags(InputStore,prfs_Store(Search)); 
  }
  
  list_Apply(clause_OrientAndReInit,ProblemClauses);

  ana_AnalyzeSortStructure(ProblemClauses);  

  if (flag_GetFlagValue(flag_AUTO)) {
    ana_ExploitSortAnalysis();
    flag_TransferSetFlags(InputStore,prfs_Store(Search)); 
  }
  prfs_SetSplitCounter(Search,flag_GetFlagValue(flag_SPLITS));

  ActBound       = flag_GetFlagValue(flag_BOUNDSTART);
  BoundMode      = flag_GetFlagValue(flag_BOUNDMODE);
  BoundLoops     = flag_GetFlagValue(flag_BOUNDLOOPS);
  *BoundApplied  = -1;

  if (flag_GetFlagValue(flag_PPROBLEM)) {
    printf("\n--------------------------SPASS-START-----------------------------");
    printf("\nInput Problem:\n");
    clause_ListPrint(ProblemClauses);
    printf("\n");
    
  }

  if (flag_GetFlagValue(flag_SORTS) != 0) {
    BOOL Strong;
    Strong     = (flag_GetFlagValue(flag_SORTS) == 2);
    for(Scan=ProblemClauses;!list_Empty(Scan);Scan=list_Cdr(Scan))
      clause_SetSortConstraint((CLAUSE)list_Car(Scan),Strong);
  }

  if (flag_GetFlagValue(flag_RINPUT)) {
    clock_StartCounter(clock_REDUCTION);
    EmptyClauses = red_ReduceInput(Search, ProblemClauses, flag_GetFlagValue(flag_DOCPROOF));
    clock_StopAddPassedTime(clock_REDUCTION);
    if (list_Empty(EmptyClauses) && flag_GetFlagValue(flag_SATINPUT))
      EmptyClauses = red_SatInput(Search, flag_GetFlagValue(flag_DOCPROOF));
  }
  else {
    for(Scan=ProblemClauses;!list_Empty(Scan);Scan=list_Cdr(Scan))
      prfs_InsertUsableClause(Search, list_Car(Scan));
  }
  Derivables = list_Nil();

  if (ana_SortRestrictions() || (ana_UnsolvedSortRestrictions() && flag_GetFlagValue(flag_SORTS) == 2)) {
    if (flag_GetFlagValue(flag_RSST))
      prfs_SetStaticSortTheory(Search,sort_ApproxStaticSortTheory(prfs_UsableClauses(Search)));
    prfs_SetDynamicSortTheory(Search,sort_TheoryCreate());
  }

  if (flag_GetFlagValue(flag_SOS)) {
    Derivables = list_Copy(prfs_UsableClauses(Search));
    for (Scan=Derivables;!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (!clause_GetFlag(list_Car(Scan), CONCLAUSE))
	prfs_MoveUsableWorkedOff(Search,list_Car(Scan));
    list_Delete(Derivables);
  }

  if (flag_GetFlagValue(flag_PPROBLEM)) {
    printf("\nProcessed Problem:\n");
    printf("\nWorked Off Clauses:\n");
    clause_ListPrint(prfs_WorkedOffClauses(Search));
    printf("\nUsable Clauses:\n");
    clause_ListPrint(prfs_UsableClauses(Search));
  }

  while ((list_Empty(EmptyClauses) || !prfs_SplitStackEmpty(Search)) && 
	 prfs_Loops(Search) != 0 &&
	 ((*BoundApplied != -1) || !list_Empty(prfs_UsableClauses(Search))) &&
	 (flag_GetFlagValue(flag_TIMELIMIT) < 0 || 
	  flag_GetFlagValue(flag_TIMELIMIT) > clock_GetSeconds(clock_OVERALL))) {

    Derivables    = list_Nil();
    SplitClause   = (CLAUSE)NULL;
    *BoundApplied = -1;

    while ((list_Empty(EmptyClauses) || !prfs_SplitStackEmpty(Search)) && 
	   prfs_Loops(Search) != 0 &&
	   (!list_Empty(prfs_UsableClauses(Search)) || !list_Empty(EmptyClauses)) &&
	   (flag_GetFlagValue(flag_TIMELIMIT) < 0 || 
	    flag_GetFlagValue(flag_TIMELIMIT) > clock_GetSeconds(clock_OVERALL))) {
      
      if (!list_Empty(EmptyClauses)) {

	clock_StartCounter(clock_BACKTRACK);
	Derivables = split_Backtrack(Search,HighestLevelEmptyClause);
	clock_StopAddPassedTime(clock_BACKTRACK);
	prfs_IncBacktrackedClauses(Search,list_Length(Derivables));
      
	if (prfs_SplitStackEmpty(Search)) {
	  Derivables = list_Nconc(EmptyClauses,Derivables);
	}
	else {
	  LIST l;
	  for(l = EmptyClauses; l != list_Nil(); l = list_Cdr(l))
	    if ((CLAUSE) list_Car(l) != HighestLevelEmptyClause)
	      clause_Delete((CLAUSE) list_Car(l));
	  prfs_InsertDocProofClause(Search, HighestLevelEmptyClause);
	  
	  if (flag_GetFlagValue(flag_DOCPROOF))
	    UsedEmptyClauses = list_Cons(HighestLevelEmptyClause, UsedEmptyClauses);
	  list_Delete(EmptyClauses);
	  if (flag_GetFlagValue(flag_DOCSPLIT))
	    printf("\n\t Split Backtracking level %d:",prfs_ValidLevel(Search));
	}
	EmptyClauses            = list_Nil();
	HighestLevelEmptyClause = (CLAUSE)NULL;
      }
      else {			

#ifdef DEBUG
	if (!prfs_Check(Search)) {
	  misc_StartErrorReport();
	  misc_ErrorReport("\n Illegal state of search in SPASS.\n");
	  misc_FinishErrorReport();
	}
	if (!ana_Equations())
	  red_CheckSplitSubsumptionCondition(Search);
#endif 

	if (flag_GetFlagValue(flag_FULLRED))
	  Derivables = top_FullReductionSelectGivenComputeDerivables(Search, &SplitClause, &Counter);
	else
	  Derivables = top_LazyReductionSelectGivenComputeDerivables(Search, &SplitClause, &Counter);
      }
      if (flag_GetFlagValue(flag_PDER))
	for(Scan=Derivables;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	  printf("\nDerived: "); 
	  clause_Print(list_Car(Scan));
	}

      Derivables = split_ExtractEmptyClauses(Derivables,&EmptyClauses);

      if (flag_GetFlagValue(flag_FULLRED)) { 
	clock_StartCounter(clock_REDUCTION);
	EmptyClauses = 
	  list_Nconc(EmptyClauses,
		     red_CompleteReductionOnDerivedClauses(Search, Derivables, 
							   flag_GetFlagValue(flag_DOCPROOF), 
							   red_ALL, ActBound, BoundMode, 
							   BoundApplied));
	clock_StopAddPassedTime(clock_REDUCTION);
      }
      else {
	clock_StartCounter(clock_REDUCTION);
	EmptyClauses = 
	  list_Nconc(EmptyClauses,
		     red_CompleteReductionOnDerivedClauses(Search, Derivables, 
							   flag_GetFlagValue(flag_DOCPROOF), 
							   red_WORKEDOFF, ActBound, BoundMode, 
							   BoundApplied));
	clock_StopAddPassedTime(clock_REDUCTION);
      }
      

      if (!list_Empty(EmptyClauses)) {
	HighestLevelEmptyClause = split_SmallestSplitLevelClause(EmptyClauses);
	if (flag_GetFlagValue(flag_PEMPTYCLAUSE)) {
	  printf("\nEmptyClause: ");
	  clause_Print(HighestLevelEmptyClause);
	}
      }
      prfs_DecLoops(Search);
    }
  
    if (ActBound != -1 && BoundMode != 0) {
      BoundLoops--;
      if (BoundLoops == 0)
	ActBound = -1;
      else
	ActBound = *BoundApplied;
      if (*BoundApplied != -1) {
	prfs_SwapIndexes(Search);
	if  (flag_GetFlagValue(flag_PBINC))
	  printf("\n\n\t ---- New Clause %s Bound: %2d ----\n",
		 (BoundMode-1 ? "Term Depth" : "Weight"),ActBound);
      }
    }
  }
  prfs_SetEmptyClauses(Search,EmptyClauses);
  prfs_SetUsedEmptyClauses(Search,UsedEmptyClauses);

  return Search;
}

void top_Flotter(int argc, char* argv[], LIST InputClauses)
{
  FILE *Output;
  LIST Scan;
  
  if (argc < opts_Indicator()+2)
    Output = stdout;
  else
    Output = misc_OpenFile(argv[opts_Indicator()+1],"w");

  fprintf(Output, "begin_problem(Unknown).\n\n");
  dfg_FPrintDescription(Output);
  fprintf(Output, "\n\n");  
  fprintf(Output, "list_of_symbols.\n");
  fol_FPrintDFGSignature(Output);
  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "list_of_clauses(axioms, cnf).\n");
  
  for (Scan=InputClauses;!list_Empty(Scan);Scan=list_Cdr(Scan))
    if (!clause_GetFlag(list_Car(Scan),CONCLAUSE))
      clause_FPrintDFG(Output,list_Car(Scan),FALSE);

  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "list_of_clauses(conjectures, cnf).\n");
  
  for (Scan=InputClauses;!list_Empty(Scan);Scan=list_Cdr(Scan))
    if (clause_GetFlag(list_Car(Scan),CONCLAUSE))
      clause_FPrintDFG(Output,list_Car(Scan),FALSE);
 
  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "\nend_problem.\n\n");

  
  printf("\nFLOTTER needed\t"); 
  clock_PrintTime(clock_INPUT);
  printf(" for the input.");
  printf("\n\t\t");
  clock_PrintTime(clock_CNF);
  printf(" for the  CNF translation.");
  

  if (Output != stdout)
    misc_CloseFile(Output,argv[opts_Indicator()+1]);
}

BOOL top_CalledFlotter(char* Call)
{
  int length;
  length = strlen(Call);
  return (strcmp((Call + (length > 7 ? length - 7 : 0)),"FLOTTER") ==0);
}







int main(int argc, char* argv[])
{
  LIST              InputClauses,Scan,Axioms,Conjectures, Sorts, QueryClauses, 
                    LabelClauses, QueryPair, ProblemClauses, Labellist, Sortlabellist, 
                    Symblist;
  PROOFSEARCH       Search, FlotterSearch;
  FLAGSTORE         InputStore;
  FILE*             InputStream; 
  HASH              TermLabelToClauselist, ClauseToTermLabellist;
  NAT               n;
  top_SEARCHRESULT  Result;

  Search = (PROOFSEARCH)NULL;
  
#if (defined(SPASS_SIGNALS))
  top_PROOFSEARCH = &Search;
  signal(SIGINT, top_SigHandler);
  signal(SIGTERM, top_SigHandler);
  signal(SIGSEGV, top_SigHandler);
  signal(SIGBUS, top_SigHandler);
#endif

  clock_Init();
  clock_StartCounter(clock_OVERALL);
  memory_Init(-1);
  
  symbol_Init(TRUE);
  stack_Init();
  hash_Init();
  sort_Init();
  term_Init();
  fol_Init(TRUE);
  cont_Init();
  unify_Init();
  flag_Init();
  subs_Init();
  clause_Init();
  red_Init();
  dp_Init();
  opts_Init();
  ana_Init();

  
  Search     = prfs_Create();
  InputStore = flag_NewStore();
  n          = top_UltraPower();

  
  
  opts_DeclareSPASSFlagsAsOptions();
  top_RemoveFileOptId = opts_Declare("rf", opts_NOARGTYPE);

  if (!opts_Read(argc, argv)) 
    return EINVAL;

  flag_InitStoreByDefaults(InputStore);   
  opts_SetFlags(InputStore);
  flag_InstallStore(InputStore);

  if (argc < opts_Indicator()+1 && !flag_GetFlagValue(flag_STDIN)) { 
    printf("\n\t          %s %s",argv[0],misc_VERSION);
    if (top_CalledFlotter(argv[0]) || flag_GetFlagValue(flag_FLOTTER))
      printf("\n\t       Usage: FLOTTER [options] [<input-file>] [<output-file>]\n\n");
    else
      printf("\n\t       Usage: SPASS [options] [<input-file>] \n\n");
    printf("Possible options:\n\n");
    opts_PrintSPASSNames(); 
    return EINVAL;
  }
  FlotterSearch = NULL;

  Axioms = Conjectures = Sorts = Labellist = Sortlabellist = list_Nil();
  
  if (flag_GetFlagValue(flag_STDIN)) {
    top_InputFile = (char*)NULL;
    InputStream   = stdin;
  } else {
    top_InputFile = argv[opts_Indicator()];
    InputStream = misc_OpenFile(top_InputFile, "r");
  }
  
  clock_StartCounter(clock_INPUT);
  flag_CleanStore(InputStore);
  flag_InstallStore(InputStore);
  
  InputClauses = dfg_DFGParser(InputStream, &Axioms, &Conjectures, &Sorts); 

  
  opts_SetFlags(InputStore);
  flag_TransferSetFlags(InputStore,prfs_Store(Search));
  flag_InstallStore(prfs_Store(Search));
  

  cnf_Init(); 

  ana_SetInputStatistic(Axioms, Conjectures, InputClauses);
  
  ana_AutoFof();

  
  if (flag_GetFlagValue(flag_INTERACTIVE)) {
    flag_SetFlagValue(flag_DOCPROOF, 1);
  }

 if (flag_GetFlagValue(flag_DOCPROOF))
    prfs_AddDocProofSharingIndex(Search);

  
  if(flag_GetFlagValue(flag_DOCPROOF)) {
    TermLabelToClauselist = hsh_Create();
    ClauseToTermLabellist = hsh_Create();
  }
  else {
    TermLabelToClauselist = NULL;
    ClauseToTermLabellist = NULL;
  }
  
  
  for(Scan=Conjectures;!list_Empty(Scan);Scan=list_Cdr(Scan))
      list_Rplacd(list_Car(Scan), (LIST)term_Create(fol_Not(), 
						    list_List(list_PairSecond(list_Car(Scan)))));

  clock_StopPassedTime(clock_INPUT);

  if (top_InputFile) {
    misc_CloseFile(InputStream,top_InputFile);
    if (opts_IsSet(top_RemoveFileOptId))
      remove(top_InputFile);
  }

  clock_StartCounter(clock_CNF);

  if (list_Empty(InputClauses)) {
    NAT Termcount;

    Termcount  = 0;
    
    
    for(Scan = Axioms; !list_Empty(Scan); Scan = list_Cdr(Scan), Termcount++) {
      if (list_PairFirst(list_Car(Scan)) == NULL) {
	char L[100];
	char* Label;
	sprintf(L, "axiom%d", Termcount);
	Label = memory_Malloc(sizeof(char) * (strlen(L) + 1));
	sprintf(Label, "%s", L);
	list_Rplaca(list_Car(Scan), Label);
	if(flag_GetFlagValue(flag_DOCPROOF) && flag_GetFlagValue(flag_PLABELS)) {
	  printf("\nAdded label %s for axiom \n", Label);
	  fol_PrettyPrintDFG((TERM) list_PairSecond(list_Car(Scan)));
	}
      }
    }
    Termcount  = 0;
    for(Scan = Sorts; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      char L[100];
      char* Label;
      sprintf(L, "declaration%d", Termcount);
      Label = memory_Malloc(sizeof(char) * (strlen(L) + 1));
      sprintf(Label, "%s", L);
      list_Rplaca(list_Car(Scan), Label);
      if(flag_GetFlagValue(flag_DOCPROOF) && flag_GetFlagValue(flag_PLABELS)) {
	printf("\nAdded label %s for declaration \n", Label);
	fol_PrettyPrintDFG((TERM) list_PairSecond(list_Car(Scan)));
      }
      Sortlabellist = list_Cons(Label, Sortlabellist);
    }
    Axioms = list_Nconc(Axioms, Sorts);


    if(flag_GetFlagValue(flag_APPLYDEFS) != 0) {
      LIST ConjectureList;


      def_ExtractDefsFromTermlist(Search, Axioms, list_Nil()); 

      

      ConjectureList = def_ApplyDefinitionToTermList(prfs_Definitions(Search), Conjectures, 
						     flag_GetFlagValue(flag_APPLYDEFS));

         
    }

    
    Symblist = list_Nil();

    
    
    FlotterSearch = cnf_Flotter(Axioms,Conjectures,&InputClauses, &Labellist,
				TermLabelToClauselist, ClauseToTermLabellist,
				flag_GetFlagValue(flag_INTERACTIVE),&Symblist);

    InputClauses   = clause_ListSortWeighed(InputClauses);
    clause_SetCounter(1);
    for(Scan=InputClauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      clause_NewNumber(list_Car(Scan));
    }
  }
  else {
    dfg_DeleteFormulaPairList(Axioms);
    dfg_DeleteFormulaPairList(Sorts);
    dfg_DeleteFormulaPairList(Conjectures);
    if(flag_GetFlagValue(flag_APPLYDEFS)) {
      
      def_ExtractDefsFromClauselist(Search, InputClauses);
      def_FlattenDefinitionsDestructive(Search);
      for(Scan=prfs_Definitions(Search); !list_Empty(Scan); Scan=list_Cdr(Scan))
	InputClauses = def_ApplyDefToClauselist(Search, (DEF) list_Car(Scan), InputClauses, TRUE);
    }
  }

  clock_StopPassedTime(clock_CNF);

  if (top_CalledFlotter(argv[0]) || flag_GetFlagValue(flag_FLOTTER)) {
    top_Flotter(argc,argv,InputClauses);
    flag_SetFlagValue(flag_TIMELIMIT,0);         
    flag_SetFlagValue(flag_INTERACTIVE,0);
    flag_SetFlagValue(flag_PPROBLEM,0);
  }

  prfs_SetPrecedence(Search,symbol_CreatePrecedence());
  memory_Restrict(flag_GetFlagValue(flag_MEMORY));

  do {
    LIST deflist;
    int  BoundApplied;
    ProblemClauses = list_Nil();
    LabelClauses   = list_Nil();
    Result         = top_RESOURCE;

    if (flag_GetFlagValue(flag_INTERACTIVE)) {
      QueryPair      = ia_GetNextRequest(InputStream); 
      
      if (list_Empty(QueryPair)) {
	break;
      }
      for(Scan=list_PairSecond(QueryPair);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	LIST l;
	l = list_PointerDeleteDuplicates(list_Copy(hsh_GetWithCompareFunc(TermLabelToClauselist, 
									  list_Car(Scan),
									  cnf_LabelEqual,
									  hsh_StringHashKey)));
	LabelClauses = list_Nconc(l,LabelClauses);
      }
      
      for(Scan=Sortlabellist; !list_Empty(Scan); Scan=list_Cdr(Scan)) {
	LIST l;
	l = list_PointerDeleteDuplicates(list_Copy(hsh_GetWithCompareFunc(TermLabelToClauselist, 
									  list_Car(Scan),
									  cnf_LabelEqual,
									  hsh_StringHashKey)));
	LabelClauses = list_Nconc(l,LabelClauses);
      }	

      
      
      for(Scan=LabelClauses; !list_Empty(Scan); Scan=list_Cdr(Scan)) {
	CLAUSE C;
	LIST l;
	C = (CLAUSE) list_Car(Scan);
	l = list_Copy(hsh_Get(ClauseToTermLabellist, C));
	l = list_DeleteDuplicates(l, cnf_LabelEqual);
	list_Rplaca(Scan, clause_Copy(C));
	hsh_PutList(ClauseToTermLabellist, list_Car(Scan), l);
      }
      QueryClauses   = cnf_QueryFlotter(FlotterSearch, 
					list_PairFirst(QueryPair),
					&Symblist);
      ProblemClauses = list_Nconc(QueryClauses,LabelClauses);
      for(Scan=list_PairSecond(QueryPair); !list_Empty(Scan); Scan= list_Cdr(Scan))
	memory_Free((char*) list_Car(Scan), strlen((char*) list_Car(Scan)) + 1);
      list_Delete(list_PairSecond(QueryPair));
      list_PairFree(QueryPair);
      clock_InitCounter(clock_OVERALL);
      clock_StartCounter(clock_OVERALL);
    }
    else {
      ProblemClauses = InputClauses;
      InputClauses   = list_Nil();
    }


    prfs_SetSplitCounter(Search,flag_GetFlagValue(flag_SPLITS));
    prfs_SetLoops(Search,flag_GetFlagValue(flag_LOOPS));
    prfs_SetBacktrackedClauses(Search,0);
    BoundApplied = -1;
    Search       = top_ProofSearch(Search,ProblemClauses,InputStore,&BoundApplied);  
  
    if ((flag_GetFlagValue(flag_TIMELIMIT) < 0 ||
	 flag_GetFlagValue(flag_TIMELIMIT) > clock_GetSeconds(clock_OVERALL)) &&
	prfs_Loops(Search) != 0 && 
	(BoundApplied == -1 || !list_Empty(prfs_EmptyClauses(Search)))) {
      if (list_Empty(prfs_EmptyClauses(Search)))
	Result = top_COMPLETION;
      else
	Result = top_PROOF;
    }
   
    if (flag_GetFlagValue(flag_TIMELIMIT) != 0) {  
      printf("\nSPASS %s ", misc_VERSION);
      switch (Result) {
      case top_RESOURCE: 
	if (prfs_Loops(Search) != 0)
	  printf("\nSPASS beiseite: Ran out of time.");
	else
	  printf("\nSPASS beiseite: Maximal number of loops exceeded.");
	break;
      case top_PROOF:
	printf("\nSPASS beiseite: Proof found.");
	break;
      default:        
	printf("\nSPASS beiseite: Completion found.");
      }
      printf("\nProblem: %s ", 
	     (top_InputFile != (char*)NULL ? top_InputFile : "Read from stdin."));
      if (flag_GetFlagValue(flag_PSTATISTIC)) {
	clock_StopPassedTime(clock_OVERALL);
	printf("\nSPASS derived %d clauses, backtracked %d clauses "
	       "and kept %d clauses.",prfs_DerivedClauses(Search),prfs_BacktrackedClauses(Search),
	       prfs_KeptClauses(Search));
	printf("\nSPASS allocated %lu KBytes.", memory_DemandedBytes()/1024);
	printf("\nSPASS spent\t"); clock_PrintTime(clock_OVERALL);
	printf(" on the problem.");
	printf("\n\t\t"); clock_PrintTime(clock_INPUT);
	printf(" for the input.");
	printf("\n\t\t"); clock_PrintTime(clock_CNF);
	printf(" for the FLOTTER CNF translation.");
	printf("\n\t\t"); clock_PrintTime(clock_INFERENCE);
	printf(" for inferences.");
	printf("\n\t\t"); clock_PrintTime(clock_BACKTRACK);
	printf(" for the backtracking.");
	printf("\n\t\t"); clock_PrintTime(clock_REDUCTION);
	printf(" for the reduction.\n");
      }
      if (Result != top_PROOF && flag_GetFlagValue(flag_FPMODEL)) {
	FILE *Output;
	char name[100];
	if (Result == top_COMPLETION)
	  sprintf(name,"%s.model",(top_InputFile != (char*)NULL ? top_InputFile : "SPASS"));
	else
	  sprintf(name,"%s.clauses",(top_InputFile != (char*)NULL ? top_InputFile : "SPASS"));
	Output = misc_OpenFile(name,"w");
	if (Result == top_COMPLETION)
	  fprintf(Output, "begin_problem(Completion_by_SPASS).\n\n");
	else
	  fprintf(Output, "begin_problem(Clauses_generated_by_SPASS).\n\n");
	dfg_FPrintDescription(Output);
	fprintf(Output, "list_of_symbols.\n");
	fol_FPrintDFGSignature(Output);
	fprintf(Output, "end_of_list.\n\n");
	if (Result == top_COMPLETION)
	  clause_FPrintCnfAsFormulasDFG(Output, (flag_GetFlagValue(flag_FPMODEL) == 2), 
			      prfs_WorkedOffClauses(Search),list_Nil());
	else
	  clause_FPrintCnfAsFormulasDFG(Output, (flag_GetFlagValue(flag_FPMODEL) == 2), 
			      prfs_WorkedOffClauses(Search), prfs_UsableClauses(Search));
	fprintf(Output, "list_of_settings(SPASS).\n{*\n");
	fol_FPrintPrecedence(Output);
	fprintf(Output, "\n*}\nend_of_list.\n\n");
	fprintf(Output, "end_problem.\n\n");
	misc_CloseFile(Output,name);
	if (Result == top_COMPLETION)
	  printf("\nCompletion printed to: %s\n",name);
	else
	  printf("\nClauses printed to: %s\n",name);
      }
    
      if (flag_GetFlagValue(flag_DOCPROOF) && Result != top_RESOURCE) {
	if (Result == top_COMPLETION) {
	  printf("\n\n The saturated set of worked-off clauses is :\n");
	  clause_ListPrint(prfs_WorkedOffClauses(Search));
	}
	else {
	  LIST UsedClauses, UsedTerms;
	  if (!top_InputFile)
	    top_InputFile = "SPASS";
	  UsedClauses = dp_PrintProof(Search, prfs_EmptyClauses(Search), top_InputFile);
	  
	  UsedTerms = list_Nil();

	  for(Scan=UsedClauses; !list_Empty(Scan); Scan = list_Cdr(Scan))
	    if (clause_IsFromInput((CLAUSE) list_Car(Scan))) {
	      LIST L;
	      L = hsh_Get(ClauseToTermLabellist, list_Car(Scan));
	      L = list_Copy(L);
	      L = list_DeleteDuplicates(L, cnf_LabelEqual);
	      UsedTerms = list_Nconc(UsedTerms, L);
	    }
	  list_Delete(UsedClauses);
	  UsedTerms = list_DeleteDuplicates(UsedTerms, cnf_LabelEqual);
	  fputs("\nFormulae used in the proof :", stdout);
	  for(Scan = UsedTerms; !list_Empty(Scan); Scan = list_Cdr(Scan)) 
	    if(!(strncmp((char*) list_Car(Scan), "_SORT_", 6) == 0))
	      printf(" %s", (char*) list_Car(Scan));
	  fputs("\n",stdout);
	  list_Delete(UsedTerms);
	}
      }
    }
    
    
    for(Scan=LabelClauses; !list_Empty(Scan); Scan=list_Cdr(Scan))
      hsh_DelItem(ClauseToTermLabellist, list_Car(Scan));
    
    list_Delete(ProblemClauses);
 
    fflush(stdout);

    
    deflist = prfs_Definitions(Search);
    prfs_SetDefinitions(Search, list_Nil());
    prfs_Clean(Search);
    prfs_SetDefinitions(Search, deflist);

    symbol_InstallDefaultPrecedence();
    if (flag_GetFlagValue(flag_PPROBLEM))
      printf("\n--------------------------SPASS-STOP------------------------------");
  } while (flag_GetFlagValue(flag_INTERACTIVE) &&
	   (flag_GetFlagValue(flag_TIMELIMIT) != 0));

  for(Scan=InputClauses; !list_Empty(Scan); Scan=list_Cdr(Scan))
    clause_OrientAndReInit((CLAUSE) list_Car(Scan));

  
  if(flag_GetFlagValue(flag_INTERACTIVE)) {
    if (flag_GetFlagValue(flag_DOCPROOF))
      list_Delete(Symblist);
    else 
      list_DeleteWithElement(Symblist,symbol_Delete); 
    
    if (FlotterSearch != NULL) 
      prfs_Delete(FlotterSearch);
  }
  if (flag_GetFlagValue(flag_PFLAGS)) {
    printf("\n");
    flag_Print();
  }
  if(flag_GetFlagValue(flag_DOCPROOF)) {
    hsh_Delete(TermLabelToClauselist);
    hsh_Delete(ClauseToTermLabellist);
  }
  
#ifdef DEBUG
  memory_Print();
#endif
  printf("\n");
  return 0;
}
